import { Component, Input, OnDestroy, OnInit } from '@angular/core';
import { Map } from 'ol';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { VmapSelectService } from 'src/modules/vmap/components/carto/services/vmap-select/vmap-select.service';
import { Subscription } from 'rxjs';

@Component({
  selector: 'app-vmap-layers-tree',
  templateUrl: './vmap-layers-tree.component.html',
  styleUrls: ['./vmap-layers-tree.component.scss']
})
export class VmapLayersTreeComponent implements OnInit, OnDestroy {

  @Input() public olMap: Map;

  public aLayersTree: any;
  public sVisibleParametersID = '';

  // Permet la création d'identifants uniques dans les noeuds de l'arbre
  private iCounter = 0;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  constructor(
    private vitisCheck: VitisCheck,
    private mapCoreManagerService: MapCoreManagerService,
    private vmapToolsService: VmapToolsService,
    private vmapSelectService: VmapSelectService
  ) { }

  ngOnInit(): void {

    // initialise l'arbre
    this.initTree_();
    this.olMap.on('change', () => {
      this.initTree_();
    });

    // Fix CSS pour la hauteur de la branche des éléments enfants
    setTimeout(() => {
      const layerTree = document.getElementById('layer-tree');
      const aChildrens = Array.from(layerTree.querySelectorAll('.childrens'));

      for (const children of aChildrens) {
        const count: number = children.childElementCount - 1;
        let height: number = 0;

        for (let i = 0; i < count; i++) {
          if (i === 0) {
            height += 30;
          } else {
            height += 50;
          }
        }

        const branch: any = children.querySelector('.childrens-branch');
        branch.style.height = height + 'px';
      }
    });

    this.subscriptions.push(this.vmapToolsService.getLeftPanelEmiter().subscribe((iNbCols: number) => {
      if (iNbCols === 0) {
        this.sVisibleParametersID = '';
      }
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Rajoute un id unique aux éléments de l'arbre
   * @param oTreeChildren
   * @returns
   */
  private formatLayersTree_(oTreeChildren: any) {
    oTreeChildren.id = this.mapCoreManagerService.getUniqueId();

    if (oTreeChildren.children) {
      for (let oChildren of oTreeChildren.children) {
        oChildren = this.formatLayersTree_(oChildren);
      }
    }

    return oTreeChildren;
  }

  /**
   * Gère la fermeture / ouverture d'un groupe de couches avec une transition
   * @param event
   */
  public toggleAccordion(event, element) {
    const layerGroup = event.target.closest('.layer-group');
    const childrens = layerGroup.querySelector('.childrens');
    element.collapsed = !element.collapsed;

    // Clean
    childrens.style.height = '';
    childrens.style.transition = 'none';

    // Hauteur initiale du groupe
    const startHeight = window.getComputedStyle(childrens).height;

    // On bascule l'état de la classe collapsed pour ouvrir / fermer le groupe de couches, on recalcule ensuite la hauteur
    childrens.classList.toggle('collapsed');

    // Hauteur finale du groupe
    const height = window.getComputedStyle(childrens).height;

    // On définit la hauteur de départ pour commencer la transition
    childrens.style.height = startHeight;

    requestAnimationFrame(() => {
      childrens.style.transition = '';

      requestAnimationFrame(() => {
        childrens.style.height = height;
      });
    });
  }

  /**
   * Retourne la visibilité de la couche dans l'arbre
   * @param layerElement couche dans l'arbre
   * @returns
   */
  public getLayerVisibility(layerElement: any): boolean {

    // Nécessaire pour les couches vectortiles qui se chargent après coup
    if (this.vitisCheck.isDefAndNotNull(layerElement.layer_id) &&
      !this.vitisCheck.isDefAndNotNull(layerElement.olLayer)) {
      const aLayers = this.mapCoreManagerService.deoptimiseLayers(this.olMap);
      for (let i = 0; i < aLayers.length; i++) {
        if (aLayers[i].get('layer_id') === layerElement.layer_id) {
          layerElement.olLayer = aLayers[i];
        }
      }
    }

    if (layerElement.layer_id && layerElement.olLayer) {
      return layerElement.olLayer.get('visible');
    }

    if (layerElement.children) {
      let visible = false;

      for (const children of layerElement.children) {
        const childrenVisibility = this.getLayerVisibility(children);

        if (childrenVisibility) {
          visible = true;
        }
      }

      return visible;
    }
  }

  /**
   * Set la visibilité d'une couche
   * @param layerElement
   * @param visibility
   */
  private setLayerVisibility(layerElement: any, visibility: boolean) {
    if (layerElement.layer_id && layerElement.is_map_hidden !== true) {
      this.mapCoreManagerService.setLayerVisible(layerElement, this.olMap, visibility);
    }

    if (layerElement.children) {
      for (const children of layerElement.children) {
        this.setLayerVisibility(children, visibility);
      }
    }
  }

  /**
   * Bascule l'état de visibilité d'une couche
   * @param id
   */
  public toggleLayerVisibility(element: any) {
    const currentVisibility = this.getLayerVisibility(element);
    this.setLayerVisibility(element, !currentVisibility);
  }

  /**
   * Bascule l'état de visibilité de la boîte de paramètres d'une couche, ou d'un groupe de couches
   * @param id
   */
  public toggleVisibleParameters(id) {
    if (id === this.sVisibleParametersID) {
      this.sVisibleParametersID = '';
    } else {
      this.sVisibleParametersID = id;
    }
  }

  /**
   * Initialise l'arbre
   */
  private initTree_() {
    this.aLayersTree = this.mapCoreManagerService.getLayersTree(this.olMap);
    this.createTreeIds_(this.aLayersTree);

    // Définit is_map_hidden pour les groupes dont toutes les couches le sont
    for (let i = 0; i < this.aLayersTree.length; i++) {
      const element = this.aLayersTree[i];
      if (element.children) {
        let bAllChildrensHidden = true;
        for (let ii = 0; ii < element.children.length; ii++) {
          const children = element.children[ii];
          if (children.is_map_hidden !== true) {
            bAllChildrensHidden = false;
          }
        }
        if (bAllChildrensHidden) {
          element.is_map_hidden = true;
        }
      }
    }
  }

  /**
   * Crée des identifiants dans les neuds de l'arbre
   * @param aTree
   */
  private createTreeIds_(aTree: any) {
    if (this.vitisCheck.isArray(aTree)) {
      for (let i = 0; i < aTree.length; i++) {
        const oNode = aTree[i];
        if (!this.vitisCheck.isDefAndNotNull(oNode.id)) {
          oNode.id = this.iCounter;
          this.iCounter++;
        }
        if (this.vitisCheck.isArray(oNode.children)) {
          this.createTreeIds_(oNode.children);
        }
      }
    }
  }
}
