import { Component, OnInit, Input, OnDestroy } from '@angular/core';
import { Map, Geolocation, Feature } from 'ol';
import { Geometry, Point } from 'ol/geom';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { Circle as CircleStyle, Fill, Stroke, Style } from 'ol/style';
import { Vector as VectorSource } from 'ol/source';
import { Vector as VectorLayer } from 'ol/layer';
import { unByKey } from 'ol/Observable';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { Subscription } from 'rxjs';

@Component({
  selector: 'app-vmap-tools-geolocation-mobile',
  templateUrl: './vmap-tools-geolocation-mobile.component.html',
  styleUrls: ['./vmap-tools-geolocation-mobile.component.scss']
})
export class VmapToolsGeolocationMobileComponent implements OnInit, OnDestroy {

  @Input() public olMap: Map;

  private geolocation: Geolocation;
  private enableGeolocation: boolean;
  private accuracyFeature: Feature<Geometry>;
  private positionFeature: Feature<Geometry>;

  private vectorLayer: VectorLayer<VectorSource<Geometry>>;

  private listenerKeyAccuracyChange;
  private listenerKeyPositionChange;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  constructor(
    private vmapToolsService: VmapToolsService,
    private mapUtils: MapCoreUtilsService,
  ) { }

  ngOnInit() {
    this.geolocation = new Geolocation({
      trackingOptions: {
        enableHighAccuracy: true,
      },
      projection: this.olMap.getView().getProjection(),
    });

    // Écoute l'ouverture des menus
    this.subscriptions.push(this.vmapToolsService.getActiveRightPanelEmiter().subscribe((sTool: string) => {
      this.enableGeolocation = (sTool === 'geolocation-mobile');
      this.geolocation.setTracking(this.enableGeolocation);
      if (this.enableGeolocation) {
        this.trackPosition();
      } else {
        this.disableTrackPosition();
      }
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Retourne la position actuelle avec son taux d'erreur
   */
  private trackPosition() {

    this.accuracyFeature = new Feature();
    this.positionFeature = new Feature();
    this.positionFeature.setStyle(
      new Style({
        image: new CircleStyle({
          radius: 6,
          fill: new Fill({
            color: '#3399CC',
          }),
          stroke: new Stroke({
            color: '#fff',
            width: 2,
          }),
        }),
      })
    );

    this.vectorLayer = new VectorLayer({
      source: new VectorSource({
        features: [this.accuracyFeature, this.positionFeature],
      }),
    });
    this.olMap.addLayer(this.vectorLayer);
    // Ajoute la couche de géolocalisation au premier plan
    this.vectorLayer.setZIndex(this.olMap.getLayers().getArray().length + 1);

    this.listenerKeyAccuracyChange = this.geolocation.on('change:accuracyGeometry', (e: any) => {
      this.accuracyFeature.setGeometry(e.target.getAccuracyGeometry());
      this.mapUtils.centerMapOnGeom(this.olMap, this.accuracyFeature.getGeometry(), [200, 0, 200, 0]);
    });

    this.listenerKeyPositionChange = this.geolocation.on('change:position', (e: any) => {
      const coordinates = e.target.getPosition();
      this.positionFeature.setGeometry(coordinates ? new Point(coordinates) : null);
    });
  }

  /**
   * Désactive le tracking en cours
   */
  private disableTrackPosition() {
    if (this.positionFeature) {
      this.positionFeature = null;
      this.accuracyFeature = null;
    }
    if (this.listenerKeyAccuracyChange) {
      unByKey(this.listenerKeyAccuracyChange);
    }
    if (this.listenerKeyPositionChange) {
      unByKey(this.listenerKeyPositionChange);
    }
    if (this.vectorLayer) {
      this.vectorLayer.getSource().clear();
      this.olMap.removeLayer(this.vectorLayer);
      this.vectorLayer = null;
    }
    this.enableGeolocation = false;
  }
}
