import { Component, OnInit, Input, OnDestroy, ViewChild, ElementRef } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { Map } from 'ol';
import { unByKey } from 'ol/Observable';
import { EventsKey } from 'ol/events';
import { MapToolDef } from '../../map-tools.model';
import { LineString, Point, Polygon, Circle } from 'ol/geom';
import { transform } from 'ol/proj';
import { TranslateService } from '@ngx-translate/core';
import { MapCoreDrawService } from 'src/app/vitis/map-core/services/map-core-draw/map-core-draw.service';
import { MapCorePopupService } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import Overlay from 'ol/Overlay';
import { Fill, Stroke, Style } from 'ol/style';
import CircleStyle from 'ol/style/Circle';
import moment from 'moment-timezone';
import { Subscription } from 'rxjs';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { NgbActiveModal, NgbModal } from '@ng-bootstrap/ng-bootstrap';

declare var $: any;

/**
 * Renvoie le nombre de décimal après la virgule
 * @param proj
 * @returns
 */
const formatDecimal = (proj) => proj === 'EPSG:4326' ? 7 : 2;

/**
  * Renvoie le format correct en km ou m
  * @param value
  * @param {number} decimal
  * @param {number} power
  * @returns string
  */
const formatMeasures = (value, decimal = 2, power: number = 1) => {
  const km = value / Math.pow(1000, power);
  if ((value < 0 && (value * -1) > 1) || km > 1) {
    return `${km.toFixed(decimal)} km`;
  }
  return `${value.toFixed(decimal)} m`;
};

@Component({
  selector: 'app-vmap-tools-measure',
  templateUrl: './vmap-tools-measure.component.html',
  styleUrls: ['./vmap-tools-measure.component.scss']
})
export class VmapToolsMeasureComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() oTool: MapToolDef;

  public bToolOpen;

  public geomtypes = ['POINT', 'LINE', 'POLYGON', 'CIRCLE'];
  public selectedRow: string;

  private interaction;
  private interactionSnap;
  private interactionSelect;

  public optionsProjection = [
    { value: 'EPSG:4326', text: '[EPSG:4326]-WGS84.LL' },
    { value: 'EPSG:2154', text: '[EPSG:2154]-RGF93.Lambert-93' },
    { value: 'EPSG:3857', text: '[EPSG:3857]-WGS84.Spherical Mercator' }
  ];
  public currentProjection = 'EPSG:4326';

  private listenerKeyAddFeature; // Event vectorSource.on('addFeature')
  private listenerKeyModifyStartFeature; // Event interaction.on('modifystart')
  private listenerKeyModifyFeature; // Event vectorSource.on('change')
  private listenerKeySelect; // Event interaction.on('select')
  private listenerKeyAddingUpdateFeature; // Event geometry.on('change')

  private infosMeasures = [];

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];
  private mapSubscriptions: EventsKey[] = [];

  // Variables pour la modale
  @ViewChild('measuresModal') public measuresModalContent: ElementRef;
  private measuresModal: NgbActiveModal;

  private olFeatureStyle = new Style({
    fill: new Fill({
      color: 'rgba(255, 255, 255, 0.2)',
    }),
    stroke: new Stroke({
      color: '#ffcc33',
      width: 2,
    }),
    image: new CircleStyle({
      radius: 7,
      fill: new Fill({
        color: '#ffcc33',
      }),
    })
  });

  private olMeasureStyle = new Style({
    fill: new Fill({
      color: 'rgba(255, 255, 255, 0.2)',
    }),
    stroke: new Stroke({
      color: 'rgba(0, 0, 0, 0.5)',
      lineDash: [10, 10],
      width: 2,
    }),
    image: new CircleStyle({
      radius: 5,
      stroke: new Stroke({
        color: 'rgba(0, 0, 0, 0.7)',
      }),
      fill: new Fill({
        color: 'rgba(255, 255, 255, 0.2)',
      }),
    }),
  });

  constructor(
    private route: ActivatedRoute,
    private translate: TranslateService,
    private mapCoreDrawService: MapCoreDrawService,
    private mapPopup: MapCorePopupService,
    private mapUtils: MapCoreUtilsService,
    private vmapToolsService: VmapToolsService,
    private modalService: NgbModal,
  ) { }

  ngOnInit(): void {
    let helpTooltipElement;
    let helpTooltip;

    this.mapSubscriptions.push(this.olMap.on('change', () => {
      // Vide les mesures et les properties au changement de carte
      this.bToolOpen = false;
      this.selectedRow = null;
      this.olMap.unset('measure_features');
      this.deleteMeasures();
      this.deleteInteractions();
      this.initDrawing();
    }));

    // Ouverture de l'outil
    this.subscriptions.push(this.oTool.onOpen.subscribe(() => {
      this.bToolOpen = true;
      if (this.olMap.get('measure_features')) {
        this.infosMeasures = this.olMap.get('measure_features');
      }
    }));

    // Fermeture de l'outil
    this.subscriptions.push(this.oTool.onClose.subscribe(() => {
      this.bToolOpen = false;
      this.selectedRow = null;
      this.olMap.set('measure_features', this.infosMeasures);
      this.deleteInteractions();
    }));

    const createHelpTooltip = () => {
      if (helpTooltipElement) {
        helpTooltipElement.parentNode.removeChild(helpTooltipElement);
      }
      helpTooltipElement = document.createElement('div');
      helpTooltipElement.className = 'ol-tooltip hidden';
      helpTooltip = new Overlay({
        element: helpTooltipElement,
        offset: [15, 0],
        positioning: 'center-left',
      });
      this.olMap.addOverlay(helpTooltip);
    };

    this.mapSubscriptions.push(this.olMap.on('pointermove', (evt) => {
      if (!helpTooltipElement) {
        createHelpTooltip();
      }
      if (this.bToolOpen && this.selectedRow) {
        let helpMsg = '';
        if (this.geomtypes.includes(this.selectedRow.toUpperCase())) {
          helpMsg = this.selectedRow === 'POINT' ? 'Cliquer pour mesurer l\'emplacement d\'un point' : 'Cliquer pour commencer à mesurer';
        } else if (this.selectedRow === 'edit') {
          helpMsg = 'Cliquer sur une mesure pour la modifier';
        } else if (this.selectedRow === 'delete') {
          helpMsg = 'Cliquer sur une mesure pour la supprimer';
        }
        if (helpMsg) {
          helpTooltipElement.innerHTML = helpMsg;
          helpTooltip.setPosition(evt.coordinate);
          helpTooltipElement.classList.remove('hidden');
        } else {
          helpTooltipElement.classList.add('hidden');
        }
      } else if (helpTooltipElement) {
        helpTooltipElement.classList.add('hidden');
      }
    }));

    this.olMap.getViewport().addEventListener('mouseout', function () {
      if (helpTooltipElement) {
        helpTooltipElement.classList.add('hidden');
      }
    });
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
    this.mapSubscriptions.forEach(mapSubscription => unByKey(mapSubscription));
  }

  /**
   * Initialise le vector Layer pour ajouter les mesures à la carte
   */
  private initDrawing(): void {
    this.olMap.set('measure_layer', this.mapCoreDrawService.initDrawLayer(this.olMap, null, this.olFeatureStyle));
  }

  /**
   * Supprime les intéractions et les listeners en cours
   */
  private deleteInteractions() {
    // Unlisten le précédent addFeature si il existe
    if (this.listenerKeyAddFeature) {
      unByKey(this.listenerKeyAddFeature);
    }
    // Unlisten le précédent change vector source si il existe
    if (this.listenerKeyModifyFeature) {
      unByKey(this.listenerKeyModifyFeature);
    }
    // Unlisten le précédent modifyStart si il existe
    if (this.listenerKeyModifyStartFeature) {
      unByKey(this.listenerKeyModifyStartFeature);
    }
    // Unlisten le précédent select il existe
    if (this.listenerKeySelect) {
      unByKey(this.listenerKeySelect);
    }
    // Unlisten le précédent change geometry il existe
    if (this.listenerKeyAddingUpdateFeature) {
      unByKey(this.listenerKeyAddingUpdateFeature);
    }
    if (this.interaction) {
      this.olMap.removeInteraction(this.interaction);
      this.interaction = null;
    }
    if (this.interactionSnap) {
      this.olMap.removeInteraction(this.interactionSnap);
      this.interactionSnap = null;
    }
    if (this.interactionSelect) {
      this.olMap.removeInteraction(this.interactionSelect);
      this.interactionSelect = null;
    }
  }

  /**
   * Function appelée lors de l'action d'ajout d'une géométrie à la carte
   */
  public addMeasure(geomtype) {
    if (!this.olMap.get('measure_layer')) {
      this.initDrawing();
    }

    this.deleteInteractions();
    if (!this.selectedRow || this.selectedRow !== geomtype) {
      this.selectedRow = geomtype;
    } else if (this.selectedRow === geomtype) {
      this.selectedRow = null;
      return;
    }

    // Set ou update l'interaction d'ajout si elle n'existe pas ou si on change de geomtype
    if (!this.interactionSnap && !this.interaction) {
      const { interaction, snap } = this.mapCoreDrawService.addDrawInteraction(geomtype.toLowerCase(), this.olMap, this.olMap.get('measure_layer'), { getInteractions: true, drawStyle: this.olMeasureStyle });
      this.interactionSnap = snap;
      this.interaction = interaction;
    }

    this.listenerKeyAddFeature = this.interaction.on('drawstart', (evt) => {
      this.addFeature(evt);
      const _this = this;
      // Unlisten le précédent change si il existe
      if (this.listenerKeyAddingUpdateFeature) {
        unByKey(this.listenerKeyAddingUpdateFeature);
      }
      let startModify = moment();
      this.listenerKeyAddingUpdateFeature = evt.feature.getGeometry().on('change', function (event) {
        const currentModify = moment();
        if (currentModify.diff(startModify, 'milliseconds') > 50) {
          startModify = currentModify;
          const infosMeasure = _this.infosMeasures.find(infos => infos.geometryUid == event.target.ol_uid);
          _this.setGeometryPopup(_this.olMap, event.target, infosMeasure);
        }
      });
    });
  }

  /**
   * Function d'ajout de la feature et de sa popup
   * @param evt
   */
  private addFeature(evt) {
    const geometryAdded: any | Polygon | Point | LineString | Circle = evt.feature.getGeometry();
    const geometryUid = geometryAdded.ol_uid; // Sauvegarde de l'uid de la géométrie pour les interactions
    let sPopupData: any; // Contenu dans la popup
    const infosMeasure: any = {
      geometry: { type: geometryAdded.getType() },
      sPopupData,
      projection: 'EPSG:4326'
    };
    switch (infosMeasure.geometry.type) {
      case 'Point':
        const coordinates = new Point(transform(geometryAdded.getCoordinates(), this.olMap.getView().getProjection(), infosMeasure.projection)).getCoordinates();
        infosMeasure.geometry.coordinates = coordinates;
        sPopupData = {
          select: {
            value: infosMeasure.projection,
            options: this.optionsProjection,
            event: (evtSelect) => {
              const projectionValue = evtSelect.target.value;
              const infosMeasureUpdate = this.infosMeasures.find(infos => infos.geometryUid == geometryUid);
              const transformCoordinates = new Point(transform(infosMeasureUpdate.geometry.coordinates, infosMeasureUpdate.projection, projectionValue)).getCoordinates();
              infosMeasureUpdate.geometry.coordinates = transformCoordinates;
              infosMeasureUpdate.projection = projectionValue;
              infosMeasureUpdate.sPopupData.X = transformCoordinates[0].toFixed(formatDecimal(projectionValue));
              infosMeasureUpdate.sPopupData.Y = transformCoordinates[1].toFixed(formatDecimal(projectionValue));
              infosMeasureUpdate.popup.setPopupMessage({ X: infosMeasureUpdate.sPopupData.X, Y: infosMeasureUpdate.sPopupData.Y });
            }
          },
          X: coordinates[0].toFixed(7),
          Y: coordinates[1].toFixed(7)
        };
        break;
      case 'LineString':
        const length = geometryAdded.getLength();
        infosMeasure.length = length;
        infosMeasure.geometry.coordinates = geometryAdded.getCoordinates();
        sPopupData = formatMeasures(length);
        break;
      case 'Polygon':
        const area = geometryAdded.getArea();
        infosMeasure.area = area;
        infosMeasure.geometry.coordinates = geometryAdded.getCoordinates();
        sPopupData = `${formatMeasures(area, 2, 2)}²`;
        break;
      case 'Circle':
        const radius = geometryAdded.getRadius();
        infosMeasure.radius = radius;
        infosMeasure.geometry.center = geometryAdded.getCenter();
        sPopupData = `Périmètre : ${formatMeasures(2 * Math.PI * radius)}`;
        break;
    }

    infosMeasure.popup = this.mapPopup.displayPopup(sPopupData, evt.feature, this.olMap, { hideCloseBtn: true, displayFeature: false });
    infosMeasure.sPopupData = sPopupData;
    infosMeasure.geometryUid = geometryUid;
    this.infosMeasures.push(infosMeasure);
  }

  /**
   * Function de modification de la feature et sa popup
   */
  public modifyMeasure() {
    this.deleteInteractions();
    if (!this.selectedRow || this.selectedRow !== 'edit') {
      this.selectedRow = 'edit';
    } else if (this.selectedRow === 'edit') {
      this.selectedRow = null;
      return;
    }

    // Set l'interaction de modification si elle n'existe pas
    if (!this.interactionSnap && !this.interaction) {
      const { select, interaction, snap } = this.mapCoreDrawService.addDrawInteraction('modify', this.olMap, this.olMap.get('measure_layer'), { getInteractions: true });
      this.interaction = interaction;
      this.interactionSelect = select;
      this.interactionSnap = snap;
    }

    this.listenerKeyModifyStartFeature = this.interaction.on('modifystart', (evt: any) => {
      // Récupération de l'uid de la géométrie en changement
      const modifiedGeometryUid = evt.features.getArray()[0].getGeometry().ol_uid;
      // Unlisten le précédent change si il existe
      if (this.listenerKeyModifyFeature) {
        unByKey(this.listenerKeyModifyFeature);
      }
      let startModify = moment();
      this.listenerKeyModifyFeature = this.olMap.get('measure_layer').getSource().on('change', () => {
        const currentModify = moment();
        if (currentModify.diff(startModify, 'milliseconds') > 50) {
          startModify = currentModify;
          const currModifiedFeature = evt.features.getArray().find(feature => feature.getGeometry().ol_uid == modifiedGeometryUid);
          const infosMeasure = this.infosMeasures.find(infos => infos.geometryUid == modifiedGeometryUid);
          this.setGeometryPopup(this.olMap, currModifiedFeature.getGeometry(), infosMeasure);
        }
      });
    });
  }

  /**
   * Function appelée lors de l'action de suppression d'une géométrie
   */
  public deleteMeasure() {
    this.deleteInteractions();
    if (!this.selectedRow || this.selectedRow !== 'delete') {
      this.selectedRow = 'delete';
    } else if (this.selectedRow === 'delete') {
      this.selectedRow = null;
      return;
    }

    // Set l'interaction de suppression si elle n'existe pas
    if (!this.interaction) {
      const { interaction } = this.mapCoreDrawService.addDrawInteraction('remove', this.olMap, this.olMap.get('measure_layer'), { getInteractions: true });
      this.interaction = interaction;
    }

    this.listenerKeySelect = this.interaction.on('select', (evt) => {
      if (evt.selected[0]) {
        const geometryIdToDel = evt.selected[0].getGeometry().ol_uid;
        // Suppression de la popup de la feature
        const indexinfosMeasure = this.infosMeasures.findIndex(infos => infos.geometryUid == geometryIdToDel);
        this.infosMeasures[indexinfosMeasure].popup.close();
        this.infosMeasures = [...this.infosMeasures.slice(0, indexinfosMeasure), ...this.infosMeasures.slice(indexinfosMeasure + 1, this.infosMeasures.length)];
      }
    });
  }

  /**
   * Function appelée lors de l'action de suppression de toutes les géométrie
   */
  public deleteMeasures() {
    this.selectedRow = null;
    this.deleteInteractions();

    // Supprime les features du vectorSource
    if (this.olMap.get('measure_layer')) {
      this.olMap.get('measure_layer').getSource().clear();
    }

    // Supprime toute les popups présent sur la couche
    this.infosMeasures.forEach(infos => infos.popup.close());
    this.infosMeasures = [];

  }

  /**
   * Affiche les infos des mesures
   */
  public showInfosMeasures() {
    if (this.getInfosMeasures().length > 0) {
      this.measuresModal = this.modalService.open(this.measuresModalContent, {
        size: 'xl',
      });
    }
  }

  /**
   * Cache les infos des mesures
   */
  public hideInfosMeasures() {
    if (this.measuresModal) {
      this.measuresModal.close();
    }
  }

  /**
   * Renvoie les infos sur les features de la couche vector layer
   */
  public getInfosMeasures() {
    let infosMeasures = [];
    if (this.olMap.get('measure_layer')) {
      infosMeasures = this.olMap.get('measure_layer').getSource().getFeatures().map(feature => {
        const geometry: any | Point | LineString | Polygon | Circle = feature.getGeometry();
        const type = geometry.getType();
        let data = {
          uid: geometry['ol_uid'],
          type,
          currentProjection: this.currentProjection,
        };
        switch (type) {
          case 'Point':
            data = {
              ...data,
              ...this.getInfosPoint(geometry)
            };
            break;
          case 'LineString':
            data = {
              ...data,
              ...this.getInfosLine(geometry)
            };
            break;
          case 'Polygon':
            data = {
              ...data,
              ...this.getInfosPolygon(geometry)
            };
            break;
          case 'Circle':
            data = {
              ...data,
              ...this.getInfosCircle(geometry)
            };
            break;
        }
        data.currentProjection = this.optionsProjection.find(option => option.value == data.currentProjection).text;
        return data;
      });
    }
    return infosMeasures;
  }

  /**
   * Renvoie les infos d'un point
   * @param geometry
   */
  private getInfosPoint(geometry: Point) {
    const projection = this.infosMeasures.find(infos => infos.geometryUid == geometry['ol_uid']).projection;
    const coordinates = new Point(transform(geometry.getCoordinates(), this.olMap.getView().getProjection(), projection)).getCoordinates();
    return {
      currentCoordinates: `[${coordinates[0].toFixed(formatDecimal(projection))}, ${coordinates[1].toFixed(formatDecimal(projection))}]`,
      currentProjection: projection
    };
  }

  /**
   * Renvoie les infos d'une ligne
   * @param geometry
   */
  private getInfosLine(geometry: LineString) {
    const pointsLineString = geometry.getCoordinates().map(coordinates => new Point(coordinates));
    return {
      currentLength: formatMeasures(geometry.getLength()),
      ...this.getInfosPointsLineString(pointsLineString)
    };
  }

  /**
   * Renvoie les infos d'un polygone
   * @param geometry
   */
  private getInfosPolygon(geometry: Polygon) {
    const pointsLineString = geometry.getCoordinates()[0].map(coordinates => new Point(coordinates));
    const infosPointsLineString = this.getInfosPointsLineString(pointsLineString);
    return {
      perimeter: infosPointsLineString.currentCoordinates[infosPointsLineString.currentCoordinates.length - 1].lengthTotal.label,
      area: formatMeasures(geometry.getArea()),
      ...infosPointsLineString
    };
  }

  /**
   * Renvoie les infos d'une cercle
   * @param geometry
   */
  private getInfosCircle(geometry: Circle) {
    const radius = geometry.getRadius();
    const diameter = radius * 2;
    const perimeter = diameter * Math.PI;
    const area = Math.PI * Math.pow(radius, 2);
    const center = new Point(transform(geometry.getCenter(), this.olMap.getView().getProjection(), this.currentProjection)).getCoordinates();
    return {
      columns: ['diameter', 'perimeter', 'area', 'center'],
      radius: formatMeasures(radius),
      diameter: formatMeasures(diameter, 4),
      perimeter: formatMeasures(perimeter, 4),
      area: `${formatMeasures(area, 4, 2)}²`,
      center: `[${center[0].toFixed(formatDecimal(this.currentProjection))}, ${center[1].toFixed(formatDecimal(this.currentProjection))}]`
    };
  }

  /**
   * Retourne les informations des différents points des segments d'une géométrie
   * @param pointsLineString Tableau de points des segments d'une géométrie
   * @returns
   */
  private getInfosPointsLineString(pointsLineString: Array<Point>) {
    const coordinatesLineString = pointsLineString.reduce((acc, point: Point, index) => {
      let lengthSegment = 0;
      let lengthTotal = 0;
      let deltaX = 0;
      let deltaY = 0;
      let angle = 0;
      if (index > 0) {
        const currentPreviousCoordinates = pointsLineString[index - 1].getCoordinates();
        const currentCoordinates = point.getCoordinates();
        lengthSegment = new LineString([currentPreviousCoordinates, currentCoordinates]).getLength();
        lengthTotal = acc[index - 1].lengthTotal.value + lengthSegment;
        const x1 = currentPreviousCoordinates[0];
        const x2 = currentCoordinates[0];
        const y1 = currentPreviousCoordinates[1];
        const y2 = currentCoordinates[1];
        deltaX = new LineString([[x1, 0], [x2, 0]]).getLength();
        if (x1 < x2) { deltaX = deltaX * -1; }
        deltaY = new LineString([[0, y1], [0, y2]]).getLength();
        if (y1 < y2) { deltaY = deltaY * -1; }
        angle = (Math.atan2((y2 - y1), (x2 - x1))) * (180 / Math.PI);
        angle = 90 - angle;
        angle < 0 ? (angle = 360 - angle * (-1)) : angle = angle;

      }
      const coordinates = new Point(transform(point.getCoordinates(), this.olMap.getView().getProjection(), this.currentProjection)).getCoordinates();
      return [
        ...acc,
        {
          point: index + 1,
          coordinates: `[${coordinates[0].toFixed(formatDecimal(this.currentProjection))}, ${coordinates[1].toFixed(formatDecimal(this.currentProjection))}]`,
          lengthSegment: formatMeasures(lengthSegment),
          lengthTotal: { value: lengthTotal, label: formatMeasures(lengthTotal) },
          deltaX: formatMeasures(deltaX),
          deltaY: formatMeasures(deltaY),
          angle: `${angle.toFixed(2)}°`
        }
      ];
    }, []);
    return {
      columns: ['point', 'coordinates', 'lengthSegment', 'lengthTotal', 'deltaX', 'deltaY', 'angle'],
      currentCoordinates: coordinatesLineString
    };
  }

  /**
   * Zoom la carte sur la géométrie sélectionné par l'action 'Voir sur la carte'
   */
  public showMeasureOnMap(geometryId) {
    const featureMeasure = this.olMap.get('measure_layer').getSource().getFeatures().find(feature => feature.getGeometry()['ol_uid'] == geometryId);
    this.mapUtils.centerMapOnGeom(this.olMap, featureMeasure.getGeometry(), this.vmapToolsService.getCenterPadding());
    this.hideInfosMeasures();
  }

  /**
   * Retourne une string format csv => colonnes + \n + values
   * @param geometry
   * @returns
   */
  private getExportCSVRowGeometry(geometry): string {
    const { uid, ...geometryInfos } = geometry;
    const keys = Object.keys(geometryInfos);
    const rowColumn = keys.join(';');
    const rowValues = keys.reduce((arrValues, key) => {
      return [
        ...arrValues,
        geometry[key]
      ];
    }, []);
    return rowColumn + '\n' + rowValues.join(';') + '\n\n';
  }

  /**
   * Création d'une string CSV avec les infos des mesures
   * @returns
   */
  private getInfosMeasureExportCSV(): string {
    return this.getInfosMeasures().reduce((acc, curr: any, index) => {
      // GEOMETRY INFORMATIONS TYPE POINT
      if (!Object.keys(curr).includes('columns')) {
        // GEOMETRY INFORMATIONS
        return acc + this.getExportCSVRowGeometry({ Id: index, ...curr }) + '\n';
      } else {
        const { currentCoordinates, columns, ...currInfos } = curr;
        // GEOMETRY INFORMATIONS
        let row = this.getExportCSVRowGeometry({ Id: index, ...currInfos });
        // GEOMETRY INFORMATIONS TYPE LINESTRING, POLYGON
        if (currentCoordinates) {
          row += columns.join(';');
          const infosLinePoint = currentCoordinates.reduce((accInfosLinePoint, currCoordinate) => {
            const values = [];
            columns.forEach(column => {
              values.push(currCoordinate[column].label || currCoordinate[column]);
            });
            return accInfosLinePoint + values.join(';') + '\n';
          }, '');
          return acc + row + '\n' + infosLinePoint + '\n\n';
        } else { // GEOMETRY INFORMATIONS TYPE CIRCLE
          return acc + row + '\n\n';
        }
      }
    }, '');
  }

  /**
   * Création et téléchargement des infos des mesures sous forme de CSV
   */
  public exportCSV() {
    const infosMeasureExportCSV = this.getInfosMeasureExportCSV();

    // Création d'un élément lien pour le téléchargement du csv
    const a = document.createElement('a');
    a.href = 'data:attachment/csv,' + encodeURIComponent(infosMeasureExportCSV);
    a.target = '_blank';
    a.download = 'Mesures.csv';
    document.body.appendChild(a);
    // Déclenche le téléchargement du csv
    a.click();
    document.body.removeChild(a);
  }

  /**
   * Met à jour le contenu et la position d'une popup
   * @param olMap
   * @param geometry
   * @param infosMeasure
   */
  private setGeometryPopup(olMap, geometry, infosMeasure) {
    let popupMessage;
    let popupPosition = geometry;
    switch (geometry.getType()) {
      case 'Point':
        const transformCoordinates = new Point(transform(geometry.getCoordinates(), olMap.getView().getProjection(), infosMeasure.projection)).getCoordinates();
        infosMeasure.geometry.coordinates = transformCoordinates;
        infosMeasure.sPopupData.X = transformCoordinates[0].toFixed(formatDecimal(infosMeasure.projection));
        infosMeasure.sPopupData.Y = transformCoordinates[1].toFixed(formatDecimal(infosMeasure.projection));
        infosMeasure.sPopupData.select.value = infosMeasure.projection;
        popupMessage = { X: infosMeasure.sPopupData.X, Y: infosMeasure.sPopupData.Y };
        break;
      case 'LineString':
        const length = geometry.getLength();
        infosMeasure.length = length;
        infosMeasure.geometry.coordinates = geometry.getCoordinates();
        popupMessage = formatMeasures(length);
        infosMeasure.sPopupData = popupMessage;
        popupPosition = new Point(geometry.getLastCoordinate());
        break;
      case 'Polygon':
        const area = geometry.getArea();
        infosMeasure.area = area;
        infosMeasure.geometry.coordinates = geometry.getCoordinates();
        popupMessage = `${formatMeasures(area, 2, 2)}²`;
        infosMeasure.sPopupData = popupMessage;
        break;
      case 'Circle':
        const radius = geometry.getRadius();
        infosMeasure.radius = radius;
        infosMeasure.geometry.center = geometry.getCenter();
        popupMessage = `Périmètre : ${formatMeasures(2 * Math.PI * radius)}`;
        infosMeasure.sPopupData = popupMessage;
        break;
    }

    infosMeasure.popup.setPopupMessage(popupMessage);
    infosMeasure.popup.setPopupPosition(popupPosition);
  }
}
