import { Component, Input, OnInit, ElementRef, ViewChild, OnDestroy } from '@angular/core';
import { Map } from 'ol';
import { MapToolDef } from '../../map-tools.model';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { MapCoreJSONService } from 'src/app/vitis/map-core/services/map-core-json/map-core-json.service';
import { PrintBoxService } from 'src/app/modules/vmap/carto/services/vmap-tools/vmap-tools-print/vmap-tools-print-box.service';
import { FormreaderComponent } from 'src/app/vitis/formreader/formreader.component';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { Subscription } from 'rxjs';
import { EventsKey } from 'ol/events';
import { unByKey } from 'ol/Observable';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import moment from 'moment-timezone';

@Component({
  selector: 'app-vmap-tools-print',
  templateUrl: './vmap-tools-print.component.html',
  styleUrls: ['./vmap-tools-print.component.scss']
})
export class VmapToolsPrintComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() oTool: MapToolDef;

  @ViewChild('map_print_template_parameters') oTemplateParamsFormReader: FormreaderComponent;

  public aTemplates: Array<any>;
  public aStyles: Array<any>;
  public aOutputFormats: Array<any>;
  public aScales: Array<number> = [
    1000000, 500000, 200000, 100000, 50000, 25000, 10000, 5000, 2500, 2000, 1000, 500, 200, 100
  ];
  public aResolutions: Array<any> = [{
    value: 1,
    label: '72 DPI'
  }, {
    value: 2,
    label: '144 DPI'
  }, {
    value: 4,
    label: '288 DPI'
  }
  ];

  public oCurrentTemplate: any;
  public mapScale: number;
  public aPendingPrintings: Array<any> = [];
  public oParamsForm: any = null;
  public iResizeFactor = 1; // facteur de resize de la zone d'impression (sert pour les petits écrans)

  // Valeurs des listes déroulantes
  public selectTemplateId: any;
  public selectStyleId: any;
  public selectScaleVal: any;
  public selectResolutionVal: any;
  public selectFormatVal: any;

  // Valeurs booléennes de vérification
  public bHasTemplates: boolean;
  public bHasStyles: boolean;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];
  private mapSubscriptions: EventsKey[] = [];

  constructor(
    private mapCoreJSONService: MapCoreJSONService,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private printBoxService: PrintBoxService,
    private mapUtils: MapCoreUtilsService,
    private vmapToolsService: VmapToolsService,
    private auth: AuthService,
  ) { }

  ngOnInit(): void {

    // Initialisation
    this.refreshScale_();

    // Rafraîchit l'échelle actuelle de la carte à son changement
    this.mapSubscriptions.push(this.olMap.on('moveend', () => {
      this.refreshScale_();
    }));

    // Si l'outil est déjà ouvert initialiser les champs
    if (this.oTool.collapsed_ === false) {
      this.initFields_().then(() => {
        if (this.bHasTemplates) {
          setTimeout(() => {
            this.refreshFields_();
            setTimeout(() => {
              this.printBoxService.initPrintBox(this.olMap, this.olMap.get('map_compare') ?
                this.getTemplateMapSize_('map_image_compare_1') : this.getTemplateMapSize_());
              this.printBoxService.showPrintBox(this.olMap);
            });
          });
        }
      });
    }

    // Ouverture de l'outil
    this.subscriptions.push(this.oTool.onOpen.subscribe(() => {
      this.initFields_().then(() => {
        if (this.bHasTemplates) {
          setTimeout(() => {
            this.refreshFields_();
            setTimeout(() => {
              this.printBoxService.initPrintBox(this.olMap, this.olMap.get('map_compare') ?
                this.getTemplateMapSize_('map_image_compare_1') : this.getTemplateMapSize_());
              this.printBoxService.showPrintBox(this.olMap);
            });
          });
        }
      });
    }));

    // Fermeture de l'outil
    this.subscriptions.push(this.oTool.onClose.subscribe(() => {
      this.printBoxService.hidePrintBox();
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
    this.mapSubscriptions.forEach(mapSubscription => unByKey(mapSubscription));
  }

  /**
   * Se déclenche quand l'utilisateur sélectionne un autre modèle ou un autre style
   */
  public onSelectionChange(value: any) {
    setTimeout(() => {
      this.refreshFields_();
      this.printBoxService.hidePrintBox();
      this.printBoxService.initPrintBox(this.olMap, this.olMap.get('map_compare') ?
        this.getTemplateMapSize_('map_image_compare_1') : this.getTemplateMapSize_());
      this.printBoxService.showPrintBox(this.olMap);
    });
  }

  /**
   * Changement d'échelle depuis la liste déroulante
   * @param iScale
   */
  public onScaleChange(iScale: number): void {
    if (this.vitisCheck.isNumber(Number(iScale))) {
      this.mapUtils.setScale(this.olMap, Number(iScale) * this.iResizeFactor);
    }
  }

  /**
   * Fonction déclenchée lors de l'appui sur le bouton "imprimer"
   */
  public submitPrintSync() {
    if (this.bHasTemplates) {
      this.printTemplate();
    }
  }

  /**
   * Retire une impression de la file d'impression
   * @param id
   */
  public removePrintHistory(id) {
    this.aPendingPrintings = this.aPendingPrintings.filter((element) => element.display_id !== id);
  }

  // public submitPrintAsync() {
  //   let userId = null;
  //   let sessionId = null;
  //   const currentTimestamp = new Date().getTime();
  //   const filename = currentTimestamp + '.' + this.oCurrentTemplate.outputformats_id;

  //   // Récupération de l'user_id de l'utilisateur actuel
  //   const userInfos = this.authService.getUserInfos();
  //   if (this.vitisCheck.isDefAndNotNull(userInfos)) {
  //     userId = userInfos.user_id;
  //   }

  //   // Récupération du session id dans le local storage
  //   sessionId = this.localStorage.get('vmap_printings_session_id');

  //   // Si le sessionId n'est pas défini, en générer un et le sauvegarder dans le local storage
  //   if (!this.vitisCheck.isDefAndNotNull(sessionId)) {
  //     console.log('Session id non défini, génération en cours...');

  //     sessionId = this.getUniqueId_(36);
  //     this.localStorage.set('vmap_printings_session_id', sessionId);
  //   }

  //   console.log('Session id trouvé : ' + sessionId);

  //   const oPrintHistory = {
  //     status: 0,
  //     user_id: userId,
  //     session_id: sessionId,
  //     filename: filename,
  //     href: '',
  //     print_start_date: new Date(),
  //     print_end_date: new Date()
  //   };

  //   // Ajout de l'impression dans l'historique d'impressions
  //   this.addPrintHistory_(oPrintHistory).then((value) => {
  //     this.aPendingPrintings.push({
  //       id: value.data.printhistory_id,
  //       display_id: this.getUniqueId_(8),
  //       printtemplate_id: this.oCurrentTemplate.printtemplate_id,
  //       title: oPrintHistory.filename,
  //       status: oPrintHistory.status,
  //       downloadLink: oPrintHistory.href,
  //     });

  //     // Tests à enelver, mais changer le statut du pending printing en 1 lorsque son impression est terminée
  //     const historyIndex = this.aPendingPrintings.findIndex((element) => element.title === oPrintHistory.filename);
  //     setTimeout(() => {
  //       this.aPendingPrintings[historyIndex].status = 1;
  //     }, 5000);
  //   });
  // }

  /**
   * Fonction asynchrone permettant de récupérer les datas de tous les champs select qui sont en base de données
   * @returns
   */
  private initFields_(): Promise<any> {
    return new Promise(async (resolve, reject) => {
      try {
        await this.loadTemplates_().then(aTemplates => {
          this.aTemplates = aTemplates;
          this.bHasTemplates = aTemplates.length > 0;
        });
        await this.loadStyles_().then(aStyles => {
          this.aStyles = aStyles;
          this.bHasStyles = aStyles.length > 0;
        });
        await this.loadOutputFormats_().then(aOutputFormats => {
          this.aOutputFormats = aOutputFormats;
        });

        // Valeurs par défaut des select
        this.selectTemplateId = null;
        if (this.aTemplates[0] &&
          this.aTemplates[0].printtemplate_id) {
          this.selectTemplateId = this.aTemplates[0].printtemplate_id;
        }
        this.selectScaleVal = 'current';
        this.selectResolutionVal = 1;
        this.selectFormatVal = 'pdf';

        resolve(true);
      } catch (error) {
        reject(error);
      }
    });
  }

  /**
   * Rafraîchit les valeurs des select selon le format ou le modèle changé
   */
  private refreshFields_() {

    if (this.selectTemplateId) {

      const iTemplateIndex = this.aTemplates.findIndex((element: any) => element.printtemplate_id === Number(this.selectTemplateId));

      // Si le template n'a pas changé, c'est que l'utilisateur a modifié le format
      // Sinon charger le format par défaut du template
      if (this.oCurrentTemplate && this.oCurrentTemplate.printtemplate_id === this.aTemplates[iTemplateIndex].printtemplate_id) {
        this.oCurrentTemplate.outputformats_id = this.selectFormatVal;
      } else {
        this.oCurrentTemplate = this.aTemplates[iTemplateIndex];
        this.selectFormatVal = this.oCurrentTemplate.outputformats_id;
      }

      // Paramètres du template
      try {
        this.oParamsForm = this.vitisCheck.isString(this.oCurrentTemplate.published_form_json_def) ?
          JSON.parse(this.oCurrentTemplate.published_form_json_def) : this.oCurrentTemplate.published_form_json_def;
        if (this.oParamsForm) {
          this.oTemplateParamsFormReader.initForm(this.oParamsForm, 'insert');
        }
      } catch (error) {
        console.error(error);
        this.oParamsForm = null;
      }
    }
  }

  /**
   * Récupère les données des modèles d'impression disponibles
   * @returns Promise
   */
  private loadTemplates_(): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/vmap/user_printtemplates`;

      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url,
        params: {
          order_by: 'printtemplate_order'
        }

      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting print templates');
          reject(false);
        }
      });
    });
  }

  /**
   * Récupère les données des styles d'impression disponibles
   * @returns Promise
   */
  private loadStyles_(): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/vmap/user_printstyles`;

      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting print templates');
          reject(false);
        }
      });
    });
  }

  /**
   * Récupère les données des formats de sortie disponibles
   * @returns Promise
   */
  private loadOutputFormats_(): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/vmap/print_outputformats`;

      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting print output formats');
          reject(false);
        }
      });
    });
  }

  /**
   * Refraichit le texte de l'échelle
   */
  private refreshScale_() {
    this.mapScale = Math.round(this.mapUtils.getScale(this.olMap, { pretty: false }));

    // Passe la valeur à "current" après un zoom manuel
    if (this.vitisCheck.isNumber(Number(this.selectScaleVal)) &&
      this.mapScale !== Number(this.selectScaleVal)) {
      this.selectScaleVal = 'current';
    }
  }

  /**
   * Fonction inachevée permettant de déclencher l'impression du modèle actuel
   */
  private printTemplate() {
    const mapAttributions = document.getElementById('map-attributions');
    const oUserInfos = this.auth.getUserInfos();

    let aScope: any = {
      map_scale: `1:${this.mapScale / this.iResizeFactor}`,
      date: moment.tz(oUserInfos.timezone_id).format(oUserInfos.formatdate_id),
      date_hour: moment.tz(oUserInfos.timezone_id).format(oUserInfos.formatdate_id + ' HH:mm:ss'),
      layer_sources: Array.from(mapAttributions.getElementsByTagName('li')).map(el => el.innerText)
    };
    const aIncludes = [];
    const mapDef = this.mapCoreJSONService.getMapDef(this.olMap);
    const measuresInfosMap = this.mapCoreJSONService.getMeasuresInfos(this.olMap);

    const aExtent = this.printBoxService.getExtent();
    const aOverviewExtent = [
      aExtent[0] - (aExtent[2] - aExtent[0]),
      aExtent[1] - (aExtent[3] - aExtent[1]),
      aExtent[2] + (aExtent[2] - aExtent[0]),
      aExtent[3] + (aExtent[3] - aExtent[1])
    ];

    // Légende de la carte
    const aLayersLegends = this.vmapToolsService.getLayersLegends(this.olMap);
    aScope.map_legend = aLayersLegends;

    const aMaps: any = [];

    const olMapCompare = this.olMap.get('map_compare');
    if (olMapCompare) { // Mode comparaison
      // Mode de comparaison activé dans le scope
      aScope.compare_mode = true;
      // Légende de la carte de comparaison
      const aLayersCompareLegends = this.vmapToolsService.getLayersLegends(olMapCompare);
      aScope.map_legend_compare = aLayersCompareLegends;
      const aMapImageCompare1Size = this.getTemplateMapSize_('map_image_compare_1');
      const aMapImageCompare2Size = this.getTemplateMapSize_('map_image_compare_2');
      aMaps.push({
        target: '#map_image_compare_1',
        map_definition: {
          map_json: mapDef,
          measure_features: measuresInfosMap,
          image_size: Math.round(aMapImageCompare1Size[0] * this.selectResolutionVal) + '|' + Math.round(aMapImageCompare1Size[1] * this.selectResolutionVal),
          resolution_coeff: this.iResizeFactor,
          extent: aExtent.join('|'),
          features: '',
          features_zoom: 100
        },
      }, {
        target: '#map_image_compare_2',
        map_definition: {
          map_json: this.mapCoreJSONService.getMapDef(olMapCompare),
          image_size: Math.round(aMapImageCompare2Size[0] * this.selectResolutionVal) + '|' + Math.round(aMapImageCompare2Size[1] * this.selectResolutionVal),
          resolution_coeff: this.iResizeFactor,
          extent: aExtent.join('|')
        }
      });
    } else { // Mode normal
      const aMapImageSize = this.getTemplateMapSize_('map_image');
      aMaps.push({
        target: '#map_image',
        map_definition: {
          map_json: mapDef,
          measure_features: measuresInfosMap,
          image_size: Math.round(aMapImageSize[0] * this.selectResolutionVal) + '|' + Math.round(aMapImageSize[1] * this.selectResolutionVal),
          resolution_coeff: this.iResizeFactor,
          extent: aExtent.join('|'),
          features: '',
          features_zoom: 100
        },
      });
    }

    const aMapOverviewImageSize = this.getTemplateMapSize_('map_overview');
    if (aMapOverviewImageSize && aMapOverviewImageSize[0] > 0 && aMapOverviewImageSize[1] > 0) {
      aMaps.push({
        target: '#map_overview',
        map_definition: {
          map_json: mapDef,
          image_size: Math.round(aMapOverviewImageSize[0] * this.selectResolutionVal) + '|' + Math.round(aMapOverviewImageSize[1] * this.selectResolutionVal),
          resolution_coeff: this.iResizeFactor,
          extent: aOverviewExtent.join('|'),
          features: '',
          features_zoom: 400
        }
      });
    }

    // Paramètres du formulaire
    if (this.oParamsForm) {
      const oParams = this.oTemplateParamsFormReader.getFormValues();
      aScope = { ...aScope, ...oParams };
    }

    const sScope = JSON.stringify(aScope);
    const sIncludes = JSON.stringify(aIncludes);
    const sMaps = JSON.stringify(aMaps);

    const data = {
      scope_json: sScope,
      includes_json: sIncludes,
      maps_json: sMaps,
      printstyle_id: this.selectStyleId
    };

    const printtemplate_id = this.oCurrentTemplate.printtemplate_id;
    const url = `${environment.properties.app_api_url}/vmap/printtemplateservice/${printtemplate_id}/print`;

    // Récupère le formulaire
    this.vitisRequest.ajaxRequestPromise({
      method: 'POST',
      url,
      data
    }).then((resp) => {
      if (resp.data &&
        resp.data.status === 1 &&
        this.vitisCheck.isString(resp.data.data)) {
        const sDownloadUrl = resp.data.data;

        this.vitisRequest.ajaxDownloadFile(sDownloadUrl, 'impression_' + aScope.date + '.pdf');
      }
    }).catch((error) => {
      console.error('Error occured while printing...');
      console.error(error);
    });
  }

  /**
   * Fonction permettant de récupérer la taille de la balise "map_image"
   * servant à définir la taille de la boîte d'impression
   * @returns
   */
  private getTemplateMapSize_(sTarget = 'map_image'): Array<number> {
    let aTemplateMapSize = [0, 0];

    const tempDiv = document.createElement('div');
    tempDiv.style.cssText = 'position: fixed; top: 9999px;';
    tempDiv.innerHTML = this.oCurrentTemplate.definition;
    document.body.appendChild(tempDiv);

    const mapElement = document.getElementById(sTarget);

    if (mapElement) {
      const style = window.getComputedStyle(mapElement, null);
      aTemplateMapSize = [
        parseFloat(style.width.replace('px', '')),
        parseFloat(style.height.replace('px', ''))
      ];
    }

    document.body.removeChild(tempDiv);

    // Taille de la carte dans le navigateur
    const aOlMapSize = this.olMap.getSize();

    // Calcul du facteur de resize de la zone d'impression
    this.iResizeFactor = this.calcResizeFactor_(aTemplateMapSize, aOlMapSize);
    aTemplateMapSize[0] = aTemplateMapSize[0] / this.iResizeFactor;
    aTemplateMapSize[1] = aTemplateMapSize[1] / this.iResizeFactor;

    return aTemplateMapSize;
  }

  /**
   * Fonction récursive qui calcule le facteur de resize de la zone d'impression
   * @param aTemplateMapSize
   * @param aOlMapSize
   * @param iResizeFactor
   * @returns
   */
  private calcResizeFactor_(aTemplateMapSize: Array<number>, aOlMapSize: Array<number>, iResizeFactor = 1): number {
    if (aOlMapSize[0] < aTemplateMapSize[0] / iResizeFactor ||
      aOlMapSize[1] < aTemplateMapSize[1] / iResizeFactor) {
      return this.calcResizeFactor_(aTemplateMapSize, aOlMapSize, iResizeFactor + 1);
    } else {
      return iResizeFactor;
    }
  }

  /**
   * Trouve un identifiant html unique
   * @returns string
   */
  private getUniqueId_(length): string {
    let id = '';
    const characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
    const charactersLength = characters.length;

    for (let i = 0; i < length; i++) {
      id += characters.charAt(Math.floor(Math.random() * charactersLength));
    }

    return id;
  }

  /**
   * Ajoute un élément d'historique en base de données
   * @param oData
   * @returns
   */
  private addPrintHistory_(oData: any): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/vmap/print_histories`;

      // Insère l'historique d'impression
      this.vitisRequest.ajaxRequestPromise({
        method: 'POST',
        data: oData,
        url
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          resolve(resp.data);
        } else {
          console.error('Error while inserting print history');
          reject(false);
        }
      });
    });
  }
}
