import { Component, Input, OnInit, OnChanges, ViewChild, OnDestroy } from '@angular/core';
import { Subscription } from 'rxjs';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { DataGridComponent } from 'src/app/vitis/data-grid/data-grid.component';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { WorkspaceListOptions } from 'src/app/vitis/workspacelist/workspacelist.model';
import { environment } from 'src/environments/environment';
import { VmapSelectService } from 'src/modules/vmap/components/carto/services/vmap-select/vmap-select.service';

declare var $: any;

@Component({
  selector: 'app-vmap-tools-requestor-modal',
  templateUrl: './vmap-tools-requestor-modal.component.html',
  styleUrls: ['./vmap-tools-requestor-modal.component.scss']
})
export class VmapToolsRequestorModalComponent implements OnInit, OnChanges, OnDestroy {

  @ViewChild(DataGridComponent, { static: true }) oDataGrid: DataGridComponent;

  @Input() sDatas;


  private aLayers: any;
  private layer: any;
  private clickedCell_: any;

  public currentLayer: string; // NAME DE LA LAYER SELECTIONNÉE

  // Largeur de page
  public pageLength: Number = environment.properties.rows_per_page;

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {},
    datagridOptions: WorkspaceListOptions
  };

  public dataTableInstance: any;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  constructor(private mapSelectService: VmapSelectService,
    private vitisRequest: VitisRequest,
    private vitisCheck: VitisCheck,
    private auth: AuthService) { }

  ngOnInit(): void {
    // Nombre d'éléments à afficher sur la grille
    this.pageLength = Number(environment.properties.rows_per_page);

    this.subscriptions.push(this.mapSelectService.getObservableSelectLayers().subscribe(aLayers => {
      this.aLayers = aLayers;
      this.oDataGrid.oDataGridOptions = this.oDataGridOptions;
    }));
  }

  ngOnChanges(changes) {
    if (changes && changes.sDatas && changes.sDatas.currentValue) {
      this.currentLayer = Object.keys(changes.sDatas.currentValue)[0];
      this.initDataGridColumns();
    }
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Charge les données des couches
   * @param layer_name
   */
  public loadLayerDatas(layer_name) {
    this.currentLayer = this.currentLayer === layer_name ? null : layer_name;
    this.initDataGridColumns();
  }

  /**
   * Initialise les colonnes
   */
  private initDataGridColumns(): void {
    if (this.aLayers && this.aLayers.length) {
      this.layer = this.aLayers.find(layer => layer.name === this.currentLayer);
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: `${environment.properties.app_api_url}/vmap/pglayers/${this.layer.layer_id}/columns`,
      }).then((resp: any) => {
        let columns = [];
        if (resp && resp.data && resp.data.data) {
          columns = resp.data.data.filter(e => e && e.name).map(column => {
            return {
              align: 'left',
              field: column.name,
              title: column.label,
              translateTitle: column.label,
              resizeable: true,
              searchable: true,
              sortable: true
            };
          });
        }
        this.oDataGridOptions.datagridOptions = {
          module: 'vmap',
          ressource: 'vmap/layers/:id/query',
          id_field: this.layer.layer_id,
          display: true,
          columns
        };
        this.initDataGrid();
      });
    }
  }

  /**
   * Initialise la datagrid
   */
  public initDataGrid(): void {

    /**
     * Réajuste les dimensions de la table
     */
    const autoAjustTables = () => {
      if ($('#' + this.oDataGrid.tableId).length) {
        this.dataTableInstance.columns.adjust();
        $('#' + this.oDataGrid.tableId + '_wrapper').find('.dataTables_scrollBody').css({ height: 'calc(100vh - 250px)' });
      }
    };

    const self = this;

    // Ordre
    const order = [2, 'desc'];

    const rowId = 'nom';

    // Paramètres de la grille de données.
    this.oDataGridOptions.datatableOptions = {
      select: {
        style: 'multi',
        selector: 'tr'
      },
      processing: true,
      serverSide: true,
      searching: false,
      lengthMenu: this.getLengthMenu_(),  // Tableau choix du nombre d'éléments par page
      pageLength: this.getPageLength_(),  // Nombre d'éléments par page
      order: order, // Ordre
      displayStart: 0,  // Offset
      rowId: rowId,
      selected: 'selected',
      ajax: (data: any, callback: any, settings: any) => {

        this.getResourceData_(data).then((oResponse: any) => {

          // Vide viewContainer
          this.oDataGrid.viewContainer.clear();

          // Remplace les données
          callback(oResponse);

        }).catch(() => {
          callback({
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          });
        });
      }
    };

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions) &&
      this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.columns)) {

      setTimeout(() => {
        // Colonnes de la liste (définies dans le mode).
        this.oDataGridOptions.datatableOptions.columns = this.getParsedColumns_();

        // Création de la liste.
        this.dataTableInstance = this.oDataGrid.initTable(this.oDataGridOptions);

        this.dataTableInstance.on('draw.dt', () => {
          // Réajuste les dimensions de la table
          autoAjustTables();
        });
        // Click event (permet de récupérer la cellule cliquée)
        this.dataTableInstance.on('click', 'td', function () {
          self.clickedCell_ = self.dataTableInstance.cell(this);
        });

      });
    }
  }

  /**
   * Retourne le tableau contenant le nombre de lignes affichables
   */
  private getLengthMenu_(): Array<Number> {

    const lengthMenu = [5, 10, 20, 50, 100];

    if (this.oDataGridOptions &&
      this.oDataGridOptions.datagridOptions &&
      this.oDataGridOptions.datagridOptions.max_rows) {

      const iMaxRows = this.oDataGridOptions.datagridOptions.max_rows;

      for (let i = lengthMenu.length - 1; i >= 0; i--) {
        if (lengthMenu[i] > iMaxRows) {
          lengthMenu.splice(i, 1);
        }
      }
    }

    return lengthMenu;
  }

  /**
   * Retourne le nombre d'éléments à afficher
   */
  private getPageLength_(): Number {
    let pageLength = this.pageLength;

    // Concordance avec le tableau de choix
    const aLengthMenu = this.getLengthMenu_();
    if (aLengthMenu.indexOf(pageLength) === -1) {
      if (pageLength > aLengthMenu[aLengthMenu.length - 1]) {
        pageLength = aLengthMenu[aLengthMenu.length - 1];
      } else {
        pageLength = 20;
      }
    }

    return pageLength;
  }

  /**
   * Récupère les colonnes définies dans le mode au format BootstrapTable
   * @returns Array
   */
  private getParsedColumns_(): Array<any> {

    const aColumns: any = [{
      targets: 0,
      data: null,
      defaultContent: '',
      orderable: false,
      className: 'select-checkbox',
      width: '40px'
    }];

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {

      // Colonne boutons
      if (this.oDataGridOptions.datagridOptions.display === true
        || this.oDataGridOptions.datagridOptions.update === true) {

        let defaultContent = '<div style="text-align: center;">';
        if (this.oDataGridOptions.datagridOptions.display === true &&
          (environment.properties.data_grid_update_btn_only === false || this.oDataGridOptions.datagridOptions.update !== true)) {
          defaultContent += '<span class="datatable-btn datatable-btn-display icon-vitis-search"></span>';
        }
        if (this.oDataGridOptions.datagridOptions.update === true) {
          defaultContent += '<span class="datatable-btn datatable-btn-edit icon-vitis-draw"></span>';
        }
        defaultContent += '</div>';

        aColumns.push({
          targets: -1,
          data: null,
          orderable: false,
          defaultContent,
          width: '90px'
        });
      }
      // Colonnes du mode
      if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.columns)) {
        const aUserPrivileges = this.auth.getPrivileges();
        for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
          const column = this.oDataGridOptions.datagridOptions.columns[i];
          // Privilège obligatoire pour afficher la colonne.
          if (this.vitisCheck.isDefAndNotNull(column.privilege)) {
            if (aUserPrivileges.indexOf(column.privilege) == -1) {
              continue;
            }
          }
          aColumns.push({
            title: column.translateTitle,
            data: column.field,
            name: column.field,
            visible: column.visible === false ? false : true,
            type: 'string',
            render: (data: any, type: any, row: any, meta: any) => {
              if (column.template) {
                data = '<field-render column_number="' + i + '">' + encodeURIComponent(JSON.stringify(row)) + '</field-render>';
              }
              return data;
            },
            orderable: column.sortable,
            // Alignement du contenu de la colonne.
            className: 'table-align-' + (column.align ? column.align : 'left')
          });
        }
      }
    }

    return aColumns;
  }

  /**
   * Génére l'url pour récupérer les données
   * @returns {string}
   */
  protected getRessourceUrl(): string {
    const sRessourceUrl = this.oDataGridOptions.datagridOptions.ressource;
    return environment.properties.app_api_url + '/' + sRessourceUrl.replace(':id', this.layer.layer_id);
  }

  /**
   * Récupère les donéées au format DataTables
   * @param  {any} params
   * @returns Promise
   */
  public getResourceData_(params: any): Promise<any> {
    return new Promise((resolve, reject) => {

      this.oDataGridOptions.loadingData_ = true;

      // Lance la requête
      this.vitisRequest.ajaxRequest({
        method: 'GET',
        url: this.getRessourceUrl(),
        ajaxLoader: false,
        // params: this.getRequestParams_(params),
        success: (resp: any) => {

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };

          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            if (this.vitisCheck.isDefAndNotNull(resp.data.status === 1)) {
              if (this.vitisCheck.isArray(resp.data.data)) {
                oReturn.data = resp.data.data;
              }
              if (this.vitisCheck.isDefAndNotNull(resp.data.total_row_number)) {
                oReturn.recordsTotal = resp.data.total_row_number;
                oReturn.recordsFiltered = resp.data.list_count;
              }
            }
          }

          resolve(oReturn);
        },
        error: (err: any) => {
          console.error('error while loading values :', err);

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };

          resolve(oReturn);
        }
      });
    });
  }

  /**
   * Ferme le requêteur
   */
  public closeRequestor(): void {
    $('#modal-requestor-layers').modal('hide');
  }

}
