import { Injectable } from '@angular/core';
import { Map } from 'ol';
import Feature from 'ol/Feature';
import Point from 'ol/geom/Point';
import { Subject } from 'rxjs';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { MapCorePopupService, Popup } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';
import { VmapLayer } from 'src/app/vitis/map-core/services/map-models.model';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { Geometry } from 'ol/geom';

export class MapSelection {
  data: any;
  clickCoords: Array<number>;
  oLayer: VmapLayer;
  popup?: Popup;
  popupParams?: any;
  geometry?: Geometry;
}

@Injectable({
  providedIn: 'root'
})
export class VmapSelectService {

  private selectedLayerObs = new Subject<VmapLayer | null>();
  private selectedLayersObs = new Subject<Array<VmapLayer> | null>();
  private selectionsObs = new Subject<Array<MapSelection> | null>();
  private oSelectedLayer: VmapLayer | null;
  private oLayerFirstBackground: VmapLayer | null;
  private oSelectedLayers: Array<VmapLayer> | null;
  private aSelections: Array<MapSelection> | null;
  private clickedCoordinate: Array<number>;
  private clickedCoordinateObs = new Subject<Array<number> | null>();
  private requestorSelectionObs = new Subject<any>();

  private selectedModeObs = new Subject<string | null>();
  private selectedMode: string | null;

  private selectionObs = new Subject<VmapLayer | Array<MapSelection>>();
  private selection: any | VmapLayer | Array<MapSelection>;

  constructor(
    private mapUtils: MapCoreUtilsService,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private mapPopup: MapCorePopupService
  ) {
    this.selectedLayerObs.subscribe((oSelectedLayer) => {
      this.oSelectedLayer = oSelectedLayer;
    });
    this.selectedLayersObs.subscribe((oSelectedLayers) => {
      this.oSelectedLayers = oSelectedLayers;
    });
    this.selectionsObs.subscribe((aSelections) => {
      this.aSelections = aSelections;
    });
    this.clickedCoordinateObs.subscribe((coordinate) => {
      this.clickedCoordinate = coordinate;
    });
    this.selectedModeObs.subscribe((mode) => {
      this.selectedMode = mode;
    });
    this.selectionObs.subscribe((selection) => {
      this.selection = selection;
    });
  }

  /**
   * Retourne l'observable de la layer de sélection
   * @returns
   */
  public getSelectLayerObs(): Subject<VmapLayer | null> {
    return this.selectedLayerObs;
  }

  /**
   * oSelectedLayer getter
   * @returns
   */
  public getSelectLayer(): VmapLayer | null {
    return this.oSelectedLayer;
  }

  /**
   * oLayerFirstBackground getter
   * Retourne la couche utiliser pour récupérer les donnée si la couche sélectionner dans la carte est 'Premier plan'
   * @returns
   */
  public getSelectLayerUnder(): VmapLayer | null {
    return this.oLayerFirstBackground;
  }

  /**
   * oSelectedLayer setter
   * @param oLayer
   */
  public setSelectLayer(oLayer: VmapLayer | null): void {
    this.selectedLayerObs.next(oLayer);
  }

  /**
   * Retourne l'observable de la sélection
   * @returns
   */
  public getSelectionsObs(): Subject<Array<MapSelection> | null> {
    return this.selectionsObs;
  }

  /**
   * aSelections getter
   * @returns
   */
  public getSelections(): Array<MapSelection> | null {
    return this.aSelections;
  }

  /**
   * aSelections setter
   * @param oLayer
   */
  public setSelections(aSelections: Array<MapSelection> | null): void {
    this.selectionsObs.next(aSelections);
  }

  /**
   * oSelectedLayers setter
   * @param oLayer
   */
  public setSelectLayers(oLayers: Array<VmapLayer> | null): void {
    this.selectedLayersObs.next(oLayers);
  }

  /**
   * Retourne l'observable des couches
   * @returns
   */
  public getObservableSelectLayers(): Subject<Array<VmapLayer> | null> {
    return this.selectedLayersObs;
  }

  /**
   * aSelections getter
   * @returns
   */
  public getSelectLayers(): Array<VmapLayer> | null {
    return this.oSelectedLayers;
  }

  /**
   * Retourne les coordonnées du dernier point cliqué sur la carte
   * @returns
   */
  public getClickedCoordinate(): Array<number> | null {
    return this.clickedCoordinate;
  }

  /**
   * SelectedMode setter
   * @param mode
   */
  public setSelectMode(mode): void {
    this.selectedModeObs.next(mode);
  }

  /**
   * SelectedMode getter
   * @returns
   */
  public getSelectMode(): string | null {
    return this.selectedMode;
  }

  /**
   * Get selectedMode observable
   * @returns
   */
  public getSelectModeObs(): Subject<string | null> {
    return this.selectedModeObs;
  }

  /**
   * SelectedMode setter
   * @param mode
   */
  public setSelection(selection): void {
    this.selectionObs.next(selection);
  }

  /**
   * SelectedMode getter
   * @returns
   */
  public getSelection(): VmapLayer | Array<MapSelection> {
    return this.selection;
  }

  /**
   * Get selectedMode observable
   * @returns
   */
  public getSelectionObs(): Subject<VmapLayer | Array<MapSelection>> {
    return this.selectionObs;
  }

  /**
   * Intersection d'une couche par coordonnées
   * @param aCoordinates
   * @param oLayer
   * @param olMap
   * @param oOptions options supplémentaires à envoyer à l'API (limit, attributs, get_geom etc.)
   * @returns
   */
  public selectByCoordinates(aCoordinates: Array<number>, oLayer: VmapLayer, olMap: Map, oOptions = {}): Promise<any> {
    // Sauvegarde les coordonnées du point quand on a cliqué sur la carte
    this.clickedCoordinateObs.next(aCoordinates);

    // Geométrie de type point
    const olPoint = new Point(aCoordinates);

    // Géométrie de type EWKT
    const sEWKTGeometry = this.mapUtils.getEWKTFromGeom(olPoint, olMap);

    // Retourne une pomesse
    return this.getLayerIntersections(olMap, oLayer, sEWKTGeometry, oOptions);
  }

  /**
   * Sélection d'un élément d'une couche par son identifiant
   * @param sId
   * @param oLayer
   * @param olMap
   * @param oOptions
   * @returns
   */
  public selectById(sId: any, oLayer: VmapLayer, olMap: Map, oOptions = {}): Promise<any> {
    return new Promise((resolve, reject) => {

      if (!this.vitisCheck.isDefAndNotNull(oLayer.idfield)) {
        console.error('This layer has no idfiled', oLayer);
      }

      // Filtre
      const filter = {
        relation: 'AND',
        operators: [{
          column: oLayer.idfield,
          compare_operator: '=',
          value: sId,
        }]
      };

      // Projection en cours
      const result_srid = olMap.getView().getProjection().getCode().substring(5);

      // Paramètres
      const oParams = {
        d: Date.now(),
        get_geom: true,
        filter,
        result_srid,
      };

      this.queryLayer_(oLayer, { ...oParams, ...oOptions }).then((resp) => {
        if (resp.status === 1 && this.vitisCheck.isDefAndNotNull(resp.data)) {
          if (resp.oLayer) {
            resolve(resp);
          }
          resolve(resp.data);
        } else {
          reject(resp);
        }
      }).catch((e) => {
        reject(e);
      });
    });
  }

  /**
   * Fait une intersection sur la couche en paramètre
   * @param olMap
   * @param oLayer
   * @param sEWKTGeometry
   * @param oOptions options supplémentaires à envoyer à l'API (limit, attributs, get_geom etc.)
   */
  public getLayerIntersections(olMap: Map, oLayer: VmapLayer, sEWKTGeometry: string, oOptions = {}): Promise<any> {
    return new Promise((resolve, reject) => {

      // Buffer d'intersection
      const intersect_buffer = this.getRealScaleBuffer_(oLayer, sEWKTGeometry, olMap);

      // Geométrie à intersecter
      const intersect_geom = sEWKTGeometry;

      // Projection en cours
      const result_srid = olMap.getView().getProjection().getCode().substring(5);

      // Paramètres
      const oParams = {
        d: Date.now(),
        get_geom: true,
        intersect_buffer,
        intersect_geom,
        result_srid
      };

      this.queryLayer_(oLayer, { ...oParams, ...oOptions }).then((resp) => {
        if (resp.status === 1 && this.vitisCheck.isDefAndNotNull(resp.data)) {
          if (resp.oLayer) {
            resolve(resp);
          }
          resolve(resp.data);
        } else {
          reject(resp);
        }
      }).catch((e) => {
        reject(e);
      });
    });
  }

  /**
   * Ferme les sélections en cours
   */
  public closeSelections() {
    if (this.selection && this.vitisCheck.isArray(this.selection)) {
      for (let i = 0; i < this.selection.length; i++) {
        const oSelection = this.selection[i];
        if (oSelection.popup) {
          oSelection.popup.close();
        }
      }
      this.selection.length = 0;
      this.setSelections(null);
    }
  }

  /**
   * Affiche les popups de la sélection
   * @param olMap
   * @param oLayer
   * @param aData
   * @param aCoordinates
   */
  public displayPopups(olMap: Map, aSelections: Array<MapSelection>) {

    for (let i = 0; i < aSelections.length; i++) {
      const oSelection: MapSelection = aSelections[i];
      if (i === 0 && oSelection.oLayer) {
        this.oLayerFirstBackground = oSelection.oLayer;
      }

      let olPopupGeom: Geometry;

      // Point où l'utilisateur a cliqué
      if (oSelection.clickCoords) {
        olPopupGeom = new Point(oSelection.clickCoords);
      }

      // Geométrie de l'objet
      if (oSelection.geometry) {
        olPopupGeom = oSelection.geometry;
      }

      // Geométrie de l'objet
      if (oSelection.oLayer.geomfield && oSelection.data[oSelection.oLayer.geomfield]) {
        olPopupGeom = this.mapUtils.getGeomFromEWKT(oSelection.data[oSelection.oLayer.geomfield], olMap);
      }

      // Champ à afficher dans la popup
      let sPopupData = null;
      if (oSelection.oLayer.select_field && oSelection.data[oSelection.oLayer.select_field]) {
        sPopupData = oSelection.data[oSelection.oLayer.select_field];
      } else if (oSelection.oLayer.idfield && oSelection.data[oSelection.oLayer.idfield]) {
        sPopupData = oSelection.data[oSelection.oLayer.idfield];
      }

      if (!oSelection.oLayer.published_form_json_def || aSelections.length === 1) {
        let oOptions: any = { hideCloseBtn: true, clickCoords: oSelection.clickCoords };
        // Si oSelection.clickCoords contient plusieurs coordonnées on ne l'ajoute pas aux options
        if (oSelection.clickCoords &&
          this.vitisCheck.isArray(oSelection.clickCoords[0])) {
          oOptions = { hideCloseBtn: true };
        }
        const popup = this.mapPopup.displayPopup(sPopupData, new Feature({ geometry: olPopupGeom }), olMap, oOptions);
        oSelection.popup = popup;
      } else {
        oSelection.popupParams = { sPopupData, olPopupGeom };
      }
    }
  }

  /**
   * SelectedMode setter
   * @param mode
   */
  public setRequestorSelection(selection): void {
    this.requestorSelectionObs.next(selection);
  }

  /**
   * Get selectedMode observable
   * @returns
   */
  public getRequestorSelectionObs(): Subject<any> {
    return this.requestorSelectionObs;
  }

  /**
   * Retourne le buffer à l'échelle actuelle
   * @param  {VmapLayer} oLayer
   * @param  {string} sEWKTGeometry
   * @param  {Map} olMap
   * @returns number
   */
  public getRealScaleBuffer_(oLayer: VmapLayer, sEWKTGeometry: string, olMap: Map): number {

    // Buffer définit par l'objet métier
    const buffer = oLayer.map_selection_buffer ? oLayer.map_selection_buffer : 5;

    // Type de géométrie
    const geomType = this.mapUtils.getGeomFromEWKT(sEWKTGeometry, olMap).getType();

    if (!this.vitisCheck.isDefAndNotNull(buffer)) {
      return null;
    }
    if (!this.vitisCheck.isDefAndNotNull(geomType)) {
      return null;
    }

    let iAPIBuffer = null;
    if (geomType === 'Point' || geomType === 'MultiPoint') {
      if (this.vitisCheck.isDefAndNotNull(buffer)) {
        const scale = this.mapUtils.getScale(olMap, { pretty: false });
        iAPIBuffer = (buffer * scale) / 1000;
      }
    }
    return iAPIBuffer;
  }

  /**
   * requête une couche
   * @param oLayer
   * @param oParams
   * @returns
   */
  private queryLayer_(oLayer: VmapLayer, oParams: any): Promise<any> {
    return new Promise((resolve, reject) => {
      let url = '';
      let params;
      // Récupère le formulaire

      if (oLayer.layer_id) { // Requête une couche
        url = `${environment.properties.app_api_url}/vmap/layers/${oLayer.layer_id}/query`;
        params = { ...oParams, limit: 2000 };
      } else { // Requête la couche premier plan => toute les couches
        const layersRequetable = oParams.idList || this.getSelectLayers().filter(layer => layer.layer_id).sort(function (a, b) {
          return b.index - a.index;
        }).map(layer => layer.layer_id);
        url = `${environment.properties.app_api_url}/vmap/layers/query/layer`;
        params = { ...oParams, idList: layersRequetable, limit: 2000 };
      }

      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url,
        params,
        httpMethodOverride: true,
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          if (resp.oLayer) {
            resolve(resp);
          }
          resolve(resp.data);
        } else {
          reject(resp);
        }
      }).catch((resp) => {
        console.error('Error getting intersections', resp);
        reject(resp);
      });

    });
  }
}
