<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\Normalizer\VitisNormalizer;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;

/**
 * Controller pour le datasource treeview de vMap.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class DatasourceTreeviewController extends AbstractVmapQueriesController {
  /**
   * @var array
   */
  protected $cannotAccessResponse = [
    'status' => 0,
    'message' => 'Cannot access to database',
    'errorMessage' => 'ERROR_POSTGRES_USER_BAD_CONNECTION',
  ];

  /**
   * Permet de récupérer les base de données locales
   * @return Response
   */
  public function getDatabases(): Response {
    $aData = $this->requestToArray();

    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $sDatabaseName = $aData['database'] ?? 'postgres';
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
      $aData = array_merge($aData, $aConnectionData);
    }

    if ($this->isConnectionDatasPresent($aData)) {
      // set une valeur par défaut pour databse s'il n'est pas défini
      $aTmpData = array_merge(['database' => 'postgres'], $aData);

      $oEm = $this->getEntityManagerFromData($aTmpData);
      $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    } else {
      // sinon on récupère un queryAccessor local pour l'utilisateur
      $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    }

    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatResponse($this->cannotAccessResponse, Response::HTTP_UNAUTHORIZED);
    }

    // Récupération des bases de données
    try {
      $oDatabases = $oQueryAccessor->getDatabases($aData);
    } catch (\Exception $e) {
      return $this->formatGenericError(
        'ERROR_POSTGRES_GET_DATABASES',
        '',
        'La connexion au serveur de base de données est impossible pour la source de données: ' . $aData['datasource_id'],
        Response::HTTP_UNAUTHORIZED
      );
    }
    if (!isset($oDatabases)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oDatabases as $oDatabase) {
      $response['data'][] = ['database' => $oDatabase['database']];
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les schémas d'une base de données
   * @param mixed $databaseName Non utilisé, mis pour correspondre au getSchemas parent
   * @return Response réponse à retourner par l'API
   */
  public function getSchemas($databaseName = null) {
    $aData = $this->requestToArray();

    $aConnectionData = null;
    $sDatabaseName = $aData['database'] ?? null;
    // si les field pour la connexion ne sont pas présent et qu'on le datasource_id
    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
      $aData = array_merge($aData, $aConnectionData);
    }

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);

    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatResponse($this->cannotAccessResponse, Response::HTTP_UNAUTHORIZED);
    }

    // Récupération des schemas
    $oSchemas = $oQueryAccessor->getSchemas($aData);
    if (!isset($oSchemas)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oSchemas as $oSchema) {
      $response['data'][] = ['schema_name' => $oSchema['schema_name']];
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les tables d'un schema d'une base de données
   * @return Response réponse à retourner par l'API
   */
  public function getTables() {
    $aData = $this->requestToArray();
    // si les field pour la connexion ne sont pas présent et qu'on le datasource_id
    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $sDatabaseName = $aData['database'] ?? null;
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
      $aData = array_merge($aData, $aConnectionData);
    }

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Récupération des tables
    $oTables = $oQueryAccessor->getOnlyTables($aData);
    if (!isset($oTables)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oTables as $oTable) {
      $response['data'][] = ['name' => $oTable['name']];
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les vues d'un schema d'une base de données
   * @return Response réponse à retourner par l'API
   */
  public function getViews() {
    $aData = $this->requestToArray();
    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $sDatabaseName = $aData['database'] ?? null;
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
      $aData = array_merge($aData, $aConnectionData);
    }

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);

    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Récupération des vues
    $oViews = $oQueryAccessor->getOnlyViews($aData);
    if (!isset($oViews)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oViews as $oView) {
      $response['data'][] = ['name' => $oView['name']];
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les vues matérialisé d'un schema d'une base de données
   * @return Response réponse à retourner par l'API
   */
  public function getMaterializedViews() {
    $aData = $this->requestToArray();
    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $sDatabaseName = $aData['database'] ?? null;
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
      $aData = array_merge($aData, $aConnectionData);
    }

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);

    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Récupération des tables
    $oMatViews = $oQueryAccessor->getOnlyMaterializedViews($aData);
    if (!isset($oMatViews)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oMatViews as $oMatView) {
      $response['data'][] = ['name' => $oMatView['name']];
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les tables, les vues et les vues matérialisé d'un schema d'une base de données
   * @return Response réponse à retourner par l'API
   */
  public function getAllTables() {
    $aData = $this->requestToArray();
    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $sDatabaseName = $aData['database'] ?? null;
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
      $aData = array_merge($aData, $aConnectionData);
    }

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);

    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Tables
    $aTables = $oQueryAccessor->getOnlyTables($aData);
    // Vues
    $aViews = $oQueryAccessor->getOnlyViews($aData);
    // Vues matérialisées
    $aMatViews = $oQueryAccessor->getOnlyMaterializedViews($aData);
    if (!isset($aTables) && !isset($aViews) && !isset($aMatViews)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    $aAllTables = array_merge(
      isset($aTables) ? $aTables : [],
      isset($aViews) ? $aViews : [],
      isset($aMatViews) ? $aMatViews : []
    );
    foreach ($aAllTables as $oTable) {
      $response['data'][] = ['name' => $oTable['name']];
    }

    return $this->formatResponse($response);
  }

  /**
   * Récupère les informations de la table en paramètre pour les affichées
   * @return Response
   */
  public function getTableInfo(): Response {
    $aData = $this->requestToArray();
    $aReturn = ['status' => 1, 'data' => []];

    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $sDatabaseName = $aData['database'] ?? null;
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
      $aData = array_merge($aData, $aConnectionData);
    }

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if (!isset($aData['type_table']) || (isset($aData['type_table']) && $aData['type_table'] !== 'view')) {
      $qb = $oQueryAccessor->createQueryBuilder();

      // Récupère le nombre d'enregistrement dans la table
      try {
        $qb->select('count(*)')->from('[SCHEMA].[TABLE]', 'a');

        $aParams = [
          'SCHEMA' => ['value' => $aData['schema'], 'type' => 'schema_name'],
          'TABLE' => ['value' => $aData['pg_table'], 'type' => 'table_name'],
        ];
        $oResult = $qb->executeWithParams($aParams)->fetchAll();
        $aReturn['data']['count_data'] = $oResult[0]['count'];
      } catch (\Exception $e) {
        $this->oLoggerManager->get('SQL_ERROR')->error('getTableInfo : can\'t get count: ' . $e->getMessage());

        $aReturn['status'] = 0;
        $aReturn['message'] = 'Error during get count';
        if (strpos($e->getMessage(), 'Insufficient privilege') === false) {
          $aReturn['errorMessage'] = 'ERROR_GENERIC_CONTROLLER_GET';
        } else {
          $aReturn['errorMessage'] = 'ERROR_GET_TABLE_INFO_RIGHTS';
        }

        return $this->formatResponse($aReturn, Response::HTTP_UNAUTHORIZED);
      }
    }

    // système de coordonnées
    $aGeomInfo = $oQueryAccessor->getFirstGeomColumInfo($aData);
    if (isset($aGeomInfo) && isset($aGeomInfo['status']) && $aGeomInfo['status'] == 0) {
      return $this->formatResponse($aGeomInfo);
    }

    if (isset($aGeomInfo)) {
      $aReturn['data']['srid'] = $aGeomInfo['srid'];
      $aReturn['data']['geom_type'] = $aGeomInfo['type'];
      $aReturn['data']['geom_column_finded'] = $aGeomInfo['f_geometry_column'];
    }

    $aReturn['data'] = json_encode($aReturn['data']);
    return $this->formatResponse($aReturn);
  }

  /**
   * Récupère la liste des couches appartenant à la même source de donnée, la source de donnée est récupéré avec l'id d'une couche
   * @param mixed $id identifiant d'une couche
   * @return Response réponse à retourner par l'API
   */
  public function getSiblingsLayers($id): Response {
    // Get datasource_id
    $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();

    try {
      $qb->select('a.datasource_id')
        ->from('[SCHEMA].layer', 'a')
        ->where($qb->expr()->eq('a.layer_id', '[LAYER_ID]'));

      $aParams = [
        'SCHEMA' => ['value' => $this->oProperties->get('schema_vmap'), 'type' => 'schema_name'],
        'LAYER_ID' => ['value' => $id, 'type' => 'number'],
      ];

      $aContent = $qb->executeWithParams($aParams)->fetchAll();
      $oContent = $aContent[0];
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('getSiblingsLayers : can\'t get datasource_id: ' . $e->getMessage());
      $aReturn = [
        'status' => 0,
        'message' => 'Error during get datasource_id',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
      ];
      return $this->formatResponse($aReturn, Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    $aData = $this->requestToArray();

    // Add filter
    $aData['filter'] = json_encode([
      'relation' => 'AND',
      'operators' => [[
        'column' => 'datasource_id',
        'compare_operator' => 'IN',
        'value' => $oContent['datasource_id'],
      ],
        [
          'column' => 'layer_id',
          'compare_operator' => '<>',
          'value' => $id,
        ]],
    ]);

    // Get Layers
    $oDataAccessor = $this->getDataAccessor();
    $aEntities = $oDataAccessor->getEntities($aData);
    if ($aEntities['status'] == 0) {
      $aEntities['message'] = 'getSiblingsLayers: Error during get';
      $aEntities['errorMessage'] = 'ERROR_GENERIC_CONTROLLER_GET';
      return $this->formatResponse($aEntities, Response::HTTP_UNAUTHORIZED);
    }
    $aSerializerContext = [];
    if (isset($aData['attributs'])) {
      $aSerializerContext[AbstractNormalizer::ATTRIBUTES] = explode('|', $aData['attributs']);
    }
    if (isset($aData['distinct'])) {
      $aSerializerContext[VitisNormalizer::DISTINCT] = ($aData['distinct'] === 'true');
    }

    // transforme le tableau de réponse dans le résultat souhaité
    $this->formatResponse($aEntities, Response::HTTP_OK, $aSerializerContext);

    return $this->oResponse;

  }

  /**
   * Récupère les valeurs disponibles dans la table et la liste des colonnes
   * @return Response réponse à retourner par l'API
   */
  public function getTableValues(): Response {
    $aData = $this->requestToArray();

    if (!isset($aData['schema']) || !isset($aData['pg_table']) || !isset($aData['datasource_id'])) {
      return $this->formatResponse([
        'status' => 0,
        'message' => 'schema or pg_table or datasource_id missing',
        'errorMessage' => 'MISSING_PARAMS',
      ]);
    }

    // Récupère les informations de connexion
    $aConnectionData = null;
    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $sDatabaseName = $aData['database'] ?? null;
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
    }

    // Retourne l'erreur si il y en a eu une
    if ($aConnectionData === null) {
      return $this->formatResponse([
        'status' => 0,
        'message' => 'Cannot find Datasource data',
        'errorMessage' => 'MISSING_PARAMS',
      ], Response::HTTP_UNAUTHORIZED);
    }
    $aData = array_merge($aData, $aConnectionData);
    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);

    // Retourne l'erreur si il y en a eu une
    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'message' => 'Cannot create a QueryAccessor',
        'errorMessage' => 'MISSING_PARAMS',
      ], Response::HTTP_UNAUTHORIZED);
    }

    // Ajout des paramètre nécéssaire pour getGeneric
    $aData['sgbd'] = DatabaseConnection::SGBD;
    $aData['table'] = $aData['pg_table'];

    // Récupère les informations sur la colonne géométrie
    // Cas get_geom_column
    if (isset($aData['get_geom_column']) && $aData['get_geom_column'] == true) {
      $aGeomInfo = $oQueryAccessor->getFirstGeomColumInfo($aData);
      if (isset($aGeomInfo) && isset($aGeomInfo['status']) && $aGeomInfo['status'] == 0) {
        return $this->formatResponse($aGeomInfo);
      }

      if (isset($aGeomInfo)) {
        $aTableinfo = [];
        $aTableinfo['srid'] = $aGeomInfo['srid'];
        $aTableinfo['geom_type'] = $aGeomInfo['type'];
        $aTableinfo['geom_column_finded'] = $aGeomInfo['f_geometry_column'];

        // ajout des paramètres à aData pour ne récupère que les données géométriques et au bon format
        $aData['columns_type'] = [
          $aGeomInfo['f_geometry_column'] => 'geom',
        ];
        $aData['attributs'] = $aGeomInfo['f_geometry_column'];
      }
    }

    // Lance le generic get pour récupérer les données
    $oResponse = $this->getGeneric($aData);
    $aContent = json_decode($oResponse->getContent(), true);

    // Si erreur lors du generic get on retourne l'erreur
    if (!isset($aContent) || !isset($aContent['status']) || $aContent['status'] == 0) {
      return $oResponse;
    }

    // Récupère la liste des colonnes
    if (isset($aData['get_columns']) && $aData['get_columns'] == true) {
      $aContent = $this->getColumns($aContent, $aData);

      // Retour en cas d'erreur
      if (!isset($aContent)) {
        return $this->formatResponse([
          'status' => 0,
          'message' => 'Cannot get columns',
          'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
        ], Response::HTTP_UNAUTHORIZED);
      }

      // Si des colonnes ont été trouvé on les ajoutes à la réponse
      if (!empty($aContent['columns'])) {
        $oResponse->setContent(json_encode($aContent));
      }
    }

    if (isset($aTableinfo) && !empty($aTableinfo)) {
      $aContent['tableinfo'] = $aTableinfo;
      $oResponse->setContent(json_encode($aContent));
    }

    return $oResponse;
  }

  /**
   * Récupère la liste des colonnes d'une table au format datatable
   * @param array $aContent
   * @param array $aData
   * @return array|null
   */
  public function getColumns(array $aContent, array $aData): ?array{
    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);

    if ($oQueryAccessor === null) {
      return null;
    }

    $aColumns = $oQueryAccessor->getColumns($aData);
    if (!isset($aColumns)) {
      return null;
    }

    // Formate les colonnes pour le datatable
    $aContent['columns'] = [];
    foreach ($aColumns as $oColumn) {
      $aColumnData = [
        'title' => $oColumn['column_name'],
        'data' => $oColumn['column_name'],
        'type' => 'text',
      ];
      if (strpos($oColumn['data_type'], 'bigint') !== false
        || strpos($oColumn['data_type'], 'integer') !== false
        || strpos($oColumn['data_type'], 'smallint') !== false
      ) {
        $aColumnData['type'] = 'number';
      } else if (strpos($oColumn['data_type'], 'time zone') !== false || strpos($oColumn['data_type'], 'date') !== false) {
        $aColumnData['type'] = 'date';
      }
      $aContent['columns'][] = $aColumnData;
    }

    return $aContent;
  }
}
