<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use App\Module\Vmap\WebService\Entity\AbstractPublication;
use App\Module\Vmap\WebService\Service\MapServerService;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller flux
 *
 * @author Laurie Ambroise <laurie.ambroise@veremes.com>
 */
class FluxController extends GenericController {
  /**
   * @var MapServerService
   */
  private $oMapServerService;

  /**
   * @var Properties
   */
  private $oPropertiesSrvc;

  /**
   * @var LocalAccessor
   */
  private $oLocalFileAccessor;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param MapServerService $oMapServerService
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, MapServerService $oMapServerService) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oMapServerService = $oMapServerService;
    $this->oPropertiesSrvc = $oPropertiesSrvc;
    $this->oLocalFileAccessor = $oFileManager->getLocalFileAccessor();
  }

  /**
   * Retourne la définition de la publication
   * @param mixed $id
   * @return array|null
   */
  private function getPublicationDef($id): ?array{
    $aPublicationDef = [];
    $oPublicationRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(AbstractPublication::class);

    if (!$oPublicationRepository instanceof GenericRepository) {
      return null;
    }

    $oPublication = $oPublicationRepository->findById($id)[0];

    // Getters de l'entité AbstractPublication
    $oPublicationRoute = $this->oRouteConfig->getRequestParameters('vmap_publication');
    $aPublicationDef = $oPublication->jsonSerialize($oPublicationRoute);

    return $aPublicationDef;
  }

  /**
   * Génère le mapfile d'un flux
   */
  public function generateMapFile($id): Response {
    $aData = $this->requestToArray();
    $oDataAccessor = $this->getDataAccessor();

    $aReturn = [
      'status' => 0,
    ];

    $msPublications = [];

    if (count($aData['publications']) > 0) {
      foreach ($aData['publications'] as $oPublication) {
        $aPublicationDef = $this->getPublicationDef($oPublication['publication_id']);
        $aPublicationDef['layertype'] = $aPublicationDef['publicationtype'];
        $msPublications[] = $aPublicationDef;
      }
    }

    // Générer le fichier fonts.list avec les données en bdd
    $this->oMapServerService->generateFontList();

    // Génération du mapfile
    $sMapfileName = $this->oMapServerService->createMapfile([
      'flux_id' => $id,
      'name' => $aData['name'],
      'debuglevel' => $this->oPropertiesSrvc->get('mapfile_debug'),
    ], $msPublications, null, true);

    // Construction de l'URL du flux
    $sOnlineResourceUrl = str_replace('@file', $sMapfileName, MapServerService::PUBLIC_WMS_MAIN_URL);
    $sOnlineResourceUrl = $this->oPropertiesSrvc->replacePropertiesInString($sOnlineResourceUrl);
    $sFluxUrl = $sOnlineResourceUrl;

    $aReturn = [
      'data' => [
        'flux_url' => $sFluxUrl,
        'publications' => $msPublications,
      ],
      'status' => 1,
    ];

    return $this->formatResponse($aReturn);
  }

  /**
   * Supprime un flux
   * @param mixed $id id de l'entité à supprimer
   * @return Response réponse à retourner par l'API
   */
  public function deleteFlux($id): Response {

    // Supprime le mapfile
    $this->deleteMapFile($id);

    // Supprime de la BDD
    return $this->deleteData($id);
  }

  /**
   * Supprime les entités d'une liste pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function deleteMultipleFlux(): Response {
    $aData = $this->massiveRequestGetIdList();
    if (empty($aData)) {
      return $this->formatGenericResponseForMissingIdList();
    }

    // Supprime le mapfile
    if (!empty($aData['idList'])) {
      foreach ($aData['idList'] as $key => $value) {
        $this->deleteMapFile($value);
      }
    }

    // Supprime de la BDD
    return $this->deleteDatas();
  }

  private function deleteMapFile($id) {
    $aData = $this->requestToArray();

    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    $oEntity = $oDataAccessor->getEntity($id, $aData);

    if ($oEntity['status'] === 0) {
      return false;
    } else if (!empty($oEntity['data']) && !empty($oEntity['data']['flux_id']) && !empty($oEntity['data']['name'])) {
      $aFlux = $oEntity['data'];

      // Suppression du mapfile
      $sMapfileName = $this->oMapServerService->deleteMapfile($aFlux, true);

    } else {
      return false;
    }
  }
}
