<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vmap\WebService\Entity\AbstractLayer;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;

/**
 * Controller pour le datasource treeview de vMap.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class LayerController extends GenericController {
  /**
   * Créé plusieurs entités pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function postDatas(): Response {
    $aData = $this->requestToArray();
    if (empty($aData)) {
      $aReturn = ['status' => 0];
      $aReturn['message'] = 'data parameter is missing';
      $aReturn['errorMessage'] = 'ERROR_MISSING_DATA_PARAMETER';
      return $this->formatResponse($aReturn, Response::HTTP_NOT_ACCEPTABLE);
    }

    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    // Sauvegarde en base de donnée
    $oEntity = $oDataAccessor->insertEntities($aData['data']); // Uniquement en ORM
    if ($oEntity['status'] == 0) {
      $oEntity['message'] = 'Error during update';
      $oEntity['errorMessage'] = 'ERROR_GENERIC_CONTROLLER_INSERT';
      return $this->formatResponse($oEntity, Response::HTTP_UNAUTHORIZED);
    }
    $this->formatResponse($oEntity);

    return $this->oResponse;
  }

  /**
   * Récupère une entité AbstractLayer
   * @return Response réponse à retourner par l'API
   */
  public function getData($id): Response {
    $aData = $this->requestToArray();

    // Attribution de l'entité oLayer grâce à son layertype
    $oLayer = $this->returnLayerClass($id)['oLayer'];

    if (empty($oLayer)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find layer with id : ' . $id);
      $oResponse = [
        'status' => 0,
        'message' => 'Error during get',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
      ];
      return $this->formatResponse($oResponse, Response::HTTP_UNAUTHORIZED);
    }

    $aReturn = [
      'data' => $oLayer,
      'status' => 1,
    ];

    $aSerializerContext = [];
    if (isset($aData['attributs'])) {
      $aSerializerContext[AbstractNormalizer::ATTRIBUTES] = explode('|', $aData['attributs']);
    }

    return $this->formatResponse($aReturn, Response::HTTP_OK, $aSerializerContext);
  }

  /**
   * Met à jour une entité AbstractLayer
   * @return Response réponse à retourner par l'API
   */
  public function updateData($id): Response {
    $aData = $this->requestToArray();
    $oLayer = $this->returnLayerClass($id);
    $oEntity = $oLayer['oLayer'];

    $oEntity->setDatemaj(new \DateTime());

    if (empty($oEntity)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find layer with id : ' . $id);
      $oResponse = [
        'status' => 0,
        'message' => 'Error during get',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
      ];
      return $this->formatResponse($oResponse, Response::HTTP_UNAUTHORIZED);
    }

    $parameters = $this->getRequestParameters();
    $oDataAccessor = $this->getDataAccessor();
    $parameters['options']['entity'] = $oLayer['class'];
    $oDataAccessor->setOptions($parameters['options']);

    // Patch booléens
    foreach ($aData as $key => $value) {
      if ($value == 'false') {
        $aData[$key] = false;
      }
      if ($value == 'true') {
        $aData[$key] = true;
      }
    }

    $oSerializer = new VitisSerializer();
    $oSerializer->deserialize(json_encode($aData), $oLayer['class'], ['object_to_populate' => $oEntity]);

    // Sauvegarde en base de donnée
    $this->oDatabaseConnection->getAdminEntityManager()->persist($oEntity);
    $this->oDatabaseConnection->getAdminEntityManager()->flush();
    // rafraichissement en cas de modifications interne à la base
    $this->oDatabaseConnection->getAdminEntityManager()->refresh($oEntity);

    $aReturn = [
      'data' => $oEntity,
      'status' => 1,
    ];

    return $this->formatResponse($aReturn, Response::HTTP_OK);
  }

  /**
   * Ajoute une entité AbstractLayer
   * @return Response réponse à retourner par l'API
   */
  public function postData(): Response {
    $aData = $this->requestToJson();
    $layertype = $this->requestToArray()['layertype'];
    if (empty($layertype)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find layertype');
      $oResponse = [
        'status' => 0,
        'message' => 'Error during post',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_POST',
      ];
      return $this->formatResponse($oResponse, Response::HTTP_UNAUTHORIZED);
    }

    $oLayer = $this->returnLayerClass(null, $layertype);

    switch ($layertype) {
      case 'postgres':
        $parameters = $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_pglayers_v2');
        break;
      case 'file':
        $parameters = $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_filelayers_v2');
        break;
      case 'xyz':
        $parameters = $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_xyzlayers_v2');
        break;
      case 'osm':
        $parameters = $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_osmlayers_v2');
        break;
      case 'wms':
        $parameters = $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_wmslayers_v2');
        break;
      case 'wmts':
        $parameters = $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_wmtslayers_v2');
        break;
      case 'vectortiles':
        $parameters = $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_vectortileslayers_v2');
        break;
      default:
        $parameters = $this->oRouteConfig->getRequestParameters();
        break;
    }
    
    $oDataAccessor = $this->getDataAccessor();
    $oDataAccessor->setOptions($parameters['options']);
    $oLayerEntity = $oDataAccessor->insertEntity($aData);

    $aReturn = $oLayerEntity;

    return $this->formatResponse($aReturn, Response::HTTP_OK);
  }

  /**
   * Récupère le repository associé au layertype de la couche puis retourne l'entité
   * @param mixed|null $id
   * @param string|null $layertype
   * @return array Tableau avec l'entité AbstractLayer et sa class
   */
  private function returnLayerClass($id = null, ?string $layertype = null): array{
    // Récupération du repository parent AbstractLayer
    $oLayerRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(AbstractLayer::class);
    if (!$oLayerRepository instanceof GenericRepository) {
      throw new \Exception("Repository is not GenericRepository", 1);
    }
    if ($id !== null) {
      // Récupération de l'entité avec le parent
      $oLayer = $oLayerRepository->findById($id)[0];

      // Attribution de la class associé au layertype
      $layertype = $oLayer->getLayertype();
    }

    $class = AbstractLayer::getLayerChildrenLayerClassName($layertype);
    $aReturn = [
      'class' => $class,
    ];

    if ($id !== null) {
      // Récupération de l'entité global avec les champs spécifique à son type
      $oLayerChildRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository($class);
      if (!$oLayerChildRepository instanceof GenericRepository) {
        throw new \Exception("Repository is not GenericRepository", 1);
      }
      $aReturn['oLayer'] = $oLayerChildRepository->findById($id)[0];
    }

    // Return de l'entité et de la class du layertype si $id existe
    return $aReturn;
  }
}
