<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use App\Module\Vmap\WebService\Entity\AbstractLayer;
use App\Module\Vmap\WebService\Service\MapServerService;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller pour le datasource treeview de vMap.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class PgLayerController extends AbstractVmapQueriesController {
  /**
   * @var MapServerService
   */
  private $oMapServerService;

  /**
   * @var Properties
   */
  private $oPropertiesSrvc;

  /**
   * @var LocalAccessor
   */
  private $oLocalFileAccessor;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param MapServerService $oMapServerService
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, MapServerService $oMapServerService) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oMapServerService = $oMapServerService;
    $this->oPropertiesSrvc = $oPropertiesSrvc;
    $this->oLocalFileAccessor = $oFileManager->getLocalFileAccessor();
  }

  /**
   * Permet de récupérer les colonnes d'une table, vue ou d'une vue matérialisée d'un schema d'une base de données
   * @return Response réponse à retourner par l'API
   */
  public function getColumns(): Response {
    $aData = $this->requestToArray();
    $aConnectionData = null;
    // si les field pour la connexion ne sont pas présent et qu'on le datasource_id
    if (!$this->isConnectionDatasPresent($aData) && !empty($aData['datasource_id'])) {
      $sDatabaseName = $aData['database'] ?? null;
      $aConnectionData = $this->getDataConnectionFromDatasource($aData['datasource_id'], true, $sDatabaseName);
    }

    if ($aConnectionData === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'MISSING_PARAMS',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    $aData = array_merge($aData, $aConnectionData);

    if (!isset($aData['table']) || empty($aData['table'])) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'MISSING_PARAMS',
      ], Response::HTTP_NOT_ACCEPTABLE);
    }

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Columns
    $aColumns = $oQueryAccessor->getColumns($aData);
    if (!isset($aColumns)) {
      return $this->formatResponse(['status' => 0]);
    }
    // Tri par le nom de la colonne.
    if (!empty($aColumns)) {
      usort($aColumns, function ($a, $b) {return strcmp($a['column_name'], $b['column_name']);});
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($aColumns as $oTable) {
      $response['data'][] = [
        'column_name' => $oTable['column_name'],
        'data_type' => $oTable['data_type'],
      ];
    }

    return $this->formatResponse($response);
  }

  /**
   * Récupère la liste des colonnes disponibles dans la table de la couche à partir de layer_id
   * @param mixed $id
   * @return Response réponse à retourner par l'API
   */
  public function getPgLayerFormColumns($id): Response {
    $aData = $this->requestToArray();

    $aConnectionData = $this->getDataConnectionFromLayer($id, 'form');

    if (empty($aConnectionData)) {
      return $this->oResponse;
    }

    $aData = array_merge($aData, $aConnectionData);

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Columns
    $aColumns = $oQueryAccessor->getColumns($aData);
    if (!isset($aColumns)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => $this->getColumnProperty($aColumns)];

    return $this->formatResponse($response);
  }

  /**
   * Return les colonnes d'une table sous format formulaire / ou name&type
   */
  private function getColumnProperty($aColumns, $fields = false): array{
    $response = [];
    foreach ($aColumns as $oColumn) {
      $aColumnData = [];
      if (strpos($oColumn['data_type'], 'bigint') !== false
        || strpos($oColumn['data_type'], 'integer') !== false
        || strpos($oColumn['data_type'], 'smallint') !== false
      ) {
        $data = [
          'type' => 'integer',
          'name' => $oColumn['column_name'],
        ];
        if (!$fields) {
          $data['label'] = $oColumn['column_name'];
          $data['nb_cols'] = 12;
        }
        $aColumnData = $data;
      } else if (strpos($oColumn['data_type'], 'time zone') !== false || strpos($oColumn['data_type'], 'date') !== false) {
        $data = [
          'type' => 'date',
          'name' => $oColumn['column_name'],
        ];
        if (!$fields) {
          $data['label'] = $oColumn['column_name'];
          $data['nb_cols'] = 12;
        }
        $aColumnData = $data;
      } else if (strpos($oColumn['data_type'], 'boolean') !== false) {
        $data = [
          'type' => 'radio',
          'name' => $oColumn['column_name'],
        ];
        if (!$fields) {
          $data['label'] = $oColumn['column_name'];
          $data['nb_cols'] = 12;
          $data['options'] = [
            'choices' => [
              [
                'label' => 'FORM_YES',
                'value' => true,
              ],
              [
                'label' => 'FORM_NO',
                'value' => false,
              ],
            ],
          ];
        }
        $aColumnData = $data;
      } else {
        $data = [
          'type' => 'string',
          'name' => $oColumn['column_name'],
        ];
        if (!$fields) {
          $data['type'] = 'text';
          $data['label'] = $oColumn['column_name'];
          $data['nb_cols'] = 12;
        }
        $aColumnData = $data;
      }
      $response[] = $aColumnData;
    }
    return $response;
  }

  /**
   * Récupère les valeurs disponibles dans la table de la couche à partir de layer_id
   * @return Response réponse à retourner par l'API
   */
  public function getLayerFormData(): Response {
    $aData = $this->requestToArray();

    if (!isset($aData['layer_id'])) {
      return $this->formatResponse([
        'status' => 0,
        'message' => 'layer_id missing',
        'errorMessage' => 'MISSING_PARAMS',
      ]);
    }

    // Récupère les informations de connexion
    $aConnectionData = $this->getDataConnectionFromLayer($aData['layer_id'], 'form');

    // Retourne l'erreur si il y en a eu une
    if (empty($aConnectionData)) {
      return $this->oResponse;
    }

    $aData = array_merge($aData, $aConnectionData);
    $aData['file_path'] = '@prop(vmap_upload_dir)/' . $aData['layer_id'] . '/documents/@id/@column/@value';

    // Lance le generic get pour récupérer les données
    return $this->getGeneric($aData);
  }

  /**
   * Récupère les valeurs disponibles dans la column de la couche à partir de layer_id
   * @return Response réponse à retourner par l'API
   */
  public function getLayerFormColumnData(): Response {
    $aData = $this->requestToArray();

    if (!isset($aData['layer_id'])) {
      return $this->formatResponse([
        'status' => 0,
        'message' => 'layer_id missing',
        'errorMessage' => 'MISSING_PARAMS',
      ]);
    }

    // Récupère les informations de connexion
    $aConnectionData = $this->getDataConnectionFromLayer($aData['layer_id'], 'form');

    // Retourne l'erreur si il y en a eu une
    if (empty($aConnectionData)) {
      return $this->oResponse;
    }

    $aData = array_merge($aData, $aConnectionData);

    // Lance le generic get pour récupérer les données
    return $this->getDatasColumn($aData);
  }

  /**
   * Requête le contenu d'une couche
   * @param mixed $id
   * @return Response
   */
  public function queryLayer($id): Response {
    $aRequestData = $this->requestToArray();
    $aLayerDef = $this->getLayerDef($id);

    $aData = $this->getDataConnectionFromLayer($id, 'form', true);

    if (empty($aData)) {
      return $this->oResponse;
    }

    $aData = array_merge($aRequestData, $aData);

    // Applique le filtre sur les couches
    if (array_key_exists('layers_filters', $aRequestData)) {
      foreach ($aRequestData['layers_filters'] as $key => $value) {
        if (array_key_exists('layer_id', $value) &&
          array_key_exists('filter', $value) &&
          $value['layer_id'] == $id) {
          if (array_key_exists('filter', $aData) &&
            !empty($aData['filter'])) {
            $aData['filter'] = [
              'relation' => 'AND',
              'operators' => [$aData['filter'], $value['filter']],
            ];
          } else {
            $aData['filter'] = $value['filter'];
          }
        }
      }
    }

    $response = $this->queryLayerDatas($aData, $aLayerDef);

    return $this->formatResponse($response);
  }

  /**
   * Retourne le premier contenu trouvé et la couche dans les couches de la carte
   */
  public function queryLayers(): Response {
    $aRequestData = $this->requestToArray();
    $response = [];
    $list_count_group_by = 0;
    $total_row_number_group_by = 0;
    if (isset($aRequestData['idList'])) {
      $idList = is_array($aRequestData['idList']) ? $aRequestData['idList'] : json_decode($aRequestData['idList']);
      foreach ($idList as $id) {
        $aLayerDef = $this->getLayerDef($id);
        $aData = $this->getDataConnectionFromLayer($id, 'form', true);

        if (empty($aData)) {
          return $this->oResponse;
        }

        $aData = array_merge($aRequestData, $aData);

        // Applique le filtre sur les couches
        if (array_key_exists('layers_filters', $aRequestData)) {
          foreach ($aRequestData['layers_filters'] as $key => $value) {
            if (array_key_exists('layer_id', $value) &&
              array_key_exists('filter', $value) &&
              $value['layer_id'] == $id) {
              if (array_key_exists('filter', $aData) &&
                !empty($aData['filter'])) {
                $aData['filter'] = [
                  'relation' => 'AND',
                  'operators' => [$aData['filter'], $value['filter']],
                ];
              } else {
                $aData['filter'] = $value['filter'];
              }
            }
          }
        }

        $query = $this->queryLayerDatas($aData, $aLayerDef);

        if (isset($aRequestData['groupBy'])) {
          $response[$aLayerDef['name']] = $query;
          $list_count_group_by += $query['list_count'];
          $total_row_number_group_by += $query['total_row_number'];
        } else if ($query && $query['list_count'] > 0) {
          $query['oLayer'] = $aLayerDef;
          return $this->formatResponse($query);
        }
      }
    }
    $return = [];
    if (isset($aRequestData['groupBy'])) {
      $return = [
        "status" => 1,
        "data" => $response,
        "list_count" => $list_count_group_by,
        "total_row_number" => $total_row_number_group_by,
      ];
    } else {
      $return = $query;
    }
    return $this->formatResponse($return);
  }

  /**
   * Renvoie le contenu d'une couche
   */
  private function queryLayerDatas($aData, $aLayerDef) {
    if (
      !isset($aData['table']) || empty($aData['table']) ||
      !isset($aData['database']) || empty($aData['database']) ||
      !isset($aData['schema']) || empty($aData['schema'])
    ) {
      return [
        'status' => 0,
        'errorMessage' => 'LAYER_MISSING_PARAMS',
      ];
    }

    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse($this->cannotAccessResponse, Response::HTTP_UNAUTHORIZED);
    }

    // Fields
    $aFields = $oQueryAccessor->getFieldsFromTable($aData);
    if ($aFields['status'] == 0) {
      return $this->formatResponse($aFields);
    }

    $options = [
      'schema' => $aData['schema'],
      'table' => $aData['table'],
      'fields' => $aFields['data'],
    ];

    // Si on connait le champ identifiant on le passe à la config et on l'ajoute au attributs
    if (isset($aData['field_id'])) {
      $options['field_id'] = $aData['field_id'];
    }

    $aAttributs = [];
    foreach ($options['fields'] as $key => $value) {
      if ($value['name'] === $aLayerDef['geomfield']) {
        if (isset($aData['get_geom'])) {
          if ($aData['get_geom'] !== "false" && $aData['get_geom'] !== false) {
            $aAttributs[] = $value['name'];
          }
        } else {
          $aAttributs[] = $value['name'];
        }
      } else {
        $aAttributs[] = $value['name'];
      }
    }

    // Paramètre intersect_geom
    if (!empty($aData['intersect_geom']) && $aLayerDef['is_geometry_layer'] !== false) {
      $aIntersectFilter = [
        'column' => $aLayerDef['geomfield'],
        "compare_operator" => "intersect",
        "compare_operator_options" => [
          "source_proj" => $aLayerDef['coordsys_id'],
          "intersect_buffer" => empty($aData['intersect_buffer']) ? 0 : $aData['intersect_buffer'],
        ],
        "value" => $aData['intersect_geom'],
      ];

      if (
        !empty($aLayerDef['ms_geomtype']) &&
        !empty($aLayerDef['ms_geomtype']['label'])
      ) {
        if ($aLayerDef['ms_geomtype']['label'] === "GEOMETRY") {
          $aIntersectFilter['compare_operator_options']['intersect_buffer_geom_type'] = "POINT";
        } else {
          $aIntersectFilter['compare_operator_options']['intersect_buffer_geom_type'] = $aLayerDef['ms_geomtype']['label'];
        }
        if ($aLayerDef['ms_geomtype']['label'] === "POLYGON") {
          $aIntersectFilter['compare_operator_options']['intersect_buffer'] = 0;
        }
      }

      if (empty($aData['filter'])) {
        $aData['filter'] = [
          'relation' => 'AND',
          'operators' => [$aIntersectFilter],
        ];
      } else {
        $aUserFilter = $aData['filter'];
        if (is_string($aData['filter'])) {
          $aUserFilter = json_decode($aData['filter'], true);
        }

        $aData['filter'] = [
          'relation' => 'AND',
          'operators' => [$aUserFilter, $aIntersectFilter],
        ];
      }
    }

    // Paramètre result_srid
    if (!empty($aData['result_srid'])) {
      foreach ($options['fields'] as $key => $value) {
        if ($value['name'] === $aLayerDef['geomfield']) {
          $options['fields'][$key]['result_srid'] = $aData['result_srid'];
        }
      }
    }

    if (empty($aData['attributs'])) {
      $aData['attributs'] = implode('|', $aAttributs);
    }

    // Connexion pour récupérer les données
    $this->oRouteConfig->setOptionsParameter($options);
    $oEm = $this->getEntityManagerFromData($aData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse($this->cannotAccessResponse, Response::HTTP_UNAUTHORIZED);
    }

    // Évite de deni de service
    if (empty($aData['limit'])) {
      $aData['limit'] = $this->oProperties->get('selection_limit') || 2000;
    }

    // Récupération des tables
    $aEntities = $oQueryAccessor->getEntities($aData);
    if (is_array($aEntities['data'])) {
      $aEntities['data'] = $this->handleFilesOnDownload($aLayerDef['layer_id'], $aLayerDef['idfield'], $aEntities['data']);
    }
    return $aEntities;
  }

  /**
   * Retourne la définition de la layer
   * @param mixed $id
   * @return array|null
   */
  private function getLayerDef($id): ?array{
    $aLayerDef = [];
    $oLayerRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(AbstractLayer::class);

    if (!$oLayerRepository instanceof GenericRepository) {
      return null;
    }

    $oLayer = $oLayerRepository->findById($id)[0];

    // Getters de l'entité AbstractLayer
    $oLayerRoute = $this->oRouteConfig->getRequestParameters('vmap_layer');
    $aLayerDef = $oLayer->jsonSerialize($oLayerRoute);

    return $aLayerDef;
  }

  /**
   * Met un jour un object dans la carte
   */
  public function updateQueryLayer($id, $object_id): Response {
    $aData = $this->requestToArray();
    $parameters = $this->getRequestParameters();
    $aConnectionData = $this->getDataConnectionFromLayer($id, 'form', true);

    if (empty($aConnectionData)) {
      return $this->oResponse;
    }

    $oEm = $this->getEntityManagerFromData($aConnectionData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Set l'option car non présente dans routes.json
    $parameters['options']['schema'] = $aConnectionData['schema'];
    $parameters['options']['table'] = $aConnectionData['table'];
    $parameters['options']['field_id'] = $aConnectionData['field_id'];

    // Columns
    $aColumns = $oQueryAccessor->getFieldsFromTable($aConnectionData);
    if (!isset($aColumns)) {
      return $this->formatResponse(['status' => 0]);
    }

    // Set fields pour mettre à jour les champs
    $parameters['options']['fields'] = $this->handleGeomSRIDs($this->handleFilesOnUpload($id, $aColumns['data'], $aData));

    $oQueryAccessor->setOptions($parameters['options']);
    $this->oRouteConfig->setOptionsParameter($parameters['options']);

    $oEntity = $oQueryAccessor->updateEntity($object_id, json_encode($aData));
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnUpdate();
    }

    return $this->formatResponse($oEntity);
  }

  /**
   * Ajoute de la donnée dans la layer
   */
  public function postQueryLayer($id): Response {
    $aData = $this->requestToArray();
    $parameters = $this->getRequestParameters();

    $aConnectionData = $this->getDataConnectionFromLayer($id, 'form', true);

    if (empty($aConnectionData)) {
      return $this->oResponse;
    }

    $oEm = $this->getEntityManagerFromData($aConnectionData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Set l'option car non présente dans routes.json
    $parameters['options']['schema'] = $aConnectionData['schema'];
    $parameters['options']['table'] = $aConnectionData['table'];
    $parameters['options']['field_id'] = $aConnectionData['field_id'];

    // Columns
    $aColumns = $oQueryAccessor->getFieldsFromTable($aConnectionData);

    if (!isset($aColumns)) {
      return $this->formatResponse(['status' => 0]);
    }

    // Set fields pour mettre à jour les champs
    $parameters['options']['fields'] = $this->handleFilesOnUpload($id, $aColumns['data'], $aData);

    $oQueryAccessor->setOptions($parameters['options']);
    $this->oRouteConfig->setOptionsParameter($parameters['options']);

    // Récupère l'entité à mettre à jour pour vérifier qu'elle existe
    $oEntity = $oQueryAccessor->insertEntity(json_encode($aData));

    if ($oEntity['status'] === 0) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'Entity not found',
      ], Response::HTTP_FOUND);
    }

    return $this->formatResponse($oEntity);
  }

  /**
   * Supprimer un object dans la carte
   */
  public function deleteQueryLayer($id, $object_id): Response {
    $aData = $this->requestToArray();
    $parameters = $this->getRequestParameters();

    $aConnectionData = $this->getDataConnectionFromLayer($id, 'form', true);

    if (empty($aConnectionData)) {
      return $this->oResponse;
    }

    $oEm = $this->getEntityManagerFromData($aConnectionData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'GENERIC_ERROR',
      ], Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Set l'option car non présente dans routes.json
    $parameters['options']['schema'] = $aConnectionData['schema'];
    $parameters['options']['table'] = $aConnectionData['table'];
    $parameters['options']['field_id'] = $aConnectionData['field_id'];
    $oQueryAccessor->setOptions($parameters['options']);

    // Récupère l'entité à mettre à jour pour vérifier qu'elle existe
    $oEntity = $oQueryAccessor->getEntity($object_id);

    if ($oEntity['status'] === 0) {
      return $this->formatResponse([
        'status' => 0,
        'errorMessage' => 'Entity not found',
      ], Response::HTTP_FOUND);
    }

    // Requête querybuilder
    $qb = $oQueryAccessor->createQueryBuilder();
    $qb->delete($aConnectionData['schema'] . '.' . $aConnectionData['table'])
      ->where($aConnectionData['field_id'] . ' = :mIdInWhere')
      ->setParameter('mIdInWhere', $object_id);

    try {
      $qb->execute();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Query Accessor : can\'t delete this object : ' . $e->getMessage());
      return $this->formatResponse(['status' => 0]);
    }

    return $this->formatResponse([
      'data' => $object_id,
      'status' => 1,
    ]);
  }

  /**
   * Retourne les colonnes d'une couche
   */
  public function getFieldsFromLayer(): Response {
    $aRequestData = $this->requestToArray();
    if (
      !isset($aRequestData['table']) || empty($aRequestData['table']) ||
      !isset($aRequestData['database']) || empty($aRequestData['database']) ||
      !isset($aRequestData['schema']) || empty($aRequestData['schema'])
    ) {
      return [
        'status' => 0,
        'errorMessage' => 'LAYER_MISSING_PARAMS',
      ];
    }

    $oEm = $this->getEntityManagerFromData($aRequestData);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    if ($oQueryAccessor === null) {
      return $this->formatResponse($this->cannotAccessResponse, Response::HTTP_UNAUTHORIZED);
    }

    // Fields
    $aFields = $oQueryAccessor->getFieldsFromTable($aRequestData);

    return $this->formatResponse($aFields['data']);
  }

  /**
   * Test d'une couche
   */
  public function testLayer($id): Response {
    $aData = $this->requestToArray();
    $oDataAccessor = $this->getDataAccessor();
    $aReturn = [
      'status' => 0,
    ];

    $aLayerDef = $this->getLayerDef($id);

    if (array_key_exists('layer_id', $aLayerDef) &&
      array_key_exists('definition_sym_json', $aData) &&
      array_key_exists('layertype', $aData)) {
      $aLayerDef['definition_sym_json'] = $aData['definition_sym_json'];
      // Retour de la définition mapserver
      switch ($aData['layertype']) {
        case 'file':
          $aMSLayerDef = $this->oMapServerService->getMSLayerRasterDef($aLayerDef);
          break;
        default:
          if (!array_key_exists('coordsys_id', $aLayerDef) ||
            !array_key_exists('ms_geomtype_id', $aLayerDef) ||
            !array_key_exists('geomfield', $aLayerDef)) {
            return $this->formatResponse($aReturn = [
              'status' => 0,
              'errorMessage' => 'ERROR_MISSING_GEOMETRY_PARAMS',
            ]);
          }

          if (array_key_exists("geomtype", $aData)) {
            $aLayerDef['ms_geomtype']['label'] = strtoupper($aData['geomtype']);
            $aLayerDef['multiGeom'] = true;
            $aMSLayerDef = $this->oMapServerService->getMSLayerPostgresDef($aLayerDef, true);
          } else {
            $aMSLayerDef = $this->oMapServerService->getMSLayerPostgresDef($aLayerDef);
          }
          // Générer le fichier fonts.list avec les données en bdd
          $this->oMapServerService->generateFontList();
      }
      $sMSLayerDef = join("\r\n", $aMSLayerDef);

      // Génération du mapfile de test
      $sMapfileName = $this->oMapServerService->createMapfile([
        'map_id' => 'test_map',
        'name' => 'test_map',
      ], [$aLayerDef], null, false);

      // Construction de l'URL du flux
      $sOnlineResourceUrl = str_replace('@file', $sMapfileName, MapServerService::PRIVATE_WMS_MAIN_URL);
      $sOnlineResourceUrl = $this->oPropertiesSrvc->replacePropertiesInString($sOnlineResourceUrl);
      $sFluxUrl = $sOnlineResourceUrl;

      $aReturn = [
        'data' => [
          'mapserver_def' => $sMSLayerDef,
          'flux_url' => $sFluxUrl,
          'layer_name' => $this->oMapServerService->cleanLayerName($aLayerDef['name']),
        ],
        'status' => 1,
      ];
    }

    return $this->formatResponse($aReturn);
  }

  /**
   * Ajoute les infos nécessaires dans les colonnes pour l'envoi de fichiers
   */
  private function handleFilesOnUpload($iLayerId, $aColumns, $aData): ?array{
    // Liste des colonnes qui sont des fichiers
    $aFileColumns = $this->getFilesColumns($aColumns, $aData);

    foreach ($aColumns as $key => $value) {
      if (in_array($value['name'], $aFileColumns)) {
        $value['type'] = 'file';
        $value['multiple'] = true;
        $value['path'] = '@prop(ws_data_dir)/vmap/layer_entities/' . $iLayerId . '/documents/@id/' . $value['name'] . '/@value';
        $aColumns[$key] = $value;
      }
    }

    return $aColumns;
  }

  /**
   * Ajoute les infos nécessaires pour le téléchargement de fichiers
   */
  private function handleFilesOnDownload($iLayerId, $sIdfield, $aData): ?array{
    $oFileAccessor = $this->oFileManager->getFileAccessor();
    for ($i = 0; $i < count($aData); $i++) {
      if (array_key_exists($sIdfield, $aData[$i]) && !empty($aData[$i][$sIdfield])) {
        $sId = $aData[$i][$sIdfield];
        foreach ($aData[$i] as $key => $value) {
          $aFileNames = explode('|', $value);
          if (count($aFileNames) > 1) {
            $aData[$i][$key] = [];
          }

          for ($ii = 0; $ii < count($aFileNames); $ii++) {
            $sFilesPath = $this->oProperties->get('ws_data_dir') . '/vmap/layer_entities/' . $iLayerId . '/documents/' . $sId . '/' . $key . '/' . $aFileNames[$ii];
            if ($oFileAccessor->file_exists($sFilesPath)) {
              if (count($aFileNames) > 1) {
                $aData[$i][$key][] = $oFileAccessor->getProxyPassUrl($sFilesPath);
              } else {
                $aData[$i][$key] = $oFileAccessor->getProxyPassUrl($sFilesPath);
              }
            }
          }
        }
      }
    }

    return $aData;
  }

  /**
   * Liste des colonnes qui sont des fichiers
   */
  private function getFilesColumns($aColumns, $aData): ?array{
    $aFileColumns = [];

    // Analyse de $_FILES
    foreach ($aColumns as $key => $value) {
      if (array_key_exists($value['name'], $_FILES)) {
        array_push($aFileColumns, $value['name']);
      }
    }

    // Analyse de vitis_deleted_files
    if (array_key_exists('vitis_deleted_files', $aData) && !empty($aData['vitis_deleted_files'])) {
      $aUnchangedFiles = json_decode($aData['vitis_deleted_files'], true);
      foreach ($aUnchangedFiles as $key => $value) {
        if (!in_array($key, $aFileColumns)) {
          array_push($aFileColumns, $key);
        }
      }
    }

    return $aFileColumns;
  }

  /**
   * Met un result_srid en cas où celui-ci soit absent et que la projection de la donnée soit autre que 2154, 3857 et 4326
   */
  private function handleGeomSRIDs($aOptionFields): ?array{
    for ($i = 0; $i < count($aOptionFields); $i++) {
      if (array_key_exists('type', $aOptionFields[$i]) &&
        array_key_exists('srid', $aOptionFields[$i]) &&
        $aOptionFields[$i]['type'] === 'geom' &&
        $aOptionFields[$i]['srid'] != 2154 &&
        $aOptionFields[$i]['srid'] != 3857 &&
        $aOptionFields[$i]['srid'] != 4326) {
        if (!array_key_exists('result_srid', $aOptionFields[$i])) {
          $aOptionFields[$i]['result_srid'] = 2154;
        }
      }
    }

    return $aOptionFields;
  }
}
