<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Service\PrintService\PrintTemplate;
use Ramsey\Uuid\Uuid;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller pour impressions
 *
 * @author Armand Bahi <armand.bahi@veremes.com>
 */
class PrintController extends GenericController {
  /**
   * @var string
   */
  private const VMAP_PRINT_DIR = '@prop(extract_dir)/vmap_print';

  /**
   * Imprime un template
   * @param mixed $id
   * @return Response
   */
  public function printTemplate($id): Response {
    $aData = $this->requestToArray();
    $aReturn = [
      'status' => 0,
    ];
    try {
      // Récupère les données du template.
      $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();
      $qb->select('*')
        ->from('[SCHEMA]."v_user_printtemplate"', 'vet')
        ->where($qb->expr()->eq('vet.printtemplate_id', '[TEMPL_ID]'));
      $aParams = [
        'SCHEMA' => ['value' => $this->oProperties->get('schema_vmap'), 'type' => 'schema_name'],
        'TEMPL_ID' => ['value' => $id, 'type' => 'integer'],
      ];
      try {
        $aResult = $qb->executeWithParams($aParams)->fetch();
      } catch (\Exception $e) {
        $this->oLoggerManager->get('SQL_ERROR')->error('Error getting template' . $e->getMessage());
        $aReturn['message'] = 'Error getting template';
        $aReturn['errorMessage'] = 'REQUEST_ERROR';
        return $this->formatResponse($aReturn);
      }

      if (
        !empty($aResult['printtemplate_id']) &&
        !empty($aResult['rt_format_id']) &&
        !empty($aResult['rt_orientation_id']) &&
        !empty($aResult['definition']) &&
        !empty($aResult['outputformats_id'])
      ) {
        // Chemin du fichier de destination
        $sDestFile = $this->getPrintFilePath();

        // Impression avec le service PrintTemplate
        $oPrintTemplate = new PrintTemplate($this->oLoggerManager, $this->oProperties, $this->oFileManager);
        $sHTMLBody = $aResult['definition'];
        $sFormat = $aResult['rt_format_id'];
        $sOutputFormat = $aResult['outputformats_id'];
        $sOrientation = $aResult['rt_orientation_id'];
        $aScope = [];
        $aIncludes = [];
        $aMaps = [];

        // Style d'impression
        if (!empty($aData['printstyle_id'])) {
          $qb->select('a.*')
            ->from('[SCHEMA]."v_user_printstyle"', 'a')
            ->where($qb->expr()->eq('a.printstyle_id', '[STYLE_ID]'));
          $aParams = [
            'SCHEMA' => ['value' => $this->oProperties->get('schema_vmap'), 'type' => 'schema_name'],
            'STYLE_ID' => ['value' => $aData['printstyle_id'], 'type' => 'integer'],
          ];
          try {
            $aStyeResult = $qb->executeWithParams($aParams)->fetch();
            if (is_string($aStyeResult['definition']) && !empty($aStyeResult['definition'])) {
              $sHTMLBody = $aStyeResult['definition'] . $sHTMLBody;
            }
          } catch (\Exception $e) {
            $this->oLoggerManager->get('SQL_ERROR')->error('Error getting print style' . $e->getMessage());
            $aReturn['message'] = 'Error getting print style';
            $aReturn['errorMessage'] = 'REQUEST_ERROR';
            return $this->formatResponse($aReturn);
          }
        }

        $user = $this->oDatabaseConnection->getCurrentUser();
        // Decode scope
        try {
          $aTmpScope = json_decode($aData['scope_json'], true);
          $aTmpScope['token'] = $this->oVitisRequest->getToken();
          if (is_array($aTmpScope) && !empty($aTmpScope)) {
            $aScope = $aTmpScope;
          }
        } catch (\Throwable $th) {
          // @ignoreException
        }
        // Decode includes
        try {
          $aTmpIncludes = $this->decodeJSONIncludes($aData['includes_json']);
          if (!empty($aTmpIncludes)) {
            $aIncludes = $aTmpIncludes;
          }
        } catch (\Throwable $th) {
          // @ignoreException
        }
        // Decode maps
        try {
          $aTmpMaps = $this->decodeMapIncludes($aData['maps_json']);
          if (!empty($aTmpMaps)) {
            $aMaps = $aTmpMaps;
          }
        } catch (\Throwable $th) {
          // @ignoreException
        }
        $sDestFilePath = $oPrintTemplate->print($sHTMLBody, $sFormat, $sOutputFormat, $sOrientation, $aScope, $aIncludes, $aMaps, $sDestFile);
        if (!empty($sDestFilePath)) {

          // URL vers le fichier
          $oLocalFileAccessor = $this->getLocalFileAccessor();
          $sDestFileUrl = $oLocalFileAccessor->getProxyPassUrl($sDestFile);

          // retour
          $aReturn['data'] = $sDestFileUrl;
          $aReturn['status'] = 1;
        } else {
          $this->oLoggerManager->get('API_ERROR')->error('Error while printing template');
          $aReturn['message'] = 'Error while printing template';
          $aReturn['errorMessage'] = 'ERROR_PRINT_REQUEST';
        }
      } else {
        $this->oLoggerManager->get('API_ERROR')->error('Error print template not valid');
        $aReturn['message'] = 'Error while printing template';
      }
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Error while printing template: ' . $e->getMessage());
      $aReturn['message'] = 'Error while printing template';
    }

    return $this->formatResponse($aReturn);
  }

  /**
   * Retourne le chemin où le fichier d'impression sera écrit
   * @return string
   */
  private function getPrintFilePath(): string {
    $sDir = $this->oProperties->replacePropertiesInString(self::VMAP_PRINT_DIR);
    $oLocalFileAccessor = $this->getLocalFileAccessor();
    if ($oLocalFileAccessor->file_exists($sDir) !== true) {
      $oLocalFileAccessor->mkdir($sDir, 0777, true);
    }
    return $sDir . '/' . Uuid::uuid1()->toString() . '.pdf';
  }

  /**
   * Decode le JSON de includes tout en encodant la partie HTML
   * @param string $sJsonIncludes
   * @return array
   */
  private function decodeJSONIncludes(string $sJsonIncludes): array{
    $aIncludes = json_decode($sJsonIncludes, true);

    for ($i = 0; $i < count($aIncludes); $i++) {
      if (!empty($aIncludes[$i]['html'])) {
        $aIncludes[$i]['html'] = rawurlencode($aIncludes[$i]['html']);
      }
    }

    return $aIncludes;
  }

  /**
   * Decode le JSON des cartes
   * @param string $sJSONMaps
   * @return array
   */
  private function decodeMapIncludes(string $sJSONMaps): array{
    $aMaps = json_decode($sJSONMaps, true);

    for ($i = 0; $i < count($aMaps); $i++) {
      if (empty($aMaps[$i]['target']) || empty($aMaps[$i]['map_definition'])) {
        continue;
      }

      // Ajout du token
      $aMaps[$i]['map_definition']['token'] = $this->oVitisRequest->getToken();
    }

    return $aMaps;
  }
}
