<?php

namespace App\Module\Vmap\WebService\Repository;

use App\Module\Vitis\WebService\Query\SortableNullValueWalker;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\DataAccessor\CommonAccessorFunctions;
use App\Module\Vmap\WebService\Entity\LayerFile;
use App\Module\Vmap\WebService\Entity\LayerOSM;
use App\Module\Vmap\WebService\Entity\LayerPostgres;
use App\Module\Vmap\WebService\Entity\LayerVectorTiles;
use App\Module\Vmap\WebService\Entity\LayerWMS;
use App\Module\Vmap\WebService\Entity\LayerWMTS;
use App\Module\Vmap\WebService\Entity\LayerXYZ;
use Doctrine\DBAL\Connection;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Tools\Pagination\Paginator;

/**
 * Repository des couches de vmap.
 *
 * @author Ludivine Premont   <ludivine.premont@veremes.com>
 */

class LayerRepository extends GenericRepository {
  /**
   * {@inheritdoc}
   */
  public function __construct(EntityManagerInterface $em, ClassMetadata $class) {
    parent::__construct($em, $class);
  }

  /**
   * Retourne un lot ou l'ensemble de couche appartenant à une carte
   *
   * @param  int   $iMapId identifiant de la carte
   * @param  array $aParams paramètre de la requête HTTP
   * @return array
   */
  public function findAllJoinedToMap(int $iMapId, array $aParams): array{
    return $this->findAllLayerWithMemberOperator($iMapId, $aParams, true);
  }
  /**
   * Retourne un lot ou l'ensemble de couche n'appartenant pas à une carte
   *
   * @param  int   $iMapId identifiant de la carte
   * @param  array $aParams paramètre de la requête HTTP
   * @return array
   */
  public function findAllNotJoinedToMap(int $iMapId, array $aParams): array{
    return $this->findAllLayerWithMemberOperator($iMapId, $aParams, false);
  }

  /**
   * @param  int   $iMapId identifiant de la carte
   * @param  array $aParams paramètre de la requête HTTP
   * @param  bool  $bInclude défini si les couche doivent appartenir ou non à la carte
   * @return array
   */
  private function findAllLayerWithMemberOperator(int $iMapId, array $aParams, bool $bInclude = true): array{
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS);

    // order_by
    if (isset($aParams['order_by'])) {
      $sSortOrder = $aParams['sort_order'] ?? 'asc';
      $qb->orderBy(self::ENTITY_ALIAS . '.' . $aParams['order_by'], $sSortOrder);
    }

    $sOperator = ($bInclude) ? 'MEMBER OF' : 'NOT MEMBER OF';
    $qb->where(':iMapId ' . $sOperator . ' ' . self::ENTITY_ALIAS . '.aMaps');
    $qb->setParameter(':iMapId', $iMapId);

    $oFullPaginator = new Paginator($qb, true);
    $oFullPaginator->setUseOutputWalkers(false);

    // offset
    if (isset($aParams['offset']) && $aParams['offset'] > 0) {
      $qb->setFirstResult($aParams['offset']);
    }

    // limit
    if (isset($aParams['limit']) && $aParams['limit'] > 0) {
      $qb->setMaxResults($aParams['limit']);
    }
    $query = $qb->getQuery();

    $aEntities = $query->getResult();

    // Formatage de la réponse
    return [
      'data' => $aEntities,
      'total_row_selected' => $oFullPaginator->count(),
      'list_count' => count($aEntities),
      'status' => 1,
    ];
  }

  /**
   * Permet de filtrer et paginer les entités Doctrine.
   * @param array $aParams Objet permettant de filtrer et paginer les entités
   *                          - $aParams.order_by     (string)        : colonne pour trier les entités
   *                          - $aParams.sort_order   (string)        : sens de tri (ASC|DESC)
   *                          - $aParams.limit        (integer)       : nombre d'entité à retourner
   *                          - $aParams.offset       (integer)       : décallage des entité par rapport à la première
   *                          - $aParams.attributs    (string)        : Liste des champs à afficher
   *                          - $aParams.filter       (string)        : Json inline permettant de générer les filtres
   *                          - $aParams.distinct     (bool)          : Pas de doublon sur les entités
   * @param CommonAccessorFunctions $oAccessorFunc service de manipulation des attributs
   * @return array Liste des entités
   */
  public function search(array $aParams, CommonAccessorFunctions $oAccessorFunc): array
  {
    // initialisation query builder
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS);
    $sOrderByColumn = null;

    // order_by
    if (isset($aParams['order_by'])) {
      $sSortOrder = (isset($aParams['sort_order'])) ? $aParams['sort_order'] : 'asc';
      $sColumnOrderBy = $oAccessorFunc->getFieldColumn($aParams['order_by']);
      if (isset($sColumnOrderBy)) {
        $sOrderByColumn = self::ENTITY_ALIAS . '.' . $sColumnOrderBy;
        $qb->orderBy(self::ENTITY_ALIAS . '.' . $sColumnOrderBy, $sSortOrder);
      }
    }

    // filter
    $aParams['filter'] = $aParams['filter'] ?? null;

    $filter = (array) json_decode($aParams['filter'], true);
    if (isset($filter['operators'])) {
      $instanceOf = '';
      $filter['operators'] = array_values(array_filter($filter['operators'], function ($operator) use (&$instanceOf) {
        if ($operator['column'] == 'layertype') {
          $instanceOf = $operator['value'];
          return false;
        }
        return true;
      }));
      // Instance of sur la bonne classe si filter layertype existe
      if ($instanceOf) {
        switch ($instanceOf) {
          case 'postgres':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, LayerPostgres::class));
            break;
          case 'osm':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, LayerOSM::class));
            break;
          case 'wms':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, LayerWMS::class));
            break;
          case 'wmts':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, LayerWMTS::class));
            break;
          case 'file':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, LayerFile::class));
            break;
          case 'xyz':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, LayerXYZ::class));
            break;
          case 'vectortiles':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, LayerVectorTiles::class));
            break;
        }
      }
    }

    $filter = json_encode($filter);
    if ($oAccessorFunc->needFilter($filter)) {
      $aFilterReturn = $oAccessorFunc->prepareAndDecodeFilter($filter);

      if (!empty($aFilterReturn['request'])) {
        $qb->andWhere($aFilterReturn['request']);

        foreach ($aFilterReturn['params'] as $sKey => $aValue) {
          if (isset($aValue['type'])) {
            $qb->setParameter($sKey, $aValue['value'], $aValue['type']);
          } else {
            $qb->setParameter($sKey, $aValue['value']);
          }
        }
      }
    }

    $oFullPaginator = new Paginator($qb, true);
    $oFullPaginator->setUseOutputWalkers(false);

    // offset
    if (isset($aParams['offset']) && $aParams['offset'] > 0) {
      $qb->setFirstResult($aParams['offset']);
    }

    // limit
    if (isset($aParams['limit']) && $aParams['limit'] > 0) {
      $qb->setMaxResults($aParams['limit']);
    }
    $query = $qb->getQuery();
    if ($sOrderByColumn !== null) {
      // ajout de NULLS LAST
      $query->setHint(SortableNullValueWalker::HINT, [
        $sOrderByColumn => SortableNullValueWalker::NULLS_LAST,
      ]);
    }

    $aEntities = $query->getResult();

    // Formatage de la réponse
    $aEntities = [
      'data' => $aEntities,
      'total_row_selected' => $oFullPaginator->count(),
      'list_count' => count($aEntities),
      'status' => 1,
    ];

    return $aEntities;
  }

  /**
   * Supprime les liens entre couche et carte puis supprime la liste des couches
   * @param array $aIds Liste des ids à supprimer
   * @return array
   */
  public function removeList(array $aIds): array{
    $oReturn = [];
    $oMetaData = $this->getClassMetadata();

    $oConnexion = $this->_em->getConnection();
    $oConnexion->setAutoCommit(false);
    // on démarre une transaction
    $oConnexion->beginTransaction();
    try {
      $qb = $oConnexion->createQueryBuilder();
      $qb->delete($oMetaData->getSchemaName() . '.layer_map')
        ->where('layer_map.layer_id IN (?)');

      $qb->getConnection()->executeQuery($qb->getSQL(), [$aIds], [Connection::PARAM_STR_ARRAY]);

      $qb = $this->createQueryBuilder(self::ENTITY_ALIAS)
        ->delete(null, self::ENTITY_ALIAS)
        ->where(self::ENTITY_ALIAS . '.' . $this->sIdentifier . ' IN (:aIds)')
        ->setParameter('aIds', $aIds);

      $query = $qb->getQuery();
      $iRemoved = $query->getResult();

      $oConnexion->commit();

      $oReturn = [
        'total_row_removed' => $iRemoved,
        'data' => $aIds,
        'status' => 1,
      ];

    } catch (\Exception $e) {
      $oConnexion->rollBack();
      error_log('Can\'t delete this objects : ' . $e->getMessage());

      $oReturn['status'] = 0;
      $oReturn['message'] = 'Error during delete';
      $oReturn['errorMessage'] = 'LIST_DELETE_ERROR';
    } finally {
      // Les transaction en cours seront automatiquement commit, ne pas oublié de commit/rollback avant ce point
      $oConnexion->setAutoCommit(true);

    }
    return $oReturn;
  }

  /**
   * {@inheritdoc}
   */
  public function removeItemById($mId): bool {
    $oMetaData = $this->getClassMetadata();
    $oReturn = false;
    $oConnexion = $this->_em->getConnection();
    $oConnexion->setAutoCommit(false);
    // on démarre une transaction
    $oConnexion->beginTransaction();
    try {
      $qb = $oConnexion->createQueryBuilder();
      $qb->delete($oMetaData->getSchemaName() . '.layer_map')
        ->where('layer_map.layer_id = ?');

      $qb->getConnection()->executeQuery($qb->getSQL(), [$mId]);

      $qb = $this->createQueryBuilder(self::ENTITY_ALIAS)
        ->delete(null, self::ENTITY_ALIAS)
        ->where(self::ENTITY_ALIAS . '.' . $this->sIdentifier . ' = :mId')
        ->setParameter('mId', $mId);

      $query = $qb->getQuery();
      $iRemoved = $query->getResult();

      $oConnexion->commit();

      $oReturn = ($iRemoved === 1);
    } catch (\Exception $e) {
      $oConnexion->rollBack();
      error_log('Can\'t delete this objects : ' . $e->getMessage());
    } finally {
      // Les transaction en cours seront automatiquement commit, ne pas oublié de commit/rollback avant ce point
      $oConnexion->setAutoCommit(true);
      return $oReturn;
    }
  }
}
