<?php

namespace App\Module\Vmap\WebService\Service;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\CommonAccessorFunctions;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\CustomFieldsManager;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use App\Module\Vitis\WebService\Service\VitisProcess;
use App\Module\Vitis\WebService\Service\VitisSystem;
use App\Module\Vmap\WebService\Entity\LayerFile;
use App\Module\Vmap\WebService\Entity\LayerPostgres;
use Symfony\Component\Process\Exception\ProcessFailedException;

/**
 * Service permettant de gérer les actions MapServer
 *
 * @author Armand Bahi <armand.bahi@veremes.com>
 */
class MapServerService {
  /**
   * Chemin par défaut des clé wms
   * @var string
   */
  private const KEY_FOLDER_PATH = '@prop(vmap_ms_dir)/wms_key';

  /**
   * Chemin par défaut des  mapfile de wms privé
   * @var string
   */
  private const PRIVATE_WMS_FOLDER_PATH = '@prop(vmap_ms_dir)/wms_private';

  /**
   * Chemin par défaut des  mapfile de wms public
   * @var string
   */
  private const PUBLIC_WMS_FOLDER_PATH = '@prop(vmap_ms_dir)/wms_public';

  /**
   * Chemin par défaut pour les logs des flux privé
   * @var string
   */
  private const PRIVATE_WMS_LOG_PATH = '@prop(log_dir)/@date/mapserver';

  /**
   * URL principale pour intérrogé un flux wfs privé
   * @var string
   */
  public const PRIVATE_WMS_MAIN_URL = '@prop(vmap_ms_url)/private/@file';

  /**
   * URL principale pour intérrogé un flux wfs public
   * @var string
   */
  public const PUBLIC_WMS_MAIN_URL = '@prop(vmap_ms_url)/public/@file';

  /**
   * Chemin par défaut du dossier de partage
   * @var string
   */
  public const PRIVATE_SHARED_FOLDER_PATH = '@prop(shared_dir)';

  /**
   * Chemin vers l'executable de ms_encrypt
   * @var array
   */
  private const MSENCRYPT_EXE_PATH = [
    'lin' => '@prop(vas_dir)/bin/mapserver/bin/msencrypt',
    'win' => '@prop(vas_dir)/bin/mapserver/msencrypt.exe',
  ];

  /**
   * @var string
   */
  private const LOGGER_KEY = 'SQL_ERROR';

  /**
   * @var Properties
   */
  private $oPropertiesSrvc;

  /**
   * @var LoggerManager
   */
  private $oLoggerManager;

  /**
   * @var LocalAccessor
   */
  private $oLocalFileAccessor;

  /**
   * @var Route
   */
  private $oRouteConfigurator;

  /**
   * @var JweTokenManager
   */
  private $oJweTokenManager;

  /**
   * @var string
   */
  private $sMsEncryptPath;

  /**
   * @var DatabaseConnection
   */
  private $oDatabaseConnection;

  /**
   * @param Properties $oPropertiesSrvc Service permettant d'accéder aux properties
   * @param LoggerManager $oLoggerManager Service permettant d'accéder aux loggers
   * @param FileManager $oFileManager Service permettant d'intéragir avec des fichiers locaux
   * @param Route $oRouteConfigurator Service permettant d'accéder à la config de la route
   * @param JweTokenManager $oJweTokenManager Service permettant d'interagir avec le contenu d'un token JWE
   */
  public function __construct(Properties $oPropertiesSrvc, LoggerManager $oLoggerManager, FileManager $oFileManager, Route $oRouteConfigurator, JweTokenManager $oJweTokenManager, DatabaseConnection $oDatabaseConnection) {
    $this->oPropertiesSrvc = $oPropertiesSrvc;
    $this->oLoggerManager = $oLoggerManager;
    $this->oRouteConfigurator = $oRouteConfigurator;
    $this->oJweTokenManager = $oJweTokenManager;
    $this->oDatabaseConnection = $oDatabaseConnection;
    $this->oLocalFileAccessor = $oFileManager->getLocalFileAccessor();

    $sMsEncryptPath = $this->processMsEncryptPath();
    if ($this->oLocalFileAccessor->file_exists($sMsEncryptPath)) {
      $this->sMsEncryptPath = $sMsEncryptPath;
    } else {
      $this->oLoggerManager->get('API_ERROR')->error('cannot find msencrypt executable (' . $sMsEncryptPath . ')');
    }
  }

  /**
   * permet de définir le chemin vers msencrypt
   * @return string|null
   */
  private function processMsEncryptPath(): ?string{
    // détermine s'il faut chercher un executable windows ou linux
    $sKey = (VitisSystem::runOnWindows()) ? 'win' : 'lin';
    // remplace les properties dans le path
    return $this->oPropertiesSrvc->replacePropertiesInString(self::MSENCRYPT_EXE_PATH[$sKey]);
  }

  /**
   * Crée le mapfile (privé) correspondant à la carte
   * @param array $aMap
   * @param array $aLayers
   * @param array $aFilter filtre à appliquer sur les couches
   * @return string
   */
  public function createMapfile(array $aMap, array $aLayers, ?array $aFilter = null, ?bool $bPublic = false): string {
    $sFileName = $this->generateMapFileName($aMap);

    $sMsDirPrivate = $this->oPropertiesSrvc->replacePropertiesInString($bPublic ? self::PUBLIC_WMS_FOLDER_PATH : self::PRIVATE_WMS_FOLDER_PATH);
    if (!$this->oLocalFileAccessor->is_dir($sMsDirPrivate)) {
      $this->oLocalFileAccessor->mkdir($sMsDirPrivate, LocalAccessor::DEFAULT_LINUX_PERMISSION, true);
    }

    $aMapDef = $this->getMapserverMapDef($aMap, $aLayers, $sFileName, $aFilter, $bPublic);
    $sMapDef = join("\r\n", $aMapDef);

    $this->oLocalFileAccessor->file_put_contents("{$sMsDirPrivate}/{$sFileName}.map", $sMapDef);

    return $sFileName;
  }

  /**
   * Supprime un mapfile
   */
  public function deleteMapfile(array $aMap, $bPublic) {
    $sFileName = $this->generateMapFileName($aMap);
    $sMsDirPrivate = $this->oPropertiesSrvc->replacePropertiesInString($bPublic ? self::PUBLIC_WMS_FOLDER_PATH : self::PRIVATE_WMS_FOLDER_PATH);
    if ($this->oLocalFileAccessor->is_dir($sMsDirPrivate)) {
      $this->oLocalFileAccessor->unlink("{$sMsDirPrivate}/{$sFileName}.map");
    }
  }

  /**
   * Nettoie le nom de la couche en enlevant les espaces et caractères spéciaux
   * @param string $string
   * @return string
   */
  public static function cleanLayerName(string $string): string {
    $string = str_replace(' ', '_', $string); // Replaces all spaces with hyphens.
    return preg_replace('/[^A-Za-z0-9\-_]/', '', $string); // Removes special chars.
  }

  /**
   * Retourne la définition Mapserver de la carte sous forme de tableau PHP
   * @param array $aMap donnée de l'objet carte
   * @param array $aLayers donnée lié à la couche
   * @param string $sMapfileName nom du mapfile (aléatoire)
   * @param array $aFilter filtre à appliquer sur les couches
   * @return array définition du mapfile
   */
  private function getMapserverMapDef(array $aMap, array $aLayers, string $sMapfileName, ?array $aFilter = null, bool $bPublic = false): array{
    $sMsDir = $this->oPropertiesSrvc->get('vmap_ms_dir');
    $sMapefileMaxsize = $this->oPropertiesSrvc->get('mapfile_maxsize');
    $sMapefileDebug = $this->oPropertiesSrvc->get('mapfile_debug');

    $aMapDef = [];
    $aMapDef[] = "MAP";
    $aMapDef[] = "NAME \"{$aMap['name']}\"";
    $aMapDef[] = "MAXSIZE {$sMapefileMaxsize}";

    // MAP CONFIG
    $aMapConfigs = $this->getMapServerConfigs($sMapfileName);
    $aMapDef = array_merge($aMapDef, $aMapConfigs);

    // MAP DEBUG
    if (array_key_exists('debuglevel', $aMap) && !empty($aMap['debuglevel'])) {
      $aMapDef[] = "DEBUG {$aMap['debuglevel']}";
    } else {
      $aMapDef[] = "DEBUG {$sMapefileDebug}";
    }

    // MAP EXTENT
    $aMapDef[] = "EXTENT -378305 6093283 1212610 7186901";

    // WEB METADATA MAP
    $aMapDef[] = 'WEB';
    $aMapDef[] = 'METADATA';
    $aMapDef[] = "\"wms_title\" \"{$aMap['name']}\"";
    $sOnlineResourceUrl = str_replace('@file', $sMapfileName, $bPublic ? self::PUBLIC_WMS_MAIN_URL : self::PRIVATE_WMS_MAIN_URL);
    $sOnlineResourceUrl = $this->oPropertiesSrvc->replacePropertiesInString($sOnlineResourceUrl);
    $aMapDef[] = "\"wms_onlineresource\"  \"{$sOnlineResourceUrl}\"";
    $aMapDef[] = '"wms_srs" "EPSG:2154"';
    $aMapDef[] = '"wms_enable_request"  "*"';
    $aMapDef[] = 'END';
    $aMapDef[] = 'END';

    // MAP DEFAULT PROJECTION
    $aMapDef[] = "PROJECTION \"+init=epsg:2154\" END";

    // MAP FONTSET LIST
    $aMapDef[] = "FONTSET \"{$sMsDir}/fonts/fonts.list\"";

    // MAP LAYERS
    $aMapLayers = $this->getMapServerLayers($aLayers, $aFilter);
    $aMapDef = array_merge($aMapDef, $aMapLayers);

    // MAP SYMBOLS LAYERS
    $aMapLayersSymbols = $this->getLayersSymbolsDef($aLayers);
    foreach ($aMapLayersSymbols as $symbolDef) {
      $aMapDef = array_merge($aMapDef, [$symbolDef]);
    }

    $aMapDef[] = 'END';

    return $aMapDef;
  }

  /**
   * Renvoie les configurations du Mapfile
   */
  private function getMapServerConfigs($sMapfileName): array{
    $sMsDir = $this->oPropertiesSrvc->get('vmap_ms_dir');
    $sLogPath = $this->createMSLogFile($sMapfileName);

    $aMapConfigs = [];

    $aMapConfigs[] = "CONFIG \"MS_ERRORFILE\" \"{$sLogPath}\"";
    $aMapConfigs[] = "CONFIG \"PROJ_LIB\" \"{$sMsDir}/proj\"";

    $sKeyPath = $this->getMsEncryptKey();
    if (!empty($sKeyPath)) {
      $aMapConfigs[] = "CONFIG \"MS_ENCRYPTION_KEY\" \"{$sKeyPath}\"";
    }

    return $aMapConfigs;
  }

  /**
   * Renvoie les couches de la map
   *
   * @param array $aLayers couches
   * @param array $aFilter filtre à appliquer sur les couches
   */
  private function getMapServerLayers($aLayers, ?array $aFilter = null): array{
    $aMapLayers = [];

    foreach ($aLayers as $key => $aLayer) {
      $aLayerDef = [];
      switch ($aLayer['layertype']) {
        case LayerPostgres::TYPE:{
            if (isset($aLayer['ms_geomtype'])) {
              if ($aLayer['ms_geomtype']['label'] === "GEOMETRY") {
                if (is_string($aLayer['definition_sym_json'])) {
                  $aLayer['definition_sym_json'] = json_decode($aLayer['definition_sym_json'], true);
                } else if (is_array($aLayer['definition_sym_json'])) {
                  $aLayer['definition_sym_json'] = $aLayer['definition_sym_json'];
                } else {
                  $aLayer['definition_sym_json'] = [];
                }
                $aLayerGeometry = $aLayer;
                foreach (['point', 'line', 'polygon'] as $geomtype) {
                  if (array_key_exists($geomtype, $aLayer['definition_sym_json'])) {
                    $aLayerGeometry['name'] = "{$aLayer['name']}_{$geomtype}";
                    $aLayerGeometry['ms_geomtype']['label'] = strtoupper($geomtype);
                    $aLayerGeometry['definition_sym_json'] = $aLayer['definition_sym_json'][$geomtype];
                    $aLayerDef = array_merge($aLayerDef, $this->getMSLayerPostgresDef($aLayerGeometry, true, $aFilter));
                  }
                }
              } else {
                $aLayerDef = $this->getMSLayerPostgresDef($aLayer, array_key_exists("multiGeom", $aLayer), $aFilter);
              }
            }
          }
          break;
        case LayerFile::TYPE:{
            $aLayerDef = $this->getMSLayerRasterDef($aLayer);
          }
          break;
      }
      if (!empty($aLayerDef)) {
        $aMapLayers = array_merge($aMapLayers, $aLayerDef);
      }
    }

    return $aMapLayers;
  }

  /**
   * Retourne la connexion d'une couche
   */
  private function getMSLayerConnexion($aLayer): array{
    $aConnection = [
      'pg_login' => $aLayer['datasource']['pg_login'],
      'pg_password' => $aLayer['datasource']['pg_password'],
      'pg_database' => $aLayer['database'],
      'pg_server' => $aLayer['datasource']['pg_server'],
      'pg_port' => $aLayer['datasource']['pg_port'],
    ];
    if ($aLayer['datasource']['pg_public'] === false) {
      $aConnection = $this->getPrivateConnection($aLayer);
    }

    $sKeyPath = '';
    if (!empty($this->sMsEncryptPath)) {
      $sKeyPath = $this->getMsEncryptKey();
      if (!empty($sKeyPath)) {
        $sEncryptedPass = $this->msEncryptPass($aConnection['pg_password'], $sKeyPath);
        if (!empty($sEncryptedPass)) {
          $aConnection['pg_password'] = '{' . $sEncryptedPass . '}';
        }
      }
    }

    return [
      "CONNECTIONTYPE postgis",
      "CONNECTION \"user='{$aConnection['pg_login']}' password='{$aConnection['pg_password']}' dbname='{$aConnection['pg_database']}' host='{$aConnection['pg_server']}' port='{$aConnection['pg_port']}'\"",
    ];
  }

  /**
   * Retourne les propriétés necessare d'une Metadata d'une couche
   */
  private function getMSLayerMetadata($aLayerName, $aLayerCleanName): array{
    return [
      'METADATA',
      "\"wms_title\"      \"{$aLayerName}\"",
      '"wms_srs"      "EPSG:3857 EPSG:2154 EPSG:4326"',
      "\"wms_name\"     \"{$aLayerCleanName}\"",
      '"wms_server_version"	"1.3.0"',
      '"wms_format" 			"image/png"',
      '"wms_enable_request"     "*"',
      '"wms_extent"			"-357823.2365 6037008.6939 1313632.3628 7230727.3772"',
      'END',
    ];
  }

  /**
   * Retourne la propriété Data de la couche si il existe une connexion externe
   */
  private function getMSLayerData($aLayer, $sGeomType, $multiGeom, ?array $aFilter = null): string {
    $sDatafilter = $this->getMSLayerDataFilter($aLayer, $aFilter);
    if ($multiGeom) {
      if ($aLayer['is_multi_edit'] === true) {
        $sGeomType = "MULTI{$sGeomType}";
      }
      $specialGeomtypes = [
        ['type' => 'POLYGON', 'ST_geometrytype' => "'ST_Polygon'"],
        ['type' => 'MULTIPOLYGON', 'ST_geometrytype' => "'ST_MultiPolygon'"],
        ['type' => 'POINT', 'ST_geometrytype' => "'ST_Point'"],
        ['type' => 'MULTIPOINT', 'ST_geometrytype' => "'ST_MultiPoint'"],
        ['type' => 'LINE', 'ST_geometrytype' => "'ST_LineString'"],
        ['type' => 'MULTILINE', 'ST_geometrytype' => "'ST_MultiLineString'"],
      ];
      foreach ($specialGeomtypes as $geomtype) {
        if ($sGeomType === $geomtype['type']) {
          if (!empty($sDatafilter)) {
            $sDatafilter = "AND (${sDatafilter})";
          }
          return "DATA \"{$aLayer['geomfield']} from (select * from \\\"{$aLayer['schema']}\\\".\\\"{$aLayer['pg_table']}\\\" where ST_geometrytype(geom) = {$geomtype['ST_geometrytype']} ${sDatafilter}) as subquery USING SRID={$aLayer['coordsys_id']} USING UNIQUE {$aLayer['idfield']}\"";
        }
      }
    } else {
      if (!empty($sDatafilter)) {
        $sDatafilter = "WHERE ${sDatafilter}";
      }

      return "DATA \"{$aLayer['geomfield']} FROM (SELECT * FROM \\\"{$aLayer['schema']}\\\".\\\"{$aLayer['pg_table']}\\\" ${sDatafilter}) as subquery USING SRID={$aLayer['coordsys_id']} USING UNIQUE {$aLayer['idfield']}\"";
    }
  }

  /**
   * Retourne la clause where en fonction d'un filtre
   */
  private function getMSLayerDataFilter($aLayer, ?array $aFilter = null) {
    $sDatafilter = '';
    if (is_array($aFilter) && count($aFilter) > 0) {
      foreach ($aFilter as $key => $value) {
        if (array_key_exists('layer_id', $value) &&
          array_key_exists('filter', $value)) {
          if ($value['layer_id'] == $aLayer['layer_id']) {
            if (array_key_exists('json_columns', $aLayer) && !empty($aLayer['json_columns'])) {
              $aColumns = $this->parseColumns($aLayer['json_columns']);

              if (is_array($aColumns) && count($aColumns) > 0) {
                // Instancie l'accessor
                $sEntityAlias = $aLayer['pg_table'];
                $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();
                $oCustomFieldsManager = new CustomFieldsManager($this->oPropertiesSrvc, $this->oRouteConfigurator, $this->oLoggerManager, $this->oJweTokenManager, $oEntityManager);
                $oAccessorFunc = new CommonAccessorFunctions($sEntityAlias, $this->oRouteConfigurator, $this->oLoggerManager->get(static::LOGGER_KEY), 'QUERY', $oEntityManager, $oCustomFieldsManager);

                // Config de la route originale (route mapjsons)
                $aOriginalFieldsConfig = $oAccessorFunc->getFieldsConfig();

                // Config de la couche
                $oAccessorFunc->setFieldsConfig($aColumns);

                // Décodage du filtre vers SQL
                $sDecoded = $oAccessorFunc->decodeJSONFilter($value['filter'], $aLayer['schema'], $aLayer['pg_table']);
                $sDatafilter = $this->filterToSQL($sDecoded, $aLayer['schema'], $aLayer['pg_table'], $aColumns);

                // Config de la couche
                $oAccessorFunc->setFieldsConfig($aOriginalFieldsConfig);
              }
            }
          }
        }
      }
    }
    return $sDatafilter;
  }

  /**
   * Retourne la définition mapserver d'une couche postgres de type de géométrie raster
   */
  public function getMSLayerRasterDef($aLayer): array{
    if (is_string($aLayer['definition_sym_json'])) {
      $aLayer['definition_sym_json'] = json_decode($aLayer['definition_sym_json'], true);
    } else if (is_array($aLayer['definition_sym_json'])) {
      $aLayer['definition_sym_json'] = $aLayer['definition_sym_json'];
    } else {
      $aLayer['definition_sym_json'] = [];
    }
    $aSymDef = $aLayer['definition_sym_json'];
    $sNameClean = $this->cleanLayerName($aLayer['name']);

    $aMSLayerDef = [];
    $aMSLayerDef[] = "LAYER";
    $aMSLayerDef[] = "NAME \"{$sNameClean}\"";

    // LAYER TYPE
    $aMSLayerDef[] = "TYPE RASTER";

    // PROJECTION
    $aMSLayerDef[] = "PROJECTION \"+init=epsg:{$aLayer['coordsys_id']}\" END";

    // DATA
    if (isset($aLayer['fileformat']) && $aLayer['fileformat']['label'] === 'shp') {
      $aMSLayerDef[] = "TILEINDEX \"{$aLayer['file_path']}\"";
      $aMSLayerDef[] = "TILEITEM \"LOCATION\"";
    } else {
      $aMSLayerDef[] = "DATA \"{$aLayer['file_path']}\"";
    }

    // METADATA
    $aLayerMetadata = $this->getMSLayerMetadata($aLayer['name'], $sNameClean);
    $aMSLayerDef = array_merge($aMSLayerDef, $aLayerMetadata);

    // OPACITY
    if (array_key_exists('opacity', $aSymDef) && is_numeric($aSymDef['opacity'])) {
      $aMSLayerDef[] = "OPACITY " . $aSymDef['opacity'] * 100;
    } else {
      $aMSLayerDef[] = "OPACITY 100";
    }

    // MAXSCALEDENOM
    if (array_key_exists('maxscale', $aSymDef) && !empty($aSymDef['maxscale'])) {
      $aMSLayerDef[] = "MAXSCALEDENOM {$aSymDef['maxscale']}";
    }

    // MINSCALEDENOM
    if (array_key_exists('minscale', $aSymDef) && !empty($aSymDef['minscale'])) {
      $aMSLayerDef[] = "MINSCALEDENOM {$aSymDef['minscale']}";
    }

    $aMSLayerDef[] = "END";

    return $aMSLayerDef;
  }

  /**
   * Retourne la définition mapserver d'une couche postgres de type de géométrie autre que raster
   */
  public function getMSLayerPostgresDef($aLayer, $multiGeom = false, ?array $aFilter = null): array{
    if (is_string($aLayer['definition_sym_json'])) {
      $aLayer['definition_sym_json'] = json_decode($aLayer['definition_sym_json'], true);
    } else if (is_array($aLayer['definition_sym_json'])) {
      $aLayer['definition_sym_json'] = $aLayer['definition_sym_json'];
    } else {
      $aLayer['definition_sym_json'] = [];
    }
    $aSymDef = $aLayer['definition_sym_json'];

    $sNameClean = $this->cleanLayerName($aLayer['name']);

    if (isset($aSymDef['geomType'])) {
      $sGeomType = strtoupper($aSymDef['geomType']);
    } else {
      $sGeomType = $aLayer['ms_geomtype']['label'];
    }

    $aMSLayerDef = [];
    $aMSLayerDef[] = "LAYER";
    $aMSLayerDef[] = "NAME \"{$sNameClean}\"";

    // LAYER TYPE
    $aMSLayerDef[] = "TYPE {$sGeomType}";

    // PROJECTION
    $aMSLayerDef[] = "PROJECTION \"+init=epsg:{$aLayer['coordsys_id']}\" END";

    // CONNEXION
    $aLayerConnexion = $this->getMSLayerConnexion($aLayer);
    $aMSLayerDef = array_merge($aMSLayerDef, $aLayerConnexion);

    // DATA
    $aMSLayerDef[] = $this->getMSLayerData($aLayer, $sGeomType, $multiGeom, $aFilter);

    // METADATA
    $aLayerMetadata = $this->getMSLayerMetadata($aLayer['name'], $sNameClean);
    $aMSLayerDef = array_merge($aMSLayerDef, $aLayerMetadata);

    // OPACITY
    if (array_key_exists('opacity', $aSymDef) && is_numeric($aSymDef['opacity'])) {
      $aMSLayerDef[] = "OPACITY " . $aSymDef['opacity'] * 100;
    } else {
      $aMSLayerDef[] = "OPACITY 100";
    }

    // MAXSCALEDENOM
    if (array_key_exists('maxscale', $aSymDef) && !empty($aSymDef['maxscale'])) {
      $aMSLayerDef[] = "MAXSCALEDENOM {$aSymDef['maxscale']}";
    }

    // MINSCALEDENOM
    if (array_key_exists('minscale', $aSymDef) && !empty($aSymDef['minscale'])) {
      $aMSLayerDef[] = "MINSCALEDENOM {$aSymDef['minscale']}";
    }

    if (!empty($aSymDef) &&
      $aSymDef['type'] === 'singleclass') {
      $aLayerClassDef = $this->getMapserverLayerClassDef($aSymDef, $aLayer, $aSymDef['type']);
      if (!empty($aLayerClassDef)) {
        $aMSLayerDef = array_merge($aMSLayerDef, $aLayerClassDef);
      }
    } else if (!empty($aSymDef) &&
      $aSymDef['type'] === 'expression' &&
      is_array($aSymDef['classes'])) {
      foreach ($aSymDef['classes'] as $oClassSymDef) {
        if (!empty($oClassSymDef['expression'])) {
          $aLayerClassDef = $this->getMapserverLayerClassDef($oClassSymDef, $aLayer, $aSymDef['type']);
          if (!empty($aLayerClassDef)) {
            $aMSLayerDef = array_merge($aMSLayerDef, $aLayerClassDef);
          }
        }
      }
    }

    $aMSLayerDef[] = "END";

    return $aMSLayerDef;
  }

  /**
   * Trouve la définition des symboles utilisés dans les couches
   */
  public function getLayersSymbolsDef($aLayers): array{
    $aSymbols = [];
    foreach ($aLayers as $aLayer) {
      if ($aLayer['layertype'] === LayerPostgres::TYPE) {
        if (isset($aLayer['ms_geomtype'])) {
          if (is_string($aLayer['definition_sym_json'])) {
            $aLayer['definition_sym_json'] = json_decode($aLayer['definition_sym_json'], true);
          } else if (is_array($aLayer['definition_sym_json'])) {
            $aLayer['definition_sym_json'] = $aLayer['definition_sym_json'];
          } else {
            $aLayer['definition_sym_json'] = [];
          }
          $aLayerSymDef = $aLayer['definition_sym_json'];
          if ($aLayer['ms_geomtype']['label'] === "GEOMETRY") {
            foreach (['point', 'line', 'polygon'] as $geomtype) {
              if (array_key_exists($geomtype, $aLayerSymDef)) {
                $aLayerGeometry['name'] = "{$aLayer['name']}_{$geomtype}";
                $aLayerGeometry['ms_geomtype']['label'] = strtoupper($geomtype);
                $aLayerGeometry['definition_sym_json'] = $aLayerSymDef[$geomtype];
                $aSymbols = array_merge($aSymbols, $this->getLayerDefSymbols($aLayerGeometry['definition_sym_json']));
              }
            }
          } else {
            $aSymbols = array_merge($aSymbols, $this->getLayerDefSymbols($aLayerSymDef));
          }
        }
      }
    }

    $aSymbolsResults = [];
    if (!empty($aSymbols)) {
      try {
        $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder('a');
        $qb->select('a.definition')
          ->from('[SCHEMA].symbol', 'a');
        $qb->where($qb->expr()->in('a.name', '[SYMBOLS]'));
        $aParams = [
          'SCHEMA' => ['value' => $this->oPropertiesSrvc->get('schema_vmap'), 'type' => 'schema_name'],
          'SYMBOLS' => ['value' => $aSymbols, 'type' => 'array'],
        ];
        $aSymbolsResults = $qb->executeWithParams($aParams)->fetchAll();
      } catch (\Exception $e) {
        $this->oLoggerManager->get('API_ERROR')->error('Cannot request symbols');
      }
    }

    $aSymbolsDef = [];
    foreach ($aSymbolsResults as $key => $symbolDef) {
      $aSymbolsDef[] = $aSymbolsResults[$key]['definition'];
    }

    return $aSymbolsDef;
  }

  /**
   * Parcourt la symbologie et retourne les symboles trouvés
   */
  public function getLayerDefSymbols($aLayerSymDef): array{
    $aSymbols = [];

    if (array_key_exists('styles', $aLayerSymDef)) {
      foreach ($aLayerSymDef['styles'] as $style) {
        if (isset($style['symbol']) && !(isset($style['hasStyle']) && $style['hasStyle'] === false)) {
          $aSymbols[] = $style['symbol'];
        }
      }
    }

    if (array_key_exists('classes', $aLayerSymDef)) {
      foreach ($aLayerSymDef['classes'] as $aClassDef) {
        $aClassSymbols = $this->getLayerDefSymbols($aClassDef);
        if (!empty($aClassSymbols)) {
          $aSymbols = array_merge($aSymbols, $aClassSymbols);
        }
      }
    }

    return $aSymbols;
  }

  /**
   * Retourne la définition d'une classe
   * @param array $aLayerLabelDef
   * @return array
   */
  private function getMapserverLayerClassDef(array $aSymDef, array $aLayer, string $sType): array{
    $sNameClean = $this->cleanLayerName($aLayer['name']);

    // Particulatités classes multiples
    if ($sType === 'expression') {
      $sNameClean = $aSymDef['legendName'];
    }

    $aResult[] = 'CLASS';
    $aResult[] = "NAME \"{$sNameClean}\"";

    if (array_key_exists('expression', $aSymDef)) {
      $sExpression = is_array($aSymDef['expression']) ? $this->parseExpression($aSymDef['expression']) : $aSymDef['expression'];
      $aResult[] = "EXPRESSION ( {$sExpression} ) ";
    }

    if (array_key_exists('maxscale', $aSymDef) && !empty($aSymDef['maxscale'])) {
      $aResult[] = "MAXSCALEDENOM " . $aSymDef['maxscale'];
    }

    if (array_key_exists('minscale', $aSymDef) && !empty($aSymDef['minscale'])) {
      $aResult[] = "MINSCALEDENOM " . $aSymDef['minscale'];
    }

    if (array_key_exists('label', $aSymDef) &&
      array_key_exists('hasLabel', $aSymDef)) {
      if ($aSymDef['label'] &&
        $aSymDef['hasLabel'] !== false) {
        $aMSLayerLabelDef = $this->getMapserverLayerLabelDef($aSymDef['label']);
        if (!empty($aMSLayerLabelDef)) {
          $aResult = array_merge($aResult, $aMSLayerLabelDef);
        }
      }
    }

    if (array_key_exists('styles', $aSymDef)) {
      if ($aSymDef['styles']) {
        if (is_array($aSymDef['styles'])) {
          foreach ($aSymDef['styles'] as $value) {
            if (!(isset($value['hasStyle']) && $value['hasStyle'] === false)) {
              $aMSLayerStyleDef = $this->getMapserverLayerStyleDef($value);
              if (!empty($aMSLayerStyleDef)) {
                $aResult = array_merge($aResult, $aMSLayerStyleDef);
              }
            }
          }
        }
      }
    }

    $aResult[] = 'END';
    return $aResult;
  }

  /**
   * Retourne la définition d'un label
   * @param array $aLayerLabelDef
   * @return array
   */
  private function getMapserverLayerLabelDef(array $aLayerLabelDef): array{
    $aResult = [];
    if (!empty($aLayerLabelDef)) {
      $aResult[] = 'LABEL';
      if (array_key_exists('text', $aLayerLabelDef) && !empty($aLayerLabelDef['text'])) {
        $aResult[] = "TEXT \"{$aLayerLabelDef['text']}\"";
      }
      if (array_key_exists('minscale', $aLayerLabelDef) && !empty($aLayerLabelDef['minscale'])) {
        $aResult[] = "MINSCALEDENOM {$aLayerLabelDef['minscale']}";
      }
      if (array_key_exists('maxscale', $aLayerLabelDef) && !empty($aLayerLabelDef['maxscale'])) {
        $aResult[] = "MAXSCALEDENOM {$aLayerLabelDef['maxscale']}";
      }
      if (array_key_exists('font', $aLayerLabelDef) && !empty($aLayerLabelDef['font'])) {
        $aResult[] = "FONT \"{$aLayerLabelDef['font']}\"";
      }
      if (array_key_exists('size', $aLayerLabelDef) && !empty($aLayerLabelDef['size'])) {
        $aResult[] = "SIZE {$aLayerLabelDef['size']}";
      }
      if (array_key_exists('fillcolor', $aLayerLabelDef) && !empty($aLayerLabelDef['fillcolor'])) {
        if (strpos($aLayerLabelDef['fillcolor'], '[') === false) {
          $aResult[] = "COLOR \"{$aLayerLabelDef['fillcolor']}\"";
        } else {
          $aResult[] = "COLOR {$aLayerLabelDef['fillcolor']}";
        }
      }
      if (array_key_exists('outlineColor', $aLayerLabelDef) && !empty($aLayerLabelDef['outlineColor'])) {
        if (strpos($aLayerLabelDef['outlineColor'], '[') === false) {
          $aResult[] = "OUTLINECOLOR \"{$aLayerLabelDef['outlineColor']}\"";
        } else {
          $aResult[] = "OUTLINECOLOR {$aLayerLabelDef['outlineColor']}";
        }
      }
      if (array_key_exists('outlineWidth', $aLayerLabelDef) && !empty($aLayerLabelDef['outlineWidth'])) {
        $aResult[] = "OUTLINEWIDTH {$aLayerLabelDef['outlineWidth']}";
      }
      if (array_key_exists('rotation', $aLayerLabelDef) && !empty($aLayerLabelDef['rotation'])) {
        $aResult[] = "ANGLE {$aLayerLabelDef['rotation']}";
      }
      if (array_key_exists('position', $aLayerLabelDef) && !empty($aLayerLabelDef['position'])) {
        $aResult[] = "POSITION {$aLayerLabelDef['position']}";
      }

      if (array_key_exists('align', $aLayerLabelDef) && !empty($aLayerLabelDef['align'])) {
        $aResult[] = "ALIGN {$aLayerLabelDef['align']}";
      }
      if (array_key_exists('type', $aLayerLabelDef) && !empty($aLayerLabelDef['type'])) {
        $aResult[] = "TYPE {$aLayerLabelDef['type']}";
      }
      if (array_key_exists('shadowcolor', $aLayerLabelDef) && !empty($aLayerLabelDef['shadowcolor'])) {
        if (strpos($aLayerLabelDef['shadowcolor'], '[') === false) {
          $aResult[] = "SHADOWCOLOR \"{$aLayerLabelDef['shadowcolor']}\"";
        } else {
          $aResult[] = "SHADOWCOLOR {$aLayerLabelDef['shadowcolor']}";
        }
      }
      $haveBackgroundColor = array_key_exists('backgroundcolor', $aLayerLabelDef) && !empty($aLayerLabelDef['backgroundcolor']);
      $haveBackgroundShadowSize = array_key_exists('backgroundshadowsize', $aLayerLabelDef) && !empty($aLayerLabelDef['backgroundshadowsize']);
      if ($haveBackgroundColor || $haveBackgroundShadowSize) {
        $aResult[] = 'STYLE';
        $aResult[] = 'GEOMTRANSFORM labelpoly';
        if ($haveBackgroundColor) {
          if (strpos($aLayerLabelDef['backgroundcolor'], '[') === false) {
            $aResult[] = "COLOR \"{$aLayerLabelDef['backgroundcolor']}\"";
          } else {
            $aResult[] = "COLOR {$aLayerLabelDef['backgroundcolor']}";
          }
        }
        if ($haveBackgroundShadowSize) {
          $aResult[] = "OFFSET {$aLayerLabelDef['backgroundshadowsize']}";
        }
        $aResult[] = 'END';
      }
      if (array_key_exists('shadowsize', $aLayerLabelDef) && !empty($aLayerLabelDef['shadowsize'])) {
        $aResult[] = "SHADOWSIZE {$aLayerLabelDef['shadowsize']}";
      }
      if (array_key_exists('priority', $aLayerLabelDef) && !empty($aLayerLabelDef['priority'])) {
        $aResult[] = "PRIORITY {$aLayerLabelDef['priority']}";
      }
      if (array_key_exists('buffer', $aLayerLabelDef) && !empty($aLayerLabelDef['buffer'])) {
        $aResult[] = "BUFFER {$aLayerLabelDef['buffer']}";
      }
      if (array_key_exists('maxlength', $aLayerLabelDef) && !empty($aLayerLabelDef['maxlength'])) {
        $aResult[] = "MAXLENGTH {$aLayerLabelDef['maxlength']}";
      }
      if (array_key_exists('maxsize', $aLayerLabelDef) && !empty($aLayerLabelDef['maxsize'])) {
        $aResult[] = "MAXSIZE {$aLayerLabelDef['maxsize']}";
      }
      if (array_key_exists('mindistance', $aLayerLabelDef) && !empty($aLayerLabelDef['mindistance'])) {
        $aResult[] = "MINDISTANCE {$aLayerLabelDef['mindistance']}";
      }
      if (array_key_exists('minfeaturesize', $aLayerLabelDef) && !empty($aLayerLabelDef['minfeaturesize'])) {
        $aResult[] = "MINFEATURESIZE {$aLayerLabelDef['minfeaturesize']}";
      }
      if (array_key_exists('minsize', $aLayerLabelDef) && !empty($aLayerLabelDef['minsize'])) {
        $aResult[] = "MINSIZE {$aLayerLabelDef['minsize']}";
      }
      if (array_key_exists('outlinewidth', $aLayerLabelDef) && !empty($aLayerLabelDef['outlinewidth'])) {
        $aResult[] = "OUTLINEWIDTH {$aLayerLabelDef['outlinewidth']}";
      }
      if (array_key_exists('repeatdistance', $aLayerLabelDef) && !empty($aLayerLabelDef['repeatdistance'])) {
        $aResult[] = "REPEATDISTANCE {$aLayerLabelDef['repeatdistance']}";
      }
      if (array_key_exists('antialias', $aLayerLabelDef) && !empty($aLayerLabelDef['antialias'])) {
        $aResult[] = "ANTIALIAS {$aLayerLabelDef['antialias']}";
      }
      if (array_key_exists('partials', $aLayerLabelDef) && !empty($aLayerLabelDef['partials'])) {
        $aResult[] = "PARTIALS {$aLayerLabelDef['partials']}";
      }
      if (array_key_exists('wrap', $aLayerLabelDef) && !empty($aLayerLabelDef['wrap'])) {
        $aResult[] = "WRAP \"{$aLayerLabelDef['wrap']}\"";
      }
      if (array_key_exists('offset', $aLayerLabelDef) && !empty($aLayerLabelDef['offset'])) {
        $aResult[] = "OFFSET {$aLayerLabelDef['offset']}";
      }
      if (array_key_exists('force', $aLayerLabelDef) && !empty($aLayerLabelDef['force'])) {
        $aResult[] = "FORCE {$aLayerLabelDef['force']}";
      }

      $aResult[] = 'END';
    }

    return $aResult;
  }

  /**
   * Retourne la définition d'un style
   * @param array $aLayerStyleDef
   * @return array
   */
  private function getMapserverLayerStyleDef(array $aLayerStyleDef): array{
    $aResult = [];

    if (!empty($aLayerStyleDef)) {
      $aResult[] = 'STYLE';
      if (array_key_exists('opacity', $aLayerStyleDef) && is_numeric($aLayerStyleDef['opacity'])) {
        $aResult[] = 'OPACITY ' . $aLayerStyleDef['opacity'] * 100;
      } else {
        $aResult[] = 'OPACITY 100';
      }
      if (array_key_exists('maxscale', $aLayerStyleDef) && !empty($aLayerStyleDef['maxscale'])) {
        $aResult[] = "MAXSCALEDENOM {$aLayerStyleDef['maxscale']}";
      }
      if (array_key_exists('minscale', $aLayerStyleDef) && !empty($aLayerStyleDef['minscale'])) {
        $aResult[] = "MINSCALEDENOM {$aLayerStyleDef['minscale']}";
      }
      if (array_key_exists('fillcolor', $aLayerStyleDef) && !empty($aLayerStyleDef['fillcolor'])) {
        if (strpos($aLayerStyleDef['fillcolor'], '[') === false) {
          $aResult[] = "COLOR \"{$aLayerStyleDef['fillcolor']}\"";
        } else {
          $aResult[] = "COLOR {$aLayerStyleDef['fillcolor']}";
        }
      }
      if (array_key_exists('symbol', $aLayerStyleDef) && !empty($aLayerStyleDef['symbol'])) {
        $aResult[] = "SYMBOL \"{$aLayerStyleDef['symbol']}\"";
      }
      if (array_key_exists('outlineColor', $aLayerStyleDef) && !empty($aLayerStyleDef['outlineColor'])) {
        if (strpos($aLayerStyleDef['outlineColor'], '[') === false) {
          $aResult[] = "OUTLINECOLOR \"{$aLayerStyleDef['outlineColor']}\"";
        } else {
          $aResult[] = "OUTLINECOLOR {$aLayerStyleDef['outlineColor']}";
        }
      }
      if (array_key_exists('width', $aLayerStyleDef) && !empty($aLayerStyleDef['width'])) {
        $aResult[] = "WIDTH {$aLayerStyleDef['width']}";
      }
      if (array_key_exists('size', $aLayerStyleDef) && !empty($aLayerStyleDef['size'])) {
        $aResult[] = "SIZE {$aLayerStyleDef['size']}";
      }
      if (array_key_exists('rotation', $aLayerStyleDef) && !empty($aLayerStyleDef['rotation'])) {
        $aResult[] = "ANGLE {$aLayerStyleDef['rotation']}";
      }
      if (array_key_exists('offset', $aLayerStyleDef) && !empty($aLayerStyleDef['offset'])) {
        $aResult[] = "OFFSET {$aLayerStyleDef['offset']}";
      }
      if (array_key_exists('minsize', $aLayerStyleDef) && !empty($aLayerStyleDef['minsize'])) {
        $aResult[] = "MINSIZE {$aLayerStyleDef['minsize']}";
      }
      if (array_key_exists('maxsize', $aLayerStyleDef) && !empty($aLayerStyleDef['maxsize'])) {
        $aResult[] = "MAXSIZE {$aLayerStyleDef['maxsize']}";
      }
      if (array_key_exists('minwidth', $aLayerStyleDef) && !empty($aLayerStyleDef['minwidth'])) {
        $aResult[] = "MINWIDTH {$aLayerStyleDef['minwidth']}";
      }
      if (array_key_exists('maxwidth', $aLayerStyleDef) && !empty($aLayerStyleDef['maxwidth'])) {
        $aResult[] = "MAXWIDTH {$aLayerStyleDef['maxwidth']}";
      }
      if (array_key_exists('geomtransform', $aLayerStyleDef) && !empty($aLayerStyleDef['geomtransform'])) {
        $aResult[] = "GEOMTRANSFORM {$aLayerStyleDef['geomtransform']}";
      }
      if (array_key_exists('gap', $aLayerStyleDef) && !empty($aLayerStyleDef['gap'])) {
        $aResult[] = "GAP {$aLayerStyleDef['gap']}";
      }
      if (array_key_exists('linejoin', $aLayerStyleDef) && !empty($aLayerStyleDef['linejoin'])) {
        $aResult[] = "LINEJOIN {$aLayerStyleDef['linejoin']}";
      }
      if (array_key_exists('initialgap', $aLayerStyleDef) && !empty($aLayerStyleDef['initialgap'])) {
        $aResult[] = "INITIALGAP {$aLayerStyleDef['initialgap']}";
      }
      if (array_key_exists('linecap', $aLayerStyleDef) && !empty($aLayerStyleDef['linecap'])) {
        $aResult[] = "LINECAP {$aLayerStyleDef['linecap']}";
      }
      if (array_key_exists('pattern', $aLayerStyleDef) && !empty($aLayerStyleDef['pattern'])) {
        $aResult[] = "PATTERN {$aLayerStyleDef['pattern']} END";
      }

      $aResult[] = 'END';
    }

    return $aResult;
  }

  /**
   * Retourne une chaine aléatoire
   * @param int $length longueur de la chaine (default 10)
   * @return string
   */
  private function generateRandomString(int $length = 10): string {
    $x = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
    $iMultiplier = (int) ceil($length / strlen($x));
    return substr(str_shuffle(str_repeat($x, $iMultiplier)), 1, $length);
  }

  /**
   * Nom du fichier Map
   */
  private function generateMapFileName($aMap): string {
    if (array_key_exists('flux_id', $aMap)) {
      return $this->cleanLayerName($aMap['name']);
    }
    $sToken = $this->oRouteConfigurator->getRequest()->getToken();
    $sMapId = $aMap['map_id'];
    return hash('sha256', $sToken . '_' . $sMapId);
  }

  /**
   * Retourne le chemin du fichier de log pour un flux mapserver
   * @param string $sFileName nom du fichier mapfile
   * @return string
   */
  private function createMSLogFile(string $sFileName): string {
    $sDateReplacer = date($this->oPropertiesSrvc->get('log_period'));
    $sErrorLogDir = $this->oPropertiesSrvc->replacePropertiesInString(self::PRIVATE_WMS_LOG_PATH);
    $sErrorLogDir = str_replace('@date', $sDateReplacer, $sErrorLogDir);

    // création du dossier pour que que php puisse écrire dans le fichier
    if (!$this->oLocalFileAccessor->is_dir($sErrorLogDir)) {
      $this->oLocalFileAccessor->mkdir($sErrorLogDir, LocalAccessor::DEFAULT_LINUX_PERMISSION, true);
    }
    return $sErrorLogDir . '/' . $sFileName . '.log';
  }

  /**
   * Retourne la connexion privée
   * @param array $aLayer
   * @return array|null
   */
  private function getPrivateConnection(array $aLayer): ?array{
    $aConnection = [
      'pg_login' => '',
      'pg_password' => '',
      'pg_database' => '',
      'pg_server' => '',
      'pg_port' => '',
    ];

    $sToken = $this->oRouteConfigurator->getRequest()->getToken();
    if (empty($sToken)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find token to generate a private connection');
      return null;
    }
    // récupérations des infos du token
    $aTokenJWEDecrypt = $this->oJweTokenManager->decryptToken($sToken);
    if (!empty($aTokenJWEDecrypt)) {
      $aConnection['pg_login'] = $aTokenJWEDecrypt['data']['user']['login'];
      $aConnection['pg_password'] = $aTokenJWEDecrypt['data']['password'];
      $aConnection['pg_database'] = $aLayer['database']; // La base de données peut être autre que celle exploitée par vMap
      $aConnection['pg_server'] = $this->oPropertiesSrvc->get('db_server');
      $aConnection['pg_port'] = $this->oPropertiesSrvc->get('db_port');
    }

    return $aConnection;
  }

  /**
   * Retourne le chemin vers la clé msencrypt ou la génère si elle n'existe pas
   * @return string
   */
  private function getMsEncryptKey(): string {
    $sKeyPath = $this->oPropertiesSrvc->replacePropertiesInString(self::KEY_FOLDER_PATH);

    if (!$this->oLocalFileAccessor->file_exists($sKeyPath)) {
      if (!empty($this->sMsEncryptPath)) {
        $this->execCommand([$this->sMsEncryptPath, "-keygen", $sKeyPath]);
      }
      if ($this->oLocalFileAccessor->file_exists($sKeyPath)) {
        return $sKeyPath;
      }
    }
    return $sKeyPath;
  }

  /**
   * Fonction de cryptage de mot de passe avec msencrypt
   * @param string $sPass
   * @param string $sKeyPath
   * @return string
   */
  private function msEncryptPass(string $sPass, string $sKeyPath): string {
    return $this->execCommand([$this->sMsEncryptPath, "-key", $sKeyPath, $sPass]);
  }

  /**
   * Fonction  générique pour éxécuter une commande
   * @param array $aCommandLine script à appeler
   * @param array $aOptions liste des options à utiliser (optional)
   * @return string dernière ligne de retour de la commande
   * @throws ProcessFailedException si la commande lève une erreur
   */
  private function execCommand(array $aCommandLine, array $aOptions = []): string {
    foreach ($aOptions as $sOption) {
      $aCommandLine[] = $sOption;
    }

    $oProcess = new VitisProcess($aCommandLine, null, $this->oPropertiesSrvc, $this->oLoggerManager);

    return $oProcess->startWithReturn();
  }

  /**
   * Transforme une expression array en texte interprétable par mapserver
   */
  private function parseExpression(array $aExpression): string {
    $sExpression = '';
    foreach ($aExpression as $key => $value) {
      // Nombres
      if ($value['operator'] === '>' || $value['operator'] === '<' || $value['operator'] === '>=' || $value['operator'] === '<=') {
        $sTable = "[{$value['table']}]";
        $sOperator = "{$value['operator']}";
        $sValue = "{$value['value']}";
      } else {
        // Autres types
        $sTable = "\"[{$value['table']}]\"";
        $sOperator = "{$value['operator']}";
        $sValue = "\"{$value['value']}\"";
      }

      $sExpression .= "( {$sTable} {$sOperator} {$sValue} )";

      if ($value['sql_operator'] === 'AND' || $value['sql_operator'] === 'OR') {
        $sExpression .= " {$value['sql_operator']} ";
      }
    }
    return $sExpression;
  }

  /**
   * Générer le fichier fonts.list avec les données en bdd
   */
  public function generateFontList() {
    $fontResults = [];
    try {
      $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder('');
      $qb->select('*')
        ->from('[SCHEMA].font');
      $aParams = [
        'SCHEMA' => ['value' => $this->oPropertiesSrvc->get('schema_vmap'), 'type' => 'schema_name'],
      ];
      $fontResults = $qb->executeWithParams($aParams)->fetchAll();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot request fonts');
    }

    if (!empty($fontResults)) {
      $sMsDirPrivate = $this->oPropertiesSrvc->replacePropertiesInString("@prop(var_dir)/mapserver/fonts");
      $index = 0;
      $contentFontList = array_reduce($fontResults, function ($acc, $font) use ($index, $fontResults, $sMsDirPrivate) {
        if (isset($font) && $font['name'] && $font['ttf']) {
          $ttfExists = $this->oLocalFileAccessor->file_exists("{$sMsDirPrivate}/{$font['font_id']}/{$font['ttf']}");
          if ($ttfExists) {
            $row = "{$font['name']} ./{$font['font_id']}/{$font['ttf']}";
            $acc[] = $index + 1 == count($fontResults) ? $row : $row . "\n";
          }
        }
        $index++;
        return $acc;
      }, []);
      $this->oLocalFileAccessor->file_put_contents("{$sMsDirPrivate}/fonts.list", $contentFontList);
    }
  }

  /**
   * Transforme le filtre Vitis au format SQL tout en protégeant contre les injections SQL
   */
  private function filterToSQL(array $aFilter, string $sSchema, string $sTable, array $aColumns): string {
    $sSQL = '';

    if (array_key_exists('request', $aFilter) &&
      array_key_exists('params', $aFilter)) {
      $sSQL = $aFilter['request'];

      foreach ($aFilter['params'] as $key => $value) {
        if (array_key_exists('value', $value) &&
          array_key_exists('column_name', $value) &&
          array_key_exists('column_type', $value)) {
          if (is_numeric($value['value']) && $value['column_type'] === 'number') {
            $val = $value['value'];
            $sSQL = str_replace(":${key}", $val, $sSQL);
          } elseif (is_string($value['value'])) {
            $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();
            $oConnection = $oEntityManager->getConnection();
            $val = $oConnection->quote($value['value']);
            $sSQL = str_replace(":${key}", "${val}", $sSQL);
          }
        }
      }
    }

    if (!empty($sSQL)) {
      $sSQL = str_replace('"' . $sSchema . '"', '\\"' . $sSchema . '\\"', $sSQL);
      $sSQL = str_replace('"' . $sTable . '"', '\\"' . $sTable . '\\"', $sSQL);
      foreach ($aColumns as $key => $value) {
        if (array_key_exists('name', $value) &&
          !empty($value['name'])) {
          $sSQL = str_replace('"' . $value['name'] . '"', '\\"' . $value['name'] . '\\"', $sSQL);
        }
      }
    }

    return $sSQL;
  }

  /**
   * Parse les colonnes définies dans json_columns et les transforme en liste des colonnes admissibles par CommonAccessorFunctions
   */
  private function parseColumns($sJSONColumns) {
    if (is_string($sJSONColumns) && !empty($sJSONColumns)) {
      $aReturn = [];
      $aColumns = json_decode($sJSONColumns, true);
      foreach ($aColumns as $key => $value) {
        if (array_key_exists('column', $value) &&
          array_key_exists('type', $value)) {
          $aReturn[] = [
            'name' => $value['column'],
            'type' => $value['type'],
          ];
        }
      }

      return $aReturn;
    }
  }
}
