import { Component, OnInit } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { environment } from 'src/environments/environment';
import { VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import SwaggerUI from 'swagger-ui';

@Component({
  selector: 'app-api-doc',
  templateUrl: './api-doc.component.html',
  styleUrls: ['./api-doc.component.scss']
})
export class ApiDocComponent implements OnInit {

  public selectedWebService: any;
  public aWebServices: Array<any> = [];
  public sBasePath: string;


  constructor(
    private route: ActivatedRoute,
    private vitisRequest: VitisRequest,
    private auth: AuthService
  ) { }

  ngOnInit() {
    this.route.data.subscribe(oParams => {
      this.aWebServices = oParams.options.web_services;
      this.sBasePath = oParams.options.base_path;
      // Affiche la première doc
      this.initApiDoc(this.aWebServices[0]);
      // Web service en cours d'utilisation
      this.selectedWebService = this.aWebServices[0];
    });
  }

  /**
   * Affichage de la doc. Swagger de l'API.
   * @param  {object} oWebService
   * @returns string
   */
  public initApiDoc(oWebService: any) {
    // Chargement du fichier de définition de l'API.
    this.loadApiDefinitionFile(oWebService.api_definition_file_path).then((oApiDefinition: any) => {
      // Conversion d'une collection Postman vers Swagger.
      this.initApiDefinition(oApiDefinition, oWebService);
      // Affichage de la doc. Swagger.
      this.initSwaggerUi(oApiDefinition);
      // Web service en cours d'utilisation
      this.selectedWebService = oWebService;
    });
  }

  /**
   * Charge le fichier json de définition de l'API au format Swagger.
   * @param  {string} sSwaggerFilePath
   * @returns string
   */
  public loadApiDefinitionFile(sSwaggerFilePath: string) {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: sSwaggerFilePath
      }).then((resp: any) => {
        if (resp.status === 200) {
          resolve(resp.data);
        } else {
          console.error('Could not get the API definition file from request');
          reject();
        }
      }).catch((err) => {
        reject();
      });
    });
  }

  /**
   * Affichage de la doc. Swagger de l'API.
   * @param  {object} oApiDefinition
   * @returns string
   */
  public initSwaggerUi(oApiDefinition: any) {
    // Test d'utilisation.
    const ui = SwaggerUI({
      dom_id: '#swagger-ui',
      layout: 'BaseLayout',
      presets: [
        SwaggerUI.presets.apis,
        SwaggerUI.SwaggerUIStandalonePreset
      ],
      spec: oApiDefinition,
      docExpansion: 'none',
      operationsSorter: 'alpha',
      tagsSorter: 'alpha',
      onComplete: function () {
      }
    });
  }

  /**
   * Initialisation des paramètres de la doc.
   * @param  {object} oApiDefinition
   * @param  {object} oWebService
   * @returns void
   */
  public initApiDefinition(oApiDefinition: any, oWebService: any) {
    // Host de l'API Rest.
    const iSlashLastIndex = environment.properties.app_api_url.lastIndexOf('/');
    oApiDefinition.host = environment.properties.app_api_url.substring(0, iSlashLastIndex).replace('https://', '');
    oApiDefinition.basePath = this.sBasePath + '/' + oWebService.name.toLowerCase();
    // Version de l'API Rest.
    oApiDefinition.info.version = environment.properties.version;
    // Requetes en https.
    oApiDefinition.schemes = ['https'];
    // Définition du paramètre 'Authorization' avec le token de l'utilisateur.
    this.setAllParameters(oApiDefinition, 'Authorization', this.auth.getToken());
    // Suppression du paramètre 'Accept'.
    this.removeAllParameters(oApiDefinition, 'Accept');
  }

  /**
   * Définition d'un paramètre dans tous les web services.
   * @param  {object} oApiDefinition
   * @param  {string} sParameterName
   * @param  {any} value
   * @returns void
   */
  public setAllParameters(oApiDefinition: any, sParameterName: string, value: any) {
    let oParameter: any, aWebServices: any, aMethods: any, aParameters: Array<any>;
    aWebServices = Object.keys(oApiDefinition.paths);
    for (let i = 0; i < aWebServices.length; i++) {
      aMethods = Object.keys(oApiDefinition.paths[aWebServices[i]]);
      for (let j = 0; j < aMethods.length; j++) {
        aParameters = oApiDefinition.paths[aWebServices[i]][aMethods[j]].parameters;
        for (let k = 0; k < aParameters.length; k++) {
          oParameter = aParameters[k];
          if (oParameter.name == sParameterName) {
            // oParameter.default = value;
            oParameter.value = value;
            break;
          }
        }
      }
    }
  }

  /**
   * Suppression d'un paramètre dans tous les web services.
   * @param  {object} oApiDefinition
   * @param  {string} sParameterName
   * @returns void
   */
  public removeAllParameters(oApiDefinition: any, sParameterName: string) {
    let aWebServices: any, aMethods: any, aParameters: Array<any>;
    aWebServices = Object.keys(oApiDefinition.paths);
    for (let i = 0; i < aWebServices.length; i++) {
      aMethods = Object.keys(oApiDefinition.paths[aWebServices[i]]);
      for (let j = 0; j < aMethods.length; j++) {
        aParameters = oApiDefinition.paths[aWebServices[i]][aMethods[j]].parameters;
        for (let k = 0; k < aParameters.length; k++) {
          // oParameter = aParameters[k];
          if (aParameters[k].name == sParameterName) {
            aParameters.splice(k, 1);
            break;
          }
        }
      }
    }
  }
}
