import { Component, AfterViewInit, OnInit, ViewChild, OnChanges, SimpleChanges, EventEmitter, OnDestroy } from '@angular/core';
import { environment } from 'src/environments/environment';
import { ActivatedRoute } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import ResizeObserver from 'resize-observer-polyfill';
import { vitisAlert, VitisCheck, VitisRequest } from '../core/services/utils/vitis-utils';
import { DataGridOptions } from '../data-grid/data-grid.model';
import { DragAndDropGridOptions } from './draganddrop-grid.model';
import { DataGridComponent } from '../data-grid/data-grid.component';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { Subscription } from 'rxjs';

declare var $: any;
@Component({
  selector: 'app-draganddrop-grid',
  templateUrl: './draganddrop-grid.component.html',
  styleUrls: ['./draganddrop-grid.component.scss']
})
export class DraganddropGridComponent implements OnInit, AfterViewInit, OnChanges, OnDestroy {

  @ViewChild(DataGridComponent, { static: true }) oDataGrid: DataGridComponent;

  // Mode en cours
  public mode: any;
  public oParams: any = null;
  public placeholder: HTMLDivElement;
  public dataTableInstance: any;
  public oFormDefinition: any;
  public oFormValues: any = {};
  public oFormOptions: any;
  public sDataGridModalTitle: string;
  public sButtonSaveText: string = 'DRAG_DROP_GRID_SAVE_BUTTON';
  public sButtonCancelText: string = 'DRAG_DROP_GRID_CANCEL_BUTTON';
  public sButtonAddText: string = 'DRAG_DROP_GRID_ADD_BUTTON';
  public sButtonDeleteText: string = 'DRAG_DROP_GRID_DELETE_BUTTON';
  public sButtonAddElementText: string = 'DRAG_DROP_GRID_ADD_ELEMENT_BUTTON';
  public sLabelNoDataText: string = 'DRAG_DROP_GRID_NO_DATA';

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {},
    datagridOptions: DragAndDropGridOptions
  };

  public oDragDropGridOptions: any = {};

  // Variable pour stocker l'ensemble des lignes sous forme d'objet avec toutes les colonnes
  public oDragAndDropGridRows: Array<any> = [];

  private sIndexColumn: string;
  private sNameColumn: string;
  private sIdColumn: string;
  private iDraggedElementIndex: number;
  private iOnDragElementIndex: number;

  // Types de filtre disponible
  public oFilterTypes: any = {
    custom: 'VITIS_WORKSPACELIST_FILTER_CUSTOM',
    form: 'VITIS_WORKSPACELIST_FILTER_FORM'
  };

  // Filtre en cours
  public oFilter: any = {
    type: 'custom',
    nodes: []
  };

  // Résultat
  public oResults: any = {
    recordsTotal: 0,
    recordsFiltered: 0,
    data: []
  };

  // Event lors que le chargement est terminé
  private oRowsLoadedEmitter: EventEmitter<any> = new EventEmitter();

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  constructor(
    private route: ActivatedRoute,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private auth: AuthService
  ) { }

  ngOnChanges(changes: SimpleChanges): void {
    this.initDataGrid();
  }

  ngOnInit(): void {
    // Permet à oDataGrid de connaitre les colonnes etc..
    // Mode en cours d'utilisation
    this.route.data.subscribe(mode => {
      this.mode = mode;
      if (mode.options) {
        this.oDragDropGridOptions = mode.options;
        this.oDataGridOptions.datagridOptions = mode.options.source;
        this.oDataGridOptions.datatableOptions = mode.options.source;
        this.sIndexColumn = this.oDragDropGridOptions.index_field;
        this.sNameColumn = this.oDragDropGridOptions.name_field;
        this.sIdColumn = this.oDragDropGridOptions.id_field;
        this.oDragDropGridOptions.columns.forEach(element => {
          if (!element.width) {
            element.width = '200px';
          }
          if (element.source) {
            if (element.source === 'current') {
              element.options = [];
              // Appelé à chaque changement de contenu
              this.subscriptions.push(this.oRowsLoadedEmitter.subscribe((aLoadedRows: any) => {
                if (this.vitisCheck.isArray(aLoadedRows)) {
                  element.options = aLoadedRows.map(data => ({ label: data.name, value: data.source_id }));
                }
              }));
            } else {
              element.options = [];
              this.vitisRequest.ajaxRequestPromise({
                method: 'GET',
                url: environment.properties.app_api_url + '/' + element.source
              }).then((resp: any) => {
                const datas = resp.data.data;
                if (datas && datas.length) {
                  element.options = datas.map(data => ({ label: data[element.name_field], value: data[element.id_field] }));
                }
              });
            }
          }
        });
        // Texte optionnel pour la modale de création (si aucun enregistrement dans la liste).
        if (this.vitisCheck.isDefAndNotNull(mode.options.no_record_modal)) {
          // Update du texte à afficher dans l'html
        }
      }
    });
    this.route.params.subscribe(oParams => {
      this.oParams = oParams;
    });
    this.oDataGrid.oDataGridOptions = this.oDataGridOptions;
    // Initialisation du datagrid
    if (this.vitisCheck.isDefAndNotNull(this.oParams) && this.vitisCheck.isDefAndNotNull(this.oDataGridOptions)) {
      this.initRows().then(() => {
        this.initDataGrid();
      });
    }
  }

  ngAfterViewInit() {
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Fonction déclenchée lors d'un début de drag and drop sur un item
   * @param event
   */
  public onDragStart(event, draggedElement) {
    // Tout dans le setTimeout (sinon bug après scroll sur chrome).
    setTimeout(() => {
      this.iDraggedElementIndex = this.oDragAndDropGridRows.indexOf(draggedElement);
      // On récupère la div element-item en cours de drag
      const draggingElement = event.target;

      // On lui applique quelques styles pour qu'il ne soit plus visible pendant le drag
      draggingElement.style.position = 'absolute';
      draggingElement.style.zIndex = -10;

      // On récupère tout le groupe de l'élément
      const element = draggingElement.parentNode;

      // On crée la div de placeholder
      this.placeholder = document.createElement('div');
      this.placeholder.classList.add('placeholder');

      // Insertion du placeholder
      element.parentNode.insertBefore(
        this.placeholder,
        element.nextSibling
      );
    });
  }

  /**
   * Fonction déclenchée lors du survol d'un élément en cours de drag and drop
   * @param event
   */
  public onDragOver(event, onDragElement) {
    let focusedElement: any;

    // const sortableNeighbors = this.getSortableNeighbors(this.sDragElementID);
    const element = event.target.closest('.element');

    // Drag au dessus d'un element
    if (this.vitisCheck.isDefAndNotNull(element) /*&& sortableNeighbors.some(e => e.id === element.id)*/) {
      focusedElement = element;
    }

    const { bottom, top } = focusedElement.getBoundingClientRect();
    const elHeigth = bottom - top;
    const offsetY = bottom - event.clientY;
    // Par en bas
    if (offsetY < elHeigth / 2) {
      focusedElement.parentNode.insertBefore(
        this.placeholder,
        focusedElement
      );
      this.iOnDragElementIndex = this.oDragAndDropGridRows.indexOf(onDragElement);

    } else if (offsetY > elHeigth / 2) {
      // Par en haut
      focusedElement.parentNode.insertBefore(
        this.placeholder,
        focusedElement.nextSibling
      );
      this.iOnDragElementIndex = this.oDragAndDropGridRows.indexOf(onDragElement);
    }
    // On met à jour les indexs dans le tableau pour suivre l'échange visuel sur la page
    if (this.vitisCheck.isDefAndNotNull(this.iOnDragElementIndex) && this.vitisCheck.isDefAndNotNull(this.iDraggedElementIndex)) {
      this.changeArrayIndex(this.oDragAndDropGridRows, this.iDraggedElementIndex, this.iOnDragElementIndex);
      this.iDraggedElementIndex = this.iOnDragElementIndex;
    }
  }

  /**
   * Fonction déclenchée lors du drop d'un élément
   * @param event
   */
  public onDragEnd(event) {
    this.iDraggedElementIndex = null;
    this.iOnDragElementIndex = null;
    const draggingElement = event.target;
    draggingElement.style.removeProperty('position');
    draggingElement.style.removeProperty('z-index');
    // On met l'élément que l'on déplacait à la place du placeholder
    this.placeholder.parentNode.replaceChild(draggingElement.parentNode, this.placeholder);
  }

  /**
   * Fonction déclenchée lorsqu'on quitte le survol d'un élément pendant un drag
   * @param $event
   */
  public onDragLeave($event) {
    // On reset les index pour ne pas changer les index alors qu'ils ne devraient pas
    // this.iDraggedElementIndex = null;
    // this.iOnDragElementIndex = null;
  }

  /**
   * Récupère les données et les ordonnes en fonction de l'index d'affichage
   */
  public initRows(): Promise<any> {
    return new Promise((resolve, reject) => {
      if (this.vitisCheck.isDefAndNotNull(this.oParams) && this.vitisCheck.isDefAndNotNull(this.oDragDropGridOptions)) {
        const sRoute = this.oDragDropGridOptions.ressource.replace(':id', this.oParams.id);
        // Récupère les données à afficher
        this.vitisRequest.ajaxRequestPromise({
          method: 'GET',
          url: environment.properties.app_api_url + '/' + sRoute
        }).then((resp: any) => {
          if (resp.data) {
            if (resp.data.status === 1) {
              let sElementName: string = '';
              let sId: string;
              let sSourceId: string;
              // Reset
              this.oDragAndDropGridRows = [];
              if (this.oDragDropGridOptions.reversed === true) {
                resp.data.data.sort((a, b) => (a[this.oDragDropGridOptions.index_field] > b[this.oDragDropGridOptions.index_field]) ? 1 : -1).reverse();
              } else {
                resp.data.data.sort((a, b) => (a[this.oDragDropGridOptions.index_field] > b[this.oDragDropGridOptions.index_field]) ? 1 : -1);
              }
              // Tri du retour en fonction de l'index
              // Pour chaque élément retourné par la requête
              for (const oElement of resp.data.data) {
                sElementName = oElement[this.sNameColumn];
                sId = oElement[this.sIdColumn];
                sSourceId = oElement[this.oDataGridOptions.datagridOptions.id_field];

                // Création de l'élément insérer
                const oElementToPush: any = {
                  name: sElementName,
                  label: sElementName,
                  id: sId,
                  source_id: sSourceId,
                  selected: false,
                  columns: {}
                };

                // Pour chaque colonne spécifiée
                for (const oColumn of this.oDragDropGridOptions.columns) {
                  // Si la colonne est connue
                  if (oColumn.field in oElement || oColumn.type === 'select') {
                    if ([this.sIdColumn, this.sNameColumn, this.sIndexColumn].indexOf(oColumn.field) === -1) {
                      if ('enabled' in oColumn) {
                        const sSourceValue = oElement[oColumn.enabled.column];
                        oElementToPush.columns[oColumn.enabled.column] = sSourceValue;
                      }
                      oElementToPush.columns[oColumn.field] = oElement[oColumn.field];
                    }
                  }
                }
                this.oDragAndDropGridRows.push(oElementToPush);
              }
            }
          }
          this.oRowsLoadedEmitter.next(this.oDragAndDropGridRows);
          resolve(this.oDragAndDropGridRows);
        });
      }
    });
  }

  /**
   * Mise à jour des données en base
   */
  public updateData() {
    const oElementToSend: Array<Object> = [];
    let i: number = 1;

    // On formate l'objet en paramètre
    this.oDragAndDropGridRows.forEach(row => {
      const oItem: any = {};
      if (this.oDragDropGridOptions.reversed === true) {
        oItem[this.sIndexColumn] = this.oDragAndDropGridRows.length - (i - 1);
      } else {
        oItem[this.sIndexColumn] = i;
      }
      oItem[this.sIdColumn] = row.id;

      this.oDragDropGridOptions.columns.forEach((column: any) => {
        oItem[column.field] = row.columns[column.field];
      });
      oElementToSend.push(oItem);
      i++;
    });
    const sRoute = this.oDragDropGridOptions.ressource.replace(':id', this.oParams.id);
    // Requête pour update
    this.vitisRequest.ajaxRequestPromise({
      method: 'PUT',
      url: environment.properties.app_api_url + '/' + sRoute,
      data: { data: oElementToSend }
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          // Pop-up ok
          this.translate.get('DRAG_DROP_GRID_EDIT_OK').subscribe(async (translation) => {
            vitisAlert.alert({
              type: 'success',
              text: translation
            });
          });
        }
      }
    });
  }

  /**
   * Reset des données affichées à partir de celles en base
   */
  public cancelChangesOnData() {
    this.initRows();
  }

  /**
   * Change l'index d'un objet dans un tableau
   * @param array Tableau d'objets
   * @param fromIndex Index de départ
   * @param toIndex Index d'arrivé
   */
  private changeArrayIndex(array, fromIndex, toIndex) {
    const element = array[fromIndex];
    array.splice(fromIndex, 1);
    array.splice(toIndex, 0, element);
  }

  /**
   * Initialise la grille de données.
   * @param {oDataGrid} DataGridComponent
   * @returns void
   */
  private initDataGrid() {
    let rowId = '';
    // Ordre
    let order = [1, 'asc'];

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datatableOptions)) {

      /**
      * Calcule la hauteur de la table
      */
      const calcDataTableHeight = () => {
        if (this.vitisCheck.isNumber(this.mode.options.height)) {
          return this.mode.options.height;
        } else {
          const dashboardHeight = $('#dashboard-content').height();
          const titleHeight = $('.mode-title').height();
          const filterHeight = $('.grid-filter').height();
          const menusHeight = 150;
          const tableHeight = dashboardHeight - (menusHeight + filterHeight + titleHeight);
          return tableHeight;
        }
      };

      /**
       * Réajuste les dimensions de la table
       */
      const autoAjustTables = () => {
        if ($('#' + this.oDataGrid.tableId).length) {
          this.dataTableInstance.columns.adjust();
          $('#' + this.oDataGrid.tableId + '_wrapper').find('.dataTables_scrollBody').css('height', calcDataTableHeight());
        }
      };

      // Options app.json
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.order_by)) {
          order = this.oDataGridOptions.datagridOptions.order_by;
        }
      }

      if (this.vitisCheck.isDefAndNotNull(this.mode.options)) {
        rowId = this.mode.options.id_field;
      }
      this.sDataGridModalTitle = this.oDataGridOptions.datagridOptions.titleModalInsert;

      this.oDataGridOptions.datatableOptions = {
        processing: true,
        serverSide: true,
        searching: false,
        paging: true,
        scrollY: calcDataTableHeight(),
        scrollX: true,
        order: order, // Ordre
        rowId: rowId,
        selected: 'selected',
        select: {
          style: 'multi'
        },
        ajax: (data: any, callback: any, settings: any) => {

          this.getDataGridValues(data).then((oResponse: any) => {

            // Set oResults
            this.oResults = oResponse;

            // Vide viewContainer
            this.oDataGrid.viewContainer.clear();
            // Remplace les données
            callback(this.oResults);
          }).catch(() => {
            this.oResults = {
              recordsTotal: 0,
              recordsFiltered: 0,
              data: []
            };
            callback(this.oResults);
          });
        }
      };

      // Initialise les boutons
      this.initButtons();

      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.columns)) {

          // Traductions
          const aTranslations = [];
          for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
            const column = this.oDataGridOptions.datagridOptions.columns[i];
            aTranslations.push(column.title);
          }

          this.translate.get(aTranslations).subscribe((res: any) => {

            // Enregistre le titre traduit
            for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
              const column = this.oDataGridOptions.datagridOptions.columns[i];
              column.translateTitle = res[column.title];
            }


            // Initialise la table
            setTimeout(() => {
              // Colonnes de la liste (définies dans le mode).
              this.oDataGridOptions.datatableOptions.columns = this.getParsedColumns_();

              // Création de la grille.
              this.dataTableInstance = this.oDataGrid.initTable(this.oDataGridOptions);

              // Draw fields event
              this.dataTableInstance.on('draw.dt', () => {
                // Format les champs avec un template
                this.oDataGrid.formatFields(this.oDataGridOptions);
                // Réajuste les dimensions de la table
                autoAjustTables();
              });

              // Réajuste les dimensions de la table lors fermeture/ouverture du dashboard
              const dDashboardContent = document.querySelector('#dashboard-content');
              const resizeObserver = new ResizeObserver(entries => {
                autoAjustTables();
              });
              resizeObserver.observe(dDashboardContent);
            });
          });
        } else {
          this.oDataGridOptions.datatableOptions.columns = [];
          console.error('Aucune colonne déclaré pour la grille');
        }
      }
    }
  }

  /**
   * Récupère les colonnes définies dans le mode au format BootstrapTable
   * @returns Array
   */
  private getParsedColumns_(): Array<any> {

    const aColumns: any = [{
      targets: 0,
      data: null,
      defaultContent: '',
      orderable: false,
      className: 'select-checkbox',
      width: '40px'
    }];

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {

      // Colonne boutons
      if (this.oDataGridOptions.datagridOptions.display === true
        || this.oDataGridOptions.datagridOptions.update === true) {

        let defaultContent = '<div style="text-align: center;">';
        if (this.oDataGridOptions.datagridOptions.display === true &&
          (environment.properties.data_grid_update_btn_only === false || this.oDataGridOptions.datagridOptions.update !== true)) {
          defaultContent += '<span class="datatable-btn datatable-btn-display icon-vitis-search"></span>';
        }
        if (this.oDataGridOptions.datagridOptions.update === true) {
          defaultContent += '<span class="datatable-btn datatable-btn-edit icon-vitis-draw"></span>';
        }
        defaultContent += '</div>';

        aColumns.push({
          targets: -1,
          data: null,
          orderable: false,
          defaultContent,
          width: '90px'
        });
      }

      // Colonnes du mode
      if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.columns)) {
        const aUserPrivileges = this.auth.getPrivileges();
        for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
          const column = this.oDataGridOptions.datagridOptions.columns[i];
          // Privilège obligatoire pour afficher la colonne.
          if (this.vitisCheck.isDefAndNotNull(column.privilege)) {
            if (aUserPrivileges.indexOf(column.privilege) == -1) {
              continue;
            }
          }
          aColumns.push({
            title: column.translateTitle,
            data: column.field,
            name: column.field,
            visible: column.visible === false ? false : true,
            type: 'string',
            render: (data: any, type: any, row: any, meta: any) => {
              if (column.template) {
                data = '<field-render column_number="' + i + '">' + encodeURIComponent(JSON.stringify(row)) + '</field-render>';
              }
              return data;
            },
            orderable: column.sortable,
            // Alignement du contenu de la colonne.
            className: 'table-align-' + (column.align ? column.align : 'left')
          });
        }
      }
    }

    return aColumns;
  }

  /**
   * Retourne les valeurs du datagrid
   * @param params
   * @returns
   */
  private getDataGridValues(params: any): Promise<any> {
    return new Promise((resolve, reject) => {
      if (this.vitisCheck.isDefAndNotNull(this.oParams) && this.vitisCheck.isDefAndNotNull(this.oDataGridOptions)) {
        // Route de la requête
        const sRoute = this.oDataGridOptions.datagridOptions.ressource.replace(':id', this.oParams.id);
        // Récupère les données à afficher
        this.vitisRequest.ajaxRequestPromise({
          method: 'GET',
          params: this.getRequestParams_(params),
          url: `${environment.properties.app_api_url}/${sRoute}`
        }).then((resp: any) => {
          if (resp.data) {
            if (resp.data.status === 1) {
              const oReturn = {
                recordsTotal: 0,
                recordsFiltered: 0,
                data: []
              };
              if (this.vitisCheck.isDefAndNotNull(resp.data.data)) {
                oReturn.data = resp.data.data;
                if (this.vitisCheck.isDefAndNotNull(resp.data.total_row_number)) {
                  oReturn.recordsTotal = resp.data.total_row_number;
                  oReturn.recordsFiltered = resp.data.total_row_number;
                }
              }
              resolve(oReturn);
            }
          }
        }).catch((err: any) => {
          console.error('error while loading values :', err);

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };

          resolve(oReturn);
        });
      }
    });
  }

  /**
   * Récupère les objets affichés dans la grille Drag and drop pour les filtrer dans le datagrid
   * @returns aRows Tableau d'identifiants
   */
  private getItemsToFilter(): Array<number | string> {
    const aRows: Array<string | number> = [];

    this.oDragAndDropGridRows.forEach(oRow => {
      aRows.push(oRow.source_id);
    });
    return aRows;
  }

  /**
 * Parse les params envoyés par DataTables pour créer des paramètres interprétables par l'API Vitis
 * @param  {any} params
 * @returns any
 */
  private getRequestParams_(params: any): any {

    const requestParams: any = {};

    // Order
    if (this.vitisCheck.isDefAndNotNull(params.order)) {
      if (this.vitisCheck.isDefAndNotNull(params.order[0])) {
        if (this.vitisCheck.isString(params.order[0].dir)) {
          requestParams.sort_order = params.order[0].dir.toUpperCase();
        }
        if (this.vitisCheck.isNumber(params.order[0].column)) {
          const iColumnId = params.order[0].column;
          if (this.vitisCheck.isDefAndNotNull(params.columns[iColumnId])) {
            if (this.vitisCheck.isString(params.columns[iColumnId].data)) {
              requestParams.order_by = params.columns[iColumnId].data;
            }
          }
        }
      }
    }

    // Limit
    if (this.vitisCheck.isNumber(params.length)) {
      requestParams.limit = params.length;
    }

    // Offset
    if (this.vitisCheck.isNumber(params.start)) {
      requestParams.offset = params.start;
    }

    const aItemsToFilter: Array<string | number> = this.getItemsToFilter();
    if (aItemsToFilter.length > 0) {
      // Filter
      const filter = {
        'relation': 'AND',
        'operators': [
          { 'column': this.oDataGridOptions.datagridOptions.id_field, 'compare_operator': 'NOT IN', 'value': this.getItemsToFilter() }
        ]
      };

      if (this.vitisCheck.isDefAndNotNull(this.oFilter.value) &&
        this.vitisCheck.isDefAndNotNull(this.oFilter.value.operators)) {
        if (this.oFilter.value.operators.length) {
          filter.operators = [...filter.operators, ...this.oFilter.value.operators];
        }
      }

      if (this.vitisCheck.isDefAndNotNull(filter)) {
        requestParams.filter = filter;
      }
    }

    // Attributs
    if (this.vitisCheck.isDefAndNotNull(params.columns)) {
      requestParams['attributs'] = params.columns.filter(
        (aColumn: any) => this.vitisCheck.isString(aColumn.data)
      ).map(
        (aColumn: any) => aColumn.data
      ).join('|');
    }
    return requestParams;
  }

  /**
   * Initialise les boutons
   * @returns void
   */
  public initButtons(): void {
    const aButtons: any = [];

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      // Bouton insert
      if (this.oDataGridOptions.datagridOptions.insert === true) {
        aButtons.unshift({
          type: 'button',
          class: 'btn-primary',
          label: 'LIST_BUTTON_ADD',
          icon: 'fa fa-plus',
          event: () => {
            this.addItemsGridToDragDropGrid();
          }
        });
      }

      // Autres boutons
      if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.buttons)) {
        // Ajoute les boutons
        for (let i = 0; i < this.oDataGridOptions.datagridOptions.buttons.length; i++) {
          const button = this.oDataGridOptions.datagridOptions.buttons[i];
          this.oDataGrid.initButtonEvent_(button, this);
          this.oDataGrid.initButtonDisplay_(button, this);
          aButtons.unshift(button);
        }
      }
    }

    this.oDataGridOptions.buttons = aButtons;
  }

  /**
   * Ajoute des éléments au datagrid
   */
  private addItemsGridToDragDropGrid(): void {
    const aIds = this.oDataGrid.getSelectedIds_();
    const oElementToSend: Array<any> = [];
    if (aIds.length > 0) {
      let i: number = 1;
      // On formate l'objet en paramètre
      aIds.forEach((iId: number) => {
        const oItem: any = {};
        // Index max
        oItem[this.sIndexColumn] = this.oDragAndDropGridRows.length + i;
        oItem[this.oDragDropGridOptions.source.id_field] = iId;
        oItem[this.oDragDropGridOptions.source.id_ressource_field] = this.oParams.id;
        oElementToSend.push(oItem);
        i++;
      });
      const sRoute = this.oDragDropGridOptions.ressource.replace(':id', this.oParams.id);
      // Requête pour update
      this.vitisRequest.ajaxRequestPromise({
        method: 'POST',
        url: environment.properties.app_api_url + '/' + sRoute,
        data: { data: oElementToSend }
      }).then((resp: any) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            // Pop-up ok
            this.translate.get('DRAG_DROP_GRID_EDIT_OK').subscribe(async (translation) => {
              vitisAlert.alert({
                type: 'success',
                text: translation
              });
            });
            // On cache la modale
            $('.bd-datagrid-modal-lg').modal('hide');
            this.initRows();
            // Si la page n'a plus d'éléments : page précédente.
            const iNbPageRows: number = this.oDataGrid.dataTableInstance.data().count() - aIds.length;
            if (iNbPageRows == 0) {
              const iNbTotalRows: number = this.oDataGrid.dataTableInstance.page.info().recordsTotal - aIds.length;
              let iPage: number = this.oDataGrid.dataTableInstance.page();
              if (iPage > 0) {
                this.oDataGrid.dataTableInstance.page(--iPage);
              }
            }
          }
        }
      });
    }
  }

  /**
   * Ouvre la modale avec la datagrid
   */
  public openDataGridModal() {
    // On recharge les données
    this.dataTableInstance.ajax.reload(null, false);
    $('.bd-datagrid-modal-lg').modal('show');
  }

  /**
   * Retire l'association entre l'item et la source
   * @returns Promise
   */
  public deleteDragDropGridItem() {
    return new Promise((resolve, reject) => {
      vitisAlert.alert({
        type: 'warning',
        title: this.translate.instant('DRAG_DROP_GRID_DELETE_MESSAGE'),
        showCancelButton: true
      }).then((result) => {
        const aRowsId: Array<string | number> = this.getSelectedRowsId();
        if (result.value && aRowsId.length > 0) {
          // Id dans la route
          const sRoute = this.oDragDropGridOptions.ressource.replace(':id', this.oParams.id);

          this.vitisRequest.ajaxRequestPromise({
            url: environment.properties.app_api_url + '/' + sRoute,
            params: {
              idList: aRowsId.join('|')
            },
            method: 'DELETE',
            displayErrorMessage: true
          }).then((resp: any) => {
            if (this.vitisCheck.isObject(resp.data)) {
              const oResult = resp.data;
              if (oResult.status == 1 && oResult.data) {
                this.translate.get('LIST_DELETE_OK').subscribe((sTranslate: string) => {
                  this.initRows();
                  vitisAlert.fireToast({
                    type: 'success',
                    title: sTranslate
                  });
                });
                resolve(oResult.data);
              } else {
                reject(oResult);
              }
            } else {
              reject(resp);
            }
          });
        }
      });

    });
  }

  /**
   * Retourne un tableau contenant les id sélectionnés dans la liste
   * @returns Array<string | number>
   */
  private getSelectedRowsId(): Array<string | number> {
    const aRowsId: Array<string | number> = [];

    this.oDragAndDropGridRows.forEach((row: any) => {
      if (row.selected) {
        aRowsId.push(row.id);
      }
    });

    return aRowsId;
  }

  /**
   * Vérifie la valeur d'un champ en fonction de ce qui est défini dans le fichier Struct
   * @param sColumnfield Champ ç vérifier
   * @param oElement Ligne à vérifier
   * @returns isDisabled Est désactivé ou non
   */
  public isPropertyDisabled(sColumnfield: string, oElement: any) {
    let isdisabled: boolean = false;
    this.oDragDropGridOptions.columns.forEach((oColumnOptions: any) => {
      if (oColumnOptions.field === sColumnfield) {
        if ('enabled' in oColumnOptions) {
          if (oElement.columns[oColumnOptions.enabled.column] !== oColumnOptions.enabled.value) {
            isdisabled = true;
            return isdisabled;
          }
        }
      }
    });
    return isdisabled;
  }

  /**
   * Lit le tableau de noeuds et retourne un filtre compatible avec l'API Vitis
   * @param  {Array<any>} aFilterNodes
   * @returns any
   */
  public getAPIFilterFromNodes_(aFilterNodes: Array<any>): any {

    let oFilter: any = {};

    if (this.vitisCheck.isArray(aFilterNodes)) {
      if (aFilterNodes.length > 0) {

        oFilter = {
          relation: 'AND',
          operators: []
        };

        for (let i = 0; i < aFilterNodes.length; i++) {
          const oNode = aFilterNodes[i];

          if (this.vitisCheck.isDefAndNotNull(oNode.column)) {
            const oNodeFilter: any = {};

            if (this.vitisCheck.isDefAndNotNull(oNode.column.field)
              && this.vitisCheck.isDefAndNotNull(oNode.compareOperator)
              && this.vitisCheck.isDefAndNotNull(oNode.value)) {
              oNodeFilter.column = oNode.column.field;
              oNodeFilter.compare_operator = oNode.compareOperator;
              oNodeFilter.value = oNode.value;

              if (oNode.compareOperator === '~') {
                oNodeFilter.compare_operator = 'LIKE';
                oNodeFilter.compare_operator_options = {
                  'case_insensitive': true
                };
                oNodeFilter.value = '%' + oNodeFilter.value + '%';

              }
              // modification de la colonne à utiliser pour filtrer
              if (this.vitisCheck.isDefAndNotNull(oNode.column.searchColumn)) {
                oNodeFilter.column = oNode.column.searchColumn;
              }

              oFilter.operators.push(oNodeFilter);
            }
          }
        }
      }
    }

    return oFilter;
  }

  /**
   * Applique le filtre à la grille
   */
  public applyFilter(): void {

    // Filtre formaté pour la requête API
    this.oFilter.value = this.getAPIFilterFromNodes_(this.oFilter.nodes);

    // Rafraichit la vue
    if (this.dataTableInstance &&
      this.dataTableInstance.ajax) {
      this.dataTableInstance.ajax.reload();
    }
  }

  /**
   * Change value du select dans la grille
   * @param element
   * @param field
   * @param event
   */
  public changeSelectElement(element, field, event) {
    element.columns[field] = event.target.value;
  }
}
