import { Component, OnInit, Input, ElementRef, ViewChild } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { vitisAlert, VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { DataGridComponent } from 'src/app/vitis/data-grid/data-grid.component';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { FormreaderComponent } from '../../formreader.component';
import { DataGridStudioService } from 'src/app/vitis/data-grid/services/data-grid-studio.service';

declare var $: any;

/**
 * Type date-grid
 *
 * Arguments:
 * type : type de champ (date, datetime)
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * options : options de la grille de données
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 *
 * Exemple:
 * {
 *   "type": "data-grid",
 *   "name": "data-grid",
 *   "label": "Grille de données",
 *   "nb_cols": 12,
 *   "options": {
 *     "searching": false,
 *     "paging": false,
 *     "info": false,
 *     "select": {
 *       "style": "multi",
 *       "selector": "tr"
 *     },
 *     "columns": [{
 *         "field": "date",
 *         "title": "FORM_HISTORY_DATE_PUBLICATION_FME_PROJECTS_METADATA",
 *         "width": "15%"
 *       },
 *       {
 *         "field": "author",
 *         "title": "FORM_HISTORY_AUTHOR_PUBLICATION_FME_PROJECTS_METADATA",
 *         "width": "20%"
 *       },
 *       {
 *         "field": "description",
 *         "title": "FORM_HISTORY_DESCRIPTION_PUBLICATION_FME_PROJECTS_METADATA",
 *         "width": "65%"
 *       }
 *     ],
 *     "data": [
 *         [ "Admin-Export", "exportv3.fmw", true, "CTLPASPAD" ],
 *         [ "Admin-Form-Checker", "formChecker.fmw", false, "2EW0ET1BDHLWC" ]
 *     ],
 *     "buttons": [{
 *         "type": "button",
 *         "class": "btn-primary",
 *         "label": "FORM_BTN_DELETE_HISTORY_METADATA_PUBLICATION_FME_PROJECTS",
 *         "icon": "fa fa-trash",
 *         "event": "deleteHistory()"
 *       },
 *       {
 *         "type": "button",
 *         "class": "btn-primary",
 *         "label": "FORM_BTN_ADD_HISTORY_METADATA_PUBLICATION_FME_PROJECTS",
 *         "icon": "fa fa-plus",
 *         "event": "addHistory()"
 *       }
 *     ]
 *   }
 * }
 */

@Component({
  selector: 'app-formreader-data-grid',
  templateUrl: './data-grid.component.html',
  styleUrls: ['./data-grid.component.scss']
})
export class FormReaderDataGridComponent extends FormreaderField implements OnInit {
  @ViewChild(FormreaderComponent, { static: false }) oRowFormReader: FormreaderComponent;

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;
  @Input() oFormReader: any;

  @ViewChild(DataGridComponent, { static: false, read: DataGridComponent }) set oDataGrid(oDataGrid: DataGridComponent) {
    setTimeout(() => {
      if (this.vitisCheck.isDefAndNotNull(oDataGrid)) {
        this.initDataGrid(oDataGrid);
        this.setValueInDataGrid();
      }
    });
  }

  // Instance de la grille de données.
  public dataTableInstance: any;
  public oRowFormOptions: any;

  // Modale de création d'une ligne.
  public sModalFormTitle: string = '';
  public oModalOptions: any = {
    submit_button: false,
    cancel_button: true,
    add_button: true,
    add_event: this.addRow
  };
  public showDataGrid: boolean = true;

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef,
    private vitisRequest: VitisRequest,
    private dataGridStudioService: DataGridStudioService,
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {
    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      if (this.vitisCheck.isString(oValue) && oValue != '') {
        oValue = JSON.parse(oValue);
      }
      const oNewValue = this.setValue(this.oFormValues, this.oField, oValue);
      this.setValueInDataGrid();
      return oNewValue;
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Set contenu manuellement
    this.oField.setContent = (aRows: Array<any>) => {
      return this.setContent(aRows);
    };

    // Operations pour studio
    if (this.vitisCheck.isDefAndNotNull(this.oField.studio_id_)) {
      this.dataGridStudioService.dataGridOptionsEventListner(this.oField.studio_id_).subscribe(() => {
        this.showDataGrid = false;
        setTimeout(() => {
          this.showDataGrid = true;
        }, 500);
      });
    }

    // Ajout fonctions
    this.oField.refreshGrid = () => { this.refreshGrid(); };
  }

  /**
   * appelle la fonction
   */
  public callEvent() {
    this.oModalOptions.add_event(this);
  }

  /**
   * Ajoute les données du formulaire dans une nouvelle ligne.
   */
  public addRow(this_) {
    const aFields = this_.oRowFormReader.getFormFields(this_.oRowFormReader.oFormDefinition, this_.oRowFormReader.sFormDefinitionName);
    // Set la valeur à vide en utilisant les setters
    const aFormValues: any = {};
    // for (let i = 0; i < aFields.length; i++) {
    for (const oField of aFields) {
      // const oField = aFields[i];
      if (this_.vitisCheck.isDefAndNotNull(oField.getDisplayedValue)) {
        aFormValues[oField.name] = oField.getDisplayedValue();
      } else if (this_.vitisCheck.isDefAndNotNull(oField.getValue)) {
        aFormValues[oField.name] = oField.getValue();

        if (oField.required === true && !aFormValues[oField.name]) {
          vitisAlert.fireToast({
            type: 'error',
            title: this_.translate.instant('DATA_GRID_FORMREADERCOMPONENT_FIELD_REQUIRED') + ' ' +
                   this_.translate.instant(oField.label),
          });
          return;
        }
      }
    }

    // Concatène la nouvelle valeur avec les anciennes puis set les valeurs dans le champs.
    let aValue = this_.oField.getValue();
    if (!this_.vitisCheck.isArray(aValue)) {
      aValue = [];
    }
    aValue.push(aFormValues);
    this_.oField.setValue(aValue, true);
    // Ferme la fenêtre modale.
    $('#datagrid_modal_row_creation').modal('hide');
  }

  /**
   * Remplit la grille
   * @param aRows
   */
  public setContent(aRows: Array<any>) {
    this.oField.dataTableInstance.clear();
    for (let i = 0; i < aRows.length; i++) {
      const oRow = this.oField.dataTableInstance.row.add(aRows[i]);
      if (aRows[i].selected === true) {
        oRow.select();
      }
    }
    this.oField.dataTableInstance.draw();
  }

  /**
   * Initialise les données dans la grille
   */
  private setValueInDataGrid() {
    if (this.vitisCheck.isDefAndNotNull(this.oField.dataTableInstance)) {
      if (this.vitisCheck.isDefAndNotNull(this.oFormValues.value[this.oField.name]) && this.oFormValues.value[this.oField.name] !== '') {
        // Mise à jour des données de la grille.
        this.oField.dataTableInstance.clear();
        this.oField.dataTableInstance.rows.add(this.oFormValues.value[this.oField.name]).draw();

        // Attend le chargement des valeurs spéciales (renders) comme par exemple les images avant de réajuster la taille des colonnes
        setTimeout(() => {
          this.oField.dataTableInstance.columns.adjust();
        }, 1000);
      }
    }
  }

  /**
   * Initialise la grille de données.
   * @param {oDataGrid} DataGridComponent
   * @returns void
   */
  private initDataGrid(oDataGrid: DataGridComponent) {

    if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
      // Trie des colonnes en fonction de l'index
      if (this.vitisCheck.isDefAndNotNull(this.oField.options.columns) && this.oField.options.columns.length > 0) {
        this.oField.options.columns = this.oField.options.columns.sort((a: any, b: any) => {
          // Les colonnes sans index sont mis au début
          const indexA = this.vitisCheck.isDefAndNotNull(a.index) ? a.index : -1;
          const indexB = this.vitisCheck.isDefAndNotNull(b.index) ? b.index : -1;
          if (indexA == indexB) {
            // Trie par ordre alphabetique les colonnes avec index
            if (indexA !== -1) {
              if (a.title < b.title) { return -1; }
              if (a.title > b.title) { return 1; }
            }
            return 0;
          }
          return indexA - indexB;
        });
      }

      // Paramètres du plugin DataTable.
      const oDataGridOptions: DataGridOptions = {
        datatableOptions: {
          searching: this.oField.options.searching,
          paging: this.oField.options.paging,
          info: this.oField.options.info,
          select: this.oField.options.select,
          columns: this.oField.options.columns,
          lengthMenu: this.oField.options.lengthMenu,
          pageLength: this.oField.options.pageLength,
          scrollX: true,
        },
        datagridOptions: {
          refresh: this.oField.options.refresh === true ? true : false,
          export: this.oField.options.export === true ? true : false,
        },
        buttons: this.vitisCheck.isArray(this.oField.options.buttons) ? this.oField.options.buttons : []
      };
      if (!this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.columns)) {
        oDataGridOptions.datatableOptions.columns = [];
        console.error('Aucune colonne déclaré pour la grille');
      }

      // Ajout 1ere colonne uniquement la première fois
      if (oDataGridOptions.datatableOptions.select !== false &&
        (!this.vitisCheck.isDefAndNotNull(this.oField.dataTableInstance) ||
        !this.vitisCheck.isDefAndNotNull(this.oField.options.columns) ||
        this.oField.options.columns.length == 0)) {
        // 1ere colonne (checkbox de sélection de la ligne).
        oDataGridOptions.datatableOptions.columns.splice(0, 0, {
          targets: 0,
          data: null,
          defaultContent: '',
          orderable: false,
          className: 'select-checkbox',
          width: '40px'
        });
      }

      // Evènements sur les boutons.
      for (let i = 0; i < oDataGridOptions.buttons.length; i++) {
        if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.buttons[i].event) && !this.vitisCheck.isFunction(oDataGridOptions.buttons[i].event)) {
          const sButtonEvent = oDataGridOptions.buttons[i].event;
          oDataGridOptions.buttons[i].event = () => {
            // Evaluation de l'évènement.
            try {
              this.oFormReader.fieldsService.callFunction(sButtonEvent, this.oFormReader.getFormValues(), [this.oFormReader, this.oFormReader.getFormContainer()]);
            } catch (e) {
              console.error(e);
            }
          };
        }
      }

      // Création de la grille.
      this.dataTableInstance = oDataGrid.initTable(oDataGridOptions);
      // Sauve l'instance dans la définition du champ.
      this.oField.dataTableInstance = this.dataTableInstance;
      // Référence du composant DataGridComponent;
      this.oField.oFormReaderDataGridComponent = this;
    }
  }

  /**
   * Affiche un formulaire dans une fenêtre modale
   * @param  {string} sFormUrl
   * @param  {string} sFormMode
   * @param  {any} oFormValues
   * @param  {any} oOptions
   */
  public displayModalForm(sFormUrl: string, sFormMode: string = 'insert', oFormValues: any = {}, oOptions: any = {}) {
    // Chargement du formulaire.
    this.vitisRequest.ajaxRequestObservable({
      method: 'GET',
      url: sFormUrl,
    }).subscribe(resp => {
      if (resp.status === 200 && resp.data) {
        // Définition et valeurs par défaut du formulaire pour le FormReader.
        const oRowFormOptions = {
          form: resp.data,
          formName: sFormMode,
          formValues: oFormValues
        };
        // Exécution du FormReader.
        this.oRowFormOptions = oRowFormOptions;
        // Titre de la modale.
        this.sModalFormTitle = this.oRowFormOptions.form[sFormMode].title;
        // Fusionne les options de la grille de données.
        this.oModalOptions = { ...this.oModalOptions, ...oOptions };
        // Affichage de la fenêtre modale du formulaire.
        $('#datagrid_modal_row_creation').modal('show');
      }
    });
  }

  /**
   * Lance le adjust de la grille pour réaligner les colonnes
   */
  public refreshGrid() {
    if (this.vitisCheck.isDefAndNotNull(this.oField) &&
      this.vitisCheck.isDefAndNotNull(this.oField.dataTableInstance) &&
      this.vitisCheck.isDefAndNotNull(this.oField.dataTableInstance.columns)) {
      this.oField.dataTableInstance.columns.adjust();
    } else {
      // On attend un peu et on retente
      setTimeout(() => {
        if (this.vitisCheck.isDefAndNotNull(this.oField) &&
          this.vitisCheck.isDefAndNotNull(this.oField.dataTableInstance) &&
          this.vitisCheck.isDefAndNotNull(this.oField.dataTableInstance.columns)) {
          this.oField.dataTableInstance.columns.adjust();
        } else {
          console.error('Datagrid n\'est pas chargé, impossible d\'ajuster les colonnes (refreshGrid)');
        }
      });
    }
  }
}