import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { FormreaderField } from '../../formreader-field';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { FormGroup, FormControl } from '@angular/forms';

/**
 * Type double liste
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * nb_rows: nombre de lignes à utiliser
 * datasource : source de données
 * datasource.datasource_id : id source de données
 * datasource.distinct : distinct
 * datasource.id_key : colonne pour valeur
 * datasource.label_key : colonne pour affichage
 * datasource.order_by : ordre
 * datasource.sort_order : ordre (ASC, DESC)
 * datasource.attributs : attributs
 * datasource.parameters : paramètres supplémentaires
 * default_value? : valeur par défaut
 *
 * Exemple:
 * {
 *    "type":"double_select",
 *    "name":"groups",
 *    "label":"FORM_LOCAL_GROUPS_USERS_USER",
 *    "datasource": {
 *        "datasource_id": "datasource_user_groups",
 *        "sort_order": "ASC",
 *        "distinct": "true",
 *        "label_key": "name",
 *        "order_by": "name",
 *        "id_key": "group_id",
 *        "attributs": "group_id|name",
 *        "parameters":{
 *            "order_by": "name"
 *        }
 *    },
 *    "nb_rows":7,
 *    "nb_cols":12
 * }
 */
@Component({
  selector: 'app-double-select',
  templateUrl: './double-select.component.html',
  styleUrls: ['./double-select.component.scss']
})
export class DoubleSelectComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  public oTemporaryForm: any;

  public loading: boolean = false;
  public aValues: Array<any> = [];
  public aLeftList: Array<any> = [];
  public aRightList: Array<any> = [];

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // Valeur par dédaut
    this.oField.nb_rows = this.vitisCheck.isDefAndNotNull(this.oField.nb_rows) ? this.oField.nb_rows : 5;

    // Observe l'arrivée des choix (pour utilisation avec datasource)
    if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
      if (this.vitisCheck.isDefAndNotNull(this.oField.options.observableChoises)) {

        // Permet de savoir qu'une requête est en cours
        this.loading = true;

        // Lorsque la requête se termine
        this.oField.options.observableChoises.subscribe((aChoises: any) => {
          this.loading = false;
          return this.setValue(this.oFormValues, this.oField, this.aValues);
        });
      }
    }

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    /**
     * Réinitialisation de la sélection.
     */
    this.oField.resetValue = () => {
      // Supprime la sélection.
      this.aValues.length = 0;
      // Rafraichit les valeurs dans le form builder
      this.refreshFormValues_();
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Création FormGroup spécifique
    this.createSpecialFormGroup_();

    // Création d'un validateur spécifique
    if (this.oField.required === true) {
      this.addRequiredValidator_(this.oField);
    }
  }

  /**
   * Retourne les id sélectionnées
   * @param oFormValues
   * @param oField
   *
   * @returns {Array<any>}
   */
  public getValue(oFormValues: any, oField: any): Array<any> {

    // Rafraichit les valeurs
    if (this.loading !== true) {
      this.refreshValues_();
    }

    // Retourne une chaine de charactères
    return this.aValues;
  }

  /**
   * Setter de valeur du champ
   * @param  {any} oFormValues
   * @param  {any} oField
   * @param  {any} oValue
   */
  public setValue(oFormValues: any, oField: any, oValue: any): void {

    // Reinit RightList and LeftList
    this.aRightList = [];
    this.aLeftList = [];

    // Choix disponibles
    let aChoises: Array<any>;

    // Remplit aChoises
    if (this.vitisCheck.isDefAndNotNull(oField)) {
      if (this.vitisCheck.isDefAndNotNull(oField.options)) {
        if (this.vitisCheck.isArray(oField.options.choises)) {
          aChoises = oField.options.choises;
        }
      }
    }

    // Remplit this.aValues
    if (this.vitisCheck.isDefAndNotNull(oValue)) {
      if (this.vitisCheck.isArray(oValue)) {
        this.aValues = oValue;
      } else {
        this.aValues = [oValue];
      }
    }

    if (this.vitisCheck.isArray(this.aValues)
      && this.vitisCheck.isArray(aChoises)) {
      for (let i = 0; i < aChoises.length; i++) {
        const oChoise = aChoises[i];
        oChoise.selected = false;

        if (this.vitisCheck.isDefAndNotNull(oChoise.value)) {

          // Si la valeur n'est pas choisie
          if (this.aValues.indexOf(oChoise.value) !== -1 ||
            this.aValues.indexOf(String(oChoise.value)) !== -1) {
            this.aRightList.push(oChoise);
          } else {
            this.aLeftList.push(oChoise);
          }
        }
      }

      // Rafaichit l'ordre des listes
      this.sortLists_();
    }

    // Rafraichit les valeurs dans le form builder
    this.refreshFormValues_();
  }

  /**
   * Retourne true si tous les étéments sont sélectionnés,
   * false si aucun élément n'est sélectionné
   * null si certains éléments sont sélectionnés
   */
  public isAllSelected(aList: Array<any>): any {

    let bAllUnSelected = true;
    let bAllSelected = true;
    if (this.vitisCheck.isArray(aList)) {
      if (aList.length === 0) {
        bAllUnSelected = true;
        bAllSelected = false;
      } else {
        for (let i = 0; i < aList.length; i++) {
          if (aList[i].selected === true) {
            bAllUnSelected = false;
          } else {
            bAllSelected = false;
          }
        }
      }
    }

    if (bAllSelected) {
      return true;
    } else if (bAllUnSelected) {
      return false;
    } else {
      return null;
    }
  }

  /**
   * Change la selection d'une ligne
   * @param row
   */
  public toggleSelectRow(row: any): void {
    row.selected = !row.selected;
  }

  /**
   * Change la sélection d'une liste
   * @param aList
   */
  public toggleSelectAll(aList: Array<any>): void {

    const bCurrentState = this.isAllSelected(aList);
    switch (bCurrentState) {
      case true:
        this.unselectAll(aList);
        break;
      case false:
      case null:
        this.selectAll(aList);
        break;
      default:
        this.selectAll(aList);
        break;
    }
  }

  /**
   * Sélectionne tous les enregistrements
   * @param  {Array<any>} aList
   * @returns void
   */
  public selectAll(aList: Array<any>): void {
    if (this.vitisCheck.isArray(aList)) {
      for (let i = 0; i < aList.length; i++) {
        aList[i].selected = true;
      }
    }
  }

  /**
   * Désélectionne tous les enregistrements
   * @param  {Array<any>} aList
   * @returns void
   */
  public unselectAll(aList: Array<any>): void {
    if (this.vitisCheck.isArray(aList)) {
      for (let i = 0; i < aList.length; i++) {
        aList[i].selected = false;
      }
    }
  }

  /**
   * Envoi les lignes sur la colonne de doite
   * @returns void
   */
  public rowsToRight(): void {

    const aRows = [];

    // Récupère les éléments
    for (let i = 0; i < this.aLeftList.length; i++) {
      const element = this.aLeftList[i];
      if (element.selected === true) {
        aRows.push(element);
      }
    }

    // Ajoute les éléments sur l'autre liste
    for (let i = 0; i < aRows.length; i++) {
      this.rowToRight_(aRows[i]);
    }

    // Rafraichit l'ordre des listes
    this.sortLists_();

    // Rafraichit les valeurs
    this.refreshValues_();
  }

  /**
   * Envoi les lignes sur la colonne de gauche
   * @returns void
   */
  public rowsToLeft(): void {

    const aRows = [];

    // Récupère les éléments
    for (let i = 0; i < this.aRightList.length; i++) {
      const element = this.aRightList[i];
      if (element.selected === true) {
        aRows.push(element);
      }
    }

    // Ajoute les éléments sur l'autre liste
    for (let i = 0; i < aRows.length; i++) {
      this.rowToLeft_(aRows[i]);
    }

    // Rafraichit l'ordre des listes
    this.sortLists_();

    // Rafraichit les valeurs
    this.refreshValues_();
  }

  /**
   * Envoi la ligne sur la colonne de gauche
   * @param  {any} oRow
   * @returns void
   */
  private rowToRight_(oRow: any): void {

    // Vérifie et supprime row de la colonne de gauche
    let isPresent = false;
    for (let i = this.aLeftList.length - 1; i >= 0; i--) {
      if (this.aLeftList[i] == oRow) {
        isPresent = true;
        this.aLeftList.splice(i, 1);
      }
    }

    if (isPresent) {
      oRow.selected = false;
      this.aRightList.push(oRow);
    }

    // Rafraichit les valeurs dans le form builder
    setTimeout(() => {
      this.refreshFormValues_();
    });
  }

  /**
   * Envoi la ligne sur la colonne de gauche
   * @param  {any} oRow
   * @returns void
   */
  private rowToLeft_(oRow: any): void {

    // Vérifie et supprime row de la colonne de droite
    let isPresent = false;
    for (let i = this.aRightList.length - 1; i >= 0; i--) {
      if (this.aRightList[i] == oRow) {
        isPresent = true;
        this.aRightList.splice(i, 1);
      }
    }

    if (isPresent) {
      oRow.selected = false;
      this.aLeftList.push(oRow);
    }

    // Rafraichit les valeurs dans le form builder
    setTimeout(() => {
      this.refreshFormValues_();
    });
  }

  /**
   * Re-ordonne les listes par ordre alpha
   * @returns void
   */
  private sortLists_(): void {
    this.aLeftList.sort((a: any, b: any) => (a.label > b.label) ? 1 : -1);
    this.aRightList.sort((a: any, b: any) => (a.label > b.label) ? 1 : -1);
  }

  /**
   * Remplit this.aValues en fonction de this.aRightList
   * @returns void
   */
  private refreshValues_(): void {

    this.aValues = [];

    for (let i = 0; i < this.aRightList.length; i++) {
      const element = this.aRightList[i];
      if (this.vitisCheck.isDefAndNotNull(element.value)
        && element.value !== '') {
        this.aValues.push(element.value);
      }
    }
  }

  /**
   * Rafraichit les valeurs dans le form builder
   */
  private refreshFormValues_(): void {
    // remplit la valeur dans oFormValues
    const oPatchValues = {};
    if (this.vitisCheck.isDefAndNotNull(this.oField)) {
      if (this.vitisCheck.isDefAndNotNull(this.oField.name)) {
        oPatchValues[this.oField.name] = this.aValues;
      }
    }
    this.oFormValues.patchValue(oPatchValues);
  }

  /**
   * Création d'un FormGroup pour connaître la valeur selectionnée afin d'utiliser les différents boutons
   * Le FormGroup oFormValues n'est utilisé que pour la boucle
   */
  private createSpecialFormGroup_(): void {
    const formControl = {};
    formControl[this.oField.name] = new FormControl({ value: '', disabled: this.oField.disabled });
    this.oTemporaryForm = new FormGroup(formControl);
  }

  /**
   * Création d'un validateur spécifique
   * @param oField
   */
  private addRequiredValidator_(oField: any) {

    let aValidators = [];

    // Récupère les validators inclus par FormReader
    if (this.vitisCheck.isArray(oField.validators_list)) {
      aValidators = oField.validators_list;
    }

    aValidators.push(this.RequiredValidator(oField));
    this.oFormValues.controls[oField.name].setValidators(aValidators);
  }

  /**
   * Validateur
   * @param oField
   */
  private RequiredValidator(oField: any) {
    return () => {
      const val = oField.getValue();
      if (val.length > 0) {
        return null;
      } else {
        return { 'required': true };
      }
    };
  }
}
