import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { Validators, AbstractControl, ValidatorFn, FormControl } from '@angular/forms';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';

/**
 * Type email
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 * maxLength? : Nombre de caractères maximum.
 *
 * Exemple:
 * {
 *   "type": "email",
 *   "name": "email",
 *   "label": "Email",
 *   "required": true,
 *   "nb_cols": 12
 * }
 */
@Component({
  selector: 'app-email',
  templateUrl: './email.component.html',
  styleUrls: ['./email.component.scss']
})
export class EmailComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  public aBadges: Array<string> = [];
  public sInput: string;

  // tslint:disable-next-line:max-line-length
  public emailRegex = /^(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/;

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {

      // Remplit les badges
      if (this.oField.multiple) {
        this.setBadgesValue(oValue);
      }

      // Set la valeur du champ caché
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Validateurs
    this.updateFieldValidator_(this.oField);
  }

  /**
   * Vérifie si la valeur est une adresse mail
   * @param val
   */
  public isEmail(val: string) {
    if (this.vitisCheck.isString(val)) {
      return this.emailRegex.test(val);
    } else {
      return false;
    }
  }

  /**
   * Fonction à appeller pour setter les badges en fonction de la valeur du formulaire
   * @param val
   */
  public setBadgesValue(val: string) {
    if (this.oField.multiple &&
      this.vitisCheck.isDefAndNotNull(val)) {

      let isJSON = false;

      // Valeur de type JSON
      try {
        const parsedVal = JSON.parse(val);
        if (this.vitisCheck.isArray(parsedVal)) {
          for (let i = 0; i < parsedVal.length; i++) {
            const email = parsedVal[i];
            if (email.trim().length > 0) {
              this.aBadges.push(email);
            }
          }
        }
        isJSON = true;
      } catch (error) {
        isJSON = false;
      }

      if (!isJSON) {
        this.onInputChange(val + ' ');
      }
    }
  }

  /**
   * Fonction utilisée en mode multiple
   * @param val
   */
  public onInputChange(val: string) {
    if (this.oField.multiple &&
      this.vitisCheck.isDefAndNotNull(val)) {

      const emails = val.split(' ');
      if (emails.length > 1) {

        // Ajoute les badges
        for (let i = 0; i < emails.length; i++) {
          const email = emails[i].trim();
          if (email.length > 0) {
            this.aBadges.push(email);
          }
        }

        // Vide la saisie
        this.sInput = emails[emails.length - 1];
      }

      // Rafraichit la valeur du champ caché
      this.refreshHiddenVal_();
    }
  }

  /**
   * Supprime un badge
   * @param badge
   */
  public popBadge(badge: any) {
    if (this.oField.multiple &&
      this.vitisCheck.isDefAndNotNull(badge)) {
      const iIdx = this.aBadges.indexOf(badge);
      // Ne fait rien si badge n'est pas été trouvé dans le tableau
      if (iIdx < 0) {
        return;
      }
      // Enleve l'élément
      this.aBadges.splice(iIdx, 1);

      // Rafraichit la valeur du champ caché
      this.refreshHiddenVal_();
    }
  }

  /**
   * Permet la validation du composant
   */
  public multiEmailValidator(): any {

    let isValid = true;
    let errorMessage = '';

    if (this.vitisCheck.isArray(this.aBadges)) {
      // Vérifie si les mails sont bons
      for (let i = 0; i < this.aBadges.length; i++) {
        const email = this.aBadges[i];
        if (!this.isEmail(email)) {
          isValid = false;
          errorMessage = `Email ${email} not valid.`;
          break;
        }
      }

      // Met à jour le message d'erreur
      if (this.oField &&
        this.oField.validation_messages) {
        for (let i = 0; i < this.oField.validation_messages.length; i++) {
          const element = this.oField.validation_messages[i];
          if (element.type === 'validateMultiEmail') {
            element.message = errorMessage;
          }
        }
      }
    }

    return isValid ? null : {
      validateMultiEmail: {
        valid: false,
      }
    };
  }

  /**
   * Rafraichit le champ hidden
   */
  private refreshHiddenVal_() {
    if (this.oField.multiple &&
      this.vitisCheck.isDefAndNotNull(this.aBadges)) {

      // Set l'élément comme modifié
      this.oFormValues.get(this.oField.name).markAsTouched();

      // Set la valeur
      this.setValue(this.oFormValues, this.oField, JSON.stringify(this.aBadges));
    }
  }

  /**
   * Création d'un validateur spécifique
   * @param oField
   */
  private updateFieldValidator_(oField: any) {

    let aValidators = [];

    // Récupère les validators inclus par FormReader
    if (this.vitisCheck.isArray(oField.validators_list)) {
      aValidators = oField.validators_list;
    }

    // Validateur
    if (!oField.multiple) {
      // Set uniquement si pattern n'est pas déjà défini
      if (!this.vitisCheck.isDefAndNotNull(oField.pattern)) {
        aValidators.push(Validators.pattern(this.emailRegex));
      }
    } else {

      // Multiple email
      const self = this;
      const validateMultiEmail: ValidatorFn = () => {
        return self.multiEmailValidator();
      };
      aValidators.push(validateMultiEmail);
    }

    // Ajoute les validateurs
    this.oFormValues.controls[oField.name].setValidators(aValidators);

    // Messages
    if (!oField.validation_messages) {
      oField.validation_messages = [];
    }
    oField.validation_messages.push({
      type: 'pattern',
      message: this.translate.instant('FORM_EMAIL_NOT_VALID') + '.'
    });
    oField.validation_messages.push({
      type: 'validateMultiEmail',
      message: this.translate.instant('FORM_EMAIL_NOT_VALID') + '.'
    });
  }
}