import { Component, OnInit, Input, Output, EventEmitter, ViewChild, ElementRef } from '@angular/core';
import { vitisAlert, VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';
import { NgbActiveModal, NgbModal } from '@ng-bootstrap/ng-bootstrap';

declare var $: any;

@Component({
  selector: 'app-global-file-picker',
  templateUrl: './global-file-picker.component.html',
  styleUrls: ['./global-file-picker.component.scss']
})
export class GlobalFilePickerComponent implements OnInit {

  @Input() public sFilePickerButton: string = 'FILE_PICKER_ADD_FILE';
  @Input() public sFilePickerId: string = 'default-file-picker';
  @Input() public sFilePickerName: string = 'default-file-picker';
  @Input() public sFilePickerType: string = 'file'; // Default type file
  @Input() public aUrls: any;
  @Input() public oField: any = {};
  @Input() public sSize: any;

  @Output() public setParentValue = new EventEmitter<any>();

  @ViewChild('inputFile') inputFile: ElementRef;

  // Variables pour la modale
  @ViewChild('filepickerModal') public filepickerModalContent: ElementRef;
  private filepickerModal: NgbActiveModal;

  public aAllFiles: Array<any> = [];
  public aNewFileList: Array<any> = [];
  public aUnchangedFileList: Array<any> = [];
  public aDeletedFileList: Array<string> = [];
  public currentImageModal: any;
  public currentImageIndex: number = 0;

  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private modalService: NgbModal,
  ) { }

  ngOnInit() {
    setTimeout(() => {
      this.initFilePickerData_();
      this.initDefaultField_();
    });
  }

  /**
   * Initialise le filepicker lançant la récupération des fichiers initiaux.
   */
  private initFilePickerData_() {
    this.getFiles_().then((aFiles) => {
      this.aUnchangedFileList = aFiles;
      this.aAllFiles = aFiles;
    });
  }

  /**
   * Si oField n'a pa été fourni on en défini un par défaut
   */
  private initDefaultField_() {
    if (!this.vitisCheck.isDefAndNotNull(this.oField)) {
      this.oField = {
        displayOnly: false,
        disabled: false,
        multiple: false,
        id: this.getUniqueId_()
      };
    }
    if (!this.vitisCheck.isDefAndNotNull(this.oField.id)) {
      this.oField.id = this.getUniqueId_();
    }
  }

  /**
   * Ouvre la modale de preview
   */
  public openImageModal(index: number) {
    this.currentImageModal = index;
    this.filepickerModal = this.modalService.open(this.filepickerModalContent);
  }

  /**
   * Ferme la modale de preview
   */
  public closeImageModal() {
    if (this.filepickerModal) {
      this.filepickerModal.close();
    }
  }

  /**
   * Récupère le blob et l'ajoute comme src pour afficher la miniature.
   * @param {any} oFile
   */
  private addSrc_(oFile: any) {
    const picReader = new FileReader();
    picReader.addEventListener('load', (event) => {
      if (this.vitisCheck.isDefAndNotNull(event.target) && this.vitisCheck.isDefAndNotNull(event.target['result'])) {
        oFile.src = event.target['result'];
      }
    });
    picReader.readAsDataURL(oFile);
  }

  /**
   * Enregistre les fichiers choisi
   * @param {FileList} files
   */
  public chooseFiles(files: FileList) {

    // Si l'input n'est pas en edition multiple et qu'un nouveau fichier est chargé, on supprime les autres fichiers.
    if (files.length > 0 && this.oField.multiple !== true) {
      this.removeAllFiles_();
    }

    for (let i = 0; i < files.length; i++) {
      const file = files[i];

      // Si le fichier n'est pas déjà chargé
      if (this.findFileInFiles_(file, this.aAllFiles) === -1) {
        // Si c'est une image on récupère le src pour l'affichage.
        if (file.type.match('image')) {
          this.addSrc_(file);
        }
        this.aNewFileList.push(file);
      }
    }

    setTimeout(() => {
      this.refreshListAllFiles_();
      this.setValue_();
    });
  }

  /**
   * Supprime un fichier
   * @param {File} file
   */
  public removeFile(file: File) {
    this.translate.get(['CONFIRM_CONFIRM', 'CONFIRM_CANCEL', 'FILE_PICKER_DELETE_FILE'])
      .subscribe((translations: any) => {
        vitisAlert.alert({
          title: translations.FILE_PICKER_DELETE_FILE,
          type: 'question',
          showCancelButton: true,
          confirmButtonColor: '#3085d6',
          cancelButtonColor: '#d33',
          confirmButtonText: translations.CONFIRM_CONFIRM,
          cancelButtonText: translations.CONFIRM_CANCEL
        }).then((result) => {
          if (result.value) {

            // Si c'est un fichier déjà upload
            const iIndexUnChanged = this.findFileInFiles_(file, this.aUnchangedFileList);
            if (iIndexUnChanged !== -1) {
              // On ajout le fichier à la liste des supprimés et on l'enlève des fichiers inchangé
              this.aDeletedFileList.push(this.aUnchangedFileList[iIndexUnChanged].name);
              this.aUnchangedFileList.splice(iIndexUnChanged, 1);

            }

            // Si c'est un fichier qui vient d'être choisi
            const iIndexNew = this.findFileInFiles_(file, this.aNewFileList);
            if (iIndexNew !== -1) {
              this.aNewFileList.splice(iIndexNew, 1);
            }

            this.refreshListAllFiles_();
            this.setValue_();
            // On vide la valeur du input pour qu'il puisse à nouveau detecter un changement
            this.inputFile.nativeElement.value = '';
          }
        });
      });
  }

  /**
   * Télécharge un fichier
   * @param {any} iFileIndex
   */
  public downloadFile(iFileIndex: any) {
    if (this.aAllFiles[iFileIndex]) {
      const oFile = this.aAllFiles[iFileIndex];
      if (this.vitisCheck.isDefAndNotNull(oFile._url)) {
        this.downloadBlobFile_(oFile._url, 5, false).then((oBlob) => {
          this.downloadFile_(oFile, oBlob);
        });
      } else {
        this.downloadFile_(oFile);
      }
    }
  }

  /**
   * Vérifie si un fichier ce trouve dans la liste de fichier
   * @param {File} file Fichier à trouver
   * @param {Array<File>} files Liste des fichiers à vérifier
   * @returns {number}
   */
  private findFileInFiles_(file: File, files: Array<File>): number {
    return files.findIndex((localFile) => localFile.name === file.name);
  }

  /**
   * Rafraichit la liste de tout les fichiers en concatenant les nouveaux fichiers avec les fichiers inchangés
   */
  private refreshListAllFiles_() {
    this.aAllFiles = [...this.aUnchangedFileList.concat(this.aNewFileList)];
  }

  /**
   * Place les valeurs de la variable des fichiers inchangés dans les fichier à supprimer
   * et vide les variables des fichiers inchangés et des nouveaux fichiers
   */
  private removeAllFiles_() {
    this.aDeletedFileList = this.aUnchangedFileList.map(this.getFileName, this);
    this.aUnchangedFileList.length = 0;
    this.aNewFileList.length = 0;
  }

  /**
   * Retourne la liste des fichiers, avec les fichiers ajoutés, supprimés et inchangés.
   * @returns {any}
   */
  public getValue(): any {
    const oFileList: any = {
      files: this.aNewFileList,
      vitis_deleted_files: this.aDeletedFileList,
      vitis_unchanged_files: this.aUnchangedFileList.filter(this.getFileName, this).map(this.getFileName, this),
      vitis_thumbnail_image_files: []
    };

    if (this.sFilePickerType == 'image') {
      oFileList.vitis_thumbnail_image_files = this.aAllFiles.filter(this.getFileName, this).map(this.getFileName, this);
    }

    return oFileList;
  }

  /**
   * Raifraichit les données
   */
  public refresh() {
    setTimeout(() => {
      this.initFilePickerData_();
    });
  }

  /**
   * Réinitilise les données
   */
  public reset() {
    this.aAllFiles = [];
    this.aNewFileList = [];
    this.aUnchangedFileList = [];
    this.aDeletedFileList = [];
    this.setValue_();
  }

  /**
   * Récupère les fichiers initiaux à partir de la liste aUrls
   */
  private getFiles_(): Promise<any> {
    return new Promise((resolve, reject) => {
      let iCounter = 0;
      const aFiles = [];
      if (this.vitisCheck.isDefAndNotNull(this.aUrls)) {
        if (this.vitisCheck.isArray(this.aUrls)) {
          for (let i = 0; i < this.aUrls.length; i++) {
            const url = this.aUrls[i];
            this.downloadBlobFile_(url, 5, true).then((file) => {
              aFiles.push(file);
              iCounter++;
              if (iCounter === this.aUrls.length) {
                resolve(aFiles);
              }
            });
          }
        } else {
          this.downloadBlobFile_(this.aUrls, 5, true).then((file) => {
            aFiles.push(file);
            resolve(aFiles);
          });
        }

      }
    });
  }

  /**
   * Récupère un fichier du serveur
   * @param fileUrl Url du fichier qu'il faut récupérer
   * @param iRetry Nombre de tentative pour récupérer correctement le fichier (Nom du fichier dans le header)
   */
  private downloadBlobFile_(fileUrl: string, iRetry: number, bOnlyMetadata = false): Promise<any> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: fileUrl,
        responseType: 'blob',
        headers: {
          Accept: 'application/octet-stream'
        },
        params: bOnlyMetadata ? { type: 'document' } : {},
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          const oBlob = resp.data;
          const name = this.findFileNameInHeaders_(resp.headers);

          if (this.vitisCheck.isDefAndNotNull(name)) {
            oBlob.name = name;
            oBlob._url = fileUrl;
            oBlob.src = fileUrl;
            resolve(oBlob);
          } else if (iRetry > 0) {
            iRetry--;
            this.downloadBlobFile_(fileUrl, iRetry, bOnlyMetadata);
          } else {
            oBlob.name = fileUrl.split('/')[fileUrl.split('/').length - 1];
            oBlob._url = fileUrl;
            oBlob.src = fileUrl;
            resolve(oBlob);
          }
        } else {
          reject();
        }
      });
    });
  }

  /**
   * Télécharge le fichier
   * @param {any} oFile
   * @param {any|undefined} oBlob
   */
  private downloadFile_(oFile: any, oBlob?: any) {
    const sFileName = oFile.name;

    if (!this.vitisCheck.isDefAndNotNull(oBlob)) {
      // Trouve le fichier correspondant
      oBlob = this.aAllFiles.filter(file => file.name == oFile.name)[0];
    }

    // Télécharge le fichier
    if (this.vitisCheck.isDefAndNotNull(oBlob)) {

      // IE
      if (window.navigator['msSaveOrOpenBlob']) {
        window.navigator['msSaveOrOpenBlob'](oBlob, sFileName);
      } else {
        const a = document.createElement('a');
        const url = window.URL.createObjectURL(oBlob);
        document.body.appendChild(a);
        a.href = url;
        a.download = sFileName;
        a.click();
        window.URL.revokeObjectURL(url);
      }
    }
  }

  /**
   * Récupère le nom du fichier depuis les headers
   *
   * @param  {any} oHeaders Headers
   * @return {string}          Nom du fichier
   */
  private findFileNameInHeaders_(oHeaders: any): string {
    let sName: string;

    if (this.vitisCheck.isDefAndNotNull(oHeaders)) {

      let sContentDisp = oHeaders['content-disposition'];

      // IE
      if (!this.vitisCheck.isDefAndNotNull(sContentDisp) &&
        this.vitisCheck.isDefAndNotNull(oHeaders['Content-disposition'])) {
        sContentDisp = oHeaders['Content-disposition'];
      }

      if (this.vitisCheck.isString(sContentDisp)) {
        const aContentDisp = sContentDisp.split(';');
        for (let i = 0; i < aContentDisp.length; i++) {
          if (aContentDisp[i].indexOf('filename=') !== -1) {
            sName = aContentDisp[i].split('=')[1].replace(/\"/g, '');
          }
          if (aContentDisp[i].indexOf('filename*=utf-8\'\'') !== -1) {
            sName = aContentDisp[i].split('=utf-8\'\'')[1];
          }
        }
      }
    }

    if (this.vitisCheck.isDefAndNotNull(sName)) {
      sName = decodeURI(sName);
    }

    return sName;
  }

  /**
   * Set la valeur avec les urls ou le nom des fichers (quand on a pas l'url)
   */
  private setValue_() {
    const tmpValues = [...this.aUnchangedFileList.concat(this.aNewFileList)];

    this.setParentValue.emit(tmpValues.map(file => {
      if (this.vitisCheck.isDefAndNotNull(file)) {
        return file;
      }
    }));
  }

  /**
   * Retourne les restrictions de type input=file
   *
   * @returns {string}
   */
  public getFormatsRestrict(): string {

    let sRestrictedFormats: string;

    if (!this.vitisCheck.isString(this.oField.formats)) {
      sRestrictedFormats = this.oField.formats.join(',');
    } else {
      sRestrictedFormats = this.oField.formats;
    }

    return sRestrictedFormats;
  }

  /**
   * Retourne le nom du fichier
   * @param {any} file
   */
  public getFileName(file: any): string {
    if (this.vitisCheck.isDefAndNotNull(file) && this.vitisCheck.isDefAndNotNull(file.name)) {
      return file.name;
    }
  }

  /**
   * Trouve un identifiant unique à un élément
   * @returns string
   */
  private getUniqueId_(): string {
    let id = '_' + Math.random().toString(36).substr(2, 9);
    if (document.getElementById(id)) {
      id = this.getUniqueId_();
    }
    return id;
  }
}
