import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';

/**
 * Type liste
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * size: nombre de lignes à utiliser
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 * datasource? : source de données
 * datasource.datasource_id : identifiant de la source de données (spécifiée dans le formulaire)
 * datasource.id_key : attribut utilisé pour la valeur du champ
 * datasource.label_key : attribut à afficher
 * datasource.attributs : attributs à demander séparés par des "|"
 * datasource.sort_order? : type de sort (ASC, DESC)
 * datasource.order_by? : attribut utilisé pour l'ordre
 * datasource.distinct? : valeurs distinctes (true, false)
 * datasource.parameters? : paramètres supplémentaires à passer dans la requêtre GET
 *
 * Exemple:
 * {
 *   "type": "list",
 *   "name": "Timezones",
 *   "label": "Liste",
 *   "required": true,
 *   "nb_cols": 12,
 *   "size": 12,
 *   "datasource": {
 *     "datasource_id": "datasource_timezone_id",
 *     "sort_order": "ASC",
 *     "distinct": "true",
 *     "label_key": "label",
 *     "order_by": "name",
 *     "id_key": "name",
 *     "attributs": "name|label",
 *     "parameters": {
 *       "filter": {"relation":"AND","operators":[{"column":"country","compare_operator":"=","value":"={{country}}"}]}
 *     }
 *   }
 * }
 */
@Component({
  selector: 'app-list',
  templateUrl: './list.component.html',
  styleUrls: ['./list.component.scss']
})
export class ListComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue();
    };

    // Fonction setValue
    this.oField.setValue = (sValue: any) => {
      return this.setValue(sValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
  }

  /**
   * Retourne les valeurs sélectionnées de la liste au format texte séparées par des "|"
   * @param oFormValues
   * @param oField
   *
   * @returns {String}
   */
  public getValue(): String {
    let sReturn: any = '';
    if (this.oFormValues) {
      if (this.oFormValues.value) {
        if (this.oFormValues.value[this.oField.name]) {
          sReturn = this.oFormValues.value[this.oField.name];

          // Retour au format string séparé par des "|"
          if (this.vitisCheck.isArray(sReturn)) {
            sReturn = sReturn.join('|');
          }
        }
      }
    }
    return sReturn;
  }

  /**
   * Setter de valeur du champ
   * @param  {string|array} oValue
   */
  public setValue(aValue: any) {

    if (this.vitisCheck.isString(aValue)) {
      aValue = aValue.split('|');
    }

    const oPatchValues = {};
    if (this.vitisCheck.isDefAndNotNull(this.oField)) {
      if (this.vitisCheck.isDefAndNotNull(this.oField.name)) {
        oPatchValues[this.oField.name] = aValue;
      }
    }

    this.oFormValues.patchValue(oPatchValues);

  }
}
