import { Component, OnInit, Input, ViewChild, ElementRef, Output, EventEmitter } from '@angular/core';
import { FormreaderField } from '../../formreader-field';
import { MapCoreComponent } from 'src/app/vitis/map-core/map-core.component';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { get as getProjection, transformExtent, transform as projTransform } from 'ol/proj';
import * as olExtent from 'ol/extent';
import Feature from 'ol/Feature';
import { MapCoreFormreaderService } from 'src/app/vitis/map-core/services/map-core-fromreader/map-core-formreader.service';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { Geometry } from 'ol/geom';

@Component({
  selector: 'app-map-formreader',
  templateUrl: './map-formreader.component.html',
  styleUrls: ['./map-formreader.component.scss']
})
export class MapFormreaderComponent extends FormreaderField implements OnInit {

  @ViewChild(MapCoreComponent) mapCoreComponent: MapCoreComponent;

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  public sGeom: string;

  @Output() public setFeatureEmitter = new EventEmitter();
  @Output() public selectFeatureEmitter = new EventEmitter();


  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    private mapUtilsService: MapCoreUtilsService,
    public element: ElementRef,
    private mapCoreManagerService: MapCoreManagerService,
    private mapCoreFormreaderService: MapCoreFormreaderService,
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any, comeFromMap: boolean = false) => {
      this.sendValueOnMap(oValue, comeFromMap);
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);


    if (this.vitisCheck.isDefAndNotNull(this.oField) && this.oField.disabled === true) {
      if (this.vitisCheck.isDefAndNotNull(this.oField.map_options)
        && this.vitisCheck.isDefAndNotNull(this.oField.map_options.interactions)) {
        this.oField.map_options.interactions.remove_all = false;
        this.oField.map_options.interactions.remove = false;
        this.oField.map_options.interactions.modify = false;
        this.oField.map_options.interactions.draw_point = false;
        this.oField.map_options.interactions.draw_line = false;
        this.oField.map_options.interactions.draw_polygon = false;
      }
    }

    // Ajout fonctions
    this.oField.setCenterAndViewFromProj = (sProj: string) => this.setCenterAndViewFromProj(sProj);
    this.oField.refreshMap = () => { this.refreshMap(); };
    this.oField.getFeatures = () => this.mapCoreComponent.getDrawFeatures();
    this.oField.getMapComponent = () => this.mapCoreComponent;
    this.oField.setExtent = (aExtent: Array<number>) => this.setExtent(aExtent);
    this.oField.transform = (coordinate: Array<number>, source: string, destination: string) => projTransform(coordinate, source, destination);
    this.oField.transformExtent = (extent: Array<number>, source: string, destination: string) => transformExtent(extent, source, destination);
  }

  /**
   * Set la valeur dans le format de sortie configurer pour ce champ
   * @param oValue Valeur en ewkt
   */
  public setValueReceived(oValue: any) {
    if (this.vitisCheck.isDefAndNotNull(this.oField.map_options) &&
      this.vitisCheck.isDefAndNotNull(this.oField.map_options.output_format)) {
      switch (this.oField.map_options.output_format) {
        case 'wkt':
          oValue = this.mapUtilsService.getWKTFromEWKT(oValue).geom;
          break;

        case 'geojson':
          oValue = JSON.stringify(this.mapUtilsService.getGeojsonFromEWKT(oValue));
          break;

        default:
          break;
      }
    }
    this.oField.setValue(oValue, true);
  }

  /**
   * Relai pour selectFeatureEmitter
   * @param oValue
   */
  public selectFeatureReciver(oValue: any) {
    this.selectFeatureEmitter.emit(oValue);
  }

  /**
   * Envoie la valeur à la carte pour créer les features
   */
  private sendValueOnMap(sGeom: string, comeFromMap: boolean) {
    if (sGeom !== '') {
      // Si le setValue ne vient pas de la carte, on ajoute la valeur a la carte
      if (!this.vitisCheck.isDefAndNotNull(comeFromMap) || comeFromMap === false) {

        // Si le format est geojson on parse la valeur
        if (this.vitisCheck.isDefAndNotNull(this.oField.map_options)
          && this.vitisCheck.isDefAndNotNull(this.oField.map_options.output_format)) {
          if (this.oField.map_options.output_format === 'geojson') {
            sGeom = JSON.parse(sGeom);
          }
        }

        setTimeout(() => {
          this.sGeom = sGeom;
        }, 500);
      }
    }
  }

  /**
   * Set proj, center et view  a partir de la projection
   * @param sProj
   * @returns oMapField
   */
  public setCenterAndViewFromProj(sProj: string): any {
    this.oField.map_options.proj = sProj;

    this.mapCoreManagerService.addCustomProjections();
    // Récupère l'étendu de la carte en fonction de la projection
    const oNewProj = getProjection(this.oField.map_options.proj);
    const aNewProjExtent = oNewProj.getExtent();

    // Set proj et étendu au map_json
    this.oField.map_options.center.extent = aNewProjExtent;
    this.oField.map_options.center.coord = olExtent.getCenter(aNewProjExtent);
    this.oField.map_options.tree.children[0].view.extent = aNewProjExtent;
    this.oField.map_options.tree.children[0].view.projection = this.oField.map_options.proj;
    return this.oField;
  }

  /**
   * Lance le reload d'une carte
   */
  public refreshMap() {
    if (this.vitisCheck.isDefAndNotNull(this.oField)) {
      // Evenement pour modifier la carte
      this.mapCoreFormreaderService.emitMapFormreaderEvent(this.oField.id, {
        action: 'reload'
      });
    }
  }

  /**
   * Fait suivre la feature
   * @param oFeature
   */
  public setFeature(oFeature: Feature<Geometry> | Array<Feature<Geometry>>) {
    this.setFeatureEmitter.emit(oFeature);
  }

  /**
   * Change l'étendue de la carte
   * @param aExtent
   */
  public setExtent(aExtent: Array<number>) {
    this.mapCoreComponent.olMap.getView().fit(aExtent);
  }
}
