import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { Validators } from '@angular/forms';

/**
 * Type mot de passe
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 * pattern? : regex à utiliser pour contraintre la saisie
 * use_restriction? : booléen pour activer le check de la valeur d'aprés la properties (non cumulable avec pattern)
 * maxLength? : Nombre de caractères maximum.
 *
 * Exemple:
 * {
 *   "type": "password",
 *   "name": "password",
 *   "label": "Mot de passe",
 *   "required": true,
 *   "pattern": "^(?=.*\d)(?=.*[a-z])(?=.*[A-Z])(?!.*\s).*$",
 *   "use_restriction": false
 *   "nb_cols": 12
 * }
 */
@Component({
  selector: 'app-password',
  templateUrl: './password.component.html',
  styleUrls: ['./password.component.scss']
})
export class PasswordComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  public bShowPassword = false;
  public bShowEye = false;
  private sStarificationChar_ = '****';

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      if (oValue !== this.sStarificationChar_ &&
        oValue !== '') {
        this.bShowEye = true;
      }
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    // si la restriction est active on ajoute le validateur
    if (this.oField.use_restriction === true) {
      this.updateFieldValidator(this.oField);
    }
  }

  /**
   * Méthode appelée au premier click sur l'élément
   */
  public handleFirstClick() {
    const sVal = this.oField.getValue();
    if (sVal === this.sStarificationChar_) {
      this.oField.setValue('');
    }
  }

  /**
   * Méthode appelée lorsuq'une lettre est tapée
   */
  public handleChange() {
    const sVal = this.oField.getValue();
    if (sVal !== this.sStarificationChar_ &&
      sVal !== '') {
      this.bShowEye = true;
    }
  }

  /**
   * Permet de visualiser le mot de passe précédemment tapé
   */
  public showPassword() {
    const sVal = this.oField.getValue();
    this.handleFirstClick();
    if (sVal !== '') {
      this.bShowPassword = !this.bShowPassword;
    }
  }

  /**
   * Création d'un validateur spécifique
   * @param oField
   */
  private updateFieldValidator(oField: any) {
    let aValidators = [];
    // Récupère les validators inclus par FormReader
    if (this.vitisCheck.isArray(oField.validators_list)) {
      aValidators = oField.validators_list;
    }

    // Expression régulière correspondant au paramètre de restriction de mot de passe (properties).
    const iPasswordRrestriction = parseInt(environment.properties.password_restriction, 10);
    let sPattern: string;
    let sErrorMsg: string;
    if (!isNaN(iPasswordRrestriction) && iPasswordRrestriction != 1) {
      switch (iPasswordRrestriction) {
        // 8 caractères.
        case 2:
          sPattern = '^(.){8,}$';
          sErrorMsg = 'FORM_PASSWORD_RESTRICTION_CONFIGURATION_8_CHAR';
          break;
        // 8 caractères dont une majuscule.
        case 3:
          sPattern = '^(?=.*[A-Z])(.){8,}$';
          sErrorMsg = 'FORM_PASSWORD_RESTRICTION_CONFIGURATION_8_CHAR_1_UP';
          break;
        // 8 caractères, dont une majuscule et un caractère spécial (compatible RGPD).
        case 4:
          sPattern = '^(?=.*[\\W])(?=.*[A-Z])(.){8,}$';
          sErrorMsg = 'FORM_PASSWORD_RESTRICTION_CONFIGURATION_8_CHAR_1_UP_1_SPEC';
          break;
        // 12 caractères, dont une majuscule et un caractère spécial (compatible RGPD).
        case 5:
          sPattern = '^(?=.*[\\W])(?=.*[A-Z])(.){12,}$';
          sErrorMsg = 'FORM_PASSWORD_RESTRICTION_CONFIGURATION_12_CHAR_1_UP_1_SPEC';
          break;
      }
      // Restriction de mot de passe dans le validateur.
      if (this.vitisCheck.isDefAndNotNull(sPattern)) {
        aValidators.push(Validators.pattern(sPattern));
        this.oFormValues.controls[oField.name].setValidators(aValidators);

        if (!oField.validation_messages) {
          oField.validation_messages = [];
        }
        oField.validation_messages.push({
          type: 'pattern',
          message: this.translate.instant('FORM_PASSWORD_RESTRICTION_CONFIGURATION_VALIDATOR') + ' ' + this.translate.instant(sErrorMsg)
        });
      }
    }
  }
}
