import { Component, OnInit, Input, ElementRef, AfterViewInit, ChangeDetectorRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';

interface SelectOption {
  value: string;
  label: string;
}

/**
 * Type liste déroulante avec recherche
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 * datasource? : source de données
 * datasource.datasource_id : identifiant de la source de données (spécifiée dans le formulaire)
 * datasource.id_key : attribut utilisé pour la valeur du champ
 * datasource.label_key : attribut à afficher
 * datasource.attributs : attributs à demander séparés par des "|"
 * datasource.sort_order? : type de sort (ASC, DESC)
 * datasource.order_by? : attribut utilisé pour l'ordre
 * datasource.distinct? : valeurs distinctes (true, false)
 * datasource.parameters? : paramètres supplémentaires à passer dans la requêtre GET
 *
 * Exemple:
 * {
 *   "type": "searchable_select",
 *   "name": "Timezone",
 *   "label": "Liste",
 *   "required": true,
 *   "nb_cols": 12,
 *   "datasource": {
 *     "datasource_id": "datasource_timezone_id",
 *     "sort_order": "ASC",
 *     "distinct": "true",
 *     "label_key": "label",
 *     "order_by": "name",
 *     "id_key": "name",
 *     "attributs": "name|label",
 *     "parameters": {
 *       "filter": {"relation":"AND","operators":[{"column":"country","compare_operator":"=","value":"={{country}}"}]}
 *     }
 *   }
 * }
 */
@Component({
  selector: 'app-searchable-select',
  templateUrl: './searchable-select.component.html',
  styleUrls: ['./searchable-select.component.scss']
})
export class SearchableSelectComponent extends FormreaderField implements OnInit, AfterViewInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;
  @Input() oFormReader: any;

  public sSelectedVal: string = '';
  public sFilter: string = '';
  public dropdownId = `dropdown-${Date.now()}`;
  public filterId = `filter-${Date.now()}`;

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef,
    private chRef: ChangeDetectorRef
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      this.setValue(this.oFormValues, this.oField, String(oValue));
      this.refreshLabel();
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Observe l'arrivée des choix (pour utilisation avec datasource)
    if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
      if (this.vitisCheck.isDefAndNotNull(this.oField.options.observableChoises)) {
        this.oField.options.observableChoises.subscribe((aChoises: any) => {

          // Remplace les valeurs des options de type nombre au format string
          this.makeOptionsAsString_(this.oField.options.choises);

          // Set valeur "" si aucune valeur
          setTimeout(() => {
            if (!this.vitisCheck.isDefAndNotNull(this.oField.getValue()) && !this.oField.disabled) {
              this.oField.setValue('');
            } else {
              this.refreshLabel();
            }
          });
        });
      }
    }
  }

  ngAfterViewInit() {
    $('#' + this.dropdownId).on('shown.bs.dropdown', () => {

      setTimeout(() => {

        // Focus sur le champ recherche
        if (document.getElementById(this.filterId)) {
          document.getElementById(this.filterId).focus();
        }

        // Gestion en cas d'overflow du composant parent
        const dropdownFieldLeft = $('#' + this.dropdownId).offset().left;
        const dropdownFieldBottom = ($('#' + this.dropdownId).offset().top + $('#' + this.dropdownId).height());
        const dropdownMenuHeight = $('#' + this.dropdownId).find('.dropdown-menu').height();
        const dropdownMenuBottom = dropdownFieldBottom + dropdownMenuHeight;
        const windowHeight = $(window).height();
        const newCss: any = {
          position: 'fixed',
          left: dropdownFieldLeft,
          top: dropdownFieldBottom,
          opacity: 1,
          width: $('#' + this.dropdownId).width()
        };
        if (dropdownMenuBottom < windowHeight) {
          newCss.transform = 'none';
        }
        $('#' + this.dropdownId).find('.dropdown-menu').css(newCss);

        // Permet de garder une hauteur fixe même quand on filtre
        $('#' + this.dropdownId).find('.dropdown-menu').find('.result-list').css({
          height: $('#' + this.dropdownId).find('.dropdown-menu').find('.result-list').height()
        });
      });

    });
    $('#' + this.dropdownId).on('hidden.bs.dropdown', () => {

      // Vide le filtre
      this.sFilter = '';
      this.chRef.detectChanges();

      // gestion UI overflow
      $('#' + this.dropdownId).find('.dropdown-menu').css({
        opacity: 0,
      });
      $('#' + this.dropdownId).find('.dropdown-menu').find('.result-list').css({
        height: 'auto'
      });
    });
  }

  /**
   * Retourne true / false suivant si il faut afficher une option vide
   * @returns boolean
   */
  public hasEmptyOption(): boolean {
    let firstOption = true;

    if (this.vitisCheck.isDefAndNotNull(this.oField)) {
      if (this.oField.required) {
        firstOption = false;
      }

      if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
        if (this.vitisCheck.isDefAndNotNull(this.oField.options.choises)) {
          if (this.vitisCheck.isDefAndNotNull(this.oField.options.choises[0])) {
            if (this.oField.options.choises[0].value === '') {
              firstOption = false;
            }
          }
        }
      }
    }

    return firstOption;
  }

  /**
   * Permet de rafraîchir le libellé affiché dans la liste déroulante
   */
  public refreshLabel() {
    const currentVal = this.getValue(this.oFormValues, this.oField);
    this.sSelectedVal = '';
    if (this.oField && this.oField.options && this.vitisCheck.isArray(this.oField.options.choises)) {
      const selectedOption = this.oField.options.choises.find(o => {
        return o.value === currentVal;
      });
      if (selectedOption && !this.vitisCheck.isEmpty(selectedOption.label)) {
        this.sSelectedVal = selectedOption.label;
      }
    }
  }

  /**
   * Appelé au click sur une des options
   * @param option
   */
  public selectValue(option: SelectOption) {
    if (option && this.vitisCheck.isDefAndNotNull(option.value)) {
      this.setValue(this.oFormValues, this.oField, String(option.value));
      this.refreshLabel();
    }
  }

  /**
   * Filtre sur les options du select
   * @param options
   * @returns
   */
  public filterOptions(options: Array<SelectOption>) {
    const result: SelectOption[] = [];

    if (options) {
      for (let i = 0; i < options.length; i++) {
        const option = options[i];
        if (option.label.toLowerCase().indexOf(this.sFilter.toLowerCase()) !== -1) {
          result.push(option);
        } else if (option.value.toLowerCase().indexOf(this.sFilter.toLowerCase()) !== -1) {
          result.push(option);
        }
      }
    }

    return result;
  }

  /**
   * Filtre pour la valeur affichée
   * @returns
   */
  public filterSelectVal() {

    let sReturn = this.sSelectedVal;

    // Si la valeur n'est pas sur la liste
    if (this.oField.options &&
      this.oField.options.choises &&
      this.oField.options.choises.find) {

      if (this.sSelectedVal && !this.oField.options.choises.find(o => {
        return o.label === this.sSelectedVal;
      })) {
        sReturn = '';
        setTimeout(() => {
          this.setEmptyVal();
        });
      }
    }

    return sReturn;
  }

  /**
   * Set valeur vide
   */
  public setEmptyVal() {
    this.selectValue({ label: '', value: '' });
  }

  /**
   * Remplace les valeurs des options de type nombre au format string
   * @param  {Array<any>} aChoises
   * @returns void
   */
  private makeOptionsAsString_(aChoises: Array<any>): void {
    for (let i = 0; i < aChoises.length; i++) {
      if (this.vitisCheck.isNumber(aChoises[i].value)) {
        aChoises[i].value = String(aChoises[i].value);
      }
    }
  }
}
