import { Component, OnInit, Input, ElementRef, AfterViewInit, Renderer2, ViewChild } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';

declare var $: any;

/**
 * Type tagse
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 * datasource? : source de données
 * datasource.datasource_id : identifiant de la source de données (spécifiée dans le formulaire)
 * datasource.id_key : attribut utilisé pour la valeur du champ
 * datasource.label_key : attribut à afficher
 * datasource.attributs : attributs à demander séparés par des "|"
 * datasource.sort_order? : type de sort (ASC, DESC)
 * datasource.order_by? : attribut utilisé pour l'ordre
 * datasource.distinct? : valeurs distinctes (true, false)
 * datasource.parameters? : paramètres supplémentaires à passer dans la requêtre GET
 *
 * Exemple:
 * {
 *     "type": "tags",
 *     "name": "name",
 *     "label": "Nom",
 *     "required": false,
 *     "datasource": {
 *       "datasource_id": "datasource_tags_id",
 *       "sort_order": "ASC",
 *       "label_key": "tag",
 *       "id_key": "tag"
 *     },
 *     "nb_cols": 12
 * }
 */

@Component({
  selector: 'app-tags',
  templateUrl: './tags.component.html',
  styleUrls: ['./tags.component.scss']
})
export class TagsComponent extends FormreaderField implements OnInit, AfterViewInit {

  @ViewChild('tagInput', { static: false }) tagInput: ElementRef;

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  public aTags: Array<string>;
  public aSuggestions: Array<Object>;
  public sTagInput: string;
  public oTagInput: HTMLInputElement;

  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef,
    private renderer: Renderer2
  ) {
    super(translate, vitisCheck);
  }

  ngOnInit() {
    // Initialisation des suggestions des tags
    this.aSuggestions = [];

    // initialisation des tags
    this.aTags = [];

    // Initialisation du champs d'ajout d'un tag
    this.sTagInput = '';

    // Observe l'arrivée des choix (pour utilisation avec datasource)
    if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
      if (this.vitisCheck.isDefAndNotNull(this.oField.options.observableChoises)) {
        // Lorsque la requête se termine
        this.oField.options.observableChoises.subscribe((aChoises: any) => {
          this.filterSuggestionsList();
        });
      }
    }

    /**
     * Fonction getValue
     */
    this.oField.getValue = () => this.getValue(this.oFormValues, this.oField);

    /**
     * Fonction setValue
     */
    this.oField.setValue = (oValue: any) => {
      // Initialise les tags
      this.aTags = (oValue.length > 0) ? oValue.split(',') : [];
      // setValue
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    /**
     * Fonction pushTag
     */
    this.oField.pushTag = (sTag: string) => {
      // Ne fait rien si ce tag est vide ou a été déjà ajouté
      if (sTag.length === 0 || this.aTags.includes(sTag)) {
        return;
      }
      let sValue = sTag;
      if (this.vitisCheck.isDefAndNotNull(this.oField.getValue()) && this.oField.getValue() != '') {
        sValue = this.oField.getValue() + ',' + sTag;
      }
      this.oField.setValue(sValue);
      // Vide TagInput
      this.sTagInput = '';
      this.filterSuggestionsList();
      // Met à jour la position des suggestions
      if (this.aSuggestions.length > 0) {
        $('#' + this.oField.id + '-tagsinput').dropdown('update');
      }
    };

    /**
     * Fonction popTag
     */
    this.oField.popTag = (sTag: string) => {
      const iIdx = this.aTags.indexOf(sTag);
      // Ne fait rien si sTag n'est pas été trouvé dans le tableau
      if (iIdx < 0) {
        return;
      }
      // Enleve l'élément
      this.aTags.splice(iIdx, 1);
      this.oField.setValue(this.aTags.join(','));
      // Simule clique pour garder la liste des suggestions ouverte
      this.element.nativeElement.click();
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
  }

  ngAfterViewInit() {
    // Get NativeElement
    this.oTagInput = this.tagInput.nativeElement;
  }

  /**
   * Redimension tagInput
   * @returns void
   */
  public onTagInputChange(): void {
    // Filtre les suggestions
    this.filterSuggestionsList();
    // Réinitialiser la taille d'input
    this.renderer.setStyle(this.oTagInput, 'width', '2em');
    // Redimensionne la taille d'input
    this.renderer.setStyle(this.oTagInput, 'width', this.oTagInput.scrollWidth + 'px');
  }

  /**
   * À la pression sur 'Enter' ajout un tag s'il n'est pas vide et désactiver l'envoi du formulaire
   * @returns boolean
   */
  public onEnterPressed(): boolean {
    if (this.sTagInput.length > 0) {
      this.oField.pushTag(this.sTagInput);
      return false;
    }
    return true;
  }

  /**
   * Filtre la liste des suggestion pour l'autocompletion
   * @returns void
   */
  public filterSuggestionsList(): void {
    const aChoises = this.oField.options.choises;
    if (aChoises.length === 0) {
      return aChoises;
    }
    // Filtre les suggestions des tags
    this.aSuggestions = aChoises.filter(oChoise => {
      if (this.aTags.includes(oChoise.value)) {
        return false;
      }
      if (this.sTagInput.length === 0) {
        return true;
      }
      return oChoise.label.indexOf(this.sTagInput) >= 0;
    });
  }
}
