import { Component, OnInit, Input, ViewChild, ElementRef } from '@angular/core';
import { Subject } from 'rxjs';
import { debounceTime, distinctUntilChanged } from 'rxjs/operators';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapJSON, GeoCoder } from 'src/app/vitis/map-core/services/map-models.model';

import { Map } from 'ol';
import { transform } from 'ol/proj.js';
import Feature from 'ol/Feature';
import GeoJSON from 'ol/format/GeoJSON';
import { MapCorePopupService, Popup } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';
import { Geometry } from 'ol/geom';

interface LocationService {
  id: string;
  label: string;
  value: string;
}

@Component({
  selector: 'app-core-controls-search',
  templateUrl: './core-controls-search.component.html',
  styleUrls: ['./core-controls-search.component.scss']
})
export class CoreControlsSearchComponent implements OnInit {

  @Input() olMap: Map;
  @ViewChild('inputSearch', { read: ElementRef }) inputSearch: ElementRef;

  public txtQueryChanged: Subject<string> = new Subject<string>();

  private oLocationGeocoders: any = {};
  public sSelectedLocationService: string;

  public sSearch: string = '';
  public noResults: string = '';
  public bSearching: boolean;
  public sLocationServiceType: string;
  public locationResults: string | any;
  public bLocationOpen: boolean = false;
  public bResultsOpen: boolean = false;
  public oLocationPopup: Popup;
  public mapJSON: MapJSON;

  private geocoders = { 'OSM': { 'title': 'OpenStreetMap', 'url': 'https://nominatim.openstreetmap.org/search?format=json&addressdetails=1&limit=[limit]&extratags=1&namedetails=1&polygon_geojson=1&countrycodes=fr&q=[search]', 'data_field': 'data', 'title_field': 'display_name', 'geojson_field': 'geojson', 'summary_fields': [{ 'key': 'address.country', 'label': 'Pays' }, { 'key': 'address.state', 'label': 'Région' }, { 'key': 'address.postcode', 'label': 'Code postal' }, { 'key': 'extratags.population', 'label': 'Population' }] }, 'BAN': { 'title': 'Base nationale', 'url': 'https://api-adresse.data.gouv.fr/search/?q=[search]&limit=[limit]', 'data_field': 'data.features', 'title_field': 'properties.label', 'geojson_field': 'geometry', 'summary_fields': [{ 'key': 'properties.context', 'label': 'Département' }, { 'key': 'properties.city', 'label': 'Ville' }, { 'key': 'properties.postcode', 'label': 'Code postal' }] } };

  /**
   * Liste des services de geocoder, trié dans l'ordre pour l'affichage
   */
  public aLocationServices: Array<LocationService> = [];

  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private mapCorePopupService: MapCorePopupService,
  ) { }

  ngOnInit(): void {

    this.mapJSON = this.olMap.get('map_json');

    this.oLocationGeocoders = this.getLocationGeoCoders_();
    this.sSelectedLocationService = this.getDefaultLocationService_();
    this.aLocationServices = this.getOrderedGeoCodersList_(this.oLocationGeocoders);

    // Evenement lancé quand on écrit dans l'input grâce au ngModelChange
    this.txtQueryChanged.pipe(
      debounceTime(500), // Attend 1s après le dernier evenement avant d'émettre le dernier evenement
      distinctUntilChanged() // emit que si la valeur est différente de la valeur précédente
    ).subscribe(sLocationSearch => {
      this.searchLocation();
    });
  }

  /**
   * Lance la recherche
   */
  public searchLocation(limit: number = 4) {
    const location = (this.vitisCheck.isDefAndNotNull(this.inputSearch)) ? this.inputSearch.nativeElement.value : '';

    // Cache la popup de résultat de la recherche
    this.bResultsOpen = false;
    // Si le champ de recherche est vide on ne poursuit pas la recherche
    if (location === '') {
      return;
    }

    // Affiche la popup pour le résultat
    this.bResultsOpen = true;

    let url = null;
    let oLocationService: GeoCoder;
    // Géocodeur paramétré
    if (this.vitisCheck.isDefAndNotNull(this.oLocationGeocoders[this.sSelectedLocationService])) {
      this.sLocationServiceType = 'geocoder';
      oLocationService = this.oLocationGeocoders[this.sSelectedLocationService];

      if (this.vitisCheck.isDefAndNotNull(oLocationService.url)) {
        url = oLocationService.url.replace('[limit]', limit.toString()).replace('[search]', location);
      }
    }

    this.bSearching = true;
    this.noResults = '';

    if (this.vitisCheck.isDefAndNotNull(url)) {
      this.vitisRequest.ajaxRequest({
        method: 'GET',
        url: url,
        timeout: 5000,
        ajaxLoader: false,
        success: (resp: any) => {
          this.bSearching = false;
          this.locationResults = [];
          this.noResults = '';

          // Vérifie si il y a des données à afficher
          if (!this.vitisCheck.isDefAndNotNull(resp.data)) {
            this.noResults = 'COMPONENT_MAP_LOCATION_NO_RESULT';
            console.error('Pas de données à afficher');
            return 0;
          }

          // Vérifie si il y a une erreur
          if (this.vitisCheck.isDefAndNotNull(resp.data.error)) {
            if (this.vitisCheck.isDefAndNotNull(resp.data.error.errorMessage)) {
              console.error('Erreur réponse localisation :' + resp.data.error.errorMessage);
              this.noResults = resp.data.error.errorMessage;
              return resp.data.error;
            }
          }

          // Géocoder
          if (this.vitisCheck.isDefAndNotNull(this.oLocationGeocoders[this.sSelectedLocationService])) {
            const aResponse = this.getFieldValue_(oLocationService.data_field, resp);

            for (let i = 0; i < aResponse.length; i++) {
              const oResult: any = {};

              // title
              oResult.title = this.getFieldValue_(oLocationService.title_field, aResponse[i]);

              // description_field
              if (this.vitisCheck.isDefAndNotNull(oLocationService.description_field)) {
                oResult.description = this.getFieldValue_(oLocationService.description_field, aResponse[i]);
              }
              if (this.vitisCheck.isDefAndNotNull(oLocationService.geojson_field)) {
                oResult.geojson = this.getFieldValue_(oLocationService.geojson_field, aResponse[i]);
              }
              if (this.vitisCheck.isArray(oLocationService.summary_fields)) {
                oResult.summary = [];
                for (let ii = 0; ii < oLocationService.summary_fields.length; ii++) {
                  oResult.summary.push({
                    label: oLocationService.summary_fields[ii].label,
                    value: this.getFieldValue_(oLocationService.summary_fields[ii].key, aResponse[i])
                  });
                }
              }

              oResult.collapsed = true;
              this.locationResults.push(oResult);
            }
          }

          if (!this.vitisCheck.isDefAndNotNull(this.locationResults) || this.locationResults.length === 0) {
            this.noResults = 'COMPONENT_MAP_LOCATION_NO_RESULT';
            return 0;
          }
        },
        error: () => {
          // Cache le gif de chargement
          this.bSearching = false;
          this.locationResults = [];
          this.noResults = 'COMPONENT_MAP_LOCATION_REQUEST_ERROR';
          console.error('Requête localisation en erreur : ', url);
        }
      });
    }

  }

  /**
   * Ferme la popup de resultat et enlève le focus du champs de recherche
   */
  public removeLocation() {
    this.bResultsOpen = false;
    this.inputSearch.nativeElement.blur();
  }

  /**
   * Zoom sur un lieu
   * @param {any} place
   * @param {Number} place.lon longitude
   * @param {Number} place.lat latitude
   * @param {string|any} place.geojson geometry
   */
  public locatePlace(place: any): void {

    if (!this.vitisCheck.isDefAndNotNull(place)) {
      console.error('Place not defined');
      return;
    }

    // Ferme le resultat de la recherche
    this.removeLocation();

    const currentProj = this.olMap.getView().getProjection();

    // Centre par rapport aux params lon/lat (au cas où la géométrie soit mal renseingée)
    if (this.vitisCheck.isDefAndNotNull(place.lon) && this.vitisCheck.isDefAndNotNull(place.lat)) {
      const projectedCoordinates = transform([parseFloat(place.lon), parseFloat(place.lat)], 'EPSG:4326', currentProj);
      this.olMap.getView().setCenter(projectedCoordinates);
    }

    if (!this.vitisCheck.isDefAndNotNull(place.geojson)) {
      return;
    }

    this.locateGeom_(place.geojson, 'EPSG:4326', place.title);
  }

  /**
   * Récupère la liste des geocoders dans le MapJSON
   */
  private getLocationGeoCoders_(): any {
    let oGeoCoders: any = {};
    if (this.vitisCheck.isDefAndNotNull(this.mapJSON.map_geocoders)) {
      oGeoCoders = this.mapJSON.map_geocoders;
    } else {
      oGeoCoders = this.geocoders;
    }
    return oGeoCoders;
  }

  /**
   * Retourne le geocodeur par défaut
   */
  private getDefaultLocationService_() {
    let defaultGeocoder = 'OSM'; // Default value

    if (this.vitisCheck.isDefAndNotNull(this.oLocationGeocoders) &&
      this.vitisCheck.isDefAndNotNull(this.mapJSON) &&
      this.vitisCheck.isDefAndNotNull(this.mapJSON.map_default_geocoders) &&
      this.vitisCheck.isDefAndNotNull(this.oLocationGeocoders[this.mapJSON.map_default_geocoders])) {
      defaultGeocoder = this.mapJSON.map_default_geocoders;
    }

    return defaultGeocoder;
  }

  /**
   * Trie la liste des geocoders et retourne un tableau de LocationService
   * @param {any} oGeocoders
   *
   * @returns {Array<LocationService>}
   */
  private getOrderedGeoCodersList_(oGeocoders: any): Array<LocationService> {
    const aReturn = [];
    const aTmp = [];

    // Tableau temporaire
    for (const key in oGeocoders) {
      if (Object.prototype.hasOwnProperty.call(oGeocoders, key)) {
        aTmp.push(oGeocoders[key]);
        aTmp[aTmp.length - 1].id = key;
      }
    }

    for (let i = 0; i < aTmp.length; i++) {
      if (aTmp[i]['id']) {
        aReturn.push({
          'id': aTmp[i]['id'],
          'label': aTmp[i]['title'],
          'value': aTmp[i]
        });
      }
    }

    return aReturn;
  }

  /**
   * Lance la fonction récursive pour récupérer les valeurs d'un objet
   * @param {string} sField
   * @param {any} oObject
   */
  private getFieldValue_(sField: string, oObject: any) {
    return this.recursiveGetObjectValue_(sField.split('.'), oObject);
  }

  /**
   * Récupère les valeurs des clés aFields d'un objet
   * @param {Array<any>} aFields
   * @param {any} oObject
   */
  private recursiveGetObjectValue_(aFields: Array<any>, oObject: any) {
    if (aFields.length === 1) {
      if (this.vitisCheck.isDefAndNotNull(oObject[aFields[0]])) {
        return oObject[aFields[0]];
      } else {
        return null;
      }
    } else if (aFields.length > 1) {
      if (this.vitisCheck.isDefAndNotNull(oObject[aFields[0]])) {
        oObject = oObject[aFields[0]];
        aFields.splice(0, 1);
        return this.recursiveGetObjectValue_(aFields, oObject);
      } else {
        return null;
      }
    } else {
      return null;
    }
  }

  /**
   * Localise une geometrie
   * @param {string|any} geojson geoJSON geometry
   * @param {string} proj geom projection (ex: EPSG:2154)
   * @param {string} message message to display on the popup
   */
  private locateGeom_(geojson: string | any, proj: string, message: string): void {

    if (!this.vitisCheck.isDefAndNotNull(geojson)) {
      console.error('Geojson not defined');
      return;
    }

    // Parse le json si besoin
    if (this.vitisCheck.isString(geojson)) {
      geojson = JSON.parse(geojson);
    }

    const currentProj = this.olMap.getView().getProjection();

    // Prépare la feature
    const oGeoJson = new GeoJSON({
      dataProjection: currentProj
    });
    const geom = oGeoJson.readGeometry(geojson, {
      dataProjection: proj,
      featureProjection: currentProj
    });
    const feature = new Feature({
      geometry: geom
    });

    this.locateFeature_(feature, message);
  }

  /**
   * Localise une feature
   * @param {Feature} feature
   * @param {string} message message to display on the popup
   */
  private locateFeature_(feature: Feature<any>, message: string): void {

    const extent = feature.getGeometry().getExtent();
    const resolution = this.olMap.getView().getResolution();

    // Localise et zoom sur la feature
    switch (feature.getGeometry().getType()) {
      case 'Point':
        this.olMap.getView().setCenter(feature.getGeometry().getCoordinates());
        break;
      case 'MultiPoint':
        this.olMap.getView().fit(extent);
        this.olMap.getView().setZoom(resolution);
        break;
      default:
        this.olMap.getView().fit(extent);
    }

    if (this.vitisCheck.isDefAndNotNull(this.oLocationPopup)) {
      this.oLocationPopup.close();
    }
    // Création d'une popup
    this.oLocationPopup = this.mapCorePopupService.displayPopup(message, feature, this.olMap);
  }
}
