import { Component, OnInit } from '@angular/core';
import { Router, ActivatedRoute } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck, VitisRequest, VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';

/**
 * Composant affichagne un fil d'arianne
 */
@Component({
  selector: 'app-breadcrumb',
  templateUrl: './breadcrumb.component.html',
  styleUrls: ['./breadcrumb.component.scss']
})
export class BreadcrumbComponent implements OnInit {

  public oMode: any;
  public oParams: any;
  public aRoutes: Array<any> = [];

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private vitisRequest: VitisRequest,
    private vitisClone: VitisClone,
    private vitisCheck: VitisCheck,
    private translate: TranslateService,
  ) { }

  ngOnInit() {
    this.route.data.subscribe(oMode => {

      // Vide les routes
      this.aRoutes = [];

      // Le timeout permet d'attendre que this.oParams soit rempli
      setTimeout(() => {
        this.oMode = oMode;

        // Initialise les routes
        if (this.oMode.name) {
          this.initRoutes();
        }
      });
    });
    this.route.params.subscribe(oParams => {
      this.aRoutes = [];
      this.oParams = oParams;
    });
  }

  /**
   * Initialise this.aRoutes
   */
  public initRoutes() {

    // Récupère les modes / objets actifs en analysant app.json
    const aModesObjects = this.recursiveGetObjectsFromModes_(environment.appStruct.modes, []);

    // Ensemble des routes
    this.aRoutes = [...aModesObjects];
  }

  /**
   * Navigue vers une route en rafraîchissant si besoin
   * @param route
   */
  public navigateRoute(route: any) {
    if (route.route) {
      this.router.navigate([route.route]);
    }
  }

  /**
   * Fonction récursive permettant de retourner les modes / objets actifs en analysant app.json
   * @param  {Array<any>} aModes
   * @param  {Array<any>} aObjects
   * @returns Array
   */
  private recursiveGetObjectsFromModes_(aModes: Array<any>, aObjects: Array<any>): Array<any> {

    for (let i = 0; i < aModes.length; i++) {
      const oMode = aModes[i];

      if (this.isRouteActive(oMode.route) || this.isRouteActive(oMode.route + '/:id/:form')) {

        // Route utilisée
        const sRoute = '/dashboard/' + oMode.route;

        // Paramètres de la route
        const oParams = this.getParamsFromRoute_(sRoute, this.router.url);

        // Ajout objet id
        if (oMode.route.indexOf(':id') !== -1) {
          if (this.vitisCheck.isDefAndNotNull(oParams.id)) {
            aObjects.push({
              type: 'id',
              label: oParams.id,
              route: this.replaceParamsFromRoute_(sRoute, oParams),
              data: oParams,
              link: true,
            });
          }
        }

        if (this.vitisCheck.isArray(oMode.objects) || this.vitisCheck.isArray(oMode.sections)) {

          const aSubObjects = this.vitisCheck.isArray(oMode.objects) ? oMode.objects : this.vitisCheck.isArray(oMode.sections) ? oMode.sections : [];

          // Mode
          const newMode = {
            type: 'object',
            label: oMode.label,
            route: sRoute,
            icon: oMode.icon,
            link: true,
            data: oMode,
          };

          // Donne comme lien celui du premier objet
          if (aSubObjects &&
            aSubObjects[0] &&
            aSubObjects[0].route) {
            const sSubRoute = '/dashboard/' + aSubObjects[0].route;
            if (sSubRoute.indexOf(':') === -1) {
              newMode.route = sSubRoute;
              newMode.link = true;
            }
          }

          // Ajoute le mode
          aObjects.push(newMode);

          // Relance la fonction de manière récursive
          aObjects = this.recursiveGetObjectsFromModes_(aSubObjects, aObjects);

        } else {

          // Ajoute l'objet avec lien
          aObjects.push({
            type: 'object',
            label: oMode.label,
            route: sRoute,
            icon: oMode.icon,
            link: oMode.route.indexOf(':') === -1 ? true : false,
            data: oMode,
          });
        }

        // route formulaire
        if (this.isRouteActive(oMode.route + '/:id/:form') && !this.vitisCheck.isArray(oMode.objects)) {

          // Route du formulaire
          const sFormRoute = this.replaceParamsFromRoute_(sRoute, oParams) + '/:id/:form';

          // Paramètres de la route du formulaire
          const oFormParams = this.getParamsFromRoute_(sFormRoute, this.router.url);

          // Ajout objet id
          if (this.vitisCheck.isDefAndNotNull(oFormParams.id)) {
            if (oFormParams.form === 'insert' || oFormParams.id === 'insert') {
              aObjects.push({
                type: 'id',
                label: this.translate.instant('LIST_BUTTON_ADD'),
                route: this.replaceParamsFromRoute_(sFormRoute, oFormParams) + '/insert',
                data: oFormParams,
                link: false
              });
            } else {
              aObjects.push({
                type: 'id',
                label: oFormParams.id,
                route: this.replaceParamsFromRoute_(sFormRoute, oFormParams),
                data: oFormParams,
                link: false,
              });
            }
          }
        }
      }
    }

    return aObjects;
  }

  /**
   * Retourne true/false si la route passée en argument est active
   * @param  {string} sRoute
   * @returns boolean
   */
  public isRouteActive(sRoute: string): boolean {

    if (!this.vitisCheck.isString(sRoute)) {
      return false;
    }

    const isActive = sRoute.indexOf('/:') !== -1 ?
      this.isRouteWithParamsActive(sRoute) :
      this.router.isActive('dashboard/' + sRoute, false);
    return isActive;
  }

  /**
   * Retourne true/false si la route passée en argument est active (en ignorant les paramètres)
   * @param  {string} sRoute
   * @returns boolean
   */
  public isRouteWithParamsActive(sRoute: string): boolean {

    if (!this.vitisCheck.isString(sRoute)) {
      return false;
    }

    let isActive = true;
    const aRoute = ('/dashboard/' + sRoute).split('/');
    const aCurrentRoute = this.router.url.split('?')[0].split('/');

    if (aRoute.length !== aCurrentRoute.length) {
      return false;
    }

    for (let i = 0; i < aRoute.length; i++) {
      if (aRoute[i].substr(0, 1) === ':') {
        continue;
      }
      if (aRoute[i] !== aCurrentRoute[i]) {
        isActive = false;
      }
    }

    return isActive;
  }

  /**
   * Retourne les paramètres utilisés dans la route
   * @param sConfigRoute Route type config utilisant /:id
   * @param sRealRoute Route réelle
   */
  private getParamsFromRoute_(sConfigRoute: string, sRealRoute: string): any {
    const aConfigRoute = sConfigRoute.split('/');
    const aRealRoute = sRealRoute.split('/');
    const oParams = {};

    for (let i = 0; i < aConfigRoute.length; i++) {

      let sRealRouteElem = this.vitisClone.clone(aRealRoute[i]);
      if (sRealRouteElem.indexOf('?') !== -1) {
        sRealRouteElem = sRealRouteElem.substr(0, sRealRouteElem.indexOf('?'));
      }

      if (aConfigRoute[i] !== sRealRouteElem &&
        aConfigRoute[i].indexOf(':') === -1) {
        console.error('routes not matching', aConfigRoute[i], sRealRouteElem);
        break;
      }

      if (aConfigRoute[i].indexOf(':') !== -1) {
        const sParam = aConfigRoute[i].substr(1);
        oParams[sParam] = sRealRouteElem;
      }
    }

    return oParams;
  }

  /**
   * Remplace les paramètres d'une route
   * @param sRoute
   * @param oParams
   */
  private replaceParamsFromRoute_(sRoute: string, oParams: any): string {
    for (const key in oParams) {
      if (Object.prototype.hasOwnProperty.call(oParams, key)) {
        sRoute = sRoute.replace(':' + key, oParams[key]);
      }
    }
    return sRoute;
  }
}
