import { Component, OnInit, Input, OnChanges } from '@angular/core';
import { environment } from 'src/environments/environment';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';

@Component({
  selector: 'app-data-source-local-table',
  templateUrl: './data-source-local-table.component.html',
  styleUrls: ['./data-source-local-table.component.scss']
})
export class StudioDataSourceLocalTableComponent implements OnInit, OnChanges {

  @Input() public sSourceManagerMode: string;
  @Input() public oSelectedDataSource: any;

  public aLocalDatabases: any = [];
  public aLocalDatabasesSchemas: any = [];
  public aLocalSchemaTables: any = [];
  public aTableAttributes: any = [];
  public aTableData: any = [];

  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService
  ) { }

  ngOnInit() {
  }

  ngOnChanges() {
    if (this.oSelectedDataSource.dataType == 'tableValue') {
      this.loadLocalDatabases();
      switch (this.sSourceManagerMode) {
        case 'overview':
        case 'edition':
          break;
        case 'creation':
          // Valeurs par défaut.
          this.oSelectedDataSource.type = 'web_service';
          this.oSelectedDataSource.ressource_id = 'vitis/genericqueries';
          this.oSelectedDataSource.parameters = {
            filter: {
              relation: 'AND',
              operators: []
            }
          };
          break;
      }
    }
  }

  /**
   * Chargement de la liste des bases de données locales.
   * @returns void
   */
  public loadLocalDatabases = function () {

    // Vide l'aperçu de la table.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Vide les données.
    this.aLocalSchemaTables.length = 0;
    this.aLocalDatabasesSchemas.length = 0;
    this.aLocalDatabases.length = 0;

    // Lancement de la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries/databases',
      params: {
        order_by: 'database'
      },
      httpMethodOverride: true,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            this.aLocalDatabases = resp.data.data;
          }
        }
      }
    });
  };

  /**
   * Chargement de la liste des schémas d'une base de données.
   * @param  {string} sDatabaseName
   * @returns void
   */
  public loadDatabasesSchemas = function (sDatabaseName: string) {

    // Vide l'aperçu de la table.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Vide les données.
    this.aLocalSchemaTables.length = 0;
    this.aLocalDatabasesSchemas.length = 0;

    // Lancement de la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries/' + sDatabaseName + '/schemas',
      params: {
        order_by: 'schema_name'
      },
      httpMethodOverride: true,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            this.aLocalDatabasesSchemas = resp.data.data;
          }
        }
      }
    });
  };

  /**
   * Chargement de la liste des tables d'un schéma.
   * @param  {string} sDatabaseName
   * @param  {string} sSchemaName
   * @returns void
   */
  public loadSchemaTables = function (sDatabaseName: string, sSchemaName: string) {

    // Vide l'aperçu de la table.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Vide les données.
    this.aLocalSchemaTables.length = 0;

    // Lancement de la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries/tables',
      params: {
        schema: sSchemaName,
        database: sDatabaseName,
        order_by: 'table_name'
      },
      httpMethodOverride: true,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data) &&
            this.vitisCheck.isDefAndNotNull(resp.data.data)) {
            this.aLocalSchemaTables = resp.data.data;
          }
        }
      }
    });
  };

  /**
   * Teste les valeurs d'une table locale.
   * @param  {string} sDatabaseName
   * @param  {string} sSchemaName
   * @param  {string} sTableName
   * @returns void
   */
  public testLocalTableValues = function (sDatabaseName: string, sSchemaName: string, sTableName: string) {

    // Vide les données.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Filtres.
    const oParams: any = {
      schema: sSchemaName,
      database: sDatabaseName,
      table: sTableName,
      limit: 100
    };
    if (this.oSelectedDataSource.parameters.filter.operators.length > 0) {
      oParams.filter = JSON.stringify(this.oSelectedDataSource.parameters.filter);
    }

    // Lancement de la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries',
      params: oParams,
      httpMethodOverride: true,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data.data)) {
            // Nom des colonnes du web service (aperçu).
            this.aTableData = resp.data.data;
            if (this.vitisCheck.isDefAndNotNull(this.aTableData) && this.vitisCheck.isDefAndNotNull(this.aTableData[0])) {
              const aTableAttributes = Object.keys(this.aTableData[0]);
              for (let i = 0; i < aTableAttributes.length; i++) {
                // Pas de données tableau ou objet.
                if (!this.vitisCheck.isArray(this.aTableData[0][aTableAttributes[i]]) &&
                  !this.vitisCheck.isObject(this.aTableData[0][aTableAttributes[i]])) {
                  this.aTableAttributes.push(aTableAttributes[i]);
                }
              }
            }
          }
        }
      }
    });
  };

  /**
   * Actions si la base de données sélectionnée change.
   * @returns void
   */
  public onDatabaseChange() {

    // Vide les données du schéma et de la table sélectionnés.
    this.aLocalDatabasesSchemas.length = 0;
    this.aLocalSchemaTables.length = 0;
    delete this.oSelectedDataSource.parameters.schema;
    delete this.oSelectedDataSource.parameters.table;

    // Charge les chémas
    if (this.vitisCheck.isString(this.oSelectedDataSource.parameters.database) &&
      this.oSelectedDataSource.parameters.database !== '') {
      this.loadDatabasesSchemas(this.oSelectedDataSource.parameters.database);
    }
  }

  /**
   * Actions si le schéma sélectionné change.
   * @returns void
   */
  public onSchemaChange() {

    // Vide les données de la table sélectionnés.
    this.aLocalSchemaTables.length = 0;
    delete this.oSelectedDataSource.parameters.table;

    // Charge les tables
    if (this.vitisCheck.isString(this.oSelectedDataSource.parameters.database) &&
      this.vitisCheck.isString(this.oSelectedDataSource.parameters.schema) &&
      this.oSelectedDataSource.parameters.database !== '' &&
      this.oSelectedDataSource.parameters.schema !== '') {
      this.loadSchemaTables(this.oSelectedDataSource.parameters.database, this.oSelectedDataSource.parameters.schema);
    }
  }

  /**
   * Actions si la table sélectionné change.
   * @returns void
   */
  public onTableChange() { }

  /**
   * Ajout d'un filtre pour la source de données.
   * @returns void
   */
  public addDataSourceFilter() {
    const aDataSourceFilter = this.oSelectedDataSource.parameters.filter.operators;
    if (aDataSourceFilter.length < 5) {

      // Supprime le dernier filtre si non rempli.
      if (aDataSourceFilter.length > 0) {
        const aLastDataSourceFilter = aDataSourceFilter[aDataSourceFilter.length - 1];
        if (aLastDataSourceFilter.column == '' || aLastDataSourceFilter.operator == '') {
          aDataSourceFilter.splice(-1, 1);
        }
      }

      // Ajoute un filtre.
      this.oSelectedDataSource.parameters.filter.operators.push({
        column: '',
        compare_operator: '',
        value: ''
      });
    }
  }

  /**
   * Suppression d'un filtre pour la source de données.
   * @param  {number} iFilterIndex
   * @returns void
   */
  public deleteDataSourceFilter(iFilterIndex: number) {
    this.oSelectedDataSource.parameters.filter.operators.splice(iFilterIndex, 1);
  }
}
