import { Component, OnInit } from '@angular/core';
import { StudioService } from '../../services/studio.service';
import { VitisCheck, VitisClone, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';

declare const $: any;

@Component({
  selector: 'app-data-source-manager',
  templateUrl: './data-source-manager.component.html',
  styleUrls: ['./data-source-manager.component.scss']
})
export class StudioDataSourceManagerComponent implements OnInit {

  public oFormDefinition: any;
  public oDatasources: any;
  public aDatasourceNames: Array<string>; // Noms des datasources
  public sSourceManagerMode: string = 'overview';
  public sSelectedDatasourceName: string;
  public oSelectedDataSource: any;

  constructor(
    private translate: TranslateService,
    private studioService: StudioService,
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
  ) { }

  ngOnInit() {

    // Observe la modification du formulaire
    this.studioService.getStudioForm().subscribe((oFormDefinition: any) => {

      this.oFormDefinition = oFormDefinition;

      if (this.vitisCheck.isDefAndNotNull(oFormDefinition.datasources)) {
        this.oDatasources = oFormDefinition.datasources;
        this.aDatasourceNames = Object.keys(oFormDefinition.datasources);
      } else {
        this.oDatasources = {};
        this.aDatasourceNames = [];
      }
    });

    // Événements à la fermeture de la modale (doit attendre que bootstrap ait fini d'interpréter)
    setTimeout(() => {

      // Observe le fermeture de la modale
      $('#studio_datasource_modal').on('hidden.bs.modal', () => {
        this.sSourceManagerMode = 'overview';
        delete this.sSelectedDatasourceName;
        delete this.oSelectedDataSource;
      });
    }, 5000);
  }


  /**
   * Set this.oSelectedDataSource en fonction de oDatasources et sSelectedDatasourceName
   */
  public setSelectedDatasource() {
    if (this.oDatasources &&
      this.oDatasources[this.sSelectedDatasourceName]) {
      this.oSelectedDataSource = this.vitisClone.clone(this.oDatasources[this.sSelectedDatasourceName]);
    } else {
      delete this.oSelectedDataSource;
    }
  }

  /**
   * Annule l'édition
   */
  public cancelDataSourceEdition() {
    this.oSelectedDataSource = this.vitisClone.clone(this.oDatasources[this.sSelectedDatasourceName]);
    this.sSourceManagerMode = 'overview';
  }

  /**
   * Valide l'édition
   */
  public saveEditedDatasource() {

    // Set le nouvelle definition
    this.oDatasources[this.sSelectedDatasourceName] = this.vitisClone.clone(this.oSelectedDataSource);

    // Vide l'édition
    this.sSourceManagerMode = 'overview';

    this.studioService.setStudioIsBeingModified(true);
    // Sauvegarde dans le formulaire
    this.studioService.setStudioForm(this.oFormDefinition);
  }

  /**
   * Supprime la source de données selectionnée
   */
  public deleteSelectedDatasource() {

    // To do : vérification de l'utilisation de la source de données

    // Modale de confirmation.
    this.translate.get(['STUDIO_DATASOURCE_MODAL_REMOVE_SOURCE_CONFIRM']).subscribe((translations) => {
      vitisAlert.alert({
        type: 'warning',
        html: translations['STUDIO_DATASOURCE_MODAL_REMOVE_SOURCE_CONFIRM'],
        showCancelButton: true
      }).then(result => {
        if (result.value) {
          this.studioService.setStudioIsBeingModified(true);
          // Supprime la source de données.
          delete this.oDatasources[this.sSelectedDatasourceName];
          delete this.oSelectedDataSource;
          this.aDatasourceNames = Object.keys(this.oFormDefinition.datasources);
          this.sSelectedDatasourceName = '';
        }
      });
    });
  }

  /**
   * Ajout d'une source de données
   */
  public addDatasource() {
    this.oSelectedDataSource = {
      dataType: 'text',
      name: '',
      description: '',
      options: ['Value one|1', 'Value two|2', 'Value three|3']
    };

    this.sSourceManagerMode = 'creation';
  }

  /**
   * Anulle la création
   */
  public cancelDataSourceCreation() {
    this.sSourceManagerMode = 'overview';
    delete this.sSelectedDatasourceName;
    delete this.oSelectedDataSource;
  }

  /**
   * Ajoute la source de données
   */
  public saveCreatedDatasource() {

    // Identifiant de la source de données
    const sDatasourceId = this.getNewDatasourceId_();
    if (!sDatasourceId) {
      console.error('could not create datasource id');
      return;
    }

    if (!this.vitisCheck.isString(this.oSelectedDataSource.name) ||
      this.oSelectedDataSource.name === '') {
      console.error('could not create datasource with empty name');
      return;
    }

    // Type utilisé par FormReader
    if (this.oSelectedDataSource.dataType === 'text') {
      this.oSelectedDataSource.type = 'text';
    } else {
      this.oSelectedDataSource.type = 'web_service';
    }

    // Set le nouvelle definition
    this.oDatasources[sDatasourceId] = this.vitisClone.clone(this.oSelectedDataSource);

    // Vide l'édition
    this.sSourceManagerMode = 'overview';

    // Met l'édition sur la source précédement créée
    this.sSelectedDatasourceName = sDatasourceId;

    this.studioService.setStudioIsBeingModified(true);
    // Sauvegarde dans le formulaire
    this.studioService.setStudioForm(this.oFormDefinition);
  }

  /**
   * Retourne un nouveau id de datasource
   */
  private getNewDatasourceId_(): string {

    let sReturn: string;

    for (let i = 0; i < 1000; i++) {
      const sDatasourceId = 'datasource_' + i;
      if (!this.vitisCheck.isDefAndNotNull(this.oDatasources[sDatasourceId])) {
        sReturn = sDatasourceId;
        break;
      }
    }

    return sReturn;
  }
}
