import { Component, OnInit } from '@angular/core';
import { StudioService } from '../../services/studio.service';
import { VitisCheck, VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { Tabs, Tab } from 'src/app/vitis/formreader/formreader.model';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';
import { TranslateService } from '@ngx-translate/core';

declare const $: any;

@Component({
  selector: 'app-tabs-manager',
  templateUrl: './tabs-manager.component.html',
  styleUrls: ['./tabs-manager.component.scss']
})
export class StudioTabsManagerComponent implements OnInit {

  public oTabs: Tabs;
  public aFieldsNames: Array<any>;

  private oFormDefinition_: any;
  private sFormName_: string;

  constructor(
    private translate: TranslateService,
    private studioService: StudioService,
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
    private formReaderService: FormreaderService,
  ) { }

  ngOnInit() {

    // Observe la modification de sFormName
    this.studioService.getStudioFormName().subscribe((sFormName: any) => {
      this.sFormName_ = sFormName;
    });

    // Observe la modification du formulaire
    this.studioService.getStudioForm().subscribe((oFormDefinition: any) => {

      // set oFormDefinition_
      this.oFormDefinition_ = oFormDefinition;

      // Initialise les variables du composant
      this.initTabsManager();
    });

    // Reinitialise les variables du composant
    setTimeout(() => {
      $('#studio_tabs_modal').on('show.bd.modal', () => {
        this.initTabsManager();
      });
    });
  }

  /**
   * Initialise les variables du composant
   */
  public initTabsManager() {
    if (this.vitisCheck.isDefAndNotNull(this.oFormDefinition_[this.sFormName_])) {

      // Onglets
      if (this.formReaderService.isTabsValid(this.oFormDefinition_[this.sFormName_].tabs)) {
        this.oTabs = this.vitisClone.clone(this.oFormDefinition_[this.sFormName_].tabs);
      } else {
        console.error('tabs not valid');
        this.oTabs = {
          'position': 'top',
          'list': [{
            'label': 'Onglet',
            'elements': []
          }]
        };
      }

      // Noms des différents champs
      this.aFieldsNames = this.formReaderService.getFormFieldsNames(this.oFormDefinition_, this.sFormName_);
    }
  }

  /**
   * Ajoute un nouvel onglet
   */
  public addNewTab() {
    if (this.oTabs &&
      this.oTabs.list) {
      this.oTabs.list.push({
        label: this.translate.instant('STUDIO_TABS_MODAL_NEW_TAB'),
        elements: []
      });
    }
  }

  /**
   * Sauvegarde les modifications effectuées
   */
  public saveTabs() {
    this.studioService.setStudioIsBeingModified(true);
    this.oFormDefinition_[this.sFormName_].tabs = this.vitisClone.clone(this.oTabs);
    this.studioService.setStudioForm(this.oFormDefinition_);
  }

  /**
   * Édite le nom d'un onglet
   * @param oTab
   */
  public editTab(oTab: Tab) {
    oTab.studio_edited_ = true;
  }

  /**
   * Valide l'édition du nom d'un onglet
   * @param oTab
   */
  public validEditTab(oTab: Tab) {
    oTab.studio_edited_ = false;
  }

  /**
   * Supprime l'onglet
   * @param oTab
   */
  public removeTab(oTab: Tab) {
    const tabIndex = this.oTabs.list.indexOf(oTab);
    if (tabIndex !== -1) {
      this.oTabs.list.splice(tabIndex, 1);
    }
  }

  /**
   * Sélectionne l'ensemble des éléments sur l'onglet
   * @param oTab
   */
  public selectTabFields(oTab: Tab) {
    oTab.elements = [...this.aFieldsNames];
  }

  /**
   * Désélectionne l'ensemble des éléments sur l'onglet
   * @param oTab
   */
  public unselectTabFields(oTab: Tab) {
    oTab.elements = [];
  }

  /**
   * Déplace l'onglet sur al gauche
   * @param oTab
   */
  public moveTabLeft(oTab: Tab) {
    const tabIndex = this.oTabs.list.indexOf(oTab);
    if (tabIndex !== -1) {
      this.arrayMove_(this.oTabs.list, tabIndex, tabIndex - 1);
    }
  }

  /**
   * Déplace l'onglet sur la droite
   * @param oTab
   */
  public moveTypeRight(oTab: Tab) {
    const tabIndex = this.oTabs.list.indexOf(oTab);
    if (tabIndex !== -1) {
      this.arrayMove_(this.oTabs.list, tabIndex, tabIndex + 1);
    }
  }

  /**
   * Déplace un élément dans un tableau
   * @param arr
   * @param iOldIndex
   * @param iNewIndex
   */
  private arrayMove_(arr: Array<any>, oldIndex: number, newIndex: number) {

    if (newIndex >= 0 && newIndex < arr.length) {

      const elem = arr[oldIndex];

      // Ajoute
      if (oldIndex < newIndex) {
        arr.splice(newIndex + 1, 0, elem);
        arr.splice(oldIndex, 1);
      } else {
        arr.splice(newIndex, 0, elem);
        arr.splice(oldIndex + 1, 1);
      }
    }
    return arr;
  }

  /**
   * Met à jour l'element dans l'onglet
   * @param {any} oTab
   * @param {string} sFieldName
   */
  public updateElementInTab(oTab: any, sFieldName: string) {
    // Si l'element existe on l'enlève
    if (oTab.elements.indexOf(sFieldName) !== -1) {
      oTab.elements.splice(oTab.elements.indexOf(sFieldName), 1);
    } else { // Si l'element n'existe pas on l'ajoute
      oTab.elements.push(sFieldName);
    }
  }
}
