import { Component, OnInit, ViewChild, OnDestroy, Input, OnChanges, SimpleChanges, Output, EventEmitter } from '@angular/core';
import { DataGridComponent } from '../data-grid/data-grid.component';
import { environment } from 'src/environments/environment';
import { WorkspaceListOptions, WorkspaceListAutorefreshOptions, WorkspaceListEmbeddedOptions } from './workspacelist.model';
import { ActivatedRoute, Router } from '@angular/router';
import { VitisCheck, LocalStorage, SessionStorage, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';
import ResizeObserver from 'resize-observer-polyfill';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { WklistDateComponent } from 'src/app/vitis/workspacelist/template/date/date.component';
import { WklistParamsComponent } from 'src/app/vitis/workspacelist/template/params/params.component';
import { WklistTooltipComponent } from 'src/app/vitis/workspacelist/template/tooltip/tooltip.component';
import { WklistBooleanIconComponent } from 'src/app/vitis/workspacelist/template/boolean-icon/boolean-icon.component';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { oWkListComponents } from 'src/app/modules/modules-components';
import moment from 'moment';
import { FormreaderComponent } from '../formreader/formreader.component';
import { WklistUrlComponent } from 'src/app/vitis/workspacelist/template/url/url.component';

declare var $: any;

@Component({
  selector: 'app-workspacelist',
  templateUrl: './workspacelist.component.html',
  styleUrls: ['./workspacelist.component.scss']
})
export class WorkspacelistComponent implements OnInit, OnDestroy, OnChanges {

  // Utilisé pour piloter workspacelist depuis un composant parent et sans passer par l'URL
  @Input() public embeddedOptions: WorkspaceListEmbeddedOptions;

  @ViewChild(DataGridComponent, { static: true }) oDataGrid: DataGridComponent;

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {},
    datagridOptions: WorkspaceListOptions
  };

  // Largeur de page
  public pageLength: Number = environment.properties.rows_per_page;

  // Instance de la grille de données.
  public dataTableInstance: any;

  // Mode en cours
  public mode: any;
  public oParams: any;

  // Filtre en cours
  public oFilter: any = {
    type: 'custom',
    nodes: []
  };

  // Types de filtre disponible
  public oFilterTypes: any = {
    custom: 'VITIS_WORKSPACELIST_FILTER_CUSTOM',
    form: 'VITIS_WORKSPACELIST_FILTER_FORM'
  };

  public sDefaultFilterColumn: string;

  // Résultat
  public oResults: any = {
    recordsTotal: 0,
    recordsFiltered: 0,
    data: []
  };

  // Paramètres passés dans l'URL
  public urlParams_: any;

  // Paramètres à passer à la requête
  public datatableRessourceParams_: any;

  public onRefresh = new EventEmitter<any>();

  // Modal avec un formreader
  public modalTitle: string = '';
  @ViewChild(FormreaderComponent, { static: false }) oFormReader: FormreaderComponent;

  // Modale affichée si aucun enregistrement dans la liste.
  public bDisplayNoRecordModal: boolean = false;
  public sNoRecordLabel: string = 'VITIS_WORKSPACELIST_NO_RECORD';
  public sInsertButtonLabel: string = 'VITIS_WORKSPACELIST_INSERT_BTN';

  // Variables privées
  private intervalId_: any;
  private localStorageValues_: any;
  private sessionStorageValues_: any;
  private clickedCell_: any;
  private queryCache_: any;

  constructor(
    private route: ActivatedRoute,
    private vitisCheck: VitisCheck,
    private router: Router,
    private localStorage: LocalStorage,
    private sessionStorage: SessionStorage,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private auth: AuthService
  ) { }

  ngOnInit() {

    // Nombre d'éléments à afficher sur la grille
    this.pageLength = Number(environment.properties.rows_per_page);

    // Permet à oDataGrid de connaitre les colonnes etc..
    this.oDataGrid.oDataGridOptions = this.oDataGridOptions;

    // Mode normal (non embarqué)
    if (!this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {

      // Paramètres de la route en cours d'utilisation
      this.route.params.subscribe(oParams => {
        this.oParams = oParams;
      });

      // Mode en cours d'utilisation
      this.route.data.subscribe(mode => {
        this.mode = mode;

        // options app.json
        if (mode.options) {
          this.oDataGridOptions.datagridOptions = mode.options;

          // Texte optionnel pour la modale de création (si aucun enregistrement dans la liste).
          if (this.vitisCheck.isDefAndNotNull(mode.options.no_record_modal)) {
            if (this.vitisCheck.isDefAndNotNull(mode.options.no_record_modal.label)) {
              this.sNoRecordLabel = mode.options.no_record_modal.label;
            }
            if (this.vitisCheck.isDefAndNotNull(mode.options.no_record_modal.button_label)) {
              this.sInsertButtonLabel = mode.options.no_record_modal.button_label;
            }
          }
        }

        // Récupère les paramètres en cookies --> dans data-grid ?
        this.initCookies();

        // Initialise les filtres
        this.initFilters();

        // Initialise la grille de données
        this.initDataGrid();
      });

      // Survéillance du filtre dans les paramètres
      this.route.queryParams.subscribe(oParams => {

        // mémorise les paramètres URL
        this.urlParams_ = oParams;

        // Applique le filtre
        if (this.oFilter &&
          this.oFilter.nodes &&
          this.oFilter.nodes.length === 0) {
          if (this.vitisCheck.isDefAndNotNull(oParams && oParams.q)) {
            const oFilterNodes = this.getFilterByQuery_(oParams.q);
            if (oFilterNodes.length > 0 &&
              this.oFilter.nodes.length === 0) {
              this.oFilter.nodes = oFilterNodes;
              this.applyFilter();
            }
          }
        }
      });
    } else {
      if (this.embeddedOptions &&
        this.embeddedOptions.events &&
        this.embeddedOptions.events.onRefresh &&
        this.embeddedOptions.events.onRefresh.emit) {
        if (this.onRefresh &&
          this.onRefresh.subscribe) {
          this.onRefresh.subscribe(() => {
            this.embeddedOptions.events.onRefresh.emit();
          });
        }
      }
    }
  }

  ngOnChanges(changes: SimpleChanges) {

    if (this.vitisCheck.isDefAndNotNull(changes.embeddedOptions)) {
      if (this.embeddedOptions.mode) {
        this.mode = this.embeddedOptions.mode;

        // options app.json
        if (this.embeddedOptions.mode.options) {
          this.oDataGridOptions.datagridOptions = this.embeddedOptions.mode.options;
        }

        // Récupère les paramètres en cookies --> dans data-grid ?
        this.initCookies();

        // Initialise les filtres
        this.initFilters();

        // Initialise la grille de données
        this.initDataGrid();
      }
    }
  }

  ngOnDestroy() {
    // Termine l'intervale de rechargement
    if (this.intervalId_) {
      clearInterval(this.intervalId_);
    }

    // Vide le filtre dans l'URL
    this.clearUrlParams_();
  }

  /**
   * Applique le filtre à la grille
   */
  public applyFilter(): void {

    // Filtre formaté pour la requête API
    this.oFilter.value = this.getAPIFilterFromNodes_(this.oFilter.nodes);

    // Rafraichit la vue
    if (this.dataTableInstance &&
      this.dataTableInstance.ajax) {
      this.dataTableInstance.ajax.reload();
    }

    // Mémorisation du filtre dans la session
    this.sessionStorage.add(this.mode.name, {
      filter: this.oFilter
    });

    // Mémorisation du filtre dans l'url
    this.setUrlParams_();
  }

  /**
   * Initialise les filtres
   */
  private initFilters(): void {

    if (this.vitisCheck.isDefAndNotNull(this.mode)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {

        this.oFilter.type = 'custom';

        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.search)) {

          // Recherche custom
          if (this.oDataGridOptions.datagridOptions.search.custom !== true) {
            delete this.oFilterTypes.custom;
          }

          // Recherche par formulaire
          if (this.oDataGridOptions.datagridOptions.search.form !== true) {
            delete this.oFilterTypes.form;
          }

          // Recherche par formulaire
          if (this.oDataGridOptions.datagridOptions.search.default === 'custom') {
            this.oFilter.type = 'custom';
          }
          if (this.oDataGridOptions.datagridOptions.search.default === 'form') {
            this.oFilter.type = 'form';
          }
        } else {
          delete this.oFilterTypes.form;
        }
      }
    }

    // Nombre de possibilités
    this.oFilter.nbFilterTypes = Object.keys(this.oFilterTypes).length;

    const oQueryParams: any = this.route.queryParams;
    if (!(oQueryParams && oQueryParams.value && oQueryParams.value.q)) {
      if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.filter)) {
        this.oFilter = this.sessionStorageValues_.filter;
      }
    }

    // Gestion du parent en cas de workspacelist dans une section
    if (this.mode.options &&
      this.mode.options.parent_id_field &&
      this.oParams &&
      this.oParams.id) {

      if (this.mode.options.disable_filter_parent !== true) {
        // Ajout d'un filtre
        if (this.vitisCheck.isArray(this.oFilter.nodes)) {
          this.oFilter.nodes.push({
            column: {
              field: this.mode.options.parent_id_field
            },
            compareOperator: '=',
            value: String(this.oParams.id)
          });
        }
      }
      this.oFilter.value = this.getAPIFilterFromNodes_(this.oFilter.nodes);
    }

    // Gestion sDefaultFilterColumn
    this.sDefaultFilterColumn = null;
    if (this.mode.options &&
      this.mode.options.search &&
      this.mode.options.search.defaultColumn) {
      this.sDefaultFilterColumn = this.mode.options.search.defaultColumn;
    }

    // Mémorisation du filtre dans l'url
    this.setUrlParams_();
  }

  /**
   * Parse les params envoyés par DataTables pour créer des paramètres interprétables par l'API Vitis
   * @param  {any} params
   * @returns any
   */
  private getRequestParams_(params: any): any {

    const requestParams: any = {};

    // Order
    if (this.vitisCheck.isDefAndNotNull(params.order)) {
      if (this.vitisCheck.isDefAndNotNull(params.order[0])) {
        if (this.vitisCheck.isString(params.order[0].dir)) {
          requestParams.sort_order = params.order[0].dir.toUpperCase();
        }
        if (this.vitisCheck.isNumber(params.order[0].column)) {
          const iColumnId = params.order[0].column;
          if (this.vitisCheck.isDefAndNotNull(params.columns[iColumnId])) {
            if (this.vitisCheck.isString(params.columns[iColumnId].data)) {
              requestParams.order_by = params.columns[iColumnId].data;
            }
          }
        }
      }
    }

    // Limit
    if (this.vitisCheck.isNumber(params.length)) {
      requestParams.limit = params.length;
    }

    // Offset
    if (this.vitisCheck.isNumber(params.start)) {
      requestParams.offset = params.start;
    }

    // Filter
    const filters = this.getFilters_();
    if (this.vitisCheck.isDefAndNotNull(filters)) {
      requestParams.filter = filters;
    }

    // Attributs
    if (this.vitisCheck.isDefAndNotNull(params.columns)) {
      // détecte la présence d'attribut supplementaire requis par la liste
      const extra_attributs = (this.oDataGridOptions.datagridOptions.extra_attributs) ? this.oDataGridOptions.datagridOptions.extra_attributs : [];
      requestParams['attributs'] = params.columns.filter(
        (aColumn: any) => this.vitisCheck.isString(aColumn.data)
      ).map(
        (aColumn: any) => aColumn.data
      ).concat(extra_attributs)
        .join('|');

    }

    return requestParams;
  }

  /**
   * Écrit le filtre dans l'URL
   */
  public setUrlParams_() {

    /**
     * Paramètres de l'url
     */
    const queryParams: any = {};

    // Filtre
    queryParams.q = null;
    if (this.oFilter) {
      const sQuery = this.getQueryByFilter_(this.oFilter);
      if (sQuery) {
        queryParams.q = sQuery;
      }
    }

    // Page
    if (this.datatableRessourceParams_) {

      // Valeurs vides
      queryParams.wkgstart = null;
      queryParams.wkglength = null;

      // Valeurs réelles
      if (this.datatableRessourceParams_.start > 0 &&
        this.datatableRessourceParams_.length > 0) {
        queryParams.wkgstart = this.datatableRessourceParams_.start;
        queryParams.wkglength = this.datatableRessourceParams_.length;
      }
    }

    // Mode normal (non embarqué)
    if (!this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {
      this.router.navigate(
        [],
        {
          relativeTo: this.route.parent,
          queryParams,
          queryParamsHandling: 'merge',
        });
    }
  }

  /**
   * Récupère les donéées au format DataTables
   * @param  {any} params
   * @returns Promise
   */
  public getResourceData_(params: any): Promise<any> {
    return new Promise((resolve, reject) => {

      // Mémorise les paramètres
      this.datatableRessourceParams_ = params;

      // Mémorisation des paramètres de la table
      this.localStorage.add(this.mode.name, {
        datatableParams: params,
      });
      this.sessionStorage.add(this.mode.name, {
        datatableParams: params,
      });

      // Set les paramètres dans l'url
      this.setUrlParams_();

      this.oDataGridOptions.loadingData_ = true;

      // Lance la requête
      this.vitisRequest.ajaxRequest({
        method: 'GET',
        url: this.getRessourceUrl(),
        ajaxLoader: false,
        params: this.getRequestParams_(params),
        success: (resp: any) => {

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };

          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            if (this.vitisCheck.isDefAndNotNull(resp.data.status === 1)) {
              if (this.vitisCheck.isArray(resp.data.data)) {
                oReturn.data = resp.data.data;
              }
              if (this.vitisCheck.isDefAndNotNull(resp.data.total_row_number)) {
                oReturn.recordsTotal = resp.data.total_row_number;
                oReturn.recordsFiltered = resp.data.total_row_number;
              }
            }
          }

          resolve(oReturn);
        },
        error: (err: any) => {
          console.error('error while loading values :', err);

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };

          resolve(oReturn);
        }
      });
    });
  }

  /**
   * Génére l'url pour récupérer les données
   * @returns {string}
   */
  protected getRessourceUrl(): string {
    let sRessourceUrl = this.oDataGridOptions.datagridOptions.ressource;

    // Gestion du parent en cas de workspacelist dans une section
    if (this.mode.options &&
      this.mode.options.parent_id_field &&
      this.oParams &&
      this.oParams.id) {

      if (sRessourceUrl.indexOf(`:${this.mode.options.parent_id_field}`) !== -1) {
        // Remplacement dans l'URL
        sRessourceUrl = sRessourceUrl.replace(`:${this.mode.options.parent_id_field}`, String(this.oParams.id));
      }
    }

    return environment.properties.app_api_url + '/' + sRessourceUrl;
  }

  /**
   * Initialise localStorageValues_ et sessionStorageValues_
   */
  public initCookies() {

    this.localStorageValues_ = this.localStorage.get(this.mode.name);
    this.sessionStorageValues_ = this.sessionStorage.get(this.mode.name);
    if (!this.vitisCheck.isDefAndNotNull(this.localStorageValues_)) {
      this.localStorageValues_ = {};
    }
    if (!this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_)) {
      this.sessionStorageValues_ = {};
    }
  }

  /**
   * Initialise la grille de données
   */
  public initDataGrid() {
    /**
     * Calcule la hauteur de la table
     */
    const calcDataTableHeight = () => {
      if (this.vitisCheck.isNumber(this.mode.options.height)) {
        return this.mode.options.height;
      } else {
        const dashboardHeight = $('#dashboard-content').height();
        const titleHeight = $('.mode-title').height();
        const filterHeight = $('.grid-filter').height();
        const menusHeight = 150;
        const tableHeight = dashboardHeight - (menusHeight + filterHeight + titleHeight);
        return tableHeight;
      }
    };

    /**
     * Réajuste les dimensions de la table
     */
    const autoAjustTables = () => {
      if ($('#' + this.oDataGrid.tableId).length) {
        this.dataTableInstance.columns.adjust();
        $('#' + this.oDataGrid.tableId + '_wrapper').find('.dataTables_scrollBody').css('height', calcDataTableHeight());
      }
    };

    const self = this;
    // Ordre
    let order = [2, 'desc'];

    // Offset
    let displayStart = 0;

    // Options de rafraichissement auto
    let autorefresh = null;

    let rowId = '';

    // Options app.json
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.order_by)) {
        order = this.oDataGridOptions.datagridOptions.order_by;
      }
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.autorefresh)) {
        autorefresh = this.oDataGridOptions.datagridOptions.autorefresh;
      }
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.id_field)) {
        rowId = this.oDataGridOptions.datagridOptions.id_field;
      }
    }

    // LocalStorage
    if (this.localStorageValues_ &&
      this.localStorageValues_.datatableParams) {

      // Order
      if (this.localStorageValues_.datatableParams.order &&
        this.localStorageValues_.datatableParams.order[0] &&
        this.localStorageValues_.datatableParams.order[0].column &&
        this.localStorageValues_.datatableParams.order[0].dir) {
        order = [this.localStorageValues_.datatableParams.order[0].column,
        this.localStorageValues_.datatableParams.order[0].dir];
      }
    }

    // SessionStorage
    if (this.sessionStorageValues_ &&
      this.sessionStorageValues_.datatableParams) {

      // start
      if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.datatableParams.start)) {
        displayStart = this.sessionStorageValues_.datatableParams.start;
      }
    }

    // Paramètres URL
    if (this.urlParams_ &&
      this.urlParams_.wkgstart &&
      this.urlParams_.wkglength) {
      displayStart = Number(this.urlParams_.wkgstart);
    }

    // Paramètres de la grille de données.
    this.oDataGridOptions.datatableOptions = {
      select: {
        style: this.mode.options.singleSelect ? 'single' : 'multi',
        selector: 'tr'
      },
      processing: true,
      serverSide: true,
      searching: false,
      lengthMenu: this.getLengthMenu_(),  // Tableau choix du nombre d'éléments par page
      pageLength: this.getPageLength_(),  // Nombre d'éléments par page
      order: order, // Ordre
      displayStart: displayStart,  // Offset
      scrollY: calcDataTableHeight(),
      scrollX: true,
      rowId: rowId,
      selected: 'selected',
      ajax: (data: any, callback: any, settings: any) => {

        // Cache utilisé pour le rafraichissement auto
        this.queryCache_ = {
          options: data,
          date: Date.now(),
        };

        this.getResourceData_(data).then((oResponse: any) => {

          // Set oResults
          this.oResults = oResponse;

          // Vide viewContainer
          this.oDataGrid.viewContainer.clear();

          // Remplace les données
          callback(this.oResults);

          // Remplit le cache
          this.queryCache_.result = oResponse;

          // event onRefresh
          if (this.onRefresh) {
            this.onRefresh.emit();
          }

          // Si aucun enregistrement : affichage de la modale de création si l'onglet à un mode "insert".
          if (this.mode.options.insert) {
            if (this.oResults.recordsTotal == 0) {
              this.bDisplayNoRecordModal = true;
            } else {
              this.bDisplayNoRecordModal = false;
            }
          }
        }).catch(() => {
          this.oResults = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };
          callback(this.oResults);
        });
      }
    };

    // Composants pour les templates des colonnes.
    this.oDataGridOptions.customTemplateComponents = [
      WklistDateComponent,
      WklistParamsComponent,
      WklistTooltipComponent,
      WklistBooleanIconComponent,
      WklistUrlComponent
    ];

    // Composants template contenus dans le modules externes
    if (this.vitisCheck.isObject(oWkListComponents)) {
      for (const key in oWkListComponents) {
        if (oWkListComponents.hasOwnProperty(key)) {
          this.oDataGridOptions.customTemplateComponents.push(oWkListComponents[key]);
        }
      }
    }

    // Initialise les boutons
    this.initButtons();

    // Charge le JS
    this.loadModeJavaScript_(this.mode);

    // Rafraichissement auto
    if (autorefresh &&
      autorefresh.interval) {
      this.intervalId_ = setInterval(() => {
        this.autoRefresh_(autorefresh);
      }, autorefresh.interval);
    }

    //
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.columns)) {

        // Traductions
        const aTranslations = [];
        for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
          const column = this.oDataGridOptions.datagridOptions.columns[i];
          aTranslations.push(column.title);
        }
        this.translate.get(aTranslations).subscribe((res: any) => {

          // Enregistre le titre traduit
          for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
            const column = this.oDataGridOptions.datagridOptions.columns[i];
            column.translateTitle = res[column.title];
          }

          // Initialise la table
          setTimeout(() => {
            // Boutons display (Evènement).
            this.oDataGridOptions.displayEvent = function (data: any) {
              self.displayRowForm_(data, 'display');
            };
            // Boutons update (Evènement).
            this.oDataGridOptions.updateEvent = function (data: any) {
              self.displayRowForm_(data, 'update');
            };
            // Colonnes de la liste (définies dans le mode).
            this.oDataGridOptions.datatableOptions.columns = this.getParsedColumns_();

            // Création de la liste.
            this.dataTableInstance = this.oDataGrid.initTable(this.oDataGridOptions);

            // Draw fields event
            this.dataTableInstance.on('draw.dt', () => {
              // Format les champs avec un template
              this.oDataGrid.formatFields(this.oDataGridOptions);
              // Réajuste les dimensions de la table
              autoAjustTables();
            });

            // Select event
            this.dataTableInstance.on('select', () => {
              if (this.embeddedOptions &&
                this.embeddedOptions.events &&
                this.embeddedOptions.events.onSelect &&
                this.embeddedOptions.events.onSelect.emit) {
                setTimeout(() => {
                  const rows: any = this.oDataGrid.getSelectedRows_();
                  rows.clickedCell_ = this.clickedCell_;
                  this.embeddedOptions.events.onSelect.emit(rows);
                });
              }
            });

            // Deselect event
            this.dataTableInstance.on('deselect', () => {
              if (this.embeddedOptions &&
                this.embeddedOptions.events &&
                this.embeddedOptions.events.onDeselect &&
                this.embeddedOptions.events.onDeselect.emit) {
                setTimeout(() => {
                  const rows: any = this.oDataGrid.getSelectedRows_();
                  rows.clickedCell_ = this.clickedCell_;
                  this.embeddedOptions.events.onDeselect.emit(rows);
                });
              }
            });

            // Click event (permet de récupérer la cellule cliquée)
            this.dataTableInstance.on('click', 'td', function () {
              self.clickedCell_ = self.dataTableInstance.cell(this);
            });

            // Réajuste les dimensions de la table lors fermeture/ouverture du dashboard
            const dDashboardContent = document.querySelector('#dashboard-content');
            const resizeObserver = new ResizeObserver(entries => {
              autoAjustTables();
            });
            resizeObserver.observe(dDashboardContent);
          });
        });
      }
    }
  }


  /**
   * Lit le tableau de noeuds et retourne un filtre compatible avec l'API Vitis
   * @param  {Array<any>} aFilterNodes
   * @returns any
   */
  public getAPIFilterFromNodes_(aFilterNodes: Array<any>): any {

    let oFilter: any = {};

    if (this.vitisCheck.isArray(aFilterNodes)) {
      if (aFilterNodes.length > 0) {

        oFilter = {
          relation: 'AND',
          operators: []
        };

        for (let i = 0; i < aFilterNodes.length; i++) {
          const oNode = aFilterNodes[i];

          if (this.vitisCheck.isDefAndNotNull(oNode.column)) {
            const oNodeFilter: any = {};

            if (this.vitisCheck.isDefAndNotNull(oNode.column.field)
              && this.vitisCheck.isDefAndNotNull(oNode.compareOperator)
              && this.vitisCheck.isDefAndNotNull(oNode.value)) {
              oNodeFilter.column = oNode.column.field;
              oNodeFilter.compare_operator = oNode.compareOperator;
              oNodeFilter.value = oNode.value;

              if (oNode.compareOperator === '~') {
                oNodeFilter.compare_operator = 'LIKE';
                oNodeFilter.compare_operator_options = {
                  'case_insensitive': true
                };
                oNodeFilter.value = '%' + oNodeFilter.value + '%';

              }
              // modification de la colonne à utiliser pour filtrer
              if (this.vitisCheck.isDefAndNotNull(oNode.column.searchColumn)) {
                oNodeFilter.column = oNode.column.searchColumn;
              }

              oFilter.operators.push(oNodeFilter);
            }
          }
        }
      }
    }

    return oFilter;
  }

  /**
   * Retourne le filtre au format URL
   * @param oFilter
   */
  public getQueryByFilter_(oFilter: any): string {

    let sQuery: any = null;
    if (oFilter.nodes) {
      for (let i = 0; i < oFilter.nodes.length; i++) {
        const node = oFilter.nodes[i];
        if (i === 0) {
          sQuery = '';
        } else {
          sQuery += '||';
        }

        let sNodeFilter = `${node.column.field}|${node.compareOperator}|${node.value}`;
        if (node.column.searchWebService && node.value != node.displayValue) {
          sNodeFilter += `|${node.displayValue}`;
        }
        sQuery += encodeURIComponent(sNodeFilter);
      }
    }

    return sQuery;
  }

  /**
   * Retourne les nodes du filtre en fonction d'un filtre URL
   * @param sQuery
   */
  public getFilterByQuery_(sQuery: string): any {

    const aNodes = [];

    if (!this.vitisCheck.isDefAndNotNull(sQuery)) {
      return;
    }

    sQuery = decodeURIComponent(sQuery);

    const aQuery = sQuery.split('||');
    for (let i = 0; i < aQuery.length; i++) {
      const aNode = aQuery[i].split('|');

      if (aNode) {
        const field = aNode[0];
        const compareOperator = aNode[1];
        const value = aNode[2];

        if (field &&
          compareOperator &&
          value) {

          const column = this.oDataGrid.getColumnByName_(field);
          let displayValue = (aNode.length === 4) ? aNode[3] : value;

          // Gestion valeur à afficher
          if (this.vitisCheck.isArray(column.searchSuggestions)) {
            for (let ii = 0; ii < column.searchSuggestions.length; ii++) {
              const suggestion = column.searchSuggestions[ii];
              if (this.vitisCheck.isDefAndNotNull(suggestion.label) &&
                this.vitisCheck.isDefAndNotNull(suggestion.value)) {
                if (suggestion.value === value) {
                  displayValue = suggestion.label;
                }
              }
            }
          }

          // Gestion dates
          if (column.searchType === 'date') {
            let sDateFormat = 'DD/MM/YYYY';
            const oUser = this.auth.getUserInfos();
            if (this.vitisCheck.isDefAndNotNull(oUser)) {
              if (this.vitisCheck.isDefAndNotNull(oUser.formatdate_id)) {
                sDateFormat = oUser.formatdate_id;
              }
            }
            displayValue = moment(Number(value)).format(sDateFormat + ' HH:mm');
          }

          if (column) {
            aNodes.push({
              column,
              compareOperator,
              value,
              displayValue
            });
          }
        }
      }
    }
    return aNodes;
  }

  /**
   * Récupère les filtres du mode
   */
  public getParsedFilters_() {
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.filter)) {
      const oFilter = this.oDataGridOptions.datagridOptions.filter;

      if (this.vitisCheck.isDefAndNotNull(oFilter.lang) && oFilter.lang === '') {
        oFilter.lang = environment.properties.language;
      }

      const oGivenFilter = {
        relation: 'AND',
        operators: []
      };
      for (const key in oFilter) {
        if (oFilter.hasOwnProperty(key)) {
          oGivenFilter.operators.push({
            column: key,
            compare_operator: '=',
            value: oFilter[key]
          });
        }
      }

      return oGivenFilter;
    }
  }

  /**
   * Lance le rafraichissement automatique
   * @param options
   */
  private autoRefresh_(options: WorkspaceListAutorefreshOptions) {

    if (options &&
      options.interval) {

      // Ne recharge pas automatiquement si la page n'est pas visible
      if (document.hidden) {
        return;
      }

      // Ne lance pas le rechargement automatique si un chargement a déjà été lancé récemment
      if (this.queryCache_ &&
        this.queryCache_.date &&
        Date.now() - this.queryCache_.date < options.interval - 1000) {
        return;
      }

      // Lance une requête et compare les résultats avant de lancer la compilation du template
      if (options.compare_columns) {
        if (this.queryCache_ &&
          this.queryCache_.options &&
          this.queryCache_.result &&
          this.queryCache_.result.data) {
          this.getResourceData_(this.queryCache_.options).then((oResponse: any) => {

            let bSameObjects = true;
            if (oResponse &&
              oResponse.data) {

              for (let i = 0; i < oResponse.data.length; i++) {
                const cacheRow = this.queryCache_.result.data[i];
                const loadedRow = oResponse.data[i];

                for (let ii = 0; ii < options.compare_columns.length; ii++) {
                  const column = options.compare_columns[ii];

                  if (cacheRow[column] !== loadedRow[column]) {
                    bSameObjects = false;
                  }
                }
              }
            }

            if (bSameObjects === false) {
              this.dataTableInstance.ajax.reload(null, false);
            }
          });
        }
      } else {
        this.dataTableInstance.ajax.reload(null, false);
      }
    }
  }

  /**
   * Affiche le formulaire correspondant
   * @param  {any} oRow
   * @param  {String} sFormName
   */
  private displayRowForm_(oRow: any, sFormName: String) {

    // Mode embedded
    if (this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {
      if (sFormName === 'display') {
        if (this.embeddedOptions.events &&
          this.embeddedOptions.events.onDisplay &&
          this.embeddedOptions.events.onDisplay.emit) {
          this.embeddedOptions.events.onDisplay.emit(oRow);
        }
      }
      if (sFormName === 'update') {
        if (this.embeddedOptions.events &&
          this.embeddedOptions.events.onEdit &&
          this.embeddedOptions.events.onEdit.emit) {
          this.embeddedOptions.events.onEdit.emit(oRow);
        }
      }
    } else {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.id_field)) {
          if (this.vitisCheck.isDefAndNotNull(oRow)) {
            if (this.vitisCheck.isDefAndNotNull(oRow[this.oDataGridOptions.datagridOptions.id_field])) {
              if (this.vitisCheck.isDefAndNotNull(sFormName)) {

                const sId = oRow[this.oDataGridOptions.datagridOptions.id_field];
                const aModeRoute = this.mode.route.split('/');

                if (this.vitisCheck.isDefAndNotNull(this.mode.sections) &&
                  this.vitisCheck.isDefAndNotNull(this.mode.options.map_field)) { // Avec sections et map_field
                  // Recupère les données afin d'avoir' la valeur du map_field
                  this.vitisRequest.ajaxRequestPromise({
                    method: 'GET',
                    url: this.vitisRequest.getApiRessourceUrl(this.mode.options.ressource, sId)
                  }).then((resp) => {
                    if (resp.data) {
                      if (resp.data.status === 1) {
                        if (resp.data.data) {
                          this.mode.options.values = resp.data.data;
                          const sections = this.mode.sections[this.mode.options.values[this.mode.options.map_field]];
                          this.navigateFirstSection(sections, sId, sFormName, aModeRoute);
                        }
                      }
                    }
                  });
                } else {
                  this.navigateFirstSection(this.mode.sections, sId, sFormName, aModeRoute);
                }
              }
            }
          }
        }
      }
    }
  }

  /**
   * // Navigue vers le bon url lors de l'event sur la grid
   * @param sections // mode sections
   * @param sId // id de l'élément
   * @param sFormName // formulaire choisi 'update|insert...'
   * @param aModeRoute
   */
  private navigateFirstSection(sections, sId: number, sFormName: any, aModeRoute: any) {
    if (sections) {
      if (sections[0] && sections[0].route) {
        const aSectionRoute = sections[0].route.split('/');
        for (let i = aSectionRoute.length - 1; i >= 0; i--) {
          if (aSectionRoute[i] === ':id') {
            aSectionRoute[i] = sId;
          }
          if (aSectionRoute[i] === ':form') {
            aSectionRoute[i] = sFormName;
          }
          if (aModeRoute.indexOf(aSectionRoute[i]) !== -1) {
            aSectionRoute.splice(i, 1);
          }
        }
        this.router.navigate(aSectionRoute, { relativeTo: this.route.parent });
      }
    } else {
      this.router.navigate([sId, sFormName], { relativeTo: this.route.parent });
    }
  }

  /**
   * Récupère les colonnes définies dans le mode au format BootstrapTable
   * @returns Array
   */
  private getParsedColumns_(): Array<any> {

    const aColumns: any = [{
      targets: 0,
      data: null,
      defaultContent: '',
      orderable: false,
      className: 'select-checkbox',
      width: '40px'
    }];

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {

      // Colonne boutons
      if (this.oDataGridOptions.datagridOptions.display === true
        || this.oDataGridOptions.datagridOptions.update === true) {

        let defaultContent = '<div style="text-align: center;">';
        if (this.oDataGridOptions.datagridOptions.display === true &&
          (environment.properties.data_grid_update_btn_only === false || this.oDataGridOptions.datagridOptions.update !== true)) {
          defaultContent += '<span class="datatable-btn datatable-btn-display icon-vitis-search"></span>';
        }
        if (this.oDataGridOptions.datagridOptions.update === true) {
          defaultContent += '<span class="datatable-btn datatable-btn-edit icon-vitis-draw"></span>';
        }
        defaultContent += '</div>';

        aColumns.push({
          targets: -1,
          data: null,
          orderable: false,
          defaultContent,
          width: '90px'
        });
      }

      // Colonnes du mode
      if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.columns)) {
        const aUserPrivileges = this.auth.getPrivileges();
        for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
          const column = this.oDataGridOptions.datagridOptions.columns[i];
          // Privilège obligatoire pour afficher la colonne.
          if (this.vitisCheck.isDefAndNotNull(column.privilege)) {
            if (aUserPrivileges.indexOf(column.privilege) == -1) {
              continue;
            }
          }
          aColumns.push({
            title: column.translateTitle,
            data: column.field,
            name: column.field,
            visible: column.visible === false ? false : true,
            type: column.template ? 'html' : 'string',
            render: (data: any, type: any, row: any, meta: any) => {
              if (column.template) {
                data = '<field-render column_number="' + i + '">' + encodeURIComponent(JSON.stringify(row)) + '</field-render>';
              }
              return data;
            },
            orderable: column.sortable,
            template: column.template,
            // Alignement du contenu de la colonne.
            className: 'table-align-' + (column.align ? column.align : 'left')
          });
        }
      }
    }

    return aColumns;
  }

  /**
   * Nettoie le filtre dans l'URL
   */
  private clearUrlParams_() {

    this.router.navigate(
      [],
      {
        relativeTo: this.route.parent,
        queryParams: {
          wkglength: null,
          wkgstart: null,
          q: null
        },
        queryParamsHandling: 'merge',
      }
    );
  }

  /**
   * Affiche le formulaire d'insertion
   * @returns void
   */
  private displayInsertForm_(): void {

    // Mode embedded
    if (this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {
      if (this.embeddedOptions.events &&
        this.embeddedOptions.events.onCreate &&
        this.embeddedOptions.events.onCreate.emit) {
        this.embeddedOptions.events.onCreate.emit();
      }
    } else {
      if (this.vitisCheck.isDefAndNotNull(this.mode.sections)) {
        const aModeRoute = this.mode.route.split('/');
        let aSectionRoute;
        if (this.vitisCheck.isDefAndNotNull(this.mode.options.map_field)) {
          if (!this.vitisCheck.isArray(this.mode.sections)) {
            aSectionRoute = this.mode.sections[Object.keys(this.mode.sections)[0]][0].route.split('/');
          }
        } else {
          aSectionRoute = this.mode.sections[0] && this.mode.sections[0].route.split('/');
        }
        if (aSectionRoute) {
          for (let i = aSectionRoute.length - 1; i >= 0; i--) {
            if (aSectionRoute[i] === ':form') {
              aSectionRoute[i] = 'insert';
            }
            if (aSectionRoute[i] === ':id') {
              aSectionRoute.splice(i, 1);
            }
            if (aModeRoute.indexOf(aSectionRoute[i]) !== -1) {
              aSectionRoute.splice(i, 1);
            }
          }
          this.router.navigate(aSectionRoute, { relativeTo: this.route.parent });
        }
      } else {
        this.router.navigate(['insert'], { relativeTo: this.route.parent });
      }
    }
  }

  /**
   * Déclenche le processus de dupplication de la ligne sélectionnée
   */
  public cloneSelectedRow(): void {

    const aIds = this.oDataGrid.getSelectedIds_();

    if (!(aIds.length > 0)) {
      vitisAlert.alert({
        type: 'error',
        title: this.translate.instant('LIST_DELETE_NO_SELECTION'),
      });
    } else if (aIds.length > 1) {
      vitisAlert.alert({
        type: 'error',
        title: this.translate.instant('MORE_THAN_ONE_SELECTION'),
      });
    } else {
      this.localStorage.set('idClone', aIds[0]);
      this.displayInsertForm_();
    }
  }

  /**
   * Déclenche le processus de suppression des lignes sélectionnées
   */
  private deleteSelectedRows_(): void {

    const aIds = this.oDataGrid.getSelectedIds_();

    if (!(aIds.length > 0)) {
      vitisAlert.alert({
        type: 'error',
        title: this.translate.instant('LIST_DELETE_NO_SELECTION'),
      });
    } else {

      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.ressource)) {

          // Chemin de la ressource REST
          let sRessourceURL = environment.properties.app_api_url;
          sRessourceURL += '/' + this.oDataGridOptions.datagridOptions.ressource;

          vitisAlert.alert({
            type: 'warning',
            title: this.translate.instant('LIST_DELETE_CONFIRM'),
            showCancelButton: true
          }).then((result) => {
            if (result.value) {

              // Mode embedded
              if (this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {
                if (this.embeddedOptions.events &&
                  this.embeddedOptions.events.onDelete &&
                  this.embeddedOptions.events.onDelete.emit) {
                  this.embeddedOptions.events.onDelete.emit(aIds);
                }
              } else {
                this.runDelete_(sRessourceURL, aIds);
              }
            }
          });
        }
      }
    }
  }

  /**
   * Lance une requête delete sur l'API
   * @param  {string} sRessourceURL
   * @param  {Array<string>} aIds
   * @returns Promise
   */
  private runDelete_(sRessourceURL: string, aIds: Array<string>): Promise<void> {
    return new Promise((resolve, reject) => {

      // Lancement de la requête DELETE
      this.vitisRequest.ajaxRequestPromise({
        method: 'DELETE',
        url: sRessourceURL,
        data: {
          idList: aIds
        }
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {

            // Message succès
            this.translate.get('LIST_DELETE_OK').subscribe((translation) => {
              vitisAlert.fireToast({
                icon: 'success',
                title: translation
              });
              resolve();
            });
          } else {
            reject();
          }

          // Recharge les données
          this.dataTableInstance.ajax.reload();
        }
      });
    });
  }

  /**
   * Initialise les boutons
   * @returns void
   */
  public initButtons(): void {
    const aButtons: any = [];

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      // Bouton insert
      if (this.oDataGridOptions.datagridOptions.insert === true) {
        aButtons.unshift({
          type: 'button',
          class: 'btn-primary',
          label: 'LIST_BUTTON_ADD',
          icon: 'fa fa-plus',
          event: () => {
            this.displayInsertForm_();
          }
        });
      }

      // Bouton delete
      if (this.oDataGridOptions.datagridOptions.delete === true) {
        aButtons.unshift({
          type: 'button',
          class: 'btn-default',
          label: 'LIST_BUTTON_DELETE',
          icon: 'fa fa-trash',
          event: () => {
            this.deleteSelectedRows_();
          }
        });
      }

      // Group bouton action
      const groupButtonAction = {
        type: 'group',
        label: 'VITIS_WORKSPACELIST_ACTIONS_BTN',
        class: 'btn-primary',
        buttons: []
      };

      // Bouton clone
      if (this.oDataGridOptions.datagridOptions.clone === true) {
        groupButtonAction.buttons.push({
          type: 'button',
          label: 'LIST_BUTTON_CLONE',
          event: () => {
            this.cloneSelectedRow();
          }
        });
      }

      if (groupButtonAction.buttons.length > 0) {
        aButtons.unshift(groupButtonAction);
      }

      // Autres boutons
      if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.buttons)) {
        // Ajoute les boutons
        for (let i = 0; i < this.oDataGridOptions.datagridOptions.buttons.length; i++) {
          const button = this.oDataGridOptions.datagridOptions.buttons[i];
          this.oDataGrid.initButtonEvent_(button, this);
          this.oDataGrid.initButtonDisplay_(button, this);
          aButtons.unshift(button);
        }
      }
    }

    this.oDataGridOptions.buttons = aButtons;
  }

  /**
   * Charge le JS associé au mode
   * @param  {any} oMode
   * @returns Promise
   */
  private loadModeJavaScript_(oMode: any): Promise<boolean> {
    return new Promise((resolve, reject) => {

      // Si le mode contient un JS associé
      if (this.vitisCheck.isDefAndNotNull(oMode.options)) {
        if (this.vitisCheck.isDefAndNotNull(oMode.options.js_path)) {

          // Charge le JS du mode
          this.vitisRequest.loadExternalJs(oMode.options.js_path).then(() => {

            // Lance le JS associé
            if ((<any>window).constructor_form) {
              (<any>window).constructor_form(this, environment, {
                vitisRequest: this.vitisRequest,
                vitisCheck: this.vitisCheck,
                vitisAlert: vitisAlert,
                auth: this.auth,
                translate: this.translate
              });
            }

            resolve(true);
          });
        } else {
          resolve(false);
        }
      } else {
        resolve(false);
      }
    });
  }

  /**
   * Récupère les filtres à appliquer
   */
  private getFilters_() {
    const defaultFilter = this.getParsedFilters_();
    let currentFilter = this.oFilter.value;

    if (this.vitisCheck.isDefAndNotNull(defaultFilter) &&
      (this.vitisCheck.isDefAndNotNull(currentFilter) && Array.isArray(currentFilter.operators))) {
      currentFilter.operators = currentFilter.operators.concat(defaultFilter.operators);
    } else if (this.vitisCheck.isDefAndNotNull(defaultFilter)) {
      currentFilter = defaultFilter;
    }

    return currentFilter;
  }

  /**
   * Retourne le nombre d'éléments à afficher
   */
  private getPageLength_(): Number {
    let pageLength = this.pageLength;

    // SessionStorage
    if (this.sessionStorageValues_ &&
      this.sessionStorageValues_.datatableParams) {

      // length
      if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.datatableParams.length)) {
        pageLength = this.sessionStorageValues_.datatableParams.length;
      }
    }

    // Paramètres URL
    if (this.urlParams_ &&
      this.urlParams_.wkgstart &&
      this.urlParams_.wkglength) {
      pageLength = Number(this.urlParams_.wkglength);
    }

    // Concordance avec le tableau de choix
    const aLengthMenu = this.getLengthMenu_();
    if (aLengthMenu.indexOf(pageLength) === -1) {
      if (pageLength > aLengthMenu[aLengthMenu.length - 1]) {
        pageLength = aLengthMenu[aLengthMenu.length - 1];
      } else {
        pageLength = 20;
      }
    }

    return pageLength;
  }

  /**
   * Retourne le tableau contenant le nombre de lignes affichables
   */
  private getLengthMenu_(): Array<Number> {

    const lengthMenu = [5, 10, 20, 50, 100];

    if (this.oDataGridOptions &&
      this.oDataGridOptions.datagridOptions &&
      this.oDataGridOptions.datagridOptions.max_rows) {

      const iMaxRows = this.oDataGridOptions.datagridOptions.max_rows;

      for (let i = lengthMenu.length - 1; i >= 0; i--) {
        if (lengthMenu[i] > iMaxRows) {
          lengthMenu.splice(i, 1);
        }
      }
    }

    return lengthMenu;
  }

  /**
   * Charge le formulaire
   * @param oConfig
   * @param oConfig.form_path chemin vers le formulaire
   * @param oConfig.mode inser/update/display
   * @param oConfig.eventContainer eventContainer
   */
  public compileFormReaderModal(oConfig: any) {
    if (!this.vitisCheck.isDefAndNotNull(oConfig.form_path)) {
      console.error('Empty form_path for modal');
      return;
    }
    const oFormMode = this.vitisCheck.isDefAndNotNull(oConfig.mode) ? oConfig.mode : 'insert';
    // Chargement formulaire
    const url = oConfig.form_path;
    // Récupère le formulaire
    this.getJSONForm_(oConfig.form_path).then((oForm) => {
      if (this.vitisCheck.isDefAndNotNull(oConfig.eventContainer)) {
        oForm[oFormMode].eventContainer = oConfig.eventContainer;
      }
      this.oFormReader.initForm(oForm, oFormMode);
    });
  }

  /**
   * affiche la modale contenant le formreader
   */
  public showFormReaderModal() {
    $('#workspacelist-form-modal').modal('show');
  }

  /**
   * affiche la modale contenant le formreader
   */
  public hideFormReaderModal() {
    $('#workspacelist-form-modal').modal('hide');
  }

  /**
 * Récupère la définition JSON d'un formulaire
 * @param  {string} url
 * @returns Promise
 */
  private getJSONForm_(url: string): Promise<any> {
    return new Promise((resolve, reject) => {
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          resolve(resp.data);
        } else {
          console.error('Error getting JSON form');
          reject(false);
        }
      });
    });
  }
}