/**
 * constructor_form
 * Fonction lancée au chargement du formulaire
 * @param {object} scope Formreader
 * @param {object} env environment
 * @param {object} services
 * @returns {undefined} 
 */
var constructor_form = function (scope, env, services) {

  /**
   * Remplit le formulaire avec les infos de l'utilisateur
   */
  scope.loadProperties = function () {
    let sRessourceURL = env.properties.app_api_url + '/vitis/properties';

    services.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: sRessourceURL,
    }).then((resp) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          // Transforme la valeur en Mo pour les champs suivant: memory_limit , max_upload_file_size , log_size
          const oProperties = this.formatProperties(resp.data.data);

          // Applique au formulaire
          setTimeout(() => {
            scope.setFormValues(oProperties);
          }, 500);
        }
      }
    });
  }

  /**
   * Before event: Prépare les données avant l'envoi du formulaire
   * @param  {any} oFormValues
   */
  scope.prepareData = function (oFormValues) {
    // Transforme la valeur en octet pour les champs suivant:

    // memory_limit
    let oMemoryLimit = scope.getFormField('memory_limit');
    oFormValues.memory_limit = scope.convertMotoOctet(oMemoryLimit.getValue());

    // max_upload_file_size
    let oMaxUploadFileSize = scope.getFormField('max_upload_file_size');
    oFormValues.max_upload_file_size = scope.convertMotoOctet(oMaxUploadFileSize.getValue());
    let oMaxUploadFileSizeRestricted = scope.getFormField('max_upload_file_size_restricted');
    oFormValues.max_upload_file_size_restricted = scope.convertMotoOctet(oMaxUploadFileSizeRestricted.getValue());
    let oTransactionalThreshold = scope.getFormField('transactional_upload_threshold');
    oFormValues.transactional_upload_threshold = scope.convertMotoOctet(oTransactionalThreshold.getValue());
    let oChunkSize = scope.getFormField('chunk_size');
    oFormValues.chunk_size = scope.convertMotoOctet(oChunkSize.getValue());

    // log_size
    let oLogSize = scope.getFormField('log_size');
    oFormValues.log_size = scope.convertMotoOctet(oLogSize.getValue());

    // forbidden_extension
    let oForbiddenExtension = scope.getFormField('forbidden_extension');
    if (oForbiddenExtension.getValue()) {
      oFormValues.forbidden_extension = oForbiddenExtension.getValue().split('|');
    }

    return true;
  }

  /**
   * Fonction de mise à jour des properties
   */
  scope.updateProperties = function (oFormValues) {
    return new Promise((resolve, reject) => {
      services.route.data.subscribe((oConfig) => {
        let sRessourceURL = env.properties.app_api_url + '/' + oConfig.options.ressource;

        let oData = {}
        for (const key in oFormValues) {
          if (services.vitisCheck.isDefAndNotNull(oFormValues[key])) {
            oData[key] = oFormValues[key];
          }
        }

        services.vitisRequest.ajaxRequestPromise({
          method: 'PUT',
          url: sRessourceURL,
          data: oData
        }).then((resp) => {
          if (resp.data) {
            if (resp.data.status === 1) {
              this.getProperties().then((oProperties) => {

                // MAJ variable properties
                env.properties = {
                  ...env.properties,
                  ...oProperties
                };

                // toast
                services.translate.get('FORM_PROPERTIES_EDIT_OK').subscribe((translation) => {
                  services.vitisAlert.fireToast({
                    type: 'success',
                    title: translation
                  });
                  resolve();
                });
              }).catch(() => {
                console.error('Could not refresh properties');
              });
            }
          }
        });
      });
    })
  }

  /**
   * Convertit les octets en MégaOctet
   * @param {number} iValue
   */
  scope.convertOctetToMo = function (iValue) {
    return iValue / 1024 / 1024;
  }

  /**
   * Convertit les MégaOctet en octet
   * @param {number} iValue
   */
  scope.convertMotoOctet = function (iValue) {
    return iValue * 1024 * 1024;
  }

  /**
   * Test SMTP
   * @param {*} oFormValues 
   */
  scope.testSmtpConnection = function (oFormValues) {
    const sRessourceURL = env.properties.app_api_url + '/vitis/smtp/check';

    const oData = {
      "mail_sender": oFormValues['mail_sender'],
      "nickname_sender": oFormValues['nickname_sender'],
      "smtp_host": oFormValues['smtp_host'],
      "smtp_port": oFormValues['smtp_port'],
      "smtp_authentification": oFormValues['smtp_authentification'],
      "smtp_protocol": oFormValues['smtp_protocol'],
      "smtp_debug_level": oFormValues['smtp_debug_level'],
      "smtp_login": oFormValues['smtp_login'],
      "smtp_password": oFormValues['smtp_password']
    }

    let sToastMode = 'error';
    let sTraductionKey = 'MODAL_CONTENT_ERROR_CHECK_SMTP_CONNECTION_CONFIGURATION';

    services.vitisRequest.ajaxRequestPromise({
      method: 'POST',
      url: sRessourceURL,
      data: oData,
      displayErrorMessage: false
    }).then((resp) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          sToastMode = 'success';
          sTraductionKey = 'MODAL_CONTENT_SUCCESS_CHECK_SMTP_CONNECTION_CONFIGURATION';
        }
      }
      services.translate.get(sTraductionKey).subscribe((translation) => {
        services.vitisAlert.fireToast({
          type: sToastMode,
          title: translation
        });
      });
    }).catch(() => {
      services.translate.get(sTraductionKey).subscribe((translation) => {
        services.vitisAlert.fireToast({
          type: sToastMode,
          title: translation
        });
      });
    });

  }

  /**
   * Get Properties
   */
  scope.getProperties = function () {
    return new Promise((resolve, reject) => {

      let sRessourceURL = env.properties.app_api_url + '/vitis/properties';
      services.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: sRessourceURL,
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            resolve(resp.data.data);
          } else {
            reject();
          }
        } else {
          reject();
        }
      });
    });
  }

  /**
   * Formette les properties pour le formulaire
   * @param {*} oProperties 
   */
  scope.formatProperties = function (oProperties) {
    const oNewProperties = {
      ...oProperties
    }

    // Transforme la valeur en Mo pour les champs suivant: memory_limit , max_upload_file_size , log_size
    oNewProperties.memory_limit = scope.convertOctetToMo(oProperties.memory_limit);
    oNewProperties.max_upload_file_size = scope.convertOctetToMo(oProperties.max_upload_file_size);
    oNewProperties.max_upload_file_size_restricted = scope.convertOctetToMo(oProperties.max_upload_file_size_restricted);
    oNewProperties.transactional_upload_threshold = scope.convertOctetToMo(oProperties.transactional_upload_threshold);
    oNewProperties.chunk_size = scope.convertOctetToMo(oProperties.chunk_size);
    oNewProperties.log_size = scope.convertOctetToMo(oProperties.log_size);
    oNewProperties.forbidden_extension = oProperties.forbidden_extension ? oProperties.forbidden_extension.join('|') : '';

    return oNewProperties;
  }
}
