# Requête SQL à lancer avec des paramètres

Le QueryAccessor de vitis peut fournir un queryBuilder qui permettra de faire exactement les mêmes requêtes qu'avec l'ancienne fonction executeWithParams (vitis 2019).

Documentation du [QueryBuilder](https://www.doctrine-project.org/projects/doctrine-dbal/en/latest/reference/query-builder.html) de DBAL.

Toutes les actions faisables avec le querybuilder de DBAL le sont aussi avec notre QueryBuilder. Nous avons ajouté plus de possibilités (utilisation de returning, simplification de l'utilisation de IN dans les where, utilisation de NULLS LAST/FIRST dans un order by, protection des noms de colonnes, schemas, tables possibles), en plus de reprendre l'ancienne notation pour rendre le passage de l'ancienne API à la nouvelle plus simple.

L'utilisation de la fonction `setParameter`, sera fait pour vous dans le cas où vous utilisez `executeWithParams` à la place de `execute`.

## Fonctions supplémentaires

### ExecuteWithParams(array $aParams)

Comme décrit ci-dessus cette fonction permettra d'exécuter une requête SQL générée via le QueryBuilder, pour communiquer avec la BDD.

Le tableau est quasi identique à ce que l'on pouvait faire dans l'ancienne version.

ex : 
```php
$aParams = [
  'BIND 1' => ['value' => $ma_valeur_1, 'type' => 'type1'],
  'BIND 1' => ['value' => $ma_valeur_2, 'type' => 'type2'],
  .... 
  'BIND X' => ['value' => $ma_valeur_X, 'type' => 'typeX'],
];
```

La clé est la balise utilisée dans la requête entre crochets (ex : `[SCHEMA].ma_table` donnera la clé `SCHEMA` dans la liste).

Le couple type/valeur est décrit ci-dessous
- **column_name** : string, sera controlé par regex
- **table_name** : string, sera controlé par regex
- **schema_name** : string, sera controlé par regex
- **geometry** : string, sera protégé comme une valeur
- **quoted_string** : string, sera protégé comme une valeur
- **double_quote** : string, sera protégé comme une colonne
- **boolean** : boolean, sous forme True/False, 0/1 
- **integer** : integer
- **number** : integer/float
- **string** : string
- **array** : array permet de passer une liste de valeurs qui sera ensuite utilisée comme dans un IN (voir exemple)

### returning(string|string[]|null $fields=null)

Permet d'ajouter une clause RETURNING aux requêtes INSERT/UPDATE.

Cette fonction s'utilise exactement de la même manière que la fonction select du querybuilder de DBAL.

Attention par rapport à la documentation de DBAL pour les insert, dans le cas de l'utilisation d'un returning, il faut traiter le résultat de execute/executeWithParams comme si c'etait un SELECT (`->fetch`, ...).

### nullsFirst() et nullsLast()

Permet de définir la position des valeurs nulls par rapport au ORDER BY.

## Exemple

### Lancement d'une requête avec le compte de l'utilisateur courant

```php
$qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();
$qb->select('context_documentation','name','body','json_scope','json_includes')
  ->from('[SCHEMA]."v_email_template"', 'vet')
  ->where($qb->expr()->eq('vet.email_template_id', '[TEMPL_ID]'));
  $aParams = [
    'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
    'TEMPL_ID' => ['value' => $id, 'type' => 'integer'],
  ];
try {
  $aResult = $qb->executeWithParams($aParams)->fetch();
} catch (\Exception $e) {
  $this->oLoggerManager->get('SQL_ERROR')->error('Error getting mail template parameters' . $e->getMessage());
  // ...
}
```

### Lancement d'une requête avec le compte superuser en properties

```php
$oEm = $this->oDatabaseConnection->getAdminEntityManager();
$qb = $this->oDatabaseConnection->getQueryAccessor($oEm)->createQueryBuilder();
$aParamsSelect = [
  'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
  'aGroupList' => ['value' => $aLowerUserGroups, 'type' => 'array'],
];
$qb->select('a.group_id')
  ->from('[SCHEMA].group', 'a')
  ->where('lower(a.name) IN ([aGroupList])');
try {
  $aGroups = $qb->executeWithParams($aParamsSelect)->fetchAll(\PDO::FETCH_COLUMN);
} catch (\Exception $e) {
  $this->oLoggerManager->get('SQL_ERROR')->error('Cannot get groups ' . $e->getMessage());
  return false;
}
```

# Interagir avec les fichiers

L'interface fichier permet d'interagir de manière quasi transparente avec le disque ou S3 en fonction de la configuration de l'application.

La plus-part des fonctions natives de php d'interaction avec le disque ont été intégré dans l'interface.

Le contrat d'interface se trouve [ici](../../vas/src/Module/Vitis/WebService/Service/FileAccessor/FileAccessorInterface.php).

## Fonctionnement générique

quatre dossier peuvent être déporter sur S3 (tous dans var):
- ws_data
- public
- upload
- shared

Le fait d'avoir ces dossiers peuvent rendre certains actions impossibles (utilisation du paramètre publié pour shared_dir par exemple).

## Interaction volontaire en local sur le disque

Il se peut que l'on veuille forcément travailler sur un fichier en local, pour générer un fichier via une commande, ...

ci dessous un exemple de code pour faire une synchro pour travailler en local suivi d'un nettoyage du disque local

```php
  // si filesystem distant on fait une copie locale du dossier
  if ($this->oProperties->get('file_system') !== Properties::LOCAL_FS) {
    $this->getFileAccessor()->copyDirDistantToLocal($sPath, $sPath);
  }

  // GetLocalAccessor et interaction du fichier, lancement de commande, ...

  // si filesystem distant on envoi sur le système de fichier
  if ($this->oProperties->get('file_system') !== Properties::LOCAL_FS) {
    $this->getFileAccessor()->copyDirLocalToDistant($sPath, $sPath);
    // on supprime le dossier local 
    $this->getLocalFileAccessor()->clearDir($sPath);
  }
```

## Génération d'un lien de téléchargement pour un fichier

```php
  $this->oFileManager->getFileAccessor()->getProxyPassUrl($sPath);
```

Cette fonction permet de générer une url de téléchargement du fichier via l'API, peut importe de l'emplacement du ficheir (disque ou S3), permet aussi de gérer tout ce qui est gestion des thumbnail, ...

# Ajout d'un contexte de mail

Le développement d'un envoi de mail doit être à associer à un contexte de mail (qui correspond à du code spécifique, scope, ...).

Il faut donc créer une classe (ORM ou non) qui implémente une interface afin de permettre au générateur de mail de fonctionner.

# Mise à jour des properties d'un module

Par défaut le controller générique vous permettra de gérer ça sans rien toucher

## Actions spécifiques 

Il est parfois nécessaire de réaliser une action suite à la mise à jour d'une properties (ajout d'une tâche planifié, activation d'un service, copie d'un fichier, ...)

ex :
```php
class CustomPropertiesController extends ApplicationData {
  
  /**
   * Map à utiliser pour lancer des actions spécifiques si une properties est modifiée
   * Les fonctions à appeler devront toujours ne pas avoir de paramètre, retourner un booléen et 
   * être public ou protected car appelé depuis le parent
   * @var array
   */
  protected const CUSTOM_ACTION_FOR_PROPERTIES = [
    'sync_domain_group' => 'manageSyncDomainGroup',
    'sync_domain_group_period' => 'manageSyncDomainGroup',
  ];

  /**
   * Traitement spécifique lié à la properties sync_domain_group
   * @return bool true si l'action s'est bien passé false sinon
   */
  protected function manageSyncDomainGroup(): bool {
    if ($this->oProperties->get('sync_domain_group') == true) {
      $sPeriod = $this->oProperties->get('sync_domain_group_period');
      if (!empty($sPeriod)) {
        return $this->oJobsService->addAutosyncgroups($sPeriod);
      } else {
        return false;
      }
    } else {
      return $this->oJobsService->removeAutosyncgroups();
    }
  }
```