import puppeteer, { PaperFormat } from 'puppeteer';
import fs from 'fs';
import { Logger } from "vitis-engine-lib";
import { Db } from "vitis-engine-lib";
import { LocalFS } from "vitis-engine-lib";
import { initMapOptions, openMapPage, openMapPageOptions } from './utils';

interface printMapArgs {
    OutputFileName: string;
    JsonMap?: string;
    oMap?: any;
    Properties?: string;
}

interface printMapResult {
    target: string,
    imageUrl: string,
}

/**
 * Imprime plusieurs cartes
 * @param oProperties 
 * @param oDb 
 * @param oLogger 
 * @param oLocalFS 
 * @param aMaps 
 * @param sOutputFolder 
 * @returns 
 */
export async function printMaps(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, aMaps: Array<any>, sOutputFolder: string): Promise<printMapResult[]> {
    const aIncludes: printMapResult[] = [];
    let page, browser;

    for (let i = 0; i < aMaps.length; i++) {
        const oMap = aMaps[i];
        const sImagePath = sOutputFolder + '/map_' + i + '.jpeg';
        const oParams: openMapPageOptions | undefined = getPrintMapParameters_(oProperties, oDb, oLogger, oLocalFS, {
            OutputFileName: sImagePath,
            oMap,
        });

        if (oParams) {

            // Ouvre la page
            const openedPage: any = await openMapPage({
                sPrintClientURL: oParams.sPrintClientURL,
                oInitMapOptions: oParams.oInitMapOptions,
                aImageSize: oParams.aImageSize,
                oProperties,
                page,
                browser,
            }, oLogger);

            // Stocke page et browser pour qu'ils soient ré-utilisés dans les autres cartes
            page = openedPage.page;
            browser = openedPage.browser;
            await page.screenshot({ path: sImagePath, fullPage: true });
        }
        // Ajoute l'image dans les includes
        if (fs.existsSync(sImagePath)) {
            const oMapInclude = {
                target: oMap.target,
                imageUrl: sImagePath,
            };
            aIncludes.push(oMapInclude);
        }
    }

    // Ferme page et browser
    if (page) {
        await page.close();
    }
    if (browser) {
        await browser.close();
    }

    return aIncludes;
}

/**
 * Imprime une carte
 */
export async function printMap(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, args: printMapArgs) {

    const sOutputFileName = args.OutputFileName;
    let sJsonMap = args.JsonMap;
    let oMap = args.oMap ? args.oMap : {};

    // URL du front (pour que ça puisse fonctionner avec PHP ce dernier doit être lancé avec apache de manière à être multi-thread)
    const sPrintClientURL = `${oProperties.app_client_url}/widget/printmap?token=${oMap.map_definition.token}`;

    // Si sJsonMap est un fichier
    if (sJsonMap && fs.existsSync(sJsonMap)) {
        sJsonMap = fs.readFileSync(sJsonMap).toString();
    }
    try {
        if (sJsonMap) {
            oMap = JSON.parse(sJsonMap);
        }
    } catch (e) { }

    if (!oMap.target) {
        oLogger.log(`Error while printing map : target not defined`, oMap, 'ERROR');
        return;
    }
    if (!oMap.map_definition) {
        oLogger.log(`Error while printing map : map_definition not defined`, oMap, 'ERROR');
        return;
    }

    const oInitMapOptions: initMapOptions = {
        token: oMap.map_definition.token,
        apiUrl: oProperties.app_api_url
    }

    if (oMap.map_definition.map_json) {
        oInitMapOptions.mapJSON = oMap.map_definition.map_json;
    } else if (oMap.map_definition.map_id) {
        oInitMapOptions.mapId = oMap.map_definition.map_id;
    } else {
        oLogger.log(`Error while printing map : neither map_id nor map_json defined`, oMap, 'ERROR');
        return;
    }

    if (oMap.map_definition.extent && oMap.map_definition.extent.length > 0) {
        oInitMapOptions.extent = oMap.map_definition.extent;
    } else if (oMap.map_definition.features && oMap.map_definition.features.length > 0) {
        oInitMapOptions.features = oMap.map_definition.features;
    } else {
        oLogger.log(`Error while printing map : neither extent nor features defined`, oMap, 'ERROR');
        return;
    }

    if (oMap.map_definition.measure_features) {
        oInitMapOptions.measureFeatures = oMap.map_definition.measure_features;
    }

    if (oMap.map_definition.features_zoom) {
        oInitMapOptions.featuresZoom = oMap.map_definition.features_zoom;
    }

    if (oMap.map_definition.resolution_coeff) {
        oInitMapOptions.resolutionCoeff = oMap.map_definition.resolution_coeff;
    }

    let aImageSize;
    if (oMap.map_definition.image_size && typeof oMap.map_definition.image_size === 'string') {
        aImageSize = oMap.map_definition.image_size.split('|');
        aImageSize[0] = parseInt(aImageSize[0]);
        aImageSize[1] = parseInt(aImageSize[1]);
    }

    oInitMapOptions.properties = oProperties;

    // Ouvre la page
    const openedPage = await openMapPage({
        sPrintClientURL,
        oInitMapOptions,
        aImageSize,
        oProperties,
    }, oLogger);
    const page = openedPage.page;
    const browser = openedPage.browser;
    await page.screenshot({ path: sOutputFileName, fullPage: true });
    await browser.close();
}

/**
 * Retrouve les paramètres d'impression en fonction de la définition d'une carte
 * @param oProperties 
 * @param oDb 
 * @param oLogger 
 * @param oLocalFS 
 * @param args 
 * @returns 
 */
function getPrintMapParameters_(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, args: printMapArgs): openMapPageOptions | undefined {

    let sJsonMap = args.JsonMap;
    let oMap = args.oMap ? args.oMap : {};

    // URL du front (pour que ça puisse fonctionner avec PHP ce dernier doit être lancé avec apache de manière à être multi-thread)
    const sPrintClientURL = `${oProperties.app_client_url}/widget/printmap?token=${oMap.map_definition.token}`;

    // Si sJsonMap est un fichier
    if (sJsonMap && fs.existsSync(sJsonMap)) {
        sJsonMap = fs.readFileSync(sJsonMap).toString();
    }
    try {
        if (sJsonMap) {
            oMap = JSON.parse(sJsonMap);
        }
    } catch (e) { }

    if (!oMap.target) {
        oLogger.log(`Error while printing map : target not defined`, oMap, 'ERROR');
        return;
    }
    if (!oMap.map_definition) {
        oLogger.log(`Error while printing map : map_definition not defined`, oMap, 'ERROR');
        return;
    }

    const oInitMapOptions: initMapOptions = {
        token: oMap.map_definition.token,
        apiUrl: oProperties.app_api_url
    }

    if (oMap.map_definition.map_json) {
        oInitMapOptions.mapJSON = oMap.map_definition.map_json;
    } else if (oMap.map_definition.map_id) {
        oInitMapOptions.mapId = oMap.map_definition.map_id;
    } else {
        oLogger.log(`Error while printing map : neither map_id nor map_json defined`, oMap, 'ERROR');
        return;
    }

    if (oMap.map_definition.extent && oMap.map_definition.extent.length > 0) {
        oInitMapOptions.extent = oMap.map_definition.extent;
    } else if (oMap.map_definition.features && oMap.map_definition.features.length > 0) {
        oInitMapOptions.features = oMap.map_definition.features;
    } else {
        oLogger.log(`Error while printing map : neither extent nor features defined`, oMap, 'ERROR');
        return;
    }

    if (oMap.map_definition.measure_features) {
        oInitMapOptions.measureFeatures = oMap.map_definition.measure_features;
    }

    if (oMap.map_definition.features_zoom) {
        oInitMapOptions.featuresZoom = oMap.map_definition.features_zoom;
    }

    if (oMap.map_definition.resolution_coeff) {
        oInitMapOptions.resolutionCoeff = oMap.map_definition.resolution_coeff;
    }

    let aImageSize;
    if (oMap.map_definition.image_size && typeof oMap.map_definition.image_size === 'string') {
        aImageSize = oMap.map_definition.image_size.split('|');
        aImageSize[0] = parseInt(aImageSize[0]);
        aImageSize[1] = parseInt(aImageSize[1]);
    }

    oInitMapOptions.properties = oProperties;

    return {
        sPrintClientURL,
        oInitMapOptions,
        aImageSize,
        oProperties,
    }
}