import fs from 'fs';
import forge from 'node-forge';
import { PlaceHolder } from './placeholder';

export interface SignatureOptions {
  asn1StrictParsing?: boolean,
  passphrase?: string,
}

export class Signature {

  private p12Buffer: Buffer;
  private options: SignatureOptions = {
    asn1StrictParsing: false,
    passphrase: '',
  }

  private beginPlaceholder?: number;
  private endPlaceholder?: number;

  constructor(sCertPath: string, options: SignatureOptions) {
    this.p12Buffer = fs.readFileSync(sCertPath);
    this.options = {
      ...this.options,
      ...options
    }
  }


  public signPdf(pdfBuffer: Buffer): Buffer {
    // récupère le buffer à hasher
    const pdf: Buffer = this.preparePdfBufferForSignature(pdfBuffer);

    // Manipulation du certificat et hashage
    let signature = this.generateSignature(pdf);

    // replacement de la signature dans le pdf
    const finalBuffer = this.replacePlaceholderBySignature(pdfBuffer, signature);

    if (pdfBuffer.length !== finalBuffer.length) {
      throw new Error('PDF Alteration')
    }

    return finalBuffer;
  }

  private generateSignature(pdf: Buffer): Buffer {
    // Convert Buffer P12 to a forge implementation.
    const forgeCert = forge.util.createBuffer(this.p12Buffer.toString('binary'));
    const p12Asn1 = forge.asn1.fromDer(forgeCert);
    const p12 = forge.pkcs12.pkcs12FromAsn1(
      p12Asn1,
      this.options.asn1StrictParsing,
      this.options.passphrase,
    );

    // Extract safe bags by type.
    // We will need all the certificates and the private key.
    const certBags = p12.getBags({
      bagType: forge.pki.oids.certBag,
    })[forge.pki.oids.certBag];
    const keyBags = p12.getBags({
      bagType: forge.pki.oids.pkcs8ShroudedKeyBag,
    })[forge.pki.oids.pkcs8ShroudedKeyBag];

    if (!keyBags) {
      throw new Error('cannot generate keybags for this key');
    }

    if (!certBags) {
      throw new Error('cannot generate certbags for this key');
    }

    const privateKey = keyBags[0].key;
    if (!privateKey) {
      throw new Error('cannot found private key for this key');
    }
    // Here comes the actual PKCS#7 signing.
    const p7 = forge.pkcs7.createSignedData();
    // Start off by setting the content.
    p7.content = forge.util.createBuffer(pdf.toString('binary'));

    // Then add all the certificates (-cacerts & -clcerts)
    // Keep track of the last found client certificate.
    // This will be the public key that will be bundled in the signature.
    let certificate;
    Object.keys(certBags).forEach((i: any) => {

      const certTmp = certBags[i].cert
      if (!certTmp) {
        throw new Error('cannot find cert');
      }

      const { publicKey } = certTmp;

      p7.addCertificate(certTmp);

      const oPrivateKey: any = privateKey;
      const oPublicKey: any = publicKey;
      // Try to find the certificate that matches the private key.
      if (oPrivateKey.n.compareTo(oPublicKey.n) === 0
        && oPrivateKey.e.compareTo(oPublicKey.e) === 0
      ) {
        certificate = certBags[i].cert;
      }
    });

    if (typeof certificate === 'undefined') {
      throw new Error('Failed to find a certificate that matches the private key.');
    }

    const signerKey: any = privateKey;
    // Add a sha256 signer. That's what Adobe.PPKLite adbe.pkcs7.detached expects.
    p7.addSigner({
      key: signerKey,
      certificate,
      digestAlgorithm: forge.pki.oids.sha256,
      authenticatedAttributes: [
        {
          type: forge.pki.oids.contentType,
          value: forge.pki.oids.data,
        }, {
          type: forge.pki.oids.messageDigest,
          // value will be auto-populated at signing time
        }, {
          type: forge.pki.oids.signingTime,
          // value can also be auto-populated at signing time
          // We may also support passing this as an option to sign().
          // Would be useful to match the creation time of the document for example.
          value: new Date().toString(),
        },
      ],
    });

    // Sign in detached mode.
    p7.sign({ detached: true });

    // Check if the PDF has a good enough placeholder to fit the signature.
    const raw = forge.asn1.toDer(p7.toAsn1()).getBytes();
    // placeholderLength represents the length of the HEXified symbols but we're
    let signature = Buffer.from(raw, 'binary').toString('hex');
    return Buffer.from(signature);
  }

  private preparePdfBufferForSignature(pdf: Buffer): Buffer {
    // dé la façon dont c'est conçu la signature à remplir devrait être la dernière
    const signaturePos = pdf.lastIndexOf('/Type /Sig')

    if (signaturePos === -1) {
      throw new Error('No signature placeholder found')
    }

    const signatureContentPos = pdf.indexOf('/Contents', signaturePos);
    this.beginPlaceholder = pdf.indexOf('<', signatureContentPos) + 1;
    this.endPlaceholder = pdf.indexOf('>', this.beginPlaceholder);

    // les chevron ne doivent pas être dnas les donnée à hasher
    return Buffer.concat([
      pdf.slice(0, this.beginPlaceholder - 1),
      pdf.slice(this.endPlaceholder + 1),
    ]);
  }

  private replacePlaceholderBySignature(pdf: Buffer, signature: Buffer): Buffer {

    if (!this.endPlaceholder || !this.beginPlaceholder) {
      throw new Error('prepare this buffer before')
    }

    // finaliser la taille de la signature avec 0000000
    const fillerLength: number = this.endPlaceholder - this.beginPlaceholder - signature.length;
    if (fillerLength < 0) {
      throw new Error('this placeholder is too short we need : ' + signature.length + ' bytes')
    }

    // slice <> pour ajouter la signature à l'intérieur 
    return Buffer.concat([
      pdf.slice(0, this.beginPlaceholder),
      signature,
      pdf.slice(this.beginPlaceholder + signature.length),
    ]);
  }
}