import puppeteer, { Browser, HTTPRequest, Page, ConsoleMessage } from 'puppeteer';
import EventEmitter from "events";
import { Logger } from 'vitis-engine-lib';


export interface openPageOptions {
    sPrintClientURL: string;
    sHTMLBody: string;
    sFormat?: string;
    sOrientation?: string;
    aIncludes?: any;
    oScope?: any;
    oProperties?: any;
}

export interface openMapPageOptions {
    sPrintClientURL: string;
    oInitMapOptions: any;
    aImageSize: Array<number>;
    oProperties?: any;
    oLogger?: any;
    page?: any; // page déjà chargée
    browser?: any; // browser déjà chargé
}

export interface initMapOptions {
    apiUrl: string;
    token: string;
    mapId?: string;
    mapJSON?: string;
    extent?: string[];
    features?: string;
    featuresZoom?: string;
    measureFeatures?: string;
    resolutionCoeff?: string;
    properties?: any;
}

export interface openPageReturn {
    page: Page,
    browser: Browser,
}

/**
 * Formats A1, A2 etc. en pixels
 */
const a4Format = { width: 793, height: 1122 };
const a3Format = { width: a4Format.height, height: a4Format.width * 2 };
const a2Format = { width: a3Format.height, height: a3Format.width * 2 };
const a1Format = { width: a2Format.height, height: a2Format.width * 2 };
const a0Format = { width: a1Format.height, height: a1Format.width * 2 };
export const formatsInPixels: any = {
    'a4': a4Format,
    'a3': a3Format,
    'a2': a2Format,
    'a1': a1Format,
    'a0': a0Format,
}

/**
 * Ouvre la page, évalue le scope etc et retourne un objet {page, browser}
 * @param oOptions 
 * @returns 
 */
export async function openPage(oOptions: openPageOptions, oLogger: Logger): Promise<openPageReturn> {
    const oBrowserOptions: any = {
        headless: true,
        ignoreHTTPSErrors: true,
    }

    if (Array.isArray(oOptions.oProperties.chromium_args)) {
        oBrowserOptions.args = oOptions.oProperties.chromium_args;
    }

    const browser = await puppeteer.launch(oBrowserOptions);
    const page = await browser.newPage();
    listenPageErrors(page, oLogger);

    if (oOptions.sFormat &&
        formatsInPixels[oOptions.sFormat.toLowerCase()]) {
        let selectedFormat = formatsInPixels[oOptions.sFormat.toLowerCase()];
        if (oOptions.sOrientation === 'paysage') {
            selectedFormat = {
                width: selectedFormat.height,
                height: selectedFormat.width,
            }
        }
        await page.setViewport(selectedFormat);
    }
    await page.goto(oOptions.sPrintClientURL, {
        waitUntil: 'networkidle0',
    });

    await page.evaluate(function (oOptions) {
        // @ts-ignore
        initRender(oOptions);
        // @ts-ignore
        copyStyles();
        // @ts-ignore
        initPaged();
    }, {
        includes: oOptions.aIncludes,
        scope: oOptions.oScope,
        htmlbody: oOptions.sHTMLBody,
        properties: oOptions.oProperties,
    });

    const iTimeout = oOptions.oProperties.loadTimeout || 1000;
    await new Promise(r => setTimeout(r, iTimeout));

    return {
        page,
        browser,
    };
}

/**
 * Ouvre la page, évalue la carte et retourne un objet {page, browser}
 * @param oOptions 
 * @returns 
 */
export async function openMapPage(oOptions: openMapPageOptions, oLogger: Logger): Promise<openPageReturn> {
    let page, browser;
    
    const oBrowserOptions: any = {
        headless: true,
        ignoreHTTPSErrors: true,
    }

    if (Array.isArray(oOptions.oProperties.chromium_args)) {
        oBrowserOptions.args = oOptions.oProperties.chromium_args;
    }

    if (oOptions.page && oOptions.browser) {
        // Si page et browser sont déjà définis alors ne les recharge pas
        page = oOptions.page;
        browser = oOptions.browser;
        await page.setViewport({
            width: oOptions.aImageSize[0],
            height: oOptions.aImageSize[1],
        });
    } else {
        // Autrement ouvre la page demandée
        browser = await puppeteer.launch(oBrowserOptions);
        page = await browser.newPage();
        listenPageErrors(page, oLogger);
        await page.setViewport({
            width: oOptions.aImageSize[0],
            height: oOptions.aImageSize[1],
        });
        await page.goto(oOptions.sPrintClientURL, {
            waitUntil: 'networkidle0',
        });
    }

    try {
        await page.evaluate(function (oOptions: any) {
            try {
                // @ts-ignore
                window.initPrintMap(oOptions.oInitMapOptions);
            } catch (error) {
                throw new Error('init map error : ' + error);
            }
        }, {
            oInitMapOptions: oOptions.oInitMapOptions
        });
    } catch (evalError) {
        oLogger.log(`init map error : ` + evalError, null, 'ERROR');
        throw new Error('init map error : ' + evalError);
    }

    if (oOptions.oProperties && oOptions.oProperties.loadTimeout) {
        oLogger.log('Wait map timeout', oOptions.oProperties.loadTimeout, 'DEBUG');
        await new Promise(resolve => setTimeout(resolve, oOptions.oProperties.loadTimeout || 1000));
    }

    const waitForIdleNetwork = createFunctionWaitForIdleNetwork(page);
    await waitForIdleNetwork(500, 50000).catch(err => { oLogger.log(`timeout error`, err, 'ERROR'); });
    return {
        page,
        browser,
    };
}

/**
 * Fonction qui permet d'attendre que le réseau soit ok
 * @param page 
 * @returns 
 */
export function createFunctionWaitForIdleNetwork(page: Page): (idleTimeout: number, failTimeout: number) => Promise<void> {
    // @ts-ignore
    class Emitter extends EventEmitter { }

    let pendingRequests: HTTPRequest[] = [];
    const emitter = new Emitter();

    function pushRequest(request: HTTPRequest) {
        pendingRequests.push(request);
        // @ts-ignore
        emitter.emit("active");
    }

    function popRequest(request: HTTPRequest) {
        pendingRequests = pendingRequests.filter((r) => r !== request);
        if (pendingRequests.length === 0) {
            // @ts-ignore
            emitter.emit("idle");
        }
    }

    page.on("request", pushRequest);
    page.on("requestfinished", popRequest);
    page.on("requestfailed", popRequest);

    /**
     * Return a promise that will resolve when the network is idle.
     *
     * @param idleTimeout
     *   The minimum amount of time that the network must be idle before the promise will resolve.
     *
     * @param failTimeout
     *   The maximum amount of time to wait for the network to become idle before rejecting.
     */
    async function waitForIdleNetwork(
        idleTimeout: number,
        failTimeout: number
    ): Promise<void> {
        let failTimer: NodeJS.Timeout;
        let idleTimer: NodeJS.Timeout;

        return new Promise((resolve, reject) => {
            function fail() {
                reject(
                    new Error(
                        `After ${failTimeout}ms, there are still ${pendingRequests.length} pending network requests.`
                    )
                );
            }

            function succeed() {
                clearTimeout(failTimer);
                resolve();
            }

            // Start failure time immediately.
            failTimer = setTimeout(fail, failTimeout);

            // Handle edge case where neither active nor idle is emitted during the lifetime of this promise.
            if (pendingRequests.length === 0) {
                idleTimer = setTimeout(succeed, idleTimeout);
            }

            // Play a game of whack-a-mole with the idle and active events.
            // @ts-ignore
            emitter.on("idle", () => {
                idleTimer = setTimeout(succeed, idleTimeout);
            });
            // @ts-ignore
            emitter.on("active", () => {
                clearTimeout(idleTimer);
            });
        });
    }

    return waitForIdleNetwork;
}

const listenPageErrors = async (page: Page, oLogger: Logger) => {
    // listen to browser console there
    page.on('console', async (message: ConsoleMessage) => {
        const sLevel: string = (['info', 'error', 'warning'].indexOf(message.type()) > -1) ? message.type().toUpperCase() : 'DEBUG';
        oLogger.log(`console message : ${message.text()}`, null, sLevel);
    }).on('pageerror', async (err) => {
        oLogger.log(`page error : ${err.toString()}`, null, 'ERROR');
    });
}