<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Service\Configuration\Module;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\AbstractDataAccessor;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileAccessorInterface;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Doctrine\ORM\EntityManager;
use InvalidArgumentException;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\BinaryFileResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\ResponseHeaderBag;

/**
 * Controller abstrait pour les autres controller de Vitis.
 * Contient la majorité des fonctions et appels aux services pour les controller.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

abstract class AbstractVitisController extends AbstractController {
  /**
   * partie de regexp à utiliser pour remplacer les * dans les masques d'IP IPV4
   * @var string
   */
  private const IPV4_STAR_REPLACER = '[0-9]{1,3}';
  /**
   * partie de regexp à utiliser pour remplacer les * dans les masques d'IP IPV6
   * @var string
   */
  private const IPV6_STAR_REPLACER = '[0-9a-fA-F]{1,4}';
  /**
   * partie de regexp à utiliser pour remplacer les . dans les IP IPV4
   * @var string
   */
  private const IPV4_DOT_REPLACER = '\\.';
  /**
   * caractère permettant de détecter s'il s'agit d'une IPV6
   * @var string
   */
  private const IPV6_DELIMITER = ':';
  /**
   * headers authorisé pour l'application
   * @var string[]
   */
  protected const HTTP_RESPONSE_ALLOWED_HEADERS = ['Content-Type', 'Authorization', 'Accept', 'X-HTTP-Method-Override'];
  /**
   * headers authorisé pour l'application
   * @var string[]
   */
  protected const HTTP_METHODS_ALLOWED = ['GET', 'POST', 'PUT', 'DELETE', 'OPTIONS'];
  /**
   * @var Route
   */
  protected $oRouteConfig;
  /**
   * @var array
   */
  private $aRequestOptions = [];
  /**
   * @var Properties
   */
  protected $oProperties;
  /**
   * @var DatabaseConnection
   */
  protected $oDatabaseConnection;
  /**
   * @var AbstractDataAccessor
   */
  private $oDataAccessor;
  /**
   * @var VitisSerializer
   */
  protected $oSerializer;
  /**
   * @var FileAccessorInterface
   */
  private $oFileInterface;
  /**
   * @var LocalAccessor
   */
  private $oLocalFileInterface;
  /**
   * @var LoggerManager
   */
  protected $oLoggerManager;
  /**
   * @var VitisRequest
   */
  protected $oVitisRequest;
  /**
   * @var Response
   */
  protected $oResponse;
  /**
   * @var FileManager
   */
  protected $oFileManager;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service permettant d'accéder aux fichiers
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager) {
    $this->oProperties = $oPropertiesSrvc;
    $this->oVitisRequest = $oVitisRequest;
    $this->oRouteConfig = $oRouteConfig;
    $this->oLoggerManager = $oLoggerManager;
    $this->oDatabaseConnection = $oDatabaseConnection;
    $this->oFileManager = $oFileManager;

    $this->oSerializer = new VitisSerializer();

    $this->aRequestOptions = $this->oRouteConfig->getRequestParameters();
    // création d'une réponse
    $this->oResponse = $this->generateBasicResponse();
  }
  /**
   * génére un objet réponse symfony et le préremplis avec les headers CORS.
   * @return Response Reponse symfony
   */
  private static function generateBasicResponse(): Response {
    $oResponse = new Response();
    $oResponse->headers->set('Access-Control-Allow-Origin', '*');
    $oResponse->headers->set('Access-Control-Allow-Credentials', 'true');
    $oResponse->headers->set('Access-Control-Allow-Methods', implode(', ', self::HTTP_METHODS_ALLOWED));
    $oResponse->headers->set('Access-Control-Allow-Headers', implode(', ', self::HTTP_RESPONSE_ALLOWED_HEADERS));
    return $oResponse;
  }

  /**
   * Retourne la configuration pour la route utilisée.
   * @throws InvalidArgumentException quand un des services utilisés est manquant
   * @return array|null array contenant la configuration de la route ou null en cas d'erreur
   */
  protected function getRequestParameters(): ?array{
    return $this->refreshRequestParameters();
  }

  /**
   * Rafraichie  et retourne la configuration pour la route utilisée.
   * @return array|null array contenant la configuration de la route ou null en cas d'erreur
   */
  protected function refreshRequestParameters(bool $bForce = false): ?array{
    if (!$this->oRouteConfig->configurationCanChangeWithRole() || $bForce) {
      return $this->aRequestOptions;
    }

    $aUser = $this->getCurrentUserEntity();

    if ($aUser === null) {
      return null;
    }

    // surcharge les options pour l'utilisateur si nécessaire
    $this->aRequestOptions = $this->oRouteConfig->reloadConfigurationForUser($aUser->getRoles());

    return $this->aRequestOptions;
  }

  /**
   * Retourne un objet implémentant l'interface DataAccessor qui utilisera un EntityManager Doctrine.
   * Si le DataAccessor n'est pas chargé il est chargé et stocké.
   * @param EntityManager $oEntityManager EntityManager pour accéder aux données
   * @throws InvalidArgumentException quand un des services utilisés est manquant
   * @return AbstractDataAccessor|null
   */
  protected function getDataAccessor(?EntityManager $oEntityManager = null): ?AbstractDataAccessor{
    $this->oDataAccessor = (!empty($this->oDataAccessor)) ? $this->oDataAccessor : $this->oDatabaseConnection->initializeDataAccessor($oEntityManager);
    return $this->oDataAccessor;
  }

  /**
   * Serialise la sortie en fonction du header accept (json|xml).
   * @param object|array $aData Donnée à serialiser, entité doctrine ou tableau associatif
   * @param integer $iCode Code HTTP à retourner, (defaut : 200) optionnel
   * @throws InvalidArgumentException quand un des services utilisés est manquant
   * @return Response réponse à retourner par l'API
   */
  protected function formatResponse($aData, int $iCode = Response::HTTP_OK, array $aSerializerContext = []): Response {
    $sContentType = $this->oVitisRequest->getRequest()->headers->get('accept');

    switch ($sContentType) {
      case 'application/xml':
        $sSerializerFormat = 'xml';
        break;
      case 'application/json':
      case 'application/x-vm-json':
      // par défaut l'application retourne du json
      default:
        $sSerializerFormat = 'json';
    }

    $this->oResponse->headers->set('Content-Type', 'application/' . $sSerializerFormat);
    $this->oResponse->setStatusCode($iCode);
    $this->oSerializer->setSerializerFormat($sSerializerFormat);
    $this->oResponse->setContent($this->oSerializer->serialize($aData, $aSerializerContext));

    return $this->oResponse;
  }

  /**
   * Fromat une erreur de manière générique
   *
   * @param  string   $sClientMessage Code de traduction de vitis (message)
   * @param  string   $sClientTitle Code de traduction de vitis (titre)
   * @param  string   $sMessage Message à afficher via l'API (et dans le logger d'api)
   * @param  int   $iHttpCode Code HTTP pour l'erreur voir la classe response de symfony
   * @return Response
   */
  protected function formatGenericError(string $sClientMessage, string $sClientTitle = '', string $sMessage = '', int $iHttpCode = Response::HTTP_INTERNAL_SERVER_ERROR): Response {
    
    $aReturn = [
      'status' => 0,
      'errorMessage' => $sClientMessage,
    ];

    if (!empty($sMessage)) {
      $this->oLoggerManager->get('API_ERROR')->error($sMessage);
      $aReturn['message'] = $sMessage;
    }

    if (!empty($sClientTitle)) {
      $aReturn['errorTitle'] = $sClientTitle;
    }
    return $this->formatResponse($aReturn, $iHttpCode);
  }

  /**
   * Extrait les paramètres de la requête.
   * @param bool $bNoDataInBody Permet de bypass la récupération du body
   * @throws InvalidArgumentException quand un des services utilisés est manquant
   * @return array Liste des paramètres de la requête
   */
  protected function requestToArray(?bool $bNoDataInBody = false): array{
    return $this->oVitisRequest->getData($bNoDataInBody);
  }

  /**
   * Extrait les paramètres de la requête et les passe en Json.
   * @return string Liste des paramètres de la requête dans un Json
   */
  protected function requestToJson(): string {
    return json_encode($this->requestToArray()); // Text Json Brut
  }

  /**
   * Récupère l'utilisateur en cours
   * @return User utilisateur en cours
   */
  protected function getCurrentUserEntity(string $sToken = ''): ?User{
    // si le token est mal formé ou si l'utilisateur n'existe pas
    $aUser = $this->oDatabaseConnection->getCurrentUser($sToken);
    if (!empty($aUser)) {
      return $aUser;
    } else {
      return null;
    }
  }

  /**
   * Serialise la sortie en fonction du header accept (json|xml).
   * $this->oResponse sera pré rempli
   * @return bool true si l'utilisateur peut se connecter, false sinon
   */
  protected function userCanUseRoute(): bool {

    if (!$this->oRouteConfig->checkUser()) {
      return true;
    }

    $sToken = $this->oVitisRequest->getToken();

    // s'il n'y a pas de header authorization
    if (empty($sToken)) {
      $this->formatGenericError('ERROR_MISSING_AUTHORIZATION_HEADER', '',
        'Authorization header is missing', Response::HTTP_UNAUTHORIZED);
      return false;
    }

    // Utilisateur en cours
    $aUser = $this->getCurrentUserEntity($sToken);
    if (empty($aUser)) {
      // session expiré ou structure du payload non conforme
      $this->formatGenericErrorSessionExpired();
      return false;
    }

    // controle de la restriction d'IP
    $sUserIpConstraint = $aUser->getIpConstraint();
    if (!$this->checkIpConstraint($sUserIpConstraint)) {
      $sClientIp = $this->oVitisRequest->getClientIp();
      // L'utilisateur ne peut pas se connecter avec cette ip
      $this->formatGenericError('ERROR_USER_IP_UNAUTHORIZED', '',
        'Connection forbidden from ip :' . $sClientIp, Response::HTTP_UNAUTHORIZED);
      return false;
    }

    // Donne accès si le role minimum n'est pas definie
    $roleAccess = $this->oRouteConfig->getRoleAccess();
    if ($roleAccess === null) {
      return true;
    }

    // Roles de l'utilisateur
    $aRoles = $aUser->getRoles();
    if (count(array_diff($roleAccess, $aRoles)) == count($roleAccess)) {
      $this->formatGenericError('ERROR_INSUFFICIENT_PRIVILEGE', '',
        'Insufficient privilege', Response::HTTP_UNAUTHORIZED);
      return false;
    }

    return true;
  }

  /**
   * controle l'ip du client et l'ip provenant de l'entité User
   * @param string $sConstraint contrainte provenant de l'entité User
   * @return bool  true si l'utilisateur peut se connecter, false sinon
   */
  protected function checkIpConstraint(?string $sConstraint): bool {
    // si l'utilisateur n'a pas de contrainte défini on retourne true
    if (empty($sConstraint)) {
      return true;
    }
    // récupération de l'ip du client
    $sIpaddress = $this->oVitisRequest->getClientIp();

    // Récupération des constantes pour générer la regexp
    $sStarReplacer = self::IPV4_STAR_REPLACER;
    if (strpos($sConstraint, self::IPV6_DELIMITER)) {
      $sStarReplacer = self::IPV6_STAR_REPLACER;
    }

    $sConstraint = str_replace('*', $sStarReplacer, $sConstraint);
    // gestion des point en deux temps
    // - on passe d'abord les points protégés en point non protégés
    // - puis on repasse tous les points en protégés
    $sConstraint = str_replace(self::IPV4_DOT_REPLACER, '.', $sConstraint);
    $sConstraint = str_replace('.', self::IPV4_DOT_REPLACER, $sConstraint);
    // concaténation de la regexp, on élimine les / en bordure si l'utilisateur en a utilisé
    $sRegExp = '/' . trim($sConstraint, '/') . '/';

    return (preg_match($sRegExp, $sIpaddress) === 1);
  }

  /**
   * Retourne l'interface fichier à utiliser pour interagir avec des fichiers.
   * @return FileAccessorInterface Objet implémentant l'interface FileAccessor
   */
  protected function getFileAccessor(?string $sPath = null): FileAccessorInterface {

    if (!empty($sPath)) {
      return $this->oFileManager->getAccessorForPath($sPath);
    }

    $this->oFileInterface = (!empty($this->oFileInterface)) ? $this->oFileInterface : $this->oFileManager->getFileAccessor();
    return $this->oFileInterface;
  }

  /**
   * Retourne l'interface fichier à utiliser pour interagir avec des fichiers localement.
   * @return LocalAccessor Objet LocalAccessor implémentant l'interface FileAccessor
   */
  protected function getLocalFileAccessor(): LocalAccessor {
    $this->oLocalFileInterface = (!empty($this->oLocalFileInterface)) ? $this->oLocalFileInterface : $this->oFileManager->getLocalFileAccessor();
    return $this->oLocalFileInterface;
  }

  /**
   * Retourne une réponse de type blob
   * @param string $sFilePath Chemin vers le fichier à retourner
   * @return Response|BinaryFileResponse Objet réponse si le fihcier existe, réponse d'erreur sinon
   */
  protected function getBlobResponse(string $sFilePath, ?string $sFileName = null, array $aHeaders = [], bool $bDeleteAfterSend = false, string $sDisposition = ResponseHeaderBag::DISPOSITION_ATTACHMENT): ?Response{
    $oLocalAccessor = $this->getLocalFileAccessor();
    // si le fichier n'existe pas en local on stop tout
    if (!$oLocalAccessor->file_exists($sFilePath)) {
      return $this->formatGenericErrorFileNotFound($sFilePath);
    }
    // Si un nom de fichier n'est pas défini on récupère le nom du fichier dans le chemin
    $sFileName = (!empty($sFileName)) ? $sFileName : $oLocalAccessor->getFileName($sFilePath);

    $oResponseBlob = new BinaryFileResponse($sFilePath);
    $oResponseBlob->deleteFileAfterSend($bDeleteAfterSend);
    $oResponseBlob->headers->set('Access-Control-Allow-Origin', '*');
    $oResponseBlob->headers->set('Access-Control-Expose-Headers', 'Content-disposition');
    // On ajoute les headers défini en paramètre
    foreach ($aHeaders as $sHeaderKey => $mHeaderValue) {
      $oResponseBlob->headers->set($sHeaderKey, $mHeaderValue);
    }

    $sContentType = $oResponseBlob->headers->get('Content-Type');
    // si pas de content type on l'ajoute automatiquement
    if (empty($sContentType)) {
      $sContentType = $oLocalAccessor->mime_content_type($sFilePath);
      $oResponseBlob->headers->set('Content-Type', $sContentType);
    }

    // si content-type = application/octet-stream et si transfer-encoding non set on le set automatiquement
    if ($sContentType === "application/octet-stream" && empty($oResponseBlob->headers->get('Content-Transfer-Encoding'))) {
      $oResponseBlob->headers->set('Content-Transfer-Encoding', 'Binary');
    }

    $oResponseBlob->setContentDisposition($sDisposition, $sFileName, mb_convert_encoding($sFileName, 'ASCII'));

    return $oResponseBlob;
  }

  /**
   * Retourne une erreur générique pour un fichier non trouvé
   *
   * @param string $sPath chemin non trouvé
   * @return Response
   */
  protected function formatGenericErrorFileNotFound(string $sPath = ''): Response {
    return $this->formatGenericError('DOWNLOADER_ERROR_FILE_NOT_FOUND', '',
      'Cannot find this file : ' . $sPath, Response::HTTP_NOT_FOUND);
  }
  /**
   * Retourne une erreur classique de session expirée
   *
   * @return Response
   */
  protected function formatGenericErrorSessionExpired(): Response {
    return $this->formatGenericError('ERROR_SESSION_EXPIRED', 'ERROR_SESSION_EXPIRED_TITLE',
      'Session expired', Response::HTTP_UNAUTHORIZED);
  }

  /**
   * Retourne une erreur générique pour dire que la fonctionnalité n'est pas active
   *
   * @return Response
   */
  protected function formatGenericFeatureNotEnabled(): Response {
    return $this->formatGenericError('FEATURE_NOT_ENABLED', '',
      'This feature is not enabled for this application', Response::HTTP_UNAUTHORIZED);
  }

  /**
   * erreur générique sur les insertions
   * @param string $sMessage
   * @return Response
   */
  protected function formatGenericResponseForErrorOnSelect(string $sMessage = 'Error during get'): Response {
    return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '', $sMessage, Response::HTTP_UNAUTHORIZED);
  }

  /**
   * erreur générique sur les insertions
   * @param string $sMessage
   * @return Response
   */
  protected function formatGenericResponseForErrorOnInsert(string $sMessage = 'Error during insert'): Response {
    return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_INSERT', '', $sMessage, Response::HTTP_UNAUTHORIZED);
  }

  /**
   * erreur générique sur les insertions
   * @param string $sMessage
   * @return Response
   */
  protected function formatGenericResponseForErrorOnUpdate(string $sMessage = 'Error during update'): Response {
    return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_UPDATE', '', $sMessage, Response::HTTP_UNAUTHORIZED);
  }

  /**
   * erreur générique sur les suppressions
   * @param string $sMessage
   * @return Response
   */
  protected function formatGenericResponseForErrorOnDelete(string $sMessage = 'Error during delete'): Response {
    return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_DELETES', '', $sMessage, Response::HTTP_UNAUTHORIZED);

  }

  /**
   * retourne une réponse standard pour l'absence du paramètre idlist
   * @param string $sMessage
   * @return Response
   */
  protected function formatGenericResponseForMissingIdList(string $sMessage = 'idList parameter is missing'): Response {
    return $this->formatGenericError('GENERIC_ERROR_MISSING_IDLIST_PARAMETER', '', $sMessage, Response::HTTP_NOT_ACCEPTABLE);
  }

  /**
   * @return Response
   */
  protected function formatPostgresBadConnection(): Response {
    return $this->formatGenericError('ERROR_POSTGRES_USER_BAD_CONNECTION', '',
      'Cannot access to database', Response::HTTP_UNAUTHORIZED);
  }

  /**
   * @param int $iHttpCode
   * @param string $sMessage
   * @return Response
   */
  protected function formatGenericErrorRequestError(string $sMessage = '', int $iHttpCode = Response::HTTP_INTERNAL_SERVER_ERROR): Response {
    return $this->formatGenericError('REQUEST_ERROR', '', $sMessage, $iHttpCode);
  }

  /**
   * @param string $sMessage
   * @return Response
   */
  protected function formatGenericErrorCannotSendMail(string $sMessage = ''): Response {
    return $this->formatGenericError('ERROR_EMAIL_CANT_SEND', '', $sMessage, Response::HTTP_INTERNAL_SERVER_ERROR);
  }


  /**
   * @param  string   $sMessage
   * @return Response
   */
  protected function formatErrorMissingParameter(string $sMessage = 'Parameters are missing'): Response {
    return $this->formatGenericError('DOWNLOADER_ERROR_MISSING_PARAMETERS', '', $sMessage, Response::HTTP_NOT_ACCEPTABLE);
  }
}
