<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use App\Module\Vitis\WebService\Service\Configuration\Module;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Mail\SMTPConnexion;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller qui permet de récupérer des informations sur le serveur et l'application.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Armand Bahi <armand.bahi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */
class ApplicationData extends AbstractVitisController {
  /**
   * Map à utiliser pour lancer des actions spécifiques si une properties est modifiée
   * Les fonctions à appeler devront toujours ne pas avoir de paramètre et retourner un booléen
   * @var array
   */
  protected const CUSTOM_ACTION_FOR_PROPERTIES = [];

  /**
   * Récupère le phpinfos du serveur applicatif
   * @return Response réponse à retourner par l'API
   */
  public function getPhpInfos(): Response {
    // preset de la structure de sortie
    $aReturn = ['status' => 1];
    // on bloque le flux de sortie
    ob_start();
    // le contenu de phpinfo est envoyé vers le flux de sortie
    phpinfo();
    // on récupère le contenu du flux de sortie
    $aReturn['phpinfo'] = ob_get_contents();
    // on nettoie le flux de sortie et on arrête le blocage
    ob_end_clean();
    // on retourne le tout
    $this->formatResponse($aReturn);

    return $this->oResponse;
  }

  /**
   * Retourne les properties à l'utilisateur
   * @return Response réponse à retourner par l'API
   */
  public function getProperties(): Response {
    // réponse par défaut
    $aReturn = [
      'data' => [],
      'status' => 0,
    ];
    // Role par défaut
    $aRoles = ['public'];

    // Roles si l'utilisateur est connecté
    $oUser = $this->getCurrentUserEntity();
    if ($oUser !== null) {
      $aRoles = array_merge($aRoles, $oUser->getRoles());
    }

    // Récupère les properties de l'utlisateur
    $aProperties = [];
    foreach ($aRoles as $role) {
      $aProperties = array_merge($aProperties, $this->oProperties->getProperties($role));
    }

    // Récupère les themes
    $oEm = $this->oDatabaseConnection->getAdminEntityManager();
    $qb = $this->oDatabaseConnection->getQueryAccessor($oEm)->createQueryBuilder();

    $aParams = [
      'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
    ];

    try {
      $qb->select('a.*')
        ->from('[SCHEMA].theme', 'a');

      $oThemes = $qb->executeWithParams($aParams)->fetchAll();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Theme : can\'t get theme: ' . $e->getMessage());
      $aReturn['message'] = 'can\'t get theme';
      return $this->formatResponse($aReturn, Response::HTTP_NOT_ACCEPTABLE);
    }

    try {
      // Récupère le css custom
      $qb->select('a.custom_css')
        ->from('[SCHEMA].custom_css', 'a')
        ->where($qb->expr()->eq('a.custom_css_id', '[ID]'));

      $aParams['ID'] = ['value' => 1, 'type' => 'integer'];

      $oCss = $qb->executeWithParams($aParams)->fetchAll();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Custom css : can\'t get Custom css: ' . $e->getMessage());
      $aReturn['message'] = 'can\'t get Custom css';
      return $this->formatResponse($aReturn, Response::HTTP_NOT_ACCEPTABLE);
    }

    $aReturn['data'] = $aProperties;
    $aReturn['status'] = 1;

    if (!empty($oThemes)) {
      $aReturn['data']['themes'] = $oThemes;
    }

    if (!empty($oCss)) {
      $aReturn['data']['custom_css'] = $oCss[0]['custom_css'];
    }

    return $this->formatResponse($aReturn);
  }

  /**
   * Met à jour les properties de l'utilisateur
   * @return Response réponse à retourner par l'API
   */
  public function setProperties(): Response {
    // Récuperation des données passées en paramètre
    $aData = $this->requestToArray();
    // Récupération du module dans la route
    $aRequestOptions = $this->oRouteConfig->getRequestParameters();
    $sModule = $aRequestOptions['options']['module_name'];

    if (!isset($sModule) || !isset($aRequestOptions['role_minimum'])) {
      return $this->formatResponse([
        'message' => 'empty request\'s body or module_name or role_minimum doesn\'t exist in route\'s configuration',
        'status' => 0,
      ], Response::HTTP_NOT_ACCEPTABLE);
    }

    $aCustomsActions = $this->hasToApplyCustomsMethods($aData);

    // Ecriture des données dans un fichier json
    if ($this->oProperties->dumpProperties($sModule, $aData, $aRequestOptions['role_minimum'])) {
      $iStatus = 1;
      if (!empty($aCustomsActions)) {
        foreach ($aCustomsActions as $sMethodName) {
          $iStatus = ($this->$sMethodName()) ? $iStatus : 0;
        }
      }
      $this->formatResponse([
        'status' => $iStatus,
      ]);
    } else {
      $this->formatResponse([
        'message' => 'Cannot update Properties',
        'status' => 0,
      ], Response::HTTP_BAD_REQUEST);
    }

    return $this->oResponse;
  }

  /**
   * Analyse les données à mettre à jour avec l'existant pour savoir s'il faut appliquer des actions spécifique
   * @param array $aData donnée arrivant de la requête
   * @return array liste des méthode à appliquer
   */
  private function hasToApplyCustomsMethods(array $aData): array{
    $aAction = [];

    foreach (static::CUSTOM_ACTION_FOR_PROPERTIES as $sKey => $sMethod) {
      if (!isset($aData[$sKey])) {
        continue;
      }

      if ($aData[$sKey] === $this->oProperties->get($sKey)) {
        continue;
      }

      $aAction[] = $sMethod;
    }

    return array_unique($aAction);
  }

  /**
   * permet de tester une connexion SMTP soit via les properties, soit le body de la requête
   * @return Response
   */
  public function testSmtpConnexion(): Response {
    try {
      // Récuperation des données passées dans la requête
      $aData = $this->requestToArray();
      // Instanciation la connexion SMTP et utilisation des paramètre de la requête
      $oSMTPConnexion = new SMTPConnexion($this->oProperties, $this->oLoggerManager);
      // on retourne le résultat du test
      return $this->formatResponse([
        "status" => ($oSMTPConnexion->setConfiguration($aData)->testConnection()) ? 1 : 0,
      ]);
    } catch (\Exception $e) {
      return $this->formatResponse([
        "status" => 0,
      ]);
    }
  }
}
