<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Entity\Domain;
use App\Module\Vitis\WebService\Entity\DomainGroup;
use App\Module\Vitis\WebService\Entity\Group;
use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Repository\DomainRepository;
use Doctrine\ORM\EntityManager;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Ldap\Adapter\CollectionInterface;
use Symfony\Component\Ldap\Entry;
use Symfony\Component\Ldap\Ldap;

/**
 * Controller Générique de Vitis.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */
class LdapController extends GenericController {
  /**
   * @var string []
   */
  protected const AUTHORIZED_KEYS_FOR_GROUPS = [
    'cn',
    'name',
    'mail',
    'member',
    'memberUid',
    'objectSid',
    'objectGUID',
    'description',
    'objectClass',
    'objectCategory',
    'sAMAccountName',
  ];
  /**
   * @var string []
   */
  protected const AUTHORIZED_KEYS_FOR_USERS = [
    'cn',
    'mail',
    'memberOf',
    'objectSid',
    'objectGUID',
    'description',
    'displayName',
    'sAMAccountName',
    'userPrincipalName',
  ];

  /**
   * @var EntityManager
   */
  protected $oAdminEntityManager;

  /**
   * @var EntityManager
   */
  protected $oUserEntityManager;

  /**
   * Créé un domaine
   * @return Response réponse à retourner par l'API
   */
  public function createDomain(Request $oRequest): Response {
    // Initialisation des valeurs par défaut
    $aData = $this->requestToArray();
    if (!isset($aData['uid'])) {
      switch ($aData['type']) {
        case 'LDAP':
          $aData['uid'] = 'sAMAccountName';
          break;
        case 'AD':
        default:
          $aData['uid'] = 'userPrincipalName';
          break;
      }
      $oRequest->request->set('uid', $aData['uid']);
    }
    if (!isset($aData['displayname'])) {
      switch ($aData['type']) {
        case 'LDAP':
          $aData['displayname'] = 'cn';
          break;
        case 'AD':
        default:
          $aData['displayname'] = 'displayname';
          break;
      }
      $oRequest->request->set('displayname', $aData['displayname']);
    }
    if (!isset($aData['mail'])) {
      $aData['mail'] = 'mail';
    }
    // Création du domaine
    $sData = $this->requestToJson();
    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    // Sauvegarde en base de donnée
    $oEntity = $oDataAccessor->insertEntity($sData);
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnInsert();
    }
    // creation d'un groupe pour les utilisateurs du domaine
    $oEntity['data']->setLoginRole('vitis_domain_' . $oEntity['data']->getDomainId());
    // Récuperation d'UserEntityManager
    $this->oUserEntityManager = $this->oDatabaseConnection->getUserEntityManager();
    $this->oUserEntityManager->persist($oEntity['data']);
    $this->oUserEntityManager->flush($oEntity['data']);
    if (!$oQueryAccessor->ifPgRoleExist($oEntity['data']->getLoginRole())) {
      $oQueryAccessor->createRole($oEntity['data']->getLoginRole());
    }
    $this->formatResponse($oEntity);
    return $this->oResponse;
  }

  /**
   * Supprime des domaines
   * @return Response réponse à retourner par l'API
   */
  public function deleteDomains(): Response {
    $aData = $this->massiveRequestGetIdList();
    if (empty($aData)) {
      return $this->formatGenericResponseForMissingIdList();
    }
    // Chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    // Récupération d'UserEntityManager
    $this->oUserEntityManager = $this->oDatabaseConnection->getUserEntityManager();
    /* @var GenericRepository $domainRepository */
    $domainRepository = $this->oUserEntityManager->getRepository(Domain::class);
    foreach ($aData['idList'] as $iDomainId) {
      /* @var $oDomain Domain */
      $oDomain = $domainRepository->find($iDomainId);
      if ($oDomain->getLoginRole() != null) {
        $oQueryAccessor->dropRole($oDomain->getLoginRole());
      }
    }
    // Sauvegarde en base de donnée
    $aEntities = $oDataAccessor->deleteEntities($aData['idList']);
    if ($aEntities['status'] == 0) {
      return $this->formatGenericResponseForErrorOnDelete();
    }
    return $this->formatResponse($aEntities);
  }

  /**
   * Récupère les groupes d'un domaine
   * @param mixed $id id de domaine
   * @return Response réponse à retourner par l'API
   */
  public function getDomainGroups($id): Response {
    // Récupération d'UserEntityManager
    $this->oUserEntityManager = $this->oDatabaseConnection->getUserEntityManager();

    /* @var GenericRepository $domainRepository */
    $domainRepository = $this->oUserEntityManager->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    // Erreur si domaine n'existe pas
    if (!isset($oDomain)) {
      return $this->formatErrorDomainNotFound();
    }
    $sSearch = $this->getLdapGroupFilter($oDomain);
    $sDnSearchUser = $oDomain->getDnSearchUser() ?? '';
    try {
      $ldap = Ldap::create('ext_ldap', $oDomain->getLdapConfig());
      $ldap->bind($oDomain->getLogin(), $oDomain->getPassword());
      $query = $ldap->query($sDnSearchUser, $sSearch, $oDomain->getLdapQueryOptions());
      $results = $query->execute();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot connect to LDAP/AD server : ' . $e->getMessage(), [$oDomain->getLdapConfig()]);
      return $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
    }
    $aAuthorizedKeys = array_unique(self::AUTHORIZED_KEYS_FOR_GROUPS);
    // Transforme le résultat en array
    $aData = $this->decodeEntries($results, $aAuthorizedKeys);
    // Gestion de la pagination
    $aRequest = $this->requestToArray();
    $iLimit = isset($aRequest['limit']) ? $aRequest['limit'] : 20;
    $iOffset = isset($aRequest['offset']) ? $aRequest['offset'] : 0;
    // Détermine le nombre de ligne à retourner
    $iListCount = intval($iOffset + $iLimit);
    if ($iListCount > count($aData)) {
      $iListCount = count($aData);
    }
    // Prépare la réponse
    $aResponse = [
      'data' => [],
      'total_row_number' => count($aData),
      'list_count' => $iListCount,
      'status' => 1,
    ];
    // Filtrage et ajout de 'domain_id'
    for ($i = $iOffset; $i < $iListCount; $i++) {
      $aResponse['data'][] = $oDomain->convertEntriesDataForGroup($aData[$i]);
    }
    $this->formatResponse($aResponse);
    return $this->oResponse;
  }

  /**
   * Récupère les utilisateur d'un domaine
   * @param mixed $id id de domaine
   * @return Response réponse à retourner par l'API
   */
  public function getDomainUsers($id): Response {
    // Récupération d'UserEntityManager
    $this->oUserEntityManager = $this->oDatabaseConnection->getUserEntityManager();

    /* @var GenericRepository $domainRepository */
    $domainRepository = $this->oUserEntityManager->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    // Erreur si domaine n'existe pas
    if (!isset($oDomain)) {
      return $this->formatErrorDomainNotFound();
    }
    $sSearch = $this->getLdapUserFilter($oDomain);
    $sDnSearchUser = $oDomain->getDnSearchUser() ?? '';
    try {
      $ldap = Ldap::create('ext_ldap', $oDomain->getLdapConfig());
      $ldap->bind($oDomain->getLogin(), $oDomain->getPassword());
      $query = $ldap->query($sDnSearchUser, $sSearch, $oDomain->getLdapQueryOptions());
      $results = $query->execute()->toArray();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot connect to LDAP/AD server : ' . $e->getMessage(), [$oDomain->getLdapConfig()]);
      return $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
    }
    $aAuthorizedKeys = array_merge(array_unique(self::AUTHORIZED_KEYS_FOR_USERS), [
      $oDomain->getUid(),
      $oDomain->getDisplayName(),
      $oDomain->getMail(),
    ]);

    // Transforme le résultat en array
    $aData = $this->decodeEntries($results, $aAuthorizedKeys);
    // Gestion de la pagination
    $aRequest = $this->requestToArray();
    $iLimit = isset($aRequest['limit']) ? $aRequest['limit'] : 20;
    $iOffset = isset($aRequest['offset']) ? $aRequest['offset'] : 0;
    $iListCount = intval($iOffset + $iLimit);
    if ($iListCount > count($aData)) {
      $iListCount = count($aData);
    }
    $aResponse = [
      'data' => [],
      'total_row_number' => count($aData),
      'list_count' => $iListCount,
      'status' => 1,
    ];
    // Filtrage et ajout de 'domain_id'
    for ($i = $iOffset; $i < $iListCount; $i++) {
      $aResponse['data'][] = $oDomain->convertEntriesDataForUser($aData[$i]);
    }
    $this->formatResponse($aResponse);
    return $this->oResponse;
  }

  /**
   * Récupère la liste des suffixes
   * @return Response réponse à retourner par l'API
   */
  public function getSuffixes(): Response {
    // Récupération d'UserEntityManager
    $this->oAdminEntityManager = $this->oDatabaseConnection->getAdminEntityManager();

    $domainRepository = $this->oAdminEntityManager->getRepository(Domain::class);
    /* @var $domainRepository \App\Module\Vitis\WebService\Repository\DomainRepository */

    if (!$domainRepository instanceof DomainRepository) {
      $aReturn = [
        'status' => 0,
        'message' => 'Cannot find Domain suffixes, bad repository',
      ];
      return $this->formatResponse($aReturn, Response::HTTP_NOT_FOUND);

    }

    $aSuffixes = $domainRepository->getSuffixes();

    $aResponse = [
      'data' => $aSuffixes,
      'total_row_number' => count($aSuffixes),
      'status' => 1,
    ];
    return $this->formatResponse($aResponse);
  }

  /**
   * Importe des groups d'un domaine
   * @param mixed $id id du Domaine
   * @return Response réponse à retourner par l'API
   */
  public function importGroups($id): Response {
    // récupération des groupes à importer
    $aData = $this->requestToArray();
    // get auto_refresh
    $bAutoRefresh = $aData['auto_refresh'] ?? false;
    // Nb de groupes à importer
    $iLength = isset($aData['length']) ? intval($aData['length']) : 0;
    if ($iLength === 0) {
      return $this->formatErrorListNoSelection();
    }
    // Récupération d'UserEntityManager
    $this->oUserEntityManager = $this->oDatabaseConnection->getUserEntityManager();
    /* @var $domainRepository GenericRepository */
    $domainRepository = $this->oUserEntityManager->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    if (!isset($oDomain)) {
      return $this->formatErrorDomainNotFound();
    }
    $aDomainGroups = [];
    for ($i = 0; $i < $iLength; $i++) {
      if (isset($aData[$i])) {
        $oDomainGroup = $this->importDomainGroup($aData[$i], $oDomain, $bAutoRefresh);
        if (isset($oDomainGroup)) {
          $aDomainGroups[] = $oDomainGroup->toArray();
        }
      }
    }
    $aResponse = [
      'data' => $aDomainGroups,
      'list_count' => count($aDomainGroups),
      'total_row_number' => count($aDomainGroups) > $iLength ? count($aDomainGroups) : $iLength,
      'status' => 1,
    ];
    $this->formatResponse($aResponse);
    return $this->oResponse;
  }

  /**
   * Importe un groupe d'un domain
   * @param array $aDomainGroup
   * @param Domain $oDomain
   * @param bool $bAutoRefresh
   * @return DomainGroup|null
   */
  private function importDomainGroup($aDomainGroup, $oDomain, $bAutoRefresh = false): ?DomainGroup{
    /* @var $GroupRepository GroupRepository */
    $GroupRepository = $this->oUserEntityManager->getRepository(Group::class);
    // get domain name
    $sName = $aDomainGroup['displayname'];
    // Check si Group exist
    $oGroup = $GroupRepository->findOneBy([
      'name' => $sName,
    ]);
    if (!isset($oGroup)) {
      // création du Group
      $oGroup = new Group();
      $oGroup->setName($sName);
      // sauvegarde dans la base de données
      $this->oUserEntityManager->persist($oGroup);
      $this->oUserEntityManager->flush();
      $this->oUserEntityManager->refresh($oGroup);
    }
    $oDomainGroup = $oGroup->getDomainGroup();
    // récupération ou creation de groupe s'il n'existe pas
    if (!isset($oDomainGroup)) {
      // creation de DomainGroup
      $oDomainGroup = new DomainGroup();
    }
    // MAJ de 'dn'
    $oDomainGroup->setDomain($oDomain);
    $oDomainGroup->setDn($aDomainGroup['dn']);
    $oDomainGroup->setVitisGroup($oGroup);
    $oDomainGroup->setAutoRefresh($bAutoRefresh);
    $this->oUserEntityManager->persist($oDomainGroup);
    $this->oUserEntityManager->flush();
    $this->oUserEntityManager->refresh($oDomainGroup);
    return $oDomainGroup;
  }

  /**
   * Synchronise des groupes de domaine avec un groupe Vitis
   * @return Response réponse à retourner par l'API
   */
  public function autoSyncGroups(): Response {
    // Récupération d'EntityManager
    $this->oAdminEntityManager = $this->oDatabaseConnection->getAdminEntityManager();
    /* @var $domainGroupRepository GenericRepository */
    $domainGroupRepository = $this->oAdminEntityManager->getRepository(DomainGroup::class);

    $aDomainGroups = $domainGroupRepository->findBy([
      'autoRefresh' => true,
    ]);
    $aUsers = [];
    foreach ($aDomainGroups as $oDomainGroup) {
      $aResult = $this->getDomainGroupsUsers($oDomainGroup->toArray(), $oDomainGroup->getDomain());
      // S'il a eu une erreur retourne la
      if (!isset($aResult)) {
        return $this->oResponse;
      }
      foreach ($aResult as $sKey => $aUser) {
        $aUsers[$sKey] = $aUser;
      }
    }

    $this->formatResponse([
      'data' => $aUsers,
      'total_row_number' => count($aUsers),
      'status' => 1,
    ]);
    return $this->oResponse;
  }

  /**
   * Synchronise un groupe de domaine avec un groupe Vitis
   * @param mixed $id id du Domaine
   * @return Response réponse à retourner par l'API
   */
  public function syncGroups($id): Response {
    // récupération des groupes à importer
    $aData = $this->requestToArray();
    // Nb de groupes à importer
    $iLength = isset($aData['length']) ? intval($aData['length']) : 0;
    if ($iLength === 0) {
      return $this->formatErrorListNoSelection();
    }
    // Récupération d'UserEntityManager
    $this->oUserEntityManager = $this->oDatabaseConnection->getUserEntityManager();
    /* @var $domainRepository GenericRepository */
    $domainRepository = $this->oUserEntityManager->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    if (!isset($oDomain)) {
      return $this->formatErrorDomainNotFound();
    }
    $aUsers = [];
    for ($i = 0; $i < $iLength; $i++) {
      if (isset($aData[$i])) {
        $aResult = $this->getDomainGroupsUsers($aData[$i], $oDomain);
        // S'il a eu une erreur retourne la
        if (!isset($aResult)) {
          return $this->oResponse;
        }
        foreach ($aResult as $sKey => $aUser) {
          $aUsers[$sKey] = $aUser;
        }
      }
    }
    $this->formatResponse([
      'data' => $aUsers,
      'list_count' => count($aUsers),
      'total_row_number' => count($aUsers),
      'status' => 1,
    ]);
    return $this->oResponse;
  }

  /**
   * Rétourne une liste des utilisateur d'un groupe de domain
   * @param array $aDomainGroup
   * @param Domain $oDomain
   * @return array|null
   */
  private function getDomainGroupsUsers($aDomainGroup, $oDomain): ?array{
    // chargement de la connexion à la base de donnée
    $oEntityManager = $this->oUserEntityManager ?? $this->oAdminEntityManager;
    /* @var $GroupRepository GroupRepository */
    $GroupRepository = $oEntityManager->getRepository(Group::class);
    // get domain name
    $sName = $aDomainGroup['displayname'];
    // Check si Group exist
    $oGroup = $GroupRepository->findOneBy([
      'name' => $sName,
    ]);
    // Si le groupe n'existe pas retourne NULL
    if (!isset($oGroup)) {
      $this->formatGenericResponseForErrorOnSelect('Group "' . $sName . '" not found');
      return null;
    }
    $oDomainGroup = $oGroup->getDomainGroup();
    // S'il n'y pas de domaine associé retourne NULL
    if (!isset($oDomainGroup)) {
      $this->formatGenericResponseForErrorOnSelect('Group "' . $sName . '" not in any domain');
      return null;
    }
    // Si le groupe n'appartiens pas au domaine
    if ($oDomain->getDomainId() != $oDomainGroup->getDomain()->getDomainId()) {
      $this->formatGenericResponseForErrorOnSelect('Group "' . $sName . '" not in domain "' . $oDomain->getName() . '"');
      return null;
    }
    // MAJ de 'dn'
    $oDomainGroup->setDn($aDomainGroup['dn']);
    $oEntityManager->persist($oDomainGroup);
    $oEntityManager->flush();
    $oEntityManager->refresh($oDomainGroup);
    // recherche des sous-éléments de type Group
    $sDnSearch = $oDomainGroup->getDn();
    $sSearch = $this->getLdapGroupFilter($oDomain);
    try {
      $ldap = Ldap::create('ext_ldap', $oDomain->getLdapConfig());
      $ldap->bind($oDomain->getLogin(), $oDomain->getPassword());
      $query = $ldap->query($sDnSearch, $sSearch, $oDomain->getLdapQueryOptions());
      $results = $query->execute();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot connect to LDAP/AD server', [$e]);
      $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
      return null;
    }
    // Les attributes autorisés pour des groupes
    $aAuthorizedKeysGroups = array_unique(self::AUTHORIZED_KEYS_FOR_GROUPS);

    // Filtre le résultat
    $aData = $this->decodeEntries($results, $aAuthorizedKeysGroups);
    // S'il n'y a pas de group
    if (count($aData) == 0) {
      return [];
    }
    // recherche des utilisateurs AD/Ldap
    $aUsers = [];
    // Les attributes autorisés pour des utilisateurs
    $aAuthorizedKeysUsers = array_merge(array_unique(self::AUTHORIZED_KEYS_FOR_USERS), [
      $oDomain->getUid(),
      $oDomain->getDisplayName(),
      $oDomain->getMail(),
    ]);

    foreach ($aData as $aGroup) {
      if (isset($aGroup['member'])) {
        $sSearch = $this->getLdapUserFilter($oDomain);
        foreach ($aGroup['member'] as $sMemberDn) {
          $sDnSearch = $sMemberDn;
          try {
            $query = $ldap->query($sDnSearch, $sSearch, $oDomain->getLdapQueryOptions());
            $results = $query->execute();
          } catch (\Exception $e) {
            $this->oLoggerManager->get('API_ERROR')->error('Search error LDAP/AD server', [$e]);
            $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
            return null;
          }
          // filtre le résultat
          $aData = $this->decodeEntries($results, $aAuthorizedKeysUsers);
          foreach ($aData as $aUser) {
            $aUsers[$aUser[$oDomain->getUid()]] = $aUser;
          }
        }
      }
      if (isset($aGroup['memberUid'])) {
        $sDnSearch = $oDomain->getDnSearchUser() ?? '';
        foreach ($aGroup['memberUid'] as $sMemberUid) {
          $sSearch = $this->getLdapUserFilter($oDomain, 'uid=' . $sMemberUid);
          try {
            $query = $ldap->query($sDnSearch, $sSearch, $oDomain->getLdapQueryOptions());
            $results = $query->execute();
          } catch (\Exception $e) {
            $this->oLoggerManager->get('API_ERROR')->error('Search error to LDAP/AD server', [$e]);
            $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
            return null;
          }
          // filtre le résultat
          $aData = $this->decodeEntries($results, $aAuthorizedKeysUsers);
          foreach ($aData as $aUser) {
            $aUsers[$aUser[$oDomain->getUid()]] = $aUser;
          }
        }
      }
      $sDnSearch = $aGroup['dn'];
      $sSearch = $this->getLdapUserFilter($oDomain);
      try {
        $query = $ldap->query($sDnSearch, $sSearch, $oDomain->getLdapQueryOptions());
        $results = $query->execute();
      } catch (\Exception $e) {
        $this->oLoggerManager->get('API_ERROR')->error('Search error to LDAP/AD server', [$e]);
        $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
        return null;
      }
      // filtre le résultat
      $aData = $this->decodeEntries($results, $aAuthorizedKeysUsers);
      foreach ($aData as $aUser) {
        $aUsers[$aUser[$oDomain->getUid()]] = $aUser;
      }
    }
    foreach ($aUsers as $sKey => $aUser) {
      $oUser = $this->importUser($aUser, $oDomain, $oGroup);
      $aUsers[$sKey]['imported'] = isset($oUser);
    }
    return $aUsers;
  }

  /**
   * Importe des utilisateurs d'un domaine
   * @param mixed $id id du Domaine
   * @return Response réponse à retourner par l'API
   */
  public function importUsers($id): Response {
    // récupération des utilisateurs à importer
    $aData = $this->requestToArray();
    // Nb d'utilisateurs à importer
    $iLength = isset($aData['length']) ? intval($aData['length']) : 0;
    if ($iLength === 0) {
      return $this->formatErrorListNoSelection();
    }
    // Récupération d'UserEntityManager
    $this->oUserEntityManager = $this->oDatabaseConnection->getUserEntityManager();
    /* @var $domainRepository GenericRepository */
    $domainRepository = $this->oUserEntityManager->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    if (!isset($oDomain)) {
      return $this->formatErrorDomainNotFound();
    }
    $aUsers = [];
    for ($i = 0; $i < $iLength; $i++) {
      if (isset($aData[$i])) {
        $oUser = $this->importUser($aData[$i], $oDomain);
        if (isset($oUser)) {
          $aUsers[] = $oUser->getUserId();
        }
      }
    }
    $this->formatResponse([
      'data' => $aUsers,
      'list_count' => count($aUsers),
      'total_row_number' => $iLength,
      'status' => 1,
    ]);
    return $this->oResponse;
  }

  /**
   * Importe un utilisateur d'un domain
   * @param array $aUser
   * @param Domain $oDomain
   * @param Group $oGroup
   * @return User|null
   */
  private function importUser(array $aUser, Domain $oDomain, Group $oGroup = null): ?User{
    // chargement de la connexion à la base de donnée
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    // récupération d'utilisateur courant
    $oCurrentUser = $this->getCurrentUserEntity();
    // récupération du nom d'uid
    $sUID = $oDomain->getUid();
    // set login en minuscule
    $sLogin = strtolower($aUser[$sUID]);
    $oEntityManager = $this->oUserEntityManager ?? $this->oAdminEntityManager;
    /* @var $userRepository UserRepository */
    $userRepository = $oEntityManager->getRepository(User::class);
    // Check si l'utilisateur BDD exist
    $oUser = $userRepository->findOneBy([
      'login' => $sLogin,
    ]);
    // S'il existe arrêt d'import
    if (isset($oUser)) {
      return null;
    } else {
      // création de l'utilisateur dans Postgres
      if (!$oQueryAccessor->createUser($sLogin)) {
        $this->oLoggerManager->get('API_ERROR')->info('Postgres role ' . $sLogin . ' already exists');
      }
      $oUser = new User();
      $oUser->setLogin($sLogin);
    }
    // Ajoute des roles au nouveau utilisateur
    $oQueryAccessor->grantRolesToRole(
      [
        $oDomain->getLoginRole(),
        'vitis_user',
      ],
      $sLogin
    );
    // Ajoute le domaine
    $oUser->setDomain($oDomain);
    // Récupère displayname d'utilisateur
    if ($oDomain->getDisplayName() != null && isset($aUser[$oDomain->getDisplayName()])) {
      $oUser->setName($aUser[$oDomain->getDisplayName()]);
    } else {
      $oUser->setName($aUser['displayname']);
    }
    // Récupère mail d'utilisateur
    if ($oDomain->getMail() != null && isset($aUser[$oDomain->getMail()])) {
      $oUser->setEmail($aUser[$oDomain->getMail()]);
    } else {
      $oUser->setEmail(isset($aUser['mail']) ? $aUser['mail'] : '');
    }
    $oUser->setTimezoneId($oCurrentUser->getTimezoneId());
    $oUser->setFormatdateId($oCurrentUser->getFormatdateId());
    $oEntityManager->persist($oUser);
    $oEntityManager->flush();
    $oEntityManager->refresh($oUser);
    if (isset($oGroup)) {
      $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();
      $qb->insert('[SCHEMA].user_group')
        ->setValue('user_id', '[USER_ID]')
        ->setValue('group_id', '[GROUP_ID]');

      $aParams = [
        'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
        'USER_ID' => ['value' => $oUser->getUserId(), 'type' => 'integer'],
        'GROUP_ID' => ['value' => $oGroup->getGroupId(), 'type' => 'integer'],
      ];
      try {
        $qb->executeWithParams($aParams);
      } catch (\Exception $e) {
        $this->oLoggerManager->get('SQL_ERROR')->error('Cannot associate group to user ' . $e->getMessage());
        return null;
      }
    }
    return $oUser;
  }

  /**
   * Decode Ldap Entry
   *
   * @param \Symfony\Component\Ldap\Adapter\CollectionInterface|array $aEntry
   * @param array $aAuthorizedKeys
   * @return array
   */
  private function decodeEntries(iterable $aEntry, array $aAuthorizedKeys): array{
    $aDecoded = [];
    foreach ($aEntry as $sKey => $sValue) {
      // decode 'objectSid'
      if ($sKey === 'objectSid') {
        $aDecoded[$sKey] = $this->decodeSID($sValue[0]);
        continue;
      }
      // decode 'objectGUID'
      if ($sKey === 'objectGUID') {
        $aDecoded[$sKey] = $this->decodeGUID($sValue[0]);
        continue;
      }
      // decode 'logonHours'
      if ($sKey === 'logonHours') {
        $aDecoded[$sKey] = [];
        $aDecoded[$sKey] = bin2hex($sValue[0]);
        continue;
      }
      // decode des membres d'un groupe
      if ($sKey === 'member' || $sKey === 'memberUid') {
        $aDecoded[$sKey] = $this->decodeEntries($sValue, $aAuthorizedKeys);
        continue;
      }
      if ($sValue instanceof Entry) {
        /* @var Entry $sValue */
        $aDecoded[$sKey] = $this->decodeEntries($sValue->getAttributes(), $aAuthorizedKeys);
        $aDecoded[$sKey]['dn'] = $sValue->getDn();
        // cn=demandeur_ldap2,ou=Groups,dc=ldap,dc=veremes,dc=net
        $sDisplayName = explode('=', explode(',', $aDecoded[$sKey]['dn'])[0])[1];
        $aDecoded[$sKey]['displayname'] = isset($aDecoded[$sKey]['displayname']) ? $aDecoded[$sKey]['displayname'] : $sDisplayName;
        continue;
      }
      // si c'est une liste
      if (is_int($sKey)) {
        $aDecoded[$sKey] = $sValue;
      }
      // Filtre les attributes envoyés par la requête LDAP
      if (!in_array($sKey, $aAuthorizedKeys)) {
        continue;
      }
      if (is_array($sValue) && count($sValue) > 1) {
        $aDecoded[$sKey] = $this->decodeEntries($sValue, $aAuthorizedKeys);
      }
      if (is_array($sValue) && count($sValue) <= 1) {
        $aDecoded[$sKey] = $this->decodeEntries($sValue, $aAuthorizedKeys)[0];
      }
      if (is_string($sValue)) {
        $aDecoded[$sKey] = $sValue;
      }
    }
    return $aDecoded;
  }

  /**
   * Decode binary SID
   *
   * @param string $bSID
   * @return string
   */
  private function decodeSID($bSID) {
    /**
     * revision - 8bit unsigned int (C1)
     * count - 8bit unsigned int (C1)
     * 2 null bytes
     * ID - 32bit unsigned long, big-endian order
     */
    $sid = unpack('C1rev/C1count/x2/N1id', $bSID);
    $subAuthorities = [];

    if (!isset($sid['id']) || !isset($sid['rev'])) {
      throw new \UnexpectedValueException(
        'The revision level or identifier authority was not found when decoding the SID.'
      );
    }

    $revisionLevel = $sid['rev'];
    $identifierAuthority = $sid['id'];
    $subs = isset($sid['count']) ? $sid['count'] : 0;

    // The sub-authorities depend on the count, so only get as many as the count, regardless of data beyond it
    for ($i = 0; $i < $subs; $i++) {
      // Each sub-auth is a 32bit unsigned long, little-endian order
      $subAuthorities[] = unpack('V1sub', hex2bin(substr(bin2hex($bSID), 16 + ($i * 8), 8)))['sub'];
    }

    // Tack on the 'S-' and glue it all together...
    return 'S-' . $revisionLevel . '-' . $identifierAuthority . implode(
      preg_filter('/^/', '-', $subAuthorities)
    );
  }

  /**
   * Decode binary GUID
   *
   * @param string $bGUID
   * @return string
   */
  private function decodeGUID($bGUID) {
    $unpacked = unpack('Va/v2b/n2c/Nd', $bGUID);
    return sprintf('%08X-%04X-%04X-%04X-%04X%08X', $unpacked['a'], $unpacked['b1'], $unpacked['b2'], $unpacked['c1'], $unpacked['c2'], $unpacked['d']);
  }

  /**
   * Construit filtre AD/Ldap Groupe
   * @param Domain $oDomain le domaine à filtré
   * @param string|null $LdapFilter filtre ldap
   * @return string filtre AD/Ldap
   */
  private function getLdapGroupFilter(Domain $oDomain, ?string $LdapFilter = null): string {
    // Filtre du domaine
    $sFilterGroup = $oDomain->getFilterGroup();
    if (empty($sFilterGroup)) {
      $sFilterGroup = '(|(&(|(objectClass=container)(objectClass=group))(!(isCriticalSystemObject=TRUE))(!(showInAdvancedViewOnly=TRUE)))(objectClass=posixGroup))';
    }
    if (!empty($LdapFilter)) {
      $sFilterGroup = '(&' . $sFilterGroup . '(' . $LdapFilter . '))';
    }
    // Initialisation du filtre
    $sFilter = null;
    // Récupération des paramètres de la requête
    $aData = $this->requestToArray();

    // Retourne filter du domain
    if (!isset($aData['filter'])) {
      return $sFilterGroup;
    }
    $aFilter = json_decode(str_replace('%', '*', $aData['filter']), true);
    // Retourne filter du domain
    if (!isset($aFilter) || count($aFilter) == 0) {
      return $sFilterGroup;
    }
    switch ($aFilter['relation']) {
      case 'AND':
      case 'OR':
        $sFilter = '&';
        break;
    }
    foreach ($aFilter['operators'] as $aColumn) {
      $sColumn = $aColumn['column'];
      if (!isset($sColumn)) {
        // TODO faire correspondance des colonnes d'un groupe
        continue;
      }
      $sFilter = $sFilter . '(' . $sColumn . '=' . $aColumn['value'] . ')';
    }
    return '(' . $sFilter . $sFilterGroup . ')';
  }

  /**
   * Construit filtre AD/Ldap User
   * @param Domain $oDomain le domaine à filtré
   * @param string $sLdapFilter filtre LDAP
   * @return string|null filtre AD/Ldap
   */
  private function getLdapUserFilter(Domain $oDomain, string $sLdapFilter = null): ?string{
    // Filtre du domaine
    $sFilterUser = $oDomain->getFilterUser();
    if (empty($oDomain->getFilterUser())) {
      $sFilterUser = '(&(' . $oDomain->getUid() . '=*)(objectclass=person))';
    }
    if (!empty($sLdapFilter)) {
      $sFilterUser = '(&' . $sFilterUser . '(' . $sLdapFilter . '))';
    }
    // Initialisation du filtre
    $sFilter = null;
    // Récupération des paramètres de la requête
    $aData = $this->requestToArray();

    // Retourne filter du domain
    if (!isset($aData['filter'])) {
      return $sFilterUser;
    }
    $aFilter = json_decode(str_replace('%', '*', $aData['filter']), true);
    // Retourne filter du domain
    if (!isset($aFilter) || count($aFilter) == 0) {
      return $sFilterUser;
    }
    switch ($aFilter['relation']) {
      case 'AND':
        $sFilter = '&';
        break;
      case 'OR':
        $sFilter = '&';
        break;
    }
    foreach ($aFilter['operators'] as $aColumn) {
      $sColumn = null;
      switch ($aColumn['column']) {
        case 'displayname':
          $sColumn = null !== $oDomain->getDisplayName() ? $oDomain->getDisplayName() : 'displayname';
          break;
        case 'mail':
          $sColumn = null !== $oDomain->getMail() ? $oDomain->getMail() : 'mail';
          break;
        case 'uid':
          $sColumn = (!empty($oDomain->getUid())) ? $oDomain->getUid() : 'uid';
          break;
      }
      if (!isset($sColumn)) {
        continue;
      }
      $sFilter = $sFilter . '(' . $sColumn . '=' . $aColumn['value'] . ')';
    }
    return '(' . $sFilter . $sFilterUser . ')';
  }

  /**
   * @return Response
   */
  protected function formatErrorDomainNotFound(): Response {
    return $this->formatGenericError('DOMAIN_NOT_FOUND', '', 'Domain not found', Response::HTTP_NOT_FOUND);
  }

  /**
   * @return Response
   */
  protected function formatErrorListNoSelection(): Response {
    return $this->formatGenericError('LIST_NO_SELECTION', '', 'Please select at least one record.', Response::HTTP_NOT_FOUND);
  }
}
