<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\FileAccessor\LfsService;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller permettant d'uploader des gros fichiers via l'API.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class LfsController extends AbstractVitisController {
  /**
   * @var LfsService
   */
  protected $oLfsService;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oLfsService = $oFileManager->getLfsService();;
  }

  /**
   * Initialise un upload transactionnel
   *
   * @return Response
   */
  public function beginTransaction(): Response {
    $aKeyNeeded = ['md5_hash_content', 'nb_chunk', 'file_name'];
    // vérification des paramètres
    $aRequestData = $this->requestToArray();
    $aJson = [];
    foreach ($aKeyNeeded as $sKey) {
      if (!array_key_exists($sKey, $aRequestData)) {
        return $this->formatErrorMissingParameter($sKey . ' parameter is missing');
      }

      // copi de la clé pour enregistrement dans le json
      $aJson[$sKey] = $aRequestData[$sKey];
    }

    // création du dossier pour la transaction
    $sTransactionId = $this->oLfsService->createTransactionId();
    // enregistrement du json de vérification
    if (!$this->oLfsService->storeTransactionData($sTransactionId, $aJson)) {
      return $this->formatGenericErrorRequestError('Cannot store transaction data');
    }

    // formatage du retour
    return $this->formatResponse([
      'status' => 1,
      'data' => [
        'transaction_id' => $sTransactionId,
      ],
    ]);
  }
  /**
   * Ajoute un tronçon à un upload de fichier
   *
   * @param  string   $transactionId id de transaction
   * @param  int   $chunckId id du troçon dans le fichier
   * @return Response
   */
  public function addChunk($transactionId, $chunckId): Response {
    $aRequestData = $this->requestToArray();

    if (!array_key_exists('md5_hash_content', $aRequestData)) {
      return $this->formatErrorMissingParameter('md5_hash_content parameter is missing');
    }

    // récupération du tronçon
    $aData = $this->oLfsService->getTransactionData($transactionId);
    if ($aData == null) {
      return $this->formatGenericError('', '', 'Missing transaction data, please start a transaction before', Response::HTTP_NOT_FOUND);
    }

    // évaluation du hash de signature du chunk depuis tmp_path
    // si ok copie vers l'emplacement de transaction
    $bValid = $this->oLfsService->checkUploadedChunk($transactionId, $chunckId, $aRequestData['md5_hash_content']);
    // sinon suppression + erreur

    if ($bValid) {
      // stockage dans le fichier json
      $this->oLfsService->storeChunkHash($transactionId, $chunckId, $aRequestData['md5_hash_content']);
    }

    return $this->formatResponse([
      'status' => $bValid ? 1 : 0,
    ]);
  }
  /**
   * Vérifie l'intégrité générale et regénère le fichier
   *
   * @param  string   $id
   * @return Response
   */
  public function completeTransaction($id): Response {
    // récupérer les données json
    if (!$this->oLfsService->checkTransactionIntegrity($id)) {
      return $this->formatGenericError('', '', 'Error with global integrity');

    }

    return $this->formatResponse(['status' => $this->oLfsService->mergeChunk($id) ? 1 : 0]);
  }


  /**
   * Supprime une transaction
   *
   * @param  string   $id
   * @return Response
   */
  public function abortTransaction($id): Response {
return $this->formatResponse(['status' => $this->oLfsService->clearTransaction($id) ? 1 : 0]);

  }

}
