<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\TokenController;
use App\Module\Vitis\WebService\Entity\User;
use Ramsey\Uuid\Uuid;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller pour gérer les jetons de connexion.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class LoginTokenController extends TokenController {
  /**
   * Insert un nouveau jeton de connexion avec un token
   * @return Response
   */
  public function insertLoginToken(): Response {
    $sToken = $this->getUserToken();

    // Si le retour de getUserToken vaut null,
    // c'est que c'est une erreur et que la réponse est préformatée
    if ($sToken === null) {
      return $this->oResponse;
    }

    $aData = $this->requestToArray();

    $oEm = $this->oDatabaseConnection->getAdminEntityManager();
    $qb = $this->oDatabaseConnection->getQueryAccessor($oEm)->createQueryBuilder();
    $qb->select('a.token_id')
      ->from('[SCHEMA].login_token', 'a')
      ->where($qb->expr()->eq('a.token_id', '[TOKEN]'));

    $aParams = [
      'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
    ];

    $i = 0;
    $iMax = 10; // Nombre de tentative max pour trouver un id unique
    do {
      $i++;
      // Attribution d'un id unique
      $aData['token_id'] = Uuid::uuid1()->toString();

      // Vérfication de son absence dans la bdd
      $aParams['TOKEN'] = ['value' => $aData['token_id'], 'type' => 'string'];
      $oToken = $qb->executeWithParams($aParams)->fetchAll();
    } while (isset($oToken[0]) && $i <= $iMax);

    if ($i == $iMax) {
      // Si la sortie du while est du au dépassement du maximal de nombre de tentative
      $aReturn = ['status' => 0];
      $aReturn['message'] = 'Can\'t create unique id for login_token';
      return $this->formatResponse($aReturn, Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    $aData['token'] = $sToken;

    $oDataAccessor = $this->getDataAccessor($this->oDatabaseConnection->getAdminEntityManager());
    // Sauvegarde en base de donnée
    $oEntity = $oDataAccessor->insertEntity(json_encode($aData));
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnInsert();
    }
    return $this->formatResponse($oEntity);
  }

  /**
   * Vérifie que les données requises soient présente
   * @param array $aData
   * @return bool
   */
  private function checkRequiredData(array $aData): bool {
    if (!isset($aData['user_id']) || $aData['user_id'] == '') {
      return false;
    }
    if (!isset($aData['password']) || $aData['password'] == '') {
      return false;
    }
    return true;
  }

  /**
   * Récupère et retourne le token après avoir vérifié qu'il n'y a pas de problème (paramètre obligatoire, user non trouvé...).
   * @return string|null
   */
  private function getUserToken(): ?string{
    // récupération
    $aData = $this->requestToArray();

    if (!$this->checkRequiredData($aData)) {
      // Des paramètres obligatoire n'ont pas été trouvé
      $this->formatGenericError('MISSING_REQUIRED_DATA', '', 'Missing required params', Response::HTTP_BAD_REQUEST);
      return null;
    }

    // chargement de la connexion à la base de donnée
    $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();

    // vérification de l'existence de l'utilisateur
    $oUser = $oEntityManager->getRepository(User::class)->findOneBy(['user_id' => $aData['user_id']]);

    if (!isset($oUser)) {
      // L'utilisateur n'a pas pu être trouvé
      $this->formatGenericError('ERROR_USER_NOT_EXIST', '', 'This user does not exist', Response::HTTP_UNAUTHORIZED);
      return null;
    }

    $sHashedPassword = $aData['password'];
    if (empty($sHashedPassword)) {
      $this->oLoggerManager->get('API_ERROR')->error('Password Hash failed for ' . $oUser->getLogin());
      // Le mot de passe n'a pas pu être encodé
      $this->formatGenericErrorRequestError('Your password cannot be hashed, contact application\'s administrator', Response::HTTP_NOT_IMPLEMENTED);
      return null;
    }

    // Initialisation de la structure de l'utilisateur
    $aPayload = $oUser->getTokenPayload();

    // Vérification de la date de validité saisie (si il y en a une)
    if ($this->haveLimitDate($aData) && intval($aData['limit_date'], 10) < time()) {
      $this->formatGenericError('FORM_LOGIN_TOKEN_ERROR_PAST_DATE', '',
        'You try to create a past limit date', Response::HTTP_UNAUTHORIZED);
      return null;
    }

    $preparePayload = [
      'user' => $aPayload,
      'password' => $sHashedPassword,
    ];

    if (isset($aData['ip_restriction'])) {
      $preparePayload['ip_restriction'] = $aData['ip_restriction'];
    }

    // Récupération du payload jwe par défaut pour le token
    $payload = $this->getJWETokenPayload($preparePayload);
    // Modification de la date d'expiration
    $payload['exp'] = null;
    if ($this->haveLimitDate($aData)) {
      $payload['exp'] = intval($aData['limit_date'], 10);
    }
    // Création du token
    $aTokenStruct = $this->getJWEToken($payload);

    // Test de connexion avec le token
    try {
      $isValidUser = $this->oDatabaseConnection->checkUserConnection($oUser->getLogin(), $sHashedPassword);

      if (!$isValidUser) {
        // L'utilisateur n'a pas pu être trouvé
        $this->formatGenericError('ERROR_USER_NOT_EXIST', '',
          'This user does not exist', Response::HTTP_UNAUTHORIZED);
        return null;
      }
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('LoginToken: Connection failed for ' . $oUser->getLogin(), [$e]);
      // L'utilisateur n'a pas pu être connecté (ex: mauvais mot de passe)
      $this->formatGenericError('FORM_LOGIN_TOKEN_ERROR_CONNECTION', '', 'Connection error', Response::HTTP_INTERNAL_SERVER_ERROR);
      return null;
    }

    if (empty($aTokenStruct['token'])) {
      // Le token n'a pas pu être généré
      $this->oLoggerManager->get('API_ERROR')->error('Token generation failed for ' . $aData['user'], $aTokenStruct);
      $this->formatGenericError('ERROR_EMPTY_TOKEN_PAYLOAD', '', 'Problem with Token Payload', Response::HTTP_SERVICE_UNAVAILABLE);
      return null;
    }

    return $aTokenStruct['token'];
  }

  /**
   * Modifie un jeton de connexion
   * @return Response
   */
  public function updateLoginToken(): Response {
    $sToken = $this->getUserToken();
    // Si le retour de getUserToken vaut null,
    // c'est que c'est une erreur et que la réponse est préformatée
    if ($sToken === null) {
      return $this->oResponse;
    }

    $aData = $this->requestToArray();
    $aData['token'] = $sToken;

    $oDataAccessor = $this->getDataAccessor($this->oDatabaseConnection->getAdminEntityManager());
    // Sauvegarde en base de donnée
    $oEntity = $oDataAccessor->updateEntity($aData['token_id'], json_encode($aData));
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnUpdate();
    }
    return $this->formatResponse($oEntity);
  }

  /**
   * Vérifie si il y a une date limite
   * @param array $aData
   * @return bool
   */
  private function haveLimitDate($aData): bool {
    if (!empty($aData['limit_date'])) {
      if (intval($aData['limit_date'], 10) > 0) {
        return true;
      }
    }
    return false;
  }
}
