<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller qui permet de récupérer et d'intéragir avec les formulaires studio.
 *
 * La route devra contenir id et form
 * id : identifiant de l'enregistrement
 * form : formulaire à retourner (default, published, custom)
 *
 * Exemple :
 * "route": "/v2/gtf/userworkspaces/{id}/form/{form}",
 *
 * Dans les options il faudra utiliser l'attribut files et y placer un tableau d'entités
 * Chaque entité est définie par name, form_file, path
 * name : nom de l'attribut en sortie
 * form_file : définit le nom de fichier pour chaque type de formulaire
 * path : chamin vers le fichier, utilisant @id et @file_name
 *
 * Exemple :
 * "files": [{
 *   "name": "json_form",
 *   "form_file": {
 *     "default": "DSubform.json",
 *     "published": "Subform.json",
 *     "custom": "WSubform.json"
 *   },
 *   "path": "@prop(ws_data_dir)/gtf/workspace/@id/form/@file_name"
 * }, {
 *   "name": "js_form",
 *   "form_file": {
 *     "default": "DSubform.js",
 *     "published": "Subform.js",
 *     "custom": "WSubform.js"
 *   },
 *   "path": "@prop(ws_data_dir)/gtf/workspace/@id/form/ressources/@file_name"
 * }, {
 *   "name": "css_form",
 *   "form_file": {
 *     "default": "DSubform.css",
 *     "published": "Subform.css",
 *     "custom": "WSubform.css"
 *   },
 *   "path": "@prop(ws_data_dir)/gtf/workspace/@id/form/ressources/@file_name"
 * }]
 */
class StudioFormController extends AbstractVitisController {
  /**
   * Clé utilisée pour l'id dans les chemins
   * @var string
   */
  private const ID_PATH_KEY = '@id';

  /**
   * Clé utilisée pour nom du fichier dans les chemins
   * @var string
   */
  private const FILE_NAME_PATH_KEY = '@file_name';

  /**
   * Récupère le contenu du formulaire
   * @param mixed $id
   * @param mixed $form
   * @return Response réponse à retourner par l'API
   */
  public function getForm($id, $form): Response {
    $aRequestOptions = $this->getRequestParameters();
    $aReturn = [];

    if (!empty($id) && !empty($form) && is_array($aRequestOptions['options']['files'])) {
      $aReturn['data'] = [];
      foreach ($aRequestOptions['options']['files'] as $i => $oFile) {
        $sFileContent = $this->getFileContent($oFile, $id, $form);
        if ($sFileContent !== null) {
          $aReturn['data'][$oFile['name']] = $sFileContent;
        }
      }
    }

    $aReturn['status'] = 1;
    $this->formatResponse($aReturn);
    return $this->oResponse;
  }

  /**
   * Retourne le contenu d'un fichier
   * @param mixed $oFile
   * @param mixed $id
   * @param mixed $form
   * @return string|null
   */
  private function getFileContent($oFile, $id, $form): ?string {
    if (!empty($oFile['name']) && !empty($oFile['form_file'][$form]) && !empty($oFile['path'])) {
      $sPath = $this->getFilePath($oFile, $id, $form);

      // Récupération du contenu
      $oFileInterface = $this->getFileAccessor($sPath);
      $sContent = $oFileInterface->file_get_contents($sPath);

      return $sContent;
    }
    return null;
  }

  /**
   * Met à jour les fichiers d'un formulaire
   * @param mixed $id
   * @param mixed $form
   * @return Response
   */
  public function updateFiles($id, $form): Response {
    $aRequestOptions = $this->getRequestParameters();
    $aData = $this->requestToArray();

    $aReturn = [];
    if (!empty($id) && !empty($form) && is_array($aRequestOptions['options']['files'])) {
      $aReturn['data'] = [];
      foreach ($aRequestOptions['options']['files'] as $i => $oFile) {
        $sFilePath = $this->getFilePath($oFile, $id, $form);
        $oFileInterface = $this->getFileAccessor($sFilePath);
        $aPathInfo = pathinfo($sFilePath);
        if (!empty($sFilePath) && isset($aData[$aPathInfo['extension']])) {
          if ($oFileInterface->file_put_contents($sFilePath, $aData[$aPathInfo['extension']]) === null) {
            return $this->formatErrorCannotUpdateForm($sFilePath . ' can\'t be writed, verify the folder\'s rights');
          }
          $sFileContent = $this->getFileContent($oFile, $id, $form);
          if ($sFileContent !== null) {
            $aReturn['data'][$oFile['name']] = $sFileContent;
          }
        }
      }
    }

    $aReturn['status'] = 1;
    $this->formatResponse($aReturn);
    return $this->oResponse;
  }

  /**
   * Retourne chemin d'un fichier
   * @param mixed $oFile
   * @param mixed $id
   * @param mixed $form
   * @return string|null
   */
  private function getFilePath($oFile, $id, $form): ?string {
    if (!empty($oFile['name']) && !empty($oFile['form_file'][$form]) && !empty($oFile['path'])) {
      $sPath = $oFile['path'];

      // Remplacement des properties
      $sPath = $this->oProperties->replacePropertiesInString($sPath);

      // Remplacement de l'id
      $sPath = str_replace(self::ID_PATH_KEY, $id, $sPath);

      // Remplacement du nom du fichier
      $sPath = str_replace(self::FILE_NAME_PATH_KEY, $oFile['form_file'][$form], $sPath);

      return $sPath;
    }
    return null;
  }

/**
 * A creuser
 *
 * @param  string   $sMessage
 * @return Response
 */
  private function formatErrorCannotUpdateForm(string $sMessage = ''): Response {
    $aReturn = [
      'status' => 0, //ajout car n'y étais pas de base
      'success' => 0,
      'message' => $sMessage,
      'errorMessage' => '',
    ];
    return $this->formatResponse($aReturn, Response::HTTP_INTERNAL_SERVER_ERROR);
  }
}
