<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Entity\Privilege;
use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Repository\PrivilegeRepository;
use App\Module\Vitis\WebService\Repository\UserRepository;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller Générique de Vitis.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */
class UserController extends GenericController {
  /**
   * @var \App\Module\Vitis\WebService\Service\DataAccessor\QueryAccessor
   */
  protected $oQueryAccessor;

  /**
   * Créé un utilisateur
   * @return Response réponse à retourner par l'API
   */
  public function insertUser(): Response {
    // récupération login password
    $aData = $this->requestToArray();
    // récupération d'un queryAccessor natif pour les opérations sur les roles
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    if (empty($this->oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }
    // Vérifie si le login n'existe pas déjà
    if ($this->isExistingUser($aData['login']) === true) {
      return $this->formatGenericError('USER_ALREADY_CREATE', '', 'user already exists', Response::HTTP_UNAUTHORIZED);
    }
    // Vérification de non-existence du rôle postgres
    if (!$this->oQueryAccessor->ifUserExist($aData['login'])) {
      // création de l'utilisateur dans Postgres
      if (!$this->oQueryAccessor->createUser($aData['login'])) {
        return $this->formatGenericError('ERROR_POSTGRES_USER_CREATION', '',
          'Cannot create Postgres User with this login', Response::HTTP_UNAUTHORIZED);
      }
    }
    // mise à jour du mot de passe
    if (!$this->updatePgUserPassword()) {
      return $this->formatGenericError('ERROR_POSTGRES_USER_PASSWORD_UPDATE', '',
        'Cannot set password for Postgres User', Response::HTTP_UNAUTHORIZED);
    }

    // insertion de l'utilisateur dans la table user
    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    // Sauvegarde en base de donnée
    $oResponse = $oDataAccessor->insertEntity($this->requestToJson());
    if ($oResponse['status'] === 0) {
      return $this->formatGenericResponseForErrorOnInsert('Cannot create Vitis User with this login');
    }

    if ($oResponse['data'] instanceof User) {
      $oUser = $oResponse['data'];
      // ajout des privilèges
      $this->updateUserPrivileges($oUser);
    }

    return $this->formatResponse($oResponse, Response::HTTP_CREATED);
  }

  /**
   * Créé un utilisateur
   * @param mixed $id
   * @return Response réponse à retourner par l'API
   */
  public function updateUser($id): Response {
    // récupération d'un queryAccessor natif pour les opérations sur les roles
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    if (empty($this->oQueryAccessor)) {
      return $this->formatPostgresBadConnection();

    }
    // mise à jour du mot de passe
    if (!$this->updatePgUserPassword()) {
      return $this->formatGenericError('ERROR_POSTGRES_USER_PASSWORD_UPDATE', '',
        'Cannot set password for Postgres User', Response::HTTP_UNAUTHORIZED);
    }

    // insertion de l'utilisateur dans la table user
    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    // Sauvegarde en base de donnée
    $oResponse = $oDataAccessor->updateEntity($id, $this->requestToJson());
    if ($oResponse['status'] === 0) {
      return $this->formatGenericResponseForErrorOnUpdate('Cannot update Vitis User with this login');
    }

    if ($oResponse['data'] instanceof User) {
      $oUser = $oResponse['data'];
      // ajout des privilèges
      $this->updateUserPrivileges($oUser);
    }

    return $this->formatResponse($oResponse);
  }

  /**
   * Supprime des utilisateurs et les roles associés
   * @return Response
   */
  public function deleteUsers(): Response {
    // récupération du contenu de la requête
    $aData = $this->massiveRequestGetIdList();

    $aUserId = ($aData === null) ? [] : $aData['idList'];

    return $this->revokeAndDeleteUsers($aUserId);
  }

  /**
   * Supprime un utilisateur et les role associé
   * @param int $id id utilisateur
   * @return Response
   */
  public function deleteUser(int $id): Response {
    // passage format array
    $aUserId = [$id];

    return $this->revokeAndDeleteUsers($aUserId);
  }

  /**
   * Supprime des utilisateurs et les roles associés
   * @param int[] $aUserList utilsateurs
   * @return Response
   */
  private function revokeAndDeleteUsers(array $aUserList): Response {
    if (empty($aUserList)) {
      return $this->formatGenericError('ERROR_VITIS_USER_DELETE_EMPTY_ID_LIST', '',
        'Cannot find Vitis User(s)', Response::HTTP_NOT_FOUND);
    }

    // récupération d'un queryAccessor natif pour les opérations sur les roles
    $this->oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();
    if (empty($this->oQueryAccessor)) {
      return $this->formatPostgresBadConnection();

    }

    // chargement de la connexion à la base de donnée
    $oRepository = $this->oDatabaseConnection->getUserEntityManager()->getRepository(User::class);
    if (!$oRepository instanceof UserRepository) {
      return $this->formatErrorUserNotfound('Cannot find Users, bad repository');
    }

    // find users
    $aEntities = $oRepository->findByIdInRange($aUserList);
    if (empty($aEntities)) {
      return $this->formatErrorUserNotfound('Cannot find Users');
    }
    // gestion des roles dans postgres
    $aLockedUsers = [];
    $aFailedDrop = [];
    $oCurrentUser = $this->getCurrentUserEntity();
    foreach ($aEntities as $oUser) {
      $sLogin = $oUser->getLogin();
      // on n'efface pas les roles vérrouiller
      if ($oCurrentUser->getUserId() === $oUser->getUserId()) {
        $aLockedUsers[] = $oUser->getUserId();
        continue;
      }
      // on revoke tous les Privilèges
      if (!$this->removePrivilegesToUser($sLogin, $oUser->getRoles())) {
        $aFailedDrop[] = $oUser->getUserId();
      }
    }
    // on retire les roles protégés pour ne pas les effacer
    $aUserList = array_diff($aUserList, $aLockedUsers);
    // on retire les roles qui n'ont pas pû être drop
    $aUserList = array_diff($aUserList, $aFailedDrop);
    // nettoyage de la table vitis
    $oDataAccessor = $this->getDataAccessor();
    $aEntities = $oDataAccessor->deleteEntities($aUserList);

    if (!empty($aFailedDrop)) {
      $sMessage = 'Cannot drop Postgres Roles for users : ' . implode(',', $aFailedDrop);
      return $this->formatGenericError('ERROR_VITIS_USER_DELETE_DROP_ROLE', '',
        $sMessage, Response::HTTP_UNAUTHORIZED);
    }
    if (!empty($aLockedUsers)) {
      return $this->formatGenericError('ERROR_VITIS_USER_DELETE_HIMSELF', '',
        'Cannot drop your own user', Response::HTTP_UNAUTHORIZED);
    }

    return $this->formatResponse($aEntities);
  }

  /**
   * Contrôle les données de la requête pour changer le mot de passe et change le mot de passe
   * @return bool True si l'update réussi false sinon
   */
  private function updatePgUserPassword(): bool {
    $aData = $this->requestToArray();
    // si pas de password on mets pas à jour
    if (!isset($aData['password'])) {
      return true;
    }

    if (!isset($aData['password_confirm'])) {
      return false;
    }

    if (empty($aData['password']) || empty($aData['password_confirm'])) {
      return false;
    }

    if ($aData['password'] !== $aData['password_confirm']) {
      return false;
    }
    // Mise à jour du mot de passe
    return $this->setUserPassword($aData['login'], $aData['password']);
  }

  /**
   * Modifie le mot de passe d'utilisateur
   * @param string $sLogin
   * @param string $sPassword
   * @return bool
   */
  private function setUserPassword(string $sLogin, string $sPassword): bool {
    // protège le mot de passe d'après la configuraton
    $sProtectedPassword = $this->oDatabaseConnection->getHashedPassword($sLogin, $sPassword);
    // mets à jour le mot de passe
    return $this->oQueryAccessor->updateRolePassword($sLogin, $sProtectedPassword);
  }

  /**
   * Au besoin met à jour les privileges.
   * @param User $oUser l'entité User à mettre à jour
   * @return bool true si la mise à jour de l'utilisateur postgres s'est bien passé false sinon
   */
  private function updateUserPrivileges(User $oUser): bool {
    $aData = $this->requestToArray();

    if (!isset($aData['privileges'])) {
      return false;
    }

    // récupère les nouveaux privilèges
    $aNewPrivileges = $aData['privileges'];
    if (!is_array($aNewPrivileges)) {
      $aNewPrivileges = explode('|', $aNewPrivileges);
    }
    // Si l'utilisateur à modifier est l'utilisateur qui fait la mise à jour
    $oCurrentUser = $this->getCurrentUserEntity();
    if ($oCurrentUser->getUserId() === $oUser->getUserId()) {
      // Preserve le privileges 'vitis_user'
      if (!in_array('vitis_user', $aNewPrivileges)) {
        $aNewPrivileges[] = 'vitis_user';
      }
      // Preserve le privileges 'vitis_admin'
      if (!in_array('vitis_admin', $aNewPrivileges)) {
        $aNewPrivileges[] = 'vitis_admin';
      }
    }
    $aRolesToDelete = array_diff($oUser->getPrivileges(), $aNewPrivileges);
    // Supprime des privileges
    if (!$this->removePrivilegesToUser($oUser->getLogin(), $aRolesToDelete)) {
      return false;
    }
    // Récupération de PrivilegeRepository
    $privilegeRepository = $this->oDatabaseConnection->getUserEntityManager()->getRepository(Privilege::class);
    if (!$privilegeRepository instanceof PrivilegeRepository) {
      return false;
    }

    // Passage du tableau de string en tableau d'entité existante
    $aPrivileges = $privilegeRepository->findByRolname($aNewPrivileges);
    if (!$this->addPrivilegesToUser($oUser, $aPrivileges)) {
      return false;
    }
    return true;
  }

  /**
   * Ajoute des privileges a un utilisateur
   * @param User $oUser
   * @param array $aPrivileges
   * @return bool true si la mise à jour est passé false sinon
   */
  public function addPrivilegesToUser(User $oUser, array $aPrivileges): bool {
    $sLogin = $oUser->getLogin();
    $aRoles = [];
    foreach ($aPrivileges as $oPrivilege) {
      $aRoles[] = $oPrivilege->getRolname();
    }
    // Récupère le domaine d'utilisateur
    $oDomain = $oUser->getDomain();
    // Preserve le privilège d'un domaine
    if ($oDomain !== null) {
      $aRoles[] = $oDomain->getLoginRole();
    }
    if (count($aRoles) === 0) {
      return false;
    }
    if (in_array('vitis_admin', $aRoles)) {
      if (!$this->oQueryAccessor->addCreateRoleToRole($sLogin)) {
        return false;
      }
    }
    return $this->oQueryAccessor->grantRolesToRole($aRoles, $sLogin);
  }

  /**
   * Supprime les privileges d'utilisateur
   * @param string $sLogin
   * @param array $aRoles
   * @return bool true si la mise à jour est passé false sinon
   */
  public function removePrivilegesToUser(string $sLogin, array $aRoles): bool {
    if (count($aRoles) === 0) {
      return true;
    }
    if (in_array('vitis_admin', $aRoles)) {
      if (!$this->oQueryAccessor->removeCreateRoleToRole($sLogin)) {
        return false;
      }
    }
    return $this->oQueryAccessor->revokeRolesFromRole($aRoles, $sLogin);
  }

  /**
   * Vérifie si un utilisateur existe
   * @param string $sLogin
   * @return bool true si l'utilisateur existe, sinon false
   */
  private function isExistingUser(string $sLogin): bool {
    $userRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(User::class);

    if (!$userRepository instanceof UserRepository) {
      return false;
    }

    $user = $userRepository->findByLogin($sLogin);
    return !empty($user);
  }

  /**
   * @param  string   $sMessage
   * @return Response
   */
  protected function formatErrorUserNotfound(string $sMessage = ''): Response {
    return $this->formatGenericError('ERROR_VITIS_USER_DELETE_NOT_FOUND', '', $sMessage, Response::HTTP_NOT_FOUND);
  }

}
