<?php

namespace App\Module\Vitis\WebService\Repository;

use App\Module\Vitis\WebService\Query\SortableNullValueWalker;
use App\Module\Vitis\WebService\Service\DataAccessor\CommonAccessorFunctions;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Tools\Pagination\Paginator;

/**
 * Repository Générique de Vitis.
 *
 * @author Anthony Borghi   <anthony.borghi@veremes.com>
 * @author Alexandre Mironenkov   <alexandre.mironenkov@veremes.com>
 * @author Ludivine Premont   <ludivine.premont@veremes.com>
 */

class GenericRepository extends EntityRepository {
  /**
   * @var string
   */
  public const ENTITY_ALIAS = 'generic_entity';

  /**
   * @var string[]
   */
  protected $aFields = [];
  /**
   * @var string
   */
  protected $sIdentifier = 'id';
  /**
   * @var string
   */
  protected $sEntityNamespace;

  /**
   * {@inheritdoc}
   */
  public function __construct(EntityManagerInterface $em, ClassMetadata $class) {
    parent::__construct($em, $class);
    $this->aFields = $class->getFieldNames();
    $aIdentifer = $class->getIdentifierFieldNames();
    $this->sEntityNamespace = $class->getName();

    if (!empty($aIdentifer)) {
      $this->sIdentifier = $aIdentifer[0];
    }
  }

  /**
   * Permet de filtrer et paginer les entités Doctrine.
   * @param array $aParams Objet permettant de filtrer et paginer les entités
   *                          - $aParams.order_by     (string)        : colonne pour trier les entités
   *                          - $aParams.sort_order   (string)        : sens de tri (ASC|DESC)
   *                          - $aParams.limit        (integer)       : nombre d'entité à retourner
   *                          - $aParams.offset       (integer)       : décallage des entité par rapport à la première
   *                          - $aParams.attributs    (string)        : Liste des champs à afficher
   *                          - $aParams.filter       (string)        : Json inline permettant de générer les filtres
   *                          - $aParams.distinct     (bool)          : Pas de doublon sur les entités
   * @param CommonAccessorFunctions $oAccessorFunc service de manipulation des attributs
   * @return array Liste des entités
   */
  public function search(array $aParams, CommonAccessorFunctions $oAccessorFunc): array{
    // initialisation query builder
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS);
    $sOrderByColumn = null;

    // order_by
    if (isset($aParams['order_by'])) {
      $sSortOrder = (isset($aParams['sort_order'])) ? $aParams['sort_order'] : 'asc';
      $sColumnOrderBy = $oAccessorFunc->getFieldColumn($aParams['order_by']);
      if (isset($sColumnOrderBy)) {
        $sOrderByColumn = self::ENTITY_ALIAS . '.' . $sColumnOrderBy;
        $qb->orderBy(self::ENTITY_ALIAS . '.' . $sColumnOrderBy, $sSortOrder);
      }
    }
    // filter
    $aParams['filter'] = $aParams['filter'] ?? null;
    if ($oAccessorFunc->needFilter($aParams['filter'])) {
      $aFilterReturn = $oAccessorFunc->prepareAndDecodeFilter($aParams['filter']);

      if (!empty($aFilterReturn['request'])) {
        $qb->where($aFilterReturn['request']);

        foreach ($aFilterReturn['params'] as $sKey => $aValue) {
          if (isset($aValue['type'])) {
            $qb->setParameter($sKey, $aValue['value'], $aValue['type']);
          } else {
            $qb->setParameter($sKey, $aValue['value']);
          }
        }
      }
    }

    $oFullPaginator = new Paginator($qb, true);
    $oFullPaginator->setUseOutputWalkers(false);

    // offset
    if (isset($aParams['offset']) && $aParams['offset'] > 0) {
      $qb->setFirstResult($aParams['offset']);
    }

    // limit
    if (isset($aParams['limit']) && $aParams['limit'] > 0) {
      $qb->setMaxResults($aParams['limit']);
    }
    $query = $qb->getQuery();
    if ($sOrderByColumn !== null) {
      // ajout de NULLS LAST
      $query->setHint(SortableNullValueWalker::HINT, [
        $sOrderByColumn => SortableNullValueWalker::NULLS_LAST,
      ]);
    }

    $aEntities = $query->getResult();

    // Formatage de la réponse
    $aEntities = [
      'data' => $aEntities,
      'total_row_selected' => $oFullPaginator->count(),
      'list_count' => count($aEntities),
      'status' => 1,
    ];

    return $aEntities;
  }

  /**
   * Remove une liste d'entité
   * @param array $aIds Liste des ids à supprimer
   * @return array
   */
  public function removeList(array $aIds): array{
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS)
      ->delete(null, self::ENTITY_ALIAS)
      ->where(self::ENTITY_ALIAS . '.' . $this->sIdentifier . ' IN (:aIds)')
      ->setParameter('aIds', $aIds);

    $query = $qb->getQuery();
    $iRemoved = $query->getResult();

    return [
      'total_row_removed' => $iRemoved,
      'data' => $aIds,
      'status' => 1,
    ];
  }

  /**
   * Supprimer un item
   *
   * @param  mixed $mId
   * @return bool
   */
  public function removeItemById($mId): bool {
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS)
      ->delete(null, self::ENTITY_ALIAS)
      ->where(self::ENTITY_ALIAS . '.' . $this->sIdentifier . ' = :mId')
      ->setParameter('mId', $mId);

    $query = $qb->getQuery();
    $iRemoved = $query->getResult();

    return ($iRemoved === 1);
  }

  /**
   * Retourne une entité filtré par id avec uniquement les colonnes demandées
   * @param mixed $mId Id de l'entité à trouver
   * @return array
   */
  public final function findById($mId): array{
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS)
      ->where(self::ENTITY_ALIAS . '.' . $this->sIdentifier . ' = :mId')
      ->setParameter('mId', $mId);

    $query = $qb->getQuery();
    return $query->getResult();
  }

  /**
   * @param mixed[] $aId liste d'id d'entité
   * @return object[] Liste d'entité
   */
  public function findByIdInRange(array $aId): array{
    if (empty($aId)) {
      return [];
    }
    // génération de la requête
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS);
    $qb->where($qb->expr()->in(self::ENTITY_ALIAS . '.' . $this->sIdentifier, $aId));

    return $qb->getQuery()->getResult();
  }

  /**
   * update une liste d'entité
   * @param array $aIds Liste des ids à mettre à jour
   * @return array
   */
  public final function updateList(array $aIds, array $aProtectedValues, CommonAccessorFunctions $oAccessorFunc): array{
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS)
      ->update($this->sEntityNamespace, 'a');

    foreach ($aProtectedValues['values'] as $sField => $sBind) {
      $qb->set('a.' . $oAccessorFunc->getFieldColumn($sField), ':' . $sField)
        ->setParameter($sField, $aProtectedValues['parameters'][$sField]);
    }

    $oExpr = $qb->expr();
    // protection des valeur numéric de type 01256
    foreach ($aIds as &$mId) {
      $mId = $oExpr->literal($mId);
    }

    $qb->where($oExpr->in('a.' . $this->sIdentifier, $aIds));

    try {
      $qb->getQuery()->execute();
    } catch (\Exception $e) {
      return [
        'status' => 0,
        'message' => $e->getMessage(),
      ];
    }

    return [
      'status' => 1,
    ];
  }
}
