<?php

namespace App\Module\Vitis\WebService\Service\Crypto;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use InvalidArgumentException;
use Jose\Component\Checker\InvalidClaimException;
use Jose\Component\Core\Algorithm;
use Jose\Component\Core\AlgorithmManager;
use Jose\Component\Core\JWK;
use Jose\Component\Encryption\Algorithm\ContentEncryption\A256CBCHS512;
use Jose\Component\Encryption\Algorithm\KeyEncryption\ECDHESA256KW;
use Jose\Component\Encryption\Compression\CompressionMethod;
use Jose\Component\Encryption\Compression\CompressionMethodManager;
use Jose\Component\Encryption\Compression\Deflate;
use Jose\Component\Encryption\JWEBuilder;
use Jose\Component\Encryption\JWEDecrypter;
use Jose\Component\Encryption\Serializer\CompactSerializer;
use Jose\Component\Encryption\Serializer\JWESerializerManager;
use Monolog\Logger;

/**
 * Le JWE est une variante du JWT où les données sont encryptées.
 * Pour plus d'informations :
 *       - JWT : https://tools.ietf.org/html/rfc7519
 *       - JWE : https://tools.ietf.org/html/rfc7516
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class JweCrypto {
  /**
   * @var string
   */
  private const KEY_PATH = '@prop(vas_dir)/.vitis';
  /**
   * @var AlgorithmManager
   */
  private $oKeyEncryptionAlgorithmManager;
  /**
   * @var AlgorithmManager
   */
  private $oContentEncryptionAlgorithmManager;
  /**
   * @var CompressionMethodManager
   */
  private $oCompressionMethodManager;
  /**
   * @var Algorithm
   */
  private $oKeyEncryptionAlgorithm;
  /**
   * @var Algorithm
   */
  private $oContentEncryptionAlgorithm;
  /**
   * @var CompressionMethod
   */
  private $oCompressionMethod;
  /**
   * @var JWK
   */
  private $oJWK;
  /**
   * @var Logger
   */
  protected $oErrorLogger;
  /**
   * @var Properties
   */
  protected $oPropertiesSrvc;

  /**
   * @param Properties $PropertiesSrvc Service de lecture des properties
   * @param LoggerManager $oLogger Service d'accès aux Logger
   */
  public function __construct(Properties $PropertiesSrvc, LoggerManager $oLogger) {
    // récupère le log pour les erreurs de l'API
    $this->oErrorLogger = $oLogger->get('API_ERROR');
    $this->oPropertiesSrvc = $PropertiesSrvc;

    // Mise en place des classes algorithm/compression
    $this->oKeyEncryptionAlgorithm = new ECDHESA256KW();
    $this->oContentEncryptionAlgorithm = new A256CBCHS512();
    $this->oCompressionMethod = new Deflate();

    // The key encryption algorithm manager with the A256KW algorithm.
    $this->oKeyEncryptionAlgorithmManager = new AlgorithmManager([
      $this->oKeyEncryptionAlgorithm,
    ]);

    // The content encryption algorithm manager with the A256CBC-HS256 algorithm.
    $this->oContentEncryptionAlgorithmManager = new AlgorithmManager([
      $this->oContentEncryptionAlgorithm,
    ]);

    // The compression method manager with the DEF (Deflate) method.
    $this->oCompressionMethodManager = new CompressionMethodManager([
      $this->oCompressionMethod,
    ]);

    // Lecture de la clé dans le fichier spécifique de vitis
    $sKeyPath = $this->oPropertiesSrvc->replacePropertiesInString(self::KEY_PATH);
    if (!\file_exists($sKeyPath)) {
      $this->oErrorLogger->error('Cannot find the encryption key file (' . $sKeyPath . ')');
    } else {
      $sJWK = \file_get_contents($sKeyPath);
      $this->oJWK = JWK::createFromJson($sJWK);
    }
  }

  /**
   * @param array $aData Données à crypter
   * @return string donnée cryptées et encodées ou null en cas d'erreur
   */
  public function encryptData(array $aData): ?string{
    $sEncryptedData = $this->encrypt($aData);
    if ($sEncryptedData == null) {
      return null;
    }

    return $this->base64encode($sEncryptedData);
  }
  /**
   * @param string $sData Données à décrypter
   * @return array|null donnée cryptées et encodées ou null en cas d'erreur
   */
  public function decryptData(string $sData): ?array{
    $sEncryptedData = $this->base64decode($sData);
    return $this->decrypt($sEncryptedData);
  }

  /**
   * @param string $sCrypted Token JWE encrypté
   * @return string donnée cryptée encodé en base 64 préparé pour passage dans une Url
   */
  private function base64encode(string $sCrypted): string {
    return rtrim(strtr(base64_encode($sCrypted), '+/', '-_'), '=');
  }
  /**
   * @param string $sData Chaine encryptées et encodés
   * @return string données cryptées
   */
  private function base64decode(string $sData): string {
    return base64_decode(str_pad(strtr($sData, '-_', '+/'), strlen($sData) % 4, '=', STR_PAD_RIGHT));
  }

  /**
   * Renvoi le token JWE contenant les données à stocker.
   *
   * @param array $aData donnée à encrypter
   *
   * @return string|null Token encrypté ou null en cas d'erreur
   */
  protected function encrypt(array $aData): ?string {
    // vérification présence de la clé
    if (empty($this->oJWK)) {
      return null;
    }
    // We instantiate our JWE Builder.
    $jweBuilder = new JWEBuilder(
      $this->oKeyEncryptionAlgorithmManager,
      $this->oContentEncryptionAlgorithmManager,
      $this->oCompressionMethodManager
    );
    $jwe = $jweBuilder
      ->create() // We want to create a new JWE
      ->withPayload(json_encode($aData)) // We set the payload
      ->withSharedProtectedHeader([
        'alg' => $this->oKeyEncryptionAlgorithm->name(), // Key Encryption Algorithm
        'enc' => $this->oContentEncryptionAlgorithm->name(), // Content Encryption Algorithm
        'zip' => $this->oCompressionMethod->name(), // We enable the compression (irrelevant as the payload is small, just for the example).
        'typ' => 'JWE', // We enable the compression (irrelevant as the payload is small, just for the example).
      ])
      ->addRecipient($this->oJWK) // We add a recipient (a shared key or public key).
      ->build(); // We build it

    $serializer = new CompactSerializer(); // The serializer
    return $serializer->serialize($jwe, 0);
  }

  /**
   * Renvoi un tableau contenant les données.
   *
   * @param string $sData Token JWE provenant d'une requête HTTP
   *
   * @return array|null Tableau contenant le payload du token ou null en cas d'erreur
   */
  protected function decrypt(string $sData): ?array{
    // vérification présence de la clé
    if (empty($this->oJWK)) {
      return null;
    }

    // We instantiate our JWE Decrypter.
    $jweDecrypter = new JWEDecrypter(
      $this->oKeyEncryptionAlgorithmManager,
      $this->oContentEncryptionAlgorithmManager,
      $this->oCompressionMethodManager
    );

    // The serializer manager. We only use the JWE Compact Serialization Mode.
    $serializerManager = new JWESerializerManager([
      new CompactSerializer(),
    ]);

    // We try to load the token.
    try {
      $jwe = $serializerManager->unserialize($sData);
      // We decrypt the token. This method does NOT check the header.
      if ($jweDecrypter->decryptUsingKey($jwe, $this->oJWK, 0)) {
        $aPayload = json_decode($jwe->getPayload(), true);
        return $aPayload;
      }
    } catch (InvalidArgumentException $e) {
      $this->oErrorLogger->error('Error with token structure : ' . $e->getMessage());
    } catch (InvalidClaimException $e) {
      $this->oErrorLogger->error('Error with token claim check : ' . $e->getMessage());
    }
    return null;
  }

}
