<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\CustomFieldsManager;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\DateField;
use Doctrine\ORM\EntityManager;
use Monolog\Logger;
use Ramsey\Uuid\Uuid;

/**
 * Permet d'utiliser les paramètres pour les requêtes de type GET.
 * attribute et filter essentiellement
 *
 * @author Anthony Borghi   <anthony.borghi@veremes.com>
 * @author Armand Bahi      <armand.bahi@veremes.com>
 * @author Ludivine Premont   <ludivine.premont@veremes.com>
 */

class CommonAccessorFunctions {
  protected const SQL_FILTER_PART = [
    "ORM" => [
      "CAST_INTEGER_PROJ" => 'CastInteger(:[SRID])',
      "CAST_FLOAT_BUFFER" => 'CastFloat(:[BUFFER])',
      "CAST_GEOGRAPHY_INTERSECT" => 'CastGeography(ST_GeomFromEWKT([GEOM]))',
      "CAST_GEOMETRY_FILTER" => 'CastGeometry(ST_Buffer([FN_PARAMS]))',
    ],
    "QUERY" => [
      "CAST_INTEGER_PROJ" => ':[SRID]::integer',
      "CAST_FLOAT_BUFFER" => ':[BUFFER]::float',
      "CAST_GEOGRAPHY_INTERSECT" => 'ST_GeomFromEWKT([GEOM])::geography',
      "CAST_GEOMETRY_FILTER" => 'ST_Buffer([FN_PARAMS])::geometry',
    ],
  ];

  /**
   * @var array ensemble des balises reconnu pour être utilisé comme fonction
   * uniquement avec les opérateur basique <,>,=
   */
  protected const SQL_FILTER_ALLOWED_FUNCTION = [
    "ORM" => [
      '@current_user' => 'CurrentUser()',
    ],
    "QUERY" => [
      '@current_user' => '"current_user"()',
    ],
  ];

  /**
   * @var string
   */
  private $sMode;
  /**
   * @var string
   */
  private $sEntityAlias = 'generic_entity';
  /**
   * @var string[]
   */
  private $aFields = [];
  /**
   * @var array
   */
  private $aFieldsConfig = [];
  /**
   * @var Logger
   */
  private $oLogger;
  /**
   * @var EntityManager
   */
  private $oEntityManager;
  /**
   * @var Route
   */
  private $oRouteConfigurator;
  /**
   * @var CustomFieldsManager
   */
  private $oCustomFieldsManager;
  /**
   * @var array|null
   */
  private $aFilter = null;

  /**
   * @param string $sEntityAlias nom générique à utiliser
   * @param Route $oRouteConfigurator Service de configuration de la route
   * @param Logger $oLogger Logger monolog à utiliser en cas d'erreur
   * @param string $sMode permet de définir un mode d'utilisation pour les contantes
   * @param EntityManager $oEntityManager Service d'accès aux entités
   */
  public function __construct(string $sEntityAlias, Route $oRouteConfigurator, Logger $oLogger, string $sMode = 'ORM', EntityManager $oEntityManager, CustomFieldsManager $oCustomFieldsManager) {
    $this->sEntityAlias = $sEntityAlias;
    $this->aFieldsConfig = $oRouteConfigurator->getFieldsList(true); // Liste des colonnes
    $this->oLogger = $oLogger;
    $this->sMode = $sMode;
    $this->oEntityManager = $oEntityManager;
    $this->oRouteConfigurator = $oRouteConfigurator;
    $this->oCustomFieldsManager = $oCustomFieldsManager;

    $this->loadFields();
  }

  /**
   * FieldsConfig getter
   * @return array
   */
  public function getFieldsConfig(): array{
    return $this->aFieldsConfig;
  }

  /**
   * Remplace le FieldsConfig de la route
   * @param array $aFieldsConfig Fields config
   * @return void
   */
  public function setFieldsConfig(array $aFieldsConfig): void {
    $this->aFieldsConfig = $aFieldsConfig;
    $this->loadFields();
  }

  /**
   * Remplace l'entity alias par défaut
   * @param string $sEntityAlias Entity alias
   * @return void
   */
  public function setEntityAlias(string $sEntityAlias): void {
    $this->sEntityAlias = $sEntityAlias;
  }

  /**
   * Charge aFields
   * @return void
   */
  private function loadFields(): void {
    $this->aFields = [];
    foreach ($this->aFieldsConfig as $aFieldConfig) {
      $this->aFields[] = $this->getFieldColumn($aFieldConfig['name']);
    }
  }

  /**
   * défini un filtre à concaténer au moment du decode
   * @param array|null $aFilter
   */
  public function setCommonFilter(?array $aFilter = null): void {
    $this->aFilter = $aFilter;
  }

  /**
   * retourne le nom de la colonne à partir du nom du champ dans route.json
   * @param string $sFieldName Nom du champ dont on veut le nom de la colonne
   * @return string|null
   */
  public function getFieldColumn(string $sFieldName): ?string {
    foreach ($this->aFieldsConfig as $aFieldConfig) {
      if ($aFieldConfig['name'] === $sFieldName) {
        return (isset($aFieldConfig['orm_col'])) ? $aFieldConfig['orm_col'] : $aFieldConfig['name'];
      }
    }
    return null;
  }

  /**
   * retourne le type de la colonne à partir du nom du champ dans route.json
   * @param string $sFieldName Nom du champ dont on veut le nom de la colonne
   * @return string|null
   */
  public function getFieldColumnType(string $sFieldName): ?string {
    foreach ($this->aFieldsConfig as $aFieldConfig) {
      if ($aFieldConfig['name'] === $sFieldName) {
        return $aFieldConfig['type'];
      }
    }
    return null;
  }

  /**
   * retourne la liste des champs à sélectionner
   * @param string $sAttributes Nom des champs à selectionner séparé par des |
   * @return string
   */
  public function attributeManager(string $sAttributes): string {
    if (!isset($this->aFields)) {
      $this->loadFields();
    }
    $aSelectedFields = $this->aFields;
    if (!empty($sAttributes)) {
      $aTmp = explode('|', $sAttributes);
      $aResultFields = [];
      foreach ($aTmp as $sField) {
        if (in_array($sField, $this->aFields)) {
          $aResultFields[] = $sField;
        }
      }

      if (!empty($aResultFields)) {
        $aSelectedFields = $aResultFields;
      }
    }
    // concaténation de la l'alias de la table devant chaque champ
    foreach ($aSelectedFields as &$sField) {
      $sField = $this->oCustomFieldsManager->getCustomFieldColumn($this->sEntityAlias, $sField, $this->getFieldColumnType($sField));
    }

    return implode(', ', $aSelectedFields);
  }

  /**
   * Analyse le filtre en paramètre et le filtre par défaut pour définir s'il faut filtrer
   * @param string|array|null $sJSONFilter Filtre à utiliser peut être une chaine json, une array ou null
   * @return bool true si un filtre est requis
   */
  public function needFilter($sJSONFilter = null): bool {
    if (is_array($sJSONFilter) || is_string($sJSONFilter)) {
      return true;
    }

    if (is_array($this->aFilter)) {
      return true;
    }

    return false;
  }

  /**
   * Transforme un filtre JSON en filtre DQL/SQL en tenant compte du filtre commun
   * @param string|array|null $sJSONFilter Filtre à utiliser peut être une chaine json, une array ou null
   * @param string $sSchema
   * @param string $sTable
   * @return array
   */
  public function prepareAndDecodeFilter($sJSONFilter = null, ?string $sSchema = null, ?string $sTable = null): array{
    // initialisation operateur
    $aOperator = [];
    // récupération du filtre de la route si défini
    if (!empty($this->aFilter)) {
      $aOperator[] = $this->aFilter;
    }

    // analyse du filtre venant de la requête si défini
    if (!empty($sJSONFilter)) {
      $aJSONFilter = is_string($sJSONFilter) ? json_decode($sJSONFilter, true) : $sJSONFilter;

      // Valeurs par défaut si juste un operator est passé en paramètre
      if (isset($aJSONFilter['column']) && isset($aJSONFilter['compare_operator'])) {
        $aJSONFilter = [
          'relation' => 'AND',
          'operators' => [$aJSONFilter],
        ];
      }

      if (empty($aJSONFilter['relation']) || !is_array($aJSONFilter['operators'])) {
        if (strlen(trim($sJSONFilter)) > 0) {
          $this->oLogger->error('FILTER: missing argument on ' . $sJSONFilter);
        }
        $aJSONFilter = null;
      }
      if (!empty($aJSONFilter)) {
        $aOperator[] = $aJSONFilter;
      }
    }

    // si un des deux opérateurs est présent on décode sinon on remonte un filtre vide
    if (!empty($aOperator)) {
      $aFinalFilter = [
        'relation' => 'AND',
        'operators' => $aOperator,
      ];

      return $this->decodeJSONFilter($aFinalFilter, $sSchema, $sTable);
    }

    return ['request' => '', 'params' => []];
  }

  /**
   * Transforme un filtre JSON en filtre DQL
   * @param string|array $sJSONFilter JSON filter, can be passed by string, object or array
   * @param string $sSchema
   * @param string $sTable
   * @return array
   */
  public function decodeJSONFilter($sJSONFilter, ?string $sSchema = null, ?string $sTable = null): array{
    $sFilter = '';
    $aParams = [];
    if (is_string($sJSONFilter)) {
      $aJSONFilter = json_decode($sJSONFilter, true);
    } else {
      $aJSONFilter = $sJSONFilter;
    }

    // Valeurs par défaut si juste un operator est passé en paramètre
    if (isset($aJSONFilter['column']) &&
      isset($aJSONFilter['compare_operator'])) {
      $aJSONFilter = [
        'relation' => 'AND',
        'operators' => [$aJSONFilter],
      ];
    }

    if (empty($aJSONFilter['relation']) ||
      !is_array($aJSONFilter['operators'])) {
      if (strlen(trim($sJSONFilter)) > 0) {
        $this->oLogger->error('FILTER: missing argument on ' . $sJSONFilter);
      }
      return ['request' => $sFilter, 'params' => $aParams];
    }

    if (is_array($aJSONFilter['operators'])) {
      // Vérifie relation
      $sRelation = $this->checkRelationOperator($aJSONFilter['relation']);
      if (empty($sRelation)) {
        $this->oLogger->error('FILTER: relation not valid');
        return ['request' => $sFilter, 'params' => $aParams];
      }

      // Pour chacun des opérateurs
      for ($i = 0; $i < count($aJSONFilter['operators']); $i++) {
        // Filtre imbriquée ?
        if (!empty($aJSONFilter['operators'][$i]['relation'])) {
          if (!empty($aJSONFilter['operators'][$i]['operators'])) {
            // Récupère récursivement le contenu du filtre imbriqué
            $aDecodedSubFilter = $this->decodeJSONFilter($aJSONFilter['operators'][$i], $sSchema, $sTable);

            // Ajout des paramètres
            foreach ($aDecodedSubFilter['params'] as $key => $value) {
              $aParams[$key] = $value;
            }

            // Ajout dans la requête
            if (strlen($sFilter) > 0) {
              $sFilter .= ' ' . $sRelation . ' ';
            } else {
              $sFilter .= '(';
            }
            $sFilter .= $aDecodedSubFilter['request'];
          }
        } else {
          $aOperator = $this->decodeOperator($aJSONFilter['operators'][$i], $sSchema, $sTable);

          if (!empty($aOperator)) {
            // Ajout des paramètres
            foreach ($aOperator['params'] as $key => $value) {
              $aParams[$key] = $value;
            }

            // Ajout dans la requête
            if (strlen($sFilter) > 0) {
              $sFilter .= ' ' . $sRelation . ' ';
            } else {
              $sFilter .= '(';
            }
            $sFilter .= $aOperator['request'];
          }
        }
      }
      if (strlen($sFilter) > 0) {
        $sFilter .= ')';
      }
    }

    return ['request' => $sFilter, 'params' => $aParams];
  }

  /**
   * Décode un opérateur et produit la partie filtre et la liste des paramètres
   * @param array $aOperator
   * @param string $sSchema
   * @param string $sTable
   * @return array|null :
   *                  - array[].request : partie de la clause where pour l'opérateur
   *                  - array[].params : paramètre à utiliser pour l'opérateur
   */
  private function decodeOperator($aOperator, $sSchema = null, $sTable = null): ?array{
    $sFilter = '';
    $aParams = [];

    $sColumnName = $this->getFieldColumn($aOperator['column']);
    $sColumnType = $this->getFieldColumnType($aOperator['column']);

    if (isset($aOperator['value'])) {
      $aOperator['value'] = $this->parseBools($aOperator['value']);

      if ($sColumnType === 'date') {
        $aOperator['value'] = $this->parseDates($aOperator['value']);
      }
    }

    // Vérifications
    if (empty($sColumnName) ||
      empty($aOperator['compare_operator'])) {
      return null;
    } else if (!in_array($sColumnName, $this->aFields)) {
      return null;
    }
    if (!isset($aOperator['value']) || $aOperator['value'] == '') {
      switch (strtoupper(trim($aOperator['compare_operator']))) {
        case 'NULL':
        case 'IS NULL':
        case 'NOT NULL':
        case 'IS NOT NULL':
          $aOperator['value'] = "";
          break;
        default:
          if (strlen(trim(json_encode($aOperator))) > 0) {
            $this->oLogger->error('FILTER: missing argument on ' . json_encode($aOperator));
          }
          return null;
      }
    }

    // Vérifie compare_operator
    $sCompare = $this->checkCompareOperator($aOperator['compare_operator']);
    if (empty($sCompare)) {
      $this->oLogger->error('FILTER: compare_operator not valid');
      return null;
    }

    // Clées à utiliser en guise de variables
    $sValueKey = 'value_' . str_replace('-', '_', Uuid::uuid1()->toString()); // Les - empèche le remplacement des binds

    switch ($sCompare) {
      case 'IS NULL':
      case 'IS NOT NULL':
        // Ajoute à la requête
        $sFilter .= $this->sEntityAlias . '.' . $sColumnName . ' ' . $sCompare;
        break;
      case 'IN':
        if (!is_array($aOperator['value'])) {
          $aOperator['value'] = [$aOperator['value']];
        }
        $qb = $this->oEntityManager->getConnection()->createQueryBuilder();
        $oExpr = $qb->expr();
        // protection des valeur numéric de type 01256
        foreach ($aOperator['value'] as &$mValue) {
          $mValue = $oExpr->literal($mValue);
        }

        $sFilter .= $oExpr->in($this->sEntityAlias . '.' . $sColumnName, $aOperator['value']);
        break;
      case 'NOT IN':
        if (!is_array($aOperator['value'])) {
          $aOperator['value'] = [$aOperator['value']];
        }
        $qb = $this->oEntityManager->getConnection()->createQueryBuilder();
        $oExpr = $qb->expr();
        // protection des valeur numéric de type 01256
        foreach ($aOperator['value'] as &$mValue) {
          $mValue = $oExpr->literal($mValue);
        }

        $sFilter .= $oExpr->notIn($this->sEntityAlias . '.' . $sColumnName, $aOperator['value']);
        break;
      case 'LIKE':
        // Ajoute les paramètres
        $aParams[$sValueKey]['value'] = $aOperator['value'];

        $sColumnFilter = $this->addColumnQuotes($this->sEntityAlias) . '.' . $this->addColumnQuotes($sColumnName);
        $sValueFilter = ':' . $sValueKey;

        if (!empty($aOperator['compare_operator_options'])) {
          if ($aOperator['compare_operator_options']['case_insensitive'] == 'true' ||
            $aOperator['compare_operator_options']['case_insensitive'] == true) {
            $sValueFilter = 'LOWER(' . $sValueFilter . ')';
            $sColumnFilter = 'LOWER(' . $sColumnFilter . ')';
          }
        }

        // Ajoute à la requête
        $sFilter .= $sColumnFilter . ' LIKE ' . $sValueFilter;

        break;
      case 'INTERSECT':
        $sGeomDefKey = 'geom_def_' . str_replace('-', '_', Uuid::uuid1()->toString());

        $aParams[$sGeomDefKey]['value'] = $aOperator['value'];

        $sSQLGeom = ":" . $sGeomDefKey;
        $sSQLColumn = $this->addColumnQuotes($this->sEntityAlias) . '.' . $this->addColumnQuotes($sColumnName);

        $aPostgisIntersectWithoutBufferTypes = [];

        // compare_operator_options
        if (!empty($aOperator['compare_operator_options'])) {
          // source_proj
          if (!empty($aOperator['compare_operator_options']['source_proj'])) {
            $sSourceProjKey = 'source_proj_' . str_replace('-', '_', Uuid::uuid1()->toString());
            $aParams[$sSourceProjKey]['value'] = (int) $aOperator['compare_operator_options']['source_proj'];
            $aParams[$sSourceProjKey]['type'] = \Doctrine\DBAL\Types\Type::INTEGER;
            // charge la fonction pour caster depuis le tableau de const
            $sSQLSRID = str_replace('[SRID]', $sSourceProjKey, self::SQL_FILTER_PART[$this->sMode]['CAST_INTEGER_PROJ']);
            $sSRID = $aOperator['compare_operator_options']['source_proj'];
          }
          // intersect_buffer
          if (!empty($aOperator['compare_operator_options']['intersect_buffer'])) {
            $sIntersectBufferKey = 'intersect_buffer' . str_replace('-', '_', Uuid::uuid1()->toString());
            $aParams[$sIntersectBufferKey]['value'] = (float) $aOperator['compare_operator_options']['intersect_buffer'];
            $aParams[$sIntersectBufferKey]['type'] = \Doctrine\DBAL\Types\Type::FLOAT;
            $sSQLBuffer = str_replace('[BUFFER]', $sIntersectBufferKey, self::SQL_FILTER_PART[$this->sMode]['CAST_FLOAT_BUFFER']);
          }
          // intersect_buffer_geom_type
          if (!empty($aOperator['compare_operator_options']['intersect_buffer_geom_type'])) {
            $aIntersectTypes = explode('|', strtoupper($aOperator['compare_operator_options']['intersect_buffer_geom_type']));

            // Types necessitant un buffer
            $aTypesWithBuffer = [];
            for ($i = 0; $i < count($aIntersectTypes); $i++) {
              switch ($aIntersectTypes[$i]) {
                case 'POINT':
                  array_push($aTypesWithBuffer, 'ST_Point');
                  array_push($aTypesWithBuffer, 'ST_MultiPoint');
                  break;
                case 'LINE':
                  array_push($aTypesWithBuffer, 'ST_Linestring');
                  array_push($aTypesWithBuffer, 'ST_MultiLinestring');
                  break;
                case 'POLYGON':
                  array_push($aTypesWithBuffer, 'ST_Polygon');
                  array_push($aTypesWithBuffer, 'ST_MultiPolygon');
                  break;
                default:
                  break;
              }
            }

            // Types qui ne necessitent pas de buffer
            $aExcludeBufferTypes = [
              'ST_Point',
              'ST_MultiPoint',
              'ST_Linestring',
              'ST_MultiLinestring',
              'ST_Polygon',
              'ST_MultiPolygon',
            ];
            for ($i = 0; $i < count($aTypesWithBuffer); $i++) {
              if (array_search($aTypesWithBuffer[$i], $aExcludeBufferTypes) !== false) {
                array_splice($aExcludeBufferTypes, array_search($aTypesWithBuffer[$i], $aExcludeBufferTypes), 1);
              }
            }

            // Ajout dans les paramètres
            for ($ii = 0; $ii < count($aExcludeBufferTypes); $ii++) {
              $sPostgisGeomTypeKey = 'postgis_geom_type_' . str_replace('-', '_', Uuid::uuid1()->toString()) . $ii;
              $aParams[$sPostgisGeomTypeKey]['value'] = $aExcludeBufferTypes[$ii];
              array_push($aPostgisIntersectWithoutBufferTypes, ':' . $sPostgisGeomTypeKey);
            }
          }
        }

        // Si le SRID source n'est pas renseigné
        if (empty($sSQLSRID)) {
          if (!empty($sSchema) && !empty($sTable)) {
            $oGeomFileManager = $this->oCustomFieldsManager->getGeomFieldManager();
            $sSRID = $oGeomFileManager->getColumnSRID($sSchema, $sTable, $sColumnName);
            $sSourceProjKey = 'source_proj_' . str_replace('-', '_', Uuid::uuid1()->toString());
            $aParams[$sSourceProjKey]['value'] = $sSRID;
            $sSQLSRID = ':' . $sSourceProjKey;
          } else {
            $this->oLogger->error('FILTER: missing $sSchema and $sTable params');
            break;
          }
        }

        if (empty($sSRID)) {
          $sSRID = 2154;
        }
        // Géométrie
        if ($sSRID === '4326') {
          $sIntersectFilter = str_replace('[GEOM]', $sSQLGeom, self::SQL_FILTER_PART[$this->sMode]['CAST_GEOGRAPHY_INTERSECT']);
        } else {
          $sIntersectFilter = 'ST_GeomFromEWKT(' . $sSQLGeom . ')';
        }

        // Buffer de sélection
        if (!empty($sSQLBuffer)) {
          // Mise en place du buffer
          if ($sSRID === '4326') {
            $sBufferFilter = str_replace('[FN_PARAMS]', $sIntersectFilter . ', ' . $sSQLBuffer, self::SQL_FILTER_PART[$this->sMode]['CAST_GEOMETRY_FILTER']);
          } else {
            $sBufferFilter = 'ST_Buffer(' . $sIntersectFilter . ', ' . $sSQLBuffer . ')';
          }
        }
        // Reprojection
        if (!empty($sSQLSRID)) {
          $sIntersectFilter = 'ST_Transform(' . $sIntersectFilter . ', ' . $sSQLSRID . ')';
          if (!empty($sBufferFilter)) {
            $sBufferFilter = 'ST_Transform(' . $sBufferFilter . ', ' . $sSQLSRID . ')';
          }
        }
        // Intersection
        if (!empty($sSQLColumn)) {
          $sIntersectFilter = 'ST_Intersects(' . $sIntersectFilter . ', ' . $sSQLColumn . ')';
          if (!empty($sBufferFilter)) {
            $sBufferFilter = 'ST_Intersects(' . $sBufferFilter . ', ' . $sSQLColumn . ')';
          }
        }

        // Buffer de sélection
        if (!empty($sBufferFilter)) {
          if (count($aPostgisIntersectWithoutBufferTypes) > 0) {
            $sNoBufferTypesFilter = '';
            for ($i = 0; $i < count($aPostgisIntersectWithoutBufferTypes); $i++) {
              if ($i > 0) {
                $sNoBufferTypesFilter .= ' AND ';
              }
              $sNoBufferTypesFilter .= 'ST_GeometryType(' . $sSQLColumn . ')!=' . $aPostgisIntersectWithoutBufferTypes[$i] . '';
            }
            $sIntersectFilter = '(' . $sIntersectFilter . ' = TRUE) OR ((' . $sNoBufferTypesFilter . ') AND ' . $sBufferFilter . ' = TRUE)';
          } else {
            $sIntersectFilter = '(' . $sIntersectFilter . ' = TRUE) OR (' . $sBufferFilter . ' = TRUE)';
          }
        }

        // Ajout au filtre
        $sFilter .= '(' . $sIntersectFilter . ')';
        break;
      default:
        // Ajoute les paramètres
        $sFilter .= $this->addColumnQuotes($this->sEntityAlias) . '.' . $this->addColumnQuotes($sColumnName) . ' ' . $sCompare . ' ';
        // détection des fonction utilisable à laplace des valeurs
        if (isset(self::SQL_FILTER_ALLOWED_FUNCTION[$this->sMode][$aOperator['value']])) {
          $sFilter .= self::SQL_FILTER_ALLOWED_FUNCTION[$this->sMode][$aOperator['value']];
        } else {
          $sFilter .= ':' . $sValueKey;
          $aParams[$sValueKey]['value'] = $aOperator['value'];
        }
        break;
    }

    // Infos complémentaires
    if (array_key_exists($sValueKey, $aParams) &&
      !empty($aParams[$sValueKey]['value'])) {
      $aParams[$sValueKey]['column_name'] = $sColumnName;
      $aParams[$sValueKey]['column_type'] = $sColumnType;
    }

    return [
      'request' => $sFilter,
      'params' => $aParams,
    ];
  }

  /**
   * Check the given relation operator and return
   * a well formed relation operator string
   * @param string $sRelationOperator
   * @return string
   */
  private function checkRelationOperator(string $sRelationOperator): string {
    $sRelation = null;
    switch (strtoupper(trim($sRelationOperator))) {
      case 'AND':
        $sRelation = 'AND';
        break;
      case 'OR':
        $sRelation = 'OR';
        break;
      default:
        break;
    }
    return $sRelation;
  }

  /**
   * Change une variable en booléen
   * @param mixed $value
   * @return mixed
   */
  private function parseBools($value) {
    if (is_bool($value)) {
      $value = $this->boolToString($value);
    }
    return $value;
  }

  /**
   * Change une variable en date
   * @param int $value
   * @return string|null
   */
  private function parseDates($value): ?string {
    if ($value > 0) {
      $value = date(DateField::DEFAULT_FORMAT, $value);
    } else {
      $value = null;
    }

    return $value;
  }

  /**
   * Change un variable booléene en string
   * @param bool $boolVal
   * @return string
   */
  private function boolToString(bool $boolVal = false): string {
    return ($boolVal ? 'true' : 'false');
  }

  /**
   * Check the given compare operator and return
   * a well formed compare operator string
   * @param string $sCompareOperator
   * @return string|null
   */
  private function checkCompareOperator(string $sCompareOperator): ?string{
    $sCompare = null;
    switch (strtoupper(trim($sCompareOperator))) {
      case '=':
        $sCompare = '=';
        break;
      case '!=':
        $sCompare = '!=';
        break;
      case '<>':
        $sCompare = '<>';
        break;
      case '>=':
        $sCompare = '>=';
        break;
      case '<=':
        $sCompare = '<=';
        break;
      case '>':
        $sCompare = '>';
        break;
      case '<':
        $sCompare = '<';
        break;
      case 'IN':
        $sCompare = 'IN';
        break;
      case 'NOT IN':
        $sCompare = 'NOT IN';
        break;
      case 'NULL':
      case 'IS NULL':
        $sCompare = 'IS NULL';
        break;
      case 'NOT NULL':
      case 'IS NOT NULL':
        $sCompare = 'IS NOT NULL';
        break;
      case 'LIKE':
        $sCompare = 'LIKE';
        break;
      case 'INTERSECT':
        $sCompare = 'INTERSECT';
        break;
      default:
        break;
    }
    return $sCompare;
  }

  /**
   * @param array $aParams Paramètre provenant de la requête HTTP qui doivent être préparé pour la requête SQL
   * @return array valeur à préparer pour le requête SQL (Insert/Update)
   */
  public function sqlValuesExtractor(array $aParams): array{
    $aParamsToInsert = ['values' => [], 'parameters' => [], 'functions' => []];
    // si un paramètre de la requête est défini comme un des champs à utiliser ne base on le bind
    foreach ($this->oRouteConfigurator->getFieldsList(false) as $sField) {
      if (array_key_exists($sField, $aParams)) {
        $aParamsToInsert['values'][$sField] = ':' . $sField;
        $aParamsToInsert['parameters'][$sField] = $this->parseBools($aParams[$sField]);

        // Utilisé pour lancer des fonctions
        $sColumnFunction = $this->oCustomFieldsManager->getCustomFieldColumnFunction($this->sEntityAlias, $sField, $this->getFieldColumnType($sField));
        if (!empty($sColumnFunction)) {
          $aParamsToInsert['functions'][$sField] = $sColumnFunction;
        }
      }
    }
    return $aParamsToInsert;
  }

  /**
   * Ajoute les double quotes aux tables, shcemas, colonnes si besoin
   */
  private function addColumnQuotes(string $skey): string {
    if ($this->sMode === 'ORM') {
      return $skey;
    } else {
      return '"' . $skey . '"';
    }
  }
}
