<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor\Fields;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\Crypto\JweCrypto;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\ArrayField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\BooleanField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\NumberField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\DateField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\FieldInterface;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\FileField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\GeomField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\PasswordField;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\ORM\EntityManager;

/**
 * Classe qui va gérer le remplacement de tous les champs.
 *
 * @author Anthony Borghi   <anthony.borghi@veremes.com>
 */
class CustomFieldsManager {

  /**
   * @var Properties
   */
  private $oProperties = null;
  /**
   * @var Route
   */
  private $oRouteConfig = null;
  /**
   * @var LoggerManager
   */
  private $oLoggerManager = null;
  /**
   * @var array
   */
  private $aOptions = [];
  /**
   * @var FieldInterface[]
   */
  private $aCustomFieldReplacer = [];
  /**
   * @var bool
   */
  private $bDataModified = false;
  /**
   * @var JweCrypto
   */
  private $oJweCrypto;
  /**
   * @var EntityManager
   */
  private $oEntityManager;
  /**
   * @param Properties $oPropertiesSrvc Un objet pour accéder aux properties de l'application
   * @param Route $oRoute Service permettant d'accéder à la configuration d'une route
   * @param LoggerManager $oLoggerManager service d'accès aux logs
   * @param JweCrypto $oJweCrypto service de crypto
   * @param EntityManager $oEntityManager EntityManager symfony
   * @throws \Exception
   */
  public function __construct(Properties $oPropertiesSrvc, Route $oRoute, LoggerManager $oLoggerManager, JweCrypto $oJweCrypto, EntityManager $oEntityManager) {
    $this->oProperties = $oPropertiesSrvc;
    $this->oRouteConfig = $oRoute;
    $this->oLoggerManager = $oLoggerManager;
    $this->oJweCrypto = $oJweCrypto;
    $this->oEntityManager = $oEntityManager;

    $aRouteConfig = $this->oRouteConfig->getRequestParameters();
    if (isset($aRouteConfig['options'])) {
      $this->aOptions = $aRouteConfig['options'];
    }
  }

  /**
   * Remplace les champ sucstom lors de la récupération de donnée
   * @param array $aEntity Retour de la base de données
   * @param array $aRequestData données provenant de la requête
   * @return array|null données modifiées ou null en cas d'erreur
   */
  public function getCustomFields(array $aEntity, array $aRequestData): ?array{
    return $this->manageCustomFields($aEntity, $aRequestData, 'getFieldValue');
  }

  /**
   * Remplace les champ sucstom avant l'insertion de donnée
   * @param array $aEntity Retour de la base de données
   * @param array $aRequestData données provenant de la requête
   * @return array|null données modifiées ou null en cas d'erreur
   */
  public function insertCustomFields(array $aEntity, array $aRequestData): ?array{
    return $this->manageCustomFields($aEntity, $aRequestData, 'insertFieldValue');
  }
  /**
   * Remplace les champ sucstom aprés l'insertion de donnée
   * @param array $aEntity Retour de la base de données
   * @param array $aRequestData données provenant de la requête
   * @return array|null données modifiées ou null en cas d'erreur
   */
  public function afterInsertCustomFields(array $aEntity, array $aRequestData): ?array{
    return $this->manageCustomFields($aEntity, $aRequestData, 'afterInsertFieldValue');
  }

  /**
   * Remplace les champ sucstom aprés l'insertion de donnée
   * @param array $aEntity Retour de la base de données
   * @param array $aRequestData données provenant de la requête
   * @return array|null données modifiées ou null en cas d'erreur
   */
  public function updateCustomFields(array $aEntity, array $aRequestData): ?array{
    return $this->manageCustomFields($aEntity, $aRequestData, 'updateFieldValue');
  }
  /**
   * Remplace les champ sucstom aprés l'insertion de donnée
   * @param array $aEntity Retour de la base de données
   * @param array $aRequestData données provenant de la requête
   * @return array|null données modifiées ou null en cas d'erreur
   */
  public function updateMassiveCustomFields(array $aEntity, array $aRequestData): ?array{
    return $this->manageCustomFields($aEntity, $aRequestData, 'updateMassiveFieldValue');
  }

  /**
   * Remplace les champ sucstom aprés l'insertion de donnée
   * @param array $aEntity Retour de la base de données
   * @param array $aRequestData données provenant de la requête
   * @return array|null données modifiées ou null en cas d'erreur
   */
  public function deleteCustomFields(array $aEntity, array $aRequestData): ?array{
    return $this->manageCustomFields($aEntity, $aRequestData, 'deleteFieldValue');
  }

  /**
   * Retourne l'expression utilisée en tant que colonne dans le requête GET
   * @param string $sSchema Schéma
   * @param string $sField Table
   * @param string $sType Type défini dans le JSON
   * @return string chaine à utiliser dans la requête
   */
  public function getCustomFieldColumn(string $sSchema, string $sField, string $sType): string {

    // s'il n'y a pas de formatter on le récupère
    if (!isset($this->aCustomFieldReplacer[$sType])) {
      $this->aCustomFieldReplacer[$sType] = $this->loadCustomFieldFormatter($sType);
    }

    // si le formatter est une instance de FieldInterface on formate le champ
    if ($this->aCustomFieldReplacer[$sType] instanceof FieldInterface) {
      return $this->aCustomFieldReplacer[$sType]->getFieldColumn($sSchema, $sField);
    } else {
      return $sSchema . '.' . $sField;
    }
  }

  /**
   * Retourne la fonction utilisée en tant que colonne dans le requête PUT ou insert
   * @param string $sSchema Schéma
   * @param string $sField Table
   * @param string $sType Type défini dans le JSON
   * @return string|null chaine à utiliser dans la requête
   */
  public function getCustomFieldColumnFunction(string $sSchema, string $sField, string $sType): ?string {

    // s'il n'y a pas de formatter on le récupère
    if (!isset($this->aCustomFieldReplacer[$sType])) {
      $this->aCustomFieldReplacer[$sType] = $this->loadCustomFieldFormatter($sType);
    }

    // si le formatter est une instance de FieldInterface on formate le champ
    if ($this->aCustomFieldReplacer[$sType] instanceof FieldInterface) {
      return $this->aCustomFieldReplacer[$sType]->getFieldColumnFunction($sField);
    } else {
      return null;
    }
  }

  /**
   * Remplace les champ custom
   * @param array $aEntity Retour de la base de données
   * @param array $aRequestData données provenant de la requête
   * @param string $sFnName Nom de la fonction à appeler
   * @throws \Exception si le json de configuration de la route manque de données
   * @return array|null données modifiées ou null en cas d'erreur
   */
  private function manageCustomFields(array $aEntity, array $aRequestData, string $sFnName): ?array{
    $this->bDataModified = false;
    if (isset($this->aOptions['fields'])) {
      // pour chaque champ spécifique
      foreach ($this->aOptions['fields'] as $aField) {
        // si la colonne n'a pas de type on lève une exception
        if (!isset($aField['name'])) {
          throw new \Exception('Cannot get field\'s name for route : ' . json_encode($this->aOptions));
        }
        if (!isset($aField['type'])) {
          throw new \Exception('Cannot get field\'s type for field : ' . json_encode($aField));
        }
        // s'il n'y a pas de formatter on le récupère
        if (!isset($this->aCustomFieldReplacer[$aField['type']])) {
          $this->aCustomFieldReplacer[$aField['type']] = $this->loadCustomFieldFormatter($aField['type']);
        }
        // si le formatter est une instance de FieldInterface on formate le champ
        if ($this->aCustomFieldReplacer[$aField['type']] instanceof FieldInterface) {
          // défini s'il faut forcer la valeur
          if ($this->aCustomFieldReplacer[$aField['type']]->forceAttribute($sFnName)) {
            $aEntity[$aField['name']] = $aEntity[$aField['name']] ?? $this->aCustomFieldReplacer[$aField['type']]->getForceValue();
          }
          // si le champ est défini dans les données d'origine
          if ($this->aCustomFieldReplacer[$aField['type']]->fieldIsSet($aEntity, $aField)) {
            // récupération de la valeur avant modif pour controle aprés
            $mOldValue = $aEntity[$aField['name']] ?? null;
            $aEntity = $this->aCustomFieldReplacer[$aField['type']]->$sFnName($aEntity, $aRequestData, $aField);
            $mNewValue = $aEntity[$aField['name']] ?? null;
            // si le champs est set et que sa valeur a changé on chanhe le paramètre de controle
            //if (isset($aEntity[$aField['name']])) {
            $this->bDataModified = ($mOldValue === $mNewValue) ? $this->bDataModified : true;
            //}
          }
        }
      }
    }
    return $aEntity;
  }

  /**
   * Charge les classes pour gérer chaque type d'attribut dans la requête
   * @return FieldInterface|null retourne un formatter implémentant FieldInterface ou null si type ne nécessite pas de champs custom
   */
  private function loadCustomFieldFormatter(string $sType): ?FieldInterface {
    switch ($sType) {
      case FileField::TYPE:
        // remplacement du champ éxistant par une url getProxyPassUrl, enregistrement/suppression des fichiers et miniatures
        return new FileField($this->oProperties, $this->oRouteConfig, $this->oLoggerManager, $this->oJweCrypto);
      case BooleanField::TYPE:
        // remplacement du champ éxistant par une valeur boolean
        return new BooleanField($this->oProperties, $this->oRouteConfig, $this->oLoggerManager);
      case NumberField::TYPE:
        // remplacement du champ éxistant par une valeur number
        return new NumberField($this->oProperties, $this->oRouteConfig, $this->oLoggerManager);
      case PasswordField::TYPE:
        // remplacement du champ éxistant par ***
        return new PasswordField($this->oProperties, $this->oRouteConfig, $this->oLoggerManager);
      case DateField::TYPE:
        // remplacement du champ éxistant par un timestamp
        return new DateField($this->oProperties, $this->oRouteConfig, $this->oLoggerManager);
      case GeomField::TYPE:
        // remplacement du format par du EWKT
        return new GeomField($this->oProperties, $this->oRouteConfig, $this->oLoggerManager, $this->oEntityManager);
      case ArrayField::TYPE:
        // gestiond des jointures
        return new ArrayField($this->oProperties, $this->oRouteConfig, $this->oLoggerManager, $this->oEntityManager);
    }
    return null;
  }
  /**
   * Remplace les champ custom lors de la récupération de donnée
   * @return bool détermine si les données ont été modifiée (true si un champs custom a modifié les données false sinon)
   */
  public function dataIsModified(): bool {
    return $this->bDataModified;
  }

  /**
   * retourge une classe de gestion des champs file
   * @return FileField|null
   */
  public function getFileFieldManager(): ?FileField {
    $oField = $this->loadCustomFieldFormatter(FileField::TYPE);
    return ($oField instanceof FileField) ? $oField : null;
  }
  /**
   * retourge une classe de gestion des champs bool
   * @return BooleanField|null
   */
  public function getBooleanFieldManager(): ?BooleanField {
    $oField = $this->loadCustomFieldFormatter(BooleanField::TYPE);
    return ($oField instanceof BooleanField) ? $oField : null;
  }
  /**
   * retourge une classe de gestion des champs password
   * @return PasswordField|null
   */
  public function getPasswordFieldManager(): ?PasswordField {
    $oField = $this->loadCustomFieldFormatter(PasswordField::TYPE);
    return ($oField instanceof PasswordField) ? $oField : null;
  }
  /**
   * retourge une classe de gestion des champs date
   * @return DateField|null
   */
  public function getDateFieldManager(): ?DateField {
    $oField = $this->loadCustomFieldFormatter(DateField::TYPE);
    return ($oField instanceof DateField) ? $oField : null;
  }
  /**
   * retourge une classe de gestion des champs geom
   * @return GeomField|null
   */
  public function getGeomFieldManager(): ?GeomField {
    $oField = $this->loadCustomFieldFormatter(GeomField::TYPE);
    return ($oField instanceof GeomField) ? $oField : null;
  }
  /**
   * retourge une classe de gestion des champs array
   * @return ArrayField|null
   */
  public function getArrayFieldManager(): ?ArrayField {
    $oField = $this->loadCustomFieldFormatter(ArrayField::TYPE);
    return ($oField instanceof ArrayField) ? $oField : null;
  }

  /**
   * Permet de forcer de nouvelles options
   * @param array $aOptions nouvel config à appliquer
   * @return void
   */
  public function setOptions(array $aOptions): void {
    $this->aOptions = $aOptions;
  }
}
