<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor\Fields;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\AbstractCustomField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\FieldInterface;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;

/**
 * Classe abstraite pour gérer un champ de type password.
 * ne fait que des return sans modification.
 *
 *  Configuration du champ
 *    "type"      (string) : prend toujours la valeur 'password' pour ce type
 *    "name"      (string) : Nom du champ
 *    "mode"      (string) : valeurs possible : unchanged, nullify (default : nullify)
 *  exemple :
 *    {
 *      "type": "password",
 *      "name": "my_field",
 *      "mode": "nullify"
 *    }
 *
 * @author Anthony Borghi   <anthony.borghi@veremes.com>
 */

class PasswordField extends AbstractCustomField implements FieldInterface {
  /**
   * @var string
   * Valeur du champs type dans route.json
   */
  public const TYPE = 'password';
  /**
   * @var string
   */
  public const MODE_UNCHANGED = 'unchanged';
  /**
   * @var string
   */
  public const MODE_NULLIFY = 'nullify';
  /**
   * @var string
   */
  public const MODE_STARIFICATION = 'star';
  /**
   * @var string
   */
  private const STARIFICATION_CHAR = '****';
  /**
   * @var string
   */
  private const MODE_DEFAULT = self::MODE_NULLIFY;
  /**
   * @var string
   */
  public const DEFAULT_GENERATOR_NUMBER = '0123456789';
  /**
   * @var string
   */
  public const DEFAULT_GENERATOR_LOWER_CHARS = 'abcdefghijklmnopqrstuvwxyz';
  /**
   * @var string
   */
  public const DEFAULT_GENERATOR_UPPER_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  /**
   * @var string
   */
  public const DEFAULT_GENERATOR_SPECIALS = '!@#$%*&~{}[]()/+-';

  /**
   * {@inheritdoc}
   */
  public function __construct(Properties $oPropertiesSrvc, Route $oRoute, LoggerManager $oLoggerManager) {
    parent::__construct($oPropertiesSrvc, $oRoute, $oLoggerManager);
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    $sMode = isset($aFieldOptions['mode']) ? $aFieldOptions['mode'] : self::MODE_DEFAULT;

    switch ($sMode) {
      case self::MODE_UNCHANGED:
        break;
      case self::MODE_STARIFICATION:
        $aEntity[$aFieldOptions['name']] = self::STARIFICATION_CHAR;
        break;
      case self::MODE_NULLIFY:
      default:
        $aEntity[$aFieldOptions['name']] = null;
        break;
    }

    return $aEntity;
  }

  /**
   * {@inheritdoc}
   */
  public function updateFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    $sMode = isset($aFieldOptions['mode']) ? $aFieldOptions['mode'] : self::MODE_DEFAULT;
    switch ($sMode) {
      case self::MODE_STARIFICATION:
        // si la chaine est le caractère de dissimulation on unset
        if ($aEntity[$aFieldOptions['name']] === self::STARIFICATION_CHAR) {
          unset($aEntity[$aFieldOptions['name']]);
        }
        break;
    }
    return $aEntity;
  }
  /**
   * {@inheritdoc}
   */
  public function updateMassiveFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    return $this->updateFieldValue($aEntity, $aRequestData, $aFieldOptions);
  }

  /**
   * génére un mot de passe en respectant certaines contraintes
   *
   * @param  int    $iLength Longueur du mot de passe
   * @param  int    $numbers nombre minimum de nombre
   * @param  int    $upper nombre minimum de majuscule
   * @param  int    $lower nombre minimum de minuscule
   * @param  int    $specials nombre minimum de caractères spéciaux (si 0 ne sera pas utilisé dans le filler)
   * @param  string $sExclude caractères à ne pas utiliser
   * @throws \Exception si le nombre de caractères en contrainte est supérieu à la longueur du mot de passe
   * @return string
   */
  public static function generate(int $iLength = 12, int $numbers = 1, int $upper = 1, int $lower = 1, int $specials = 1, string $sExclude = ''): string {
    $password = [];

    if ($numbers + $upper + $lower + $specials > $iLength) {
      throw new \Exception('Requirement greater than length');
    }

    $sNumberList = self::DEFAULT_GENERATOR_NUMBER;
    $sCUpperList = self::DEFAULT_GENERATOR_UPPER_CHARS;
    $sCLowerList = self::DEFAULT_GENERATOR_LOWER_CHARS;
    $sSpecialList = self::DEFAULT_GENERATOR_SPECIALS;

    if (!empty($sExclude)) {
      $aExclude = str_split($sExclude);
      $sNumberList = str_replace($aExclude, '', $sNumberList);
      $sCUpperList = str_replace($aExclude, '', $sCUpperList);
      $sCLowerList = str_replace($aExclude, '', $sCLowerList);
      $sSpecialList = str_replace($aExclude, '', $sSpecialList);
    }

    for ($i = 0; $i < $numbers; $i++) {
      $password[] = self::getRandomCharInstring($sNumberList);
    }
    for ($i = 0; $i < $upper; $i++) {
      $password[] = self::getRandomCharInstring($sCUpperList);
    }
    for ($i = 0; $i < $lower; $i++) {
      $password[] = self::getRandomCharInstring($sCLowerList);
    }
    for ($i = 0; $i < $specials; $i++) {
      $password[] = self::getRandomCharInstring($sSpecialList);
    }

    $remaining = $iLength - count($password);
    $sAll = $sNumberList . $sCUpperList . $sCLowerList;

    if ($specials > 0) {
      $sAll .= $sSpecialList;
    }

    for ($i = 0; $i < $remaining; $i++) {
      $password[] = self::getRandomCharInstring($sAll);
    }

    // on mélange le tableau de char
    shuffle($password);
    // on implode
    return implode('', $password);
  }

  /**
   * retourne un cractère au hasard dans une chaine
   *
   * @param  string $sString
   * @return string
   */
  public static function getRandomCharInstring(string $sString): string {
    return $sString[random_int(0, strlen($sString) - 1)];
  }
}
