<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\CommonAccessorFunctions;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use Doctrine\ORM\EntityManager;
use Monolog\Logger;

/**
 * Permet d'accéder aux données d'une connexion Oracle
 *
 * @author Ludivine Premont   <ludivine.premont@veremes.com>
 */
class OracleAccessor extends AbstractDataAccessor implements DataAccessorInterface {
  /**
   * @var string
   */
  protected const ENTITY_ALIAS = 'generic_row';
  /**
   * @var string
   */
  protected const MODE = 'QUERY';
  /**
   * @var string
   */
  protected const LOGGER_KEY = 'SQL_ERROR';
  /**
   * @var string
   */
  protected const RETURNING_STATEMENT = ' RETURNING *;';

  /**
   * Le nom du compteur des enregistrements
   * @var string
   */
  private $sRecordsCount;
  /**
   * Le nom de la table
   * @var string
   */
  private $sTableName;

  /**
   * @param EntityManager $oEntityManager Gestionnaire d'entité Doctrine
   * @param Route $oRouteConfigurator service de configuration de la route
   * @param VitisSerializer $oSerializer Objet permettant la désearialization dans une couche ORM
   * @param LoggerManager $oLogger Service d'accès aux logger
   * @param Properties $oProperties Service d'accès aux properties
   * @param JweTokenManager $oJweTokenManager accès au Token et encryptage
   */
  public function __construct(EntityManager $oEntityManager, Route $oRouteConfigurator, VitisSerializer $oSerializer, LoggerManager $oLogger, Properties $oProperties, JweTokenManager $oJweTokenManager) {
    parent::__construct($oEntityManager, $oRouteConfigurator, $oSerializer, $oLogger, $oProperties, $oJweTokenManager);

    $this->sRecordsCount = self::ENTITY_ALIAS;
    if (isset($this->oOptions['field_id'])) {
      $this->sRecordsCount .= '.' . $this->oOptions['field_id'];
    }

    if (isset($this->oOptions['table'])) {
      $this->sTableName = $this->oOptions['table'];
      if (isset($this->oOptions['schema'])) {
        $this->sTableName = $this->oOptions['schema'] . '.' . $this->oOptions['table'];
      }
    }
  }

  /**
   * Permet de récupérer les tables d'un schema d'une base de données locale
   * @param mixed $aData
   * @return array
   */
  public function getTables($aData) {
    $qb = $this->oEntityManager->getConnection()->createQueryBuilder();

    $qb->select('tables.OBJECT_NAME AS "table_name"')
      ->from('ALL_OBJECTS', 'tables')
      ->where($qb->expr()->eq('tables.OWNER', ':OWNER'))
      ->andWhere($qb->expr()->orX(
        $qb->expr()->eq('tables.OBJECT_TYPE', ':OBJECT_TABLE'),
        $qb->expr()->eq('tables.OBJECT_TYPE', ':OBJECT_VIEW')
      )
      )
      ->setParameter('OWNER', $aData['schema'])
      ->setParameter('OBJECT_TABLE', 'TABLE')
      ->setParameter('OBJECT_VIEW', 'VIEW');

    if (isset($aData['order_by'])) {
      $qb->orderBy('"' . $aData['order_by'] . '"');
    }

    try {
      $oTables = $qb->execute()->fetchAll();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get tables: ' . $e->getMessage());
      return ['status' => 0];
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oTables as $oTable) {
      $response['data'][] = $oTable['table_name'];
    }

    return $response;
  }

  /**
   * Permet de récupérer les columns d'une table pour la configuration
   * @param mixed $aData
   * @return array
   */
  public function getFieldsFromTable($aData) {
    $qb = $this->oEntityManager->getConnection()->createQueryBuilder();

    $qb->select('columns.COLUMN_NAME AS "column_name", columns.DATA_TYPE AS "data_type"')
      ->from('ALL_TAB_COLUMNS', 'columns')
      ->where($qb->expr()->eq('columns.OWNER', ':owner'))
      ->andWhere($qb->expr()->eq('columns.TABLE_NAME', ':table_name'))
      ->setParameter('owner', $aData['schema'])
      ->setParameter('table_name', $aData['table'])
      ->orderBy('"column_name"');

    try {
      $oTables = $qb->execute()->fetchAll();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get tables: ' . $e->getMessage());
      return ['status' => 0];
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oTables as $oTable) {
      $aField = ['name' => $oTable['column_name']];

      switch ($oTable['data_type']) {
        case 'INTEGER':
        case 'SHORTINTEGER':
        case 'LONGINTEGER':
        case 'DECIMAL':
        case 'SHORTDECIMAL':
        case 'NUMBER':
          $aField['type'] = 'integer';
          break;
        case 'BOOLEAN':
          $aField['type'] = 'boolean';
          break;
        default:
          $aField['type'] = 'string';
          break;
      }

      $response['data'][] = $aField;
    }

    return $response;
  }
}
