<?php

namespace App\Module\Vitis\WebService\Service\FileAccessor;

/**
 * Interface utilisé pour interagir avec les fichiers peu importe le système de fichier.
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

interface FileAccessorInterface {
  /****************************************************************
   *                                                              *
   *                            Files                             *
   *                     PHP native functions                     *
   *                                                              *
   ****************************************************************/

  /**
   * Vérifie si un fichier ou un dossier existe
   * https://www.php.net/manual/fr/function.file-exists.php
   * @param string $sFilePath Chemin vers le fichier/dossier à tester
   * @return bool TRUE si le fichier/dossier existe et FALSE s'il n'existe pas
   */
  public function file_exists(string $sFilePath): bool;
  /**
   * Écrit des données dans un fichier
   * https://www.php.net/manual/fr/function.file-put-contents.php
   * @param string $sFilePath Chemin vers le fichier destination
   * @param mixed $sData Donnée à écrire dans le fichier
   * @param int $iFlags voir https://www.php.net/manual/fr/function.file-put-contents.php
   * @param resource|null $mContext voir https://www.php.net/manual/fr/function.file-put-contents.php
   * @return int|null Retourne le nombre d'octets qui ont été écrits au fichier, ou NULL si une erreur survient.
   */
  public function file_put_contents(string $sFilePath, $sData, int $iFlags = 0, $mContext = null): ?int;
  /**
   * Lit tout un fichier dans une chaîne
   * https://www.php.net/manual/fr/function.file-get-contents.php
   * @param string $sFilePath Chemin vers le fichier destination
   * @param bool $bUseIncludePath voir https://www.php.net/manual/fr/function.file-get-contents.php
   * @param resource|null $mContext voir https://www.php.net/manual/fr/function.file-get-contents.php
   * @param int $iOffset voir https://www.php.net/manual/fr/function.file-get-contents.php
   * @param int $iMaxLen La taille maximale des données à lire. Le comportement par défaut est de lire jusqu'à la fin du fichier. Ce paramètre s'applique sur le flux traité par les filtres.
   * @return string|null Retourne les données lues ou NULL si une erreur survient.
   */
  public function file_get_contents(string $sFilePath, bool $bUseIncludePath = false, $mContext = null, int $iOffset = 0, int $iMaxLen = -1): ?string;
  /**
   * Copie un fichier
   * https://www.php.net/manual/fr/function.copy.php
   * @param string $sSourceFilePath Chemin vers le fichier à copier
   * @param string $sDestFilePath Chemin vers le fichier destination
   * @return bool|null Retourne TRUE si la copie se passe bien ou NULL si une erreur survient.
   */
  public function copy(string $sSourceFilePath, string $sDestFilePath): ?bool;
  /**
   * Lit la taille d'un fichier
   * https://www.php.net/manual/fr/function.filesize.php
   * @param string $sFilePath Chemin vers le fichier
   * @return int|null Renvoie la taille du fichier filename en octets, ou NULL (et génère une erreur de niveau E_WARNING) en cas d'erreur.
   */
  public function filesize(string $sFilePath): ?int;
  /**
   * Lit un fichier et l'envoie dans le buffer de sortie.
   * https://www.php.net/manual/fr/function.readfile.php
   * @param string $sFilePath Chemin vers le fichier à lire
   * @param bool $bUseIncludePath voir https://www.php.net/manual/fr/function.readfile.php
   * @param resource $mContext voir https://www.php.net/manual/fr/function.readfile.php
   * @return int|null Retourne le nombre d'octets lus depuis le fichier en cas de succès, ou NULL si une erreur survient
   */
  public function readfile(string $sFilePath, bool $bUseIncludePath = false, $mContext = null): ?int;
  /**
   * Renomme un fichier ou un dossier.
   * https://www.php.net/manual/fr/function.rename.php
   * @param string $sSourceFilePath Chemin vers le fichier à renommer
   * @param string $sDestFilePath Chemin du fichier à utiliser comme destination
   * @param resource $mContext voir https://www.php.net/manual/fr/function.rename.php
   * @return bool|null Retourne TRUE si la copie se passe bien ou NULL si une erreur survient
   */
  public function rename(string $sSourceFilePath, string $sDestFilePath, $mContext = null): ?bool;
  /**
   * Efface un fichier.
   * https://www.php.net/manual/fr/function.unlink.php
   * @param string $sFilePath Chemin vers le fichier à supprimer
   * @param resource $mContext voir https://www.php.net/manual/fr/function.unlink.php
   * @return bool|null Cette fonction retourne TRUE en cas de succès ou NULL si une erreur survient.
   */
  public function unlink(string $sFilePath, $mContext = null): ?bool;
  /**
   * Lit la date de dernière modification du fichier.
   * https://www.php.net/manual/fr/function.filemtime.php
   * @param string $sFilePath Chemin vers le fichier
   * @return int|null Renvoie le timestamp Unix de dernière modification du fichier filename ou NULL si une erreur survient. Utilisez date() sur ce résultat pour obtenir une date de modification humainement lisible.
   */
  public function filemtime(string $sFilePath): ?int;
  /**
   * Détecte le type de contenu d'un fichier.
   * @param string $sFilePath Chemin vers le fichier
   * @return string Retourne le type de contenu au format MIME, comme text/plain ou application/octet-stream s'il ne connait pas le MIME type, NULL si le fichier n'exite pas.
   */
  public function mime_content_type(string $sFilePath): ?string;
  /**
   * Retourne le chemin canonique absolu.
   * A la différence de la fonction native de PHP dans le cas de windows la fonction retourne le chemin avec des / et pas des \
   * https://www.php.net/manual/fr/function.realpath.php
   * @param string $sFilePath Chemin vers le fichier
   * @return string Retourne le chemin canonique absolu vers le fichier (sans ./ /.. ou //), NULL si le fichier n'exite pas.
   */
  public function realpath(string $sFilePath): ?string;
  /****************************************************************
   *                                                              *
   *                            Files                             *
   *                       Veremes functions                      *
   *                                                              *
   ****************************************************************/

  /**
   * Lit les metadonnées d'un fichier.
   * @param string $sFilePath Chemin du fichier.
   * @param array<string> $aKeepOnlyField Liste des attributs à conserver (si vide retourne tous les attributs).
   * @return array|null Les infos du fichier demandé ou NULL en cas d'erreur.
   * ex :
   *                  [
   *                      filename            (string) : nom du fichier
   *                      size                (string) : taille du fichier (format humain ex : 125Ko)
   *                      lastModification    (string) : date de dernière modification (format 05/02/2019 12:15:22)
   *                      href                (string) : url d'accès au fichier (généré par getProxyPassUrl)
   *                  ]
   */
  public function getFileInfos(string $sFilePath, array $aKeepOnlyField = []): ?array;
  /**
   * retourne l'extension d'un fichier.
   * @param string $sFilePath Chemin du fichier.
   * @return string
   */
  public function getExtension(string $sFilePath): string;
  /**
   * retourne le nom d'un fichier.
   * @param string $sFilePath Chemin du fichier.
   * @return string
   */
  public function getFileName(string $sFilePath): string;
  /**
   * retourne une url d'accés au fichier avec controle du hash du fichier.
   * @param string $sFilePath Chemin du fichier.
   * @return string|null
   */
  public function getProxyPassUrl(string $sFilePath): ?string;
  /**
   * retourne le hash de controle d'un fichier
   * @param string $sFilePath Chemin du fichier.
   * @return string|null hash de controle ou NULL si le fichier n'existe pas
   */
  public function getFileEtag(string $sFilePath): ?string;
  /**
   * Lit tout un fichier dans une chaîne en controlant le hash de controle
   * https://www.php.net/manual/fr/function.file-get-contents.php
   * @param string $sFilePath Chemin vers le fichier destination
   * @param string $sEtag hash de controle à vérifier
   * @param bool $bUseIncludePath voir https://www.php.net/manual/fr/function.file-get-contents.php
   * @param resource|null $mContext voir https://www.php.net/manual/fr/function.file-get-contents.php
   * @param int $iOffset voir https://www.php.net/manual/fr/function.file-get-contents.php
   * @param int $iMaxLen La taille maximale des données à lire. Le comportement par défaut est de lire jusqu'à la fin du fichier. Ce paramètre s'applique sur le flux traité par les filtres.
   * @return string|null Retourne les données lues ou NULL si une erreur survient.
   */
  public function file_get_contents_with_etag(string $sFilePath, string $sEtag, bool $bUseIncludePath = false, $mContext = null, int $iOffset = 0, int $iMaxLen = -1);
  /**
   * Lit la date de dernière modification du fichier et la retourne au format string.
   * https://www.php.net/manual/fr/function.filemtime.php
   * @param string $sFilePath Chemin vers le fichier
   * @param string $sFormat Format de la date à récupérer (voir https://www.php.net/manual/fr/function.date.php)
   * @return string|null Renvoie le timestamp Unix de dernière modification du fichier filename ou NULL si une erreur survient. Utilisez date() sur ce résultat pour obtenir une date de modification humainement lisible.
   */
  public function filemtime_formated(string $sFilePath, string $sFormat): ?string;
  /**
   * Copie un fichier du disque local vers le système de fichier en configuration.
   * @param string $sSourceFilePath Fichier à copier sur le disque local.
   * @param string $sDestFilePath Destination sur le système de fichier.
   * @return bool|null
   */
  public function copyLocalToDistant(string $sSourceFilePath, string $sDestFilePath): ?bool;
  /**
   * Copie un fichier du système de fichier en configuration vers le disque local .
   * @param string $sSourceFilePath Fichier à copier sur le système de fichier.
   * @param string $sDestFilePath Destination sur le disque local.
   * @return bool|null
   */
  public function copyDistantToLocal(string $sSourceFilePath, string $sDestFilePath): ?bool;

  /**
   * Permet de redimensionner et de rééchantillonne rune image
   * @param string $sFilePath Chemin vers le fichier.
   * @param int $iDstWidth Largeur finale en pixel.
   * @param int $iDstHeight Largeur finale en pixel.
   * @param int $iCropX Number of pixel to crop (horizontal).
   * @param int $iCropY Number of pixel to crop (vertical).
   * @param int $iOffsetX Number of pixel to begin the resampling (horizontal).
   * @param int $iOffsetY Number of pixel to begin the resampling (vertical).
   * @param string $sType type de fichier pour lire l'image.
   * @return string|null $sPath chemin du fichier aprés modification ou null en cas d'erreur.
   */
  public function pictureResampler(string $sFilePath, int $iDstWidth, int $iDstHeight, int $iCropX = 0, int $iCropY = 0, int $iOffsetX = 0, int $iOffsetY = 0, ?string $sType = null): ?string;

  /**
   * Génére une miniature à partir de l'image d'origine
   * @param string $sFilePath chemin vers le fichier à minifier
   * @return string|null chemin vers la miniature
   */
  public function thumbnailResampler(string $sFilePath): ?string;

  /**
   * Teste la taille du fichier par rapport à la properties qui défin la taille max des fichiers
   * @param string $sFilePath Chemin du fichier
   * @param int $iLimitation paramètre permettant d'appliquer une limitation spécifique
   * @return bool
   */
  public function checkFileSize(string $sFilePath, int $iLimitation = -1): bool;
  
  /**
   * teste l'extension du fichier pour savoir si elle est autorisée
   * @param string $sFilePath chemin du fichier à tester
   * @return bool
   */
  public function checkExtension(string $sFilePath): bool;

  /****************************************************************
   *                                                              *
   *                            Folders                           *
   *                     PHP native functions                     *
   *                                                              *
   ****************************************************************/

  /**
   * Liste les fichiers et dossiers dans un dossier.
   * https://www.php.net/manual/fr/function.scandir.php
   * @param string $sDirPath Chemin vers le dossier à scanner
   * @param int $iSortOrder voir https://www.php.net/manual/fr/function.scandir.php
   * @param resource|null $mContext voir https://www.php.net/manual/fr/function.scandir.php
   * @return array|null Retourne un tableau des fichiers en cas de succès ou NULL en cas d'échec.
   */
  public function scandir(string $sDirPath, int $iSortOrder = SCANDIR_SORT_ASCENDING, $mContext = null): ?array;
  /**
   * Crée un dossier.
   * https://www.php.net/manual/fr/function.mkdir.php
   * @param string $sDirPath Chemin vers le dossier à scanner
   * @param int $iMode voir https://www.php.net/manual/fr/function.chmod.php
   * @param bool $bRecursive Permet la création de répertoires imbriqués spécifiés dans $sDirPath.
   * @param resource|null $mContext voir https://www.php.net/manual/fr/function.mkdir.php
   * @return bool|null Cette fonction retourne TRUE en cas de succès ou NULL si une erreur survient.
   */
  public function mkdir(string $sDirPath, int $iMode = 0777, bool $bRecursive = false, $mContext = null): ?bool;
  /**
   * Efface un dossier.
   * https://www.php.net/manual/fr/function.rmdir.php
   * @param string $sDirPath Chemin vers le dossier à supprimer
   * @param resource|null $mContext voir https://www.php.net/manual/fr/function.rmdir.php
   * @return bool|null Cette fonction retourne TRUE en cas de succès ou NULL si une erreur survient.
   */
  public function rmdir(string $sDirPath, $mContext = null): ?bool;
  /**
   * Indique si le fichier est un dossier.
   * https://www.php.net/manual/fr/function.is-dir.php
   * @param string $sDirPath Chemin vers le fichier/dossier
   * @return bool Retourne TRUE si le nom de fichier existe et que c'est un dossier, FALSE sinon.
   */
  public function is_dir(string $sDirPath): bool;

  /****************************************************************
   *                                                              *
   *                            Folders                           *
   *                       Veremes functions                      *
   *                                                              *
   ****************************************************************/

  /**
   * Scan un dossier en profondeur et retourne les metadonnées pour chaque élément enfant.
   * @param string $sDirPath Dossier à scanner
   * @param array<string> $aKeepOnlyExt Liste des extensions en lowercase à garder (si array vide on prend tout)
   * @param array<string> $aKeepOnlyField Liste des champs à garder (si array vide on prend tout)
   * @param bool $bRemoveEmptyFolder Supprime les répertoires vides de la liste
   * @param int $iDepth Niveau de profondeur à récupérer (-1 pour toute l'arborescence)
   * @return array|null Arbre :
   *                  - Dossier : retour getFileInfos + attribut content qui est un tableau de fichier/sous-dossier
   *                  - Fichier : retour getFileInfos
   */
  public function getFolderInfos(string $sDirPath, array $aKeepOnlyExt = [], array $aKeepOnlyField = [], bool $bRemoveEmptyFolder = false, int $iDepth = -1): ?array;
  /**
   * Supprime un dossier et son sontenu.
   * @param string $sDirPath Dossier à supprimer.
   * @return bool|null Retourne TRUE si le dossier a pu être supprimé, NULL sinon.
   */
  public function clearDir(string $sDirPath): ?bool;
  /**
   * Copie un dossier et son contenu.
   * @param string $sDirPathSrc Dossier à copier.
   * @param string $sDirPathDest Destination.
   * @return bool|null Retourne TRUE si le dossier a pu être supprimé, NULL sinon.
   */
  public function copyDirectory(string $sDirPathSrc, string $sDirPathDest): ?bool;
  /**
   * Copie un dossier et son arborescence du disque local vers système de fichier en configuration.
   * @param string $sDirPathSrc Dossier à copier sur le disque local.
   * @param string $sDirPathDest Destination sur le système de fichier.
   * @param string[] $aIgnoredItems Tableau contenant les fichiers/dossiers à ignorer (utiliable uniquement pour S3)
   * @return bool|null Retourne TRUE si le dossier a pu être supprimé, NULL sinon.
   */
  public function copyDirLocalToDistant(string $sDirPathSrc, string $sDirPathDest, array $aIgnoredItems = []): ?bool;
  /**
   * Copie un dossier et son arborescence du système de fichier en configuration vers le disque local .
   * @param string $sDirPathSrc Dossier à copier sur le système de fichier.
   * @param string $sDirPathDest Destination sur le disque local.
   * @return bool|null Retourne TRUE si le dossier a pu être supprimé, NULL sinon.
   */
  public function copyDirDistantToLocal(string $sDirPathSrc, string $sDirPathDest): ?bool;
  /**
   * Créer un zip sur le système de fichier en configuration à partir d'un dossier sur le système.
   * @param string $sFolder Dossier à zipper sur le système de fichier.
   * @param string $sDestination Destination sur le système de fichier.
   * @param string $sExtensionToExclude Extensions de fichier à ignorer séparées par des pipes.
   * @param string $sPassword mot de passe pour encrypter l'archive.
   * @param bool $bDeleteFiles Suppression du dossier aprés la compression.
   * @return bool|null Retourne TRUE si le dossier a pu être zippé, NULL sinon.
   */
  public function createZip(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool;
  /**
   * Encrypte une fichier zip existant
   * @param string $sZipFilePath Chemin vers le fichier zip existant
   * @param string $sPassword Mot de passe à utiliser pour encrypter l'archive
   * @return bool|null Retourne TRUE si l'archive a pu être encrypté, NULL sinon
   */
  public function encryptZip(string $sZipFilePath, string $sPassword): ?bool;
  /**
   * Décompresse une archive zip du système de fichier en configuration.
   * @param string $sZipFile Archive à décompresser sur le système de fichier.
   * @param string $sDestination Destination sur le système de fichier.
   * @return bool|null Retourne TRUE si le dossier a pu être dézippé, NULL sinon.
   */
  public function unZip(string $sZipFile, string $sDestination): ?bool;
  /**
   * Créer un zip sur le système de fichier en configuration à partir d'un dossier sur le disque local.
   * @param string $sFolder Dossier à zipper sur le disque locale.
   * @param string $sDestination Destination sur le système de fichier.
   * @param string $sExtensionToExclude Extensions de fichier à ignorer séparées par des pipes.
   * @param string $sPassword mot de passe pour encrypter l'archive.
   * @param bool $bDeleteFiles Suppression du dossier aprés la compression.
   * @return bool|null Retourne TRUE si le dossier a pu être zippé, NULL sinon.
   */
  public function zipLocalDirToDistant(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool;
  /**
   * Créer un zip sur le disque local à partir d'un dossier sur le système de fichier.
   * @param string $sFolder Dossier à zipper sur le système de fichier.
   * @param string $sDestination Destination sur le dossier local.
   * @param string $sExtensionToExclude Extensions de fichier à ignorer séparées par des pipes.
   * @param string $sPassword mot de passe pour encrypter l'archive.
   * @param bool $bDeleteFiles Suppression du dossier aprés la compression.
   * @return bool|null Retourne TRUE si le dossier a pu être zippé, NULL sinon.
   */
  public function zipDistantDirToLocal(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool;
  /**
   * Décompresse une archive zip du disque local sur le système de fichier en configuration.
   * @param string $sZipFile Archive à décompresser sur le disque local.
   * @param string $sDestination Destination sur le système de fichier en configuration.
   * @return bool|null Retourne TRUE si le dossier a pu être dézippé, NULL sinon.
   */
  public function unzipLocalToDistantDir(string $sZipFile, string $sDestination): ?bool;
  /**
   * Décompresse une archive zip du système de fichier en configuration sur le disque local.
   * @param string $sZipFile Archive à décompresser sur le système de fichier.
   * @param string $sDestination Destination sur le disque local.
   * @return bool|null Retourne TRUE si le dossier a pu être dézippé, NULL sinon.
   */
  public function unzipDistantToLocalDir(string $sZipFile, string $sDestination): ?bool;
}
