<?php

namespace App\Module\Vitis\WebService\Service\FileAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Crypto\JweCrypto;
use App\Module\Vitis\WebService\Service\FileAccessor\FileAccessorInterface;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\FileAccessor\S3Accessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;

/**
 * Service permettant d'accéder au accesseur de fichier.
 *
 * @author Anthony Borghi   <anthony.borghi@veremes.com>
 */
class FileManager {

  /**
   * @var string[]
   */
  public const MOVABLE_PATHS = [
    'ws_data_dir',
    'public_dir',
    'upload_dir',
    'shared_dir',
  ];

  /**
   * @var Properties
   */
  private $oProperties;
  /**
   * @var LocalAccessor
   */
  private $oLocalFileInterface;
  /**
   * @var FileAccessorInterface
   */
  private $oFileInterface;
  /**
   * @var LfsService
   */
  private $oLfsService;
  /**
   * @var \Monolog\Logger
   */
  protected $oLogger;

  /**
   * @param Properties $properties service d'accés aux properties
   * @param LoggerManager $oLogger service d'accés aux logger
   * @param JweCrypto $oJweCrypto service de cryptage des données
   */
  public function __construct(Properties $properties, LoggerManager $oLogger, JweCrypto $oJweCrypto) {
    // set du service de properties
    $this->oProperties = $properties;
    $this->oLogger = $oLogger->get('API_ERROR');
    // set des interface fichier
    $this->oLocalFileInterface = new LocalAccessor($this->oProperties, $oLogger, $oJweCrypto);
    switch ($this->oProperties->get('file_system')) {
      case 's3':
        $this->oFileInterface = new S3Accessor($this->oProperties, $oLogger, $oJweCrypto);
        break;
      case 'local':
        $this->oFileInterface = $this->oLocalFileInterface;
        break;
      default:
        $this->oLogger->warning('This file system doesn\'t exist : ' . $this->oProperties->get('file_system') . '.');
        $this->oFileInterface = $this->oLocalFileInterface;
        break;
    }
  }

  public function getLfsService () : LfsService {
    return $this->oLfsService ?? new LfsService($this->oProperties, $this->oLogger, $this->oLocalFileInterface);
  }

  /**
   * Retourne une interface fichier permettant d'interagir avec le système de fichier configuré
   * @return FileAccessorInterface une classe implémentant l'interface FileAccessor
   */
  public function getFileAccessor(): FileAccessorInterface {
    return $this->oFileInterface;
  }

  /**
   * Retourne une interface fichier permettant d'interagir avec le disque local
   * @return LocalAccessor une classe implémentant l'interface FileAccessor
   */
  public function getLocalFileAccessor(): LocalAccessor {
    return $this->oLocalFileInterface;
  }

  /**
   * Analyse le chemin passé en paramètre et retourne l'accesseur en fonction de l'emplacement du fichier
   * @param string $sPath Chemin du fichier/dossier à traiter
   * @return FileAccessorInterface Interface pour travailler à cet emplacement
   */
  public function getAccessorForPath(string $sPath): FileAccessorInterface {

    foreach (self::MOVABLE_PATHS as $sPropertiesKey) {
      $sPropPath = $this->oProperties->get($sPropertiesKey);
      if (\strpos($sPath, $sPropPath) === 0) {
        return $this->oFileInterface;
      }
    }

    return $this->oLocalFileInterface;
  }
}
