<?php

namespace App\Module\Vitis\WebService\Service\FileAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use Monolog\Logger;
use Ramsey\Uuid\Uuid;

/**
 * Service permettant la gestion des gros fichiers via l'API.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class LfsService {
  /**
   * @var string nom du fichier json qui contient la donnée
   */
  private const DATA_FILE_NAME = 'lfs_data.json';
  /**
   * @var int
   */
  private const CHUNK_HASH_PART = 256;

  /**
   * @var Properties
   */
  private $oProperties;
  /**
   * @var LocalAccessor
   */
  private $oLocalAccessor;
  /**
   * @var Logger
   */
  protected $oLogger;

  /**
   * @param Properties $properties service d'accés aux properties
   * @param Logger $oLogger service d'accés aux logger
   * @param LocalAccessor $oLocalAccessor service d'accés aux fichiers du disque
   */
  public function __construct(Properties $properties, Logger $oLogger, LocalAccessor $oLocalAccessor) {
    // set du service de properties
    $this->oProperties = $properties;
    $this->oLocalAccessor = $oLocalAccessor;
    $this->oLogger = $oLogger;
  }

  /**
   * @param  string $sTransactionId
   * @return string chemin vers le dossier de la transaction
   */
  private function getTransactionPath(string $sTransactionId): string {
    return $this->oProperties->get('lfs_dir') . '/' . $sTransactionId;
  }

  /**
   * @param  string $sTransactionId
   * @return string chemin vers le dossier de la transaction
   */
  private function getTransactionDataPath(string $sTransactionId): string {
    return $this->getTransactionPath($sTransactionId) . '/' . self::DATA_FILE_NAME;
  }

  /**
   * @param  string $sTransactionId
   * @return array|null donnée lié à la transaction
   */
  public function getTransactionData(string $sTransactionId): ?array{
    // si le dossier n'existe pas on stop
    if ($this->checkTransactionPath($sTransactionId) === null) {
      return null;
    }
    $sPath = $this->getTransactionDataPath($sTransactionId);
    $sData = $this->oLocalAccessor->file_get_contents($sPath);

    return json_decode($sData, true) ?? null;
  }

  /**
   * enregistre les données dans le fichier json de la transaction
   * @param  string $sTransactionId id de transaction
   * @param  array $aTransactionData donnée à stocker
   * @return bool true si tout va bien
   */
  public function storeTransactionData(string $sTransactionId, array $aTransactionData): bool {
    // si le dossier n'existe pas on stop
    if ($this->checkTransactionPath($sTransactionId) === null) {
      return false;
    }
    $sPath = $this->getTransactionDataPath($sTransactionId);
    $sData = json_encode($aTransactionData);

    return ($this->oLocalAccessor->file_put_contents($sPath, $sData) != null);

  }

  /**
   * Trouve un id de transaction disponible et prépare le dossier
   *
   * @return string Id de la transaction
   */
  public function createTransactionId(): ?string{
    $sTransactionId = '';
    // on cherche un transactionid dispo
    do {
      $sTransactionId = Uuid::uuid1()->toString(); //ex : e4eaaaf2-d142-11e1-b3e4-080027620cdd
    } while ($this->checkTransactionPath($sTransactionId) !== null || empty($sTransactionId));
    // on récupère le chemin
    $sPath = $this->getTransactionPath($sTransactionId);
    // on créer l'espace pour la transaction
    $this->oLocalAccessor->mkdir($sPath, LocalAccessor::DEFAULT_LINUX_PERMISSION, true);

    return $sTransactionId;
  }

  /**
   * Teste si un espace de transaction existe
   *
   * @param  string      $sTransactionId
   * @return string|null chemin vers le dossier de la transaction ou null si n'existe pas
   */
  private function checkTransactionPath(string $sTransactionId): ?string{
    $sPath = $this->getTransactionPath($sTransactionId);
    return ($this->oLocalAccessor->file_exists($sPath)) ? $sPath : null;
  }

  public function checkUploadedChunk(string $sTransactionId, int $index, string $sHash): ?bool {
    // check si le fichier existe
    if (empty($_FILES['chunk'])) {
      return false;
    }

    $sProcessedHash = $this->hashPart($_FILES['chunk']['tmp_name']);

    if ($sProcessedHash != $sHash) {
      return false;
    }

    return $this->oLocalAccessor->copy($_FILES['chunk']['tmp_name'], $this->getChunkPath($sTransactionId, $index));
  }

  /**
   * Undocumented function
   *
   * @param  string $sTransactionId
   * @param  int    $iIndex
   * @return string
   */
  private function getChunkPath(string $sTransactionId, int $iIndex): string {
    $sPath = $this->getTransactionPath($sTransactionId);
    return $sPath . '/' . $iIndex . '.chk';
  }

  /**
   * lit une portion de fichier la passe en hexa, calcul un sha256 dessus et l'encode en base64
   *
   * @param  string $sPath
   * @return string hash d'un fichier
   */
  private function hashPart(string $sPath): string {
    $sFile = $this->oLocalAccessor->file_get_contents($sPath, false, null, 0, self::CHUNK_HASH_PART);
    // passage en hexa, le binaire trop pénible coté client
    $sFile = bin2hex($sFile);

    return $this->hashText($sFile, true);
  }

  /**
   * Undocumented function
   *
   * @param  string $sText
   * @param  bool   $bBinary
   * @return string
   */
  private function hashText(string $sText, bool $bBinary = false): string {
    return base64_encode(hash('sha256', $sText, $bBinary));
  }

  public function storeChunkHash(string $sTransactionId, int $iChunkId, string $sHash): void {
    // stockage dans le fichier json
    $aTrData = $this->getTransactionData($sTransactionId);
    if (!array_key_exists('chunks', $aTrData)) {
      $aTrData['chunks'] = [];
    }

    $aTrData['chunks']['chk_' . $iChunkId] = $sHash;

    $this->storeTransactionData($sTransactionId, $aTrData);
  }

  /**
   * Undocumented function
   *
   * @param  mixed $a
   * @param  mixed $b
   * @return int
   */
  private function key_sorter ($a, $b) : int {

    $a_int = intval(str_replace('chk_', '', $a));
    $b_int = intval(str_replace('chk_', '', $b));

    return $a_int - $b_int;
  }

  /**
   * Undocumented function
   *
   * @param  string $sTransactionId
   * @return bool
   */
  public function checkTransactionIntegrity(string $sTransactionId): bool {
    // récupérer les données json
    $aTrData = $this->getTransactionData($sTransactionId);

    uksort($aTrData['chunks'], [$this, 'key_sorter']);

    $sFileHash = '';

    // parcourir la liste des chunks dans l'ordre
    foreach ($aTrData['chunks'] as $sIndex => $sStoredHash) {
      $iIndex = intval(str_replace('chk_', '', $sIndex));
      $sChunkPath = $this->getChunkPath($sTransactionId, $iIndex);

      $sProcessedHash = $this->hashPart($sChunkPath);

      // revérifier les hash
      if ($sProcessedHash != $sStoredHash) {
        return false;
      }

      if (empty($sFileHash)) {
        $sFileHash = $sStoredHash;
      } else {
        $sFileHash = $this->hashText($sFileHash . $sStoredHash, true);
      }

    }

    // vérifier le hash global
    return ($sFileHash == $aTrData['md5_hash_content']);
  }

  /**
   * Undocumented function
   *
   * @param  string $sTransactionId
   * @param  bool $bCheckExist
   * @return string
   */
  public function getFilePath(string $sTransactionId, bool $bCheckExist = true): ?string{
    $aTrData = $this->getTransactionData($sTransactionId);
    $sPath = $this->getTransactionPath($sTransactionId);

    $sFinalPath = $sPath . '/' . $aTrData['file_name'];

    if (!$this->oLocalAccessor->file_exists($sFinalPath) && $bCheckExist) {
      return null;
    }

    return $sFinalPath;
  }

  /**
   * Undocumented function
   *
   * @param  string      $sTransactionId
   * @return string|null
   */
  public function getFileName(string $sTransactionId): ?string{
    $aTrData = $this->getTransactionData($sTransactionId);

    return $aTrData['file_name'] ?? null;
  }

  /**
   * Undocumented function
   *
   * @param  string      $sTransactionId
   * @return int|null
   */
  public function getFileSize(string $sTransactionId): ?int{
    $sPath = $this->getFilePath($sTransactionId);

    if (empty($sPath)){
      return null;
    }

    return $this->oLocalAccessor->filesize($sPath);
  }
  /**
   * Undocumented function
   *
   * @param  string      $sTransactionId
   * @return string|null
   */
  public function getFileMimeType(string $sTransactionId): ?string{
    $sPath = $this->getFilePath($sTransactionId);

    if (empty($sPath)){
      return null;
    }

    return $this->oLocalAccessor->mime_content_type($sPath);
  }

  /**
   * Undocumented function
   *
   * @param  string $sTransactionId
   * @return bool
   */
  public function mergeChunk(string $sTransactionId): bool {
    $aTrData = $this->getTransactionData($sTransactionId);
    uksort($aTrData['chunks'], [$this, 'key_sorter']);
    $sFinalPath = $this->getFilePath($sTransactionId, false);
    // si tout est bon alors on aggrège les tronçons pour reconstruire le fichier
    // parcourir la liste des chunks dans l'ordre
    foreach ($aTrData['chunks'] as $sIndex => $sStoredHash) {
      $iIndex = intval(str_replace('chk_', '', $sIndex));
      $sChunkPath = $this->getChunkPath($sTransactionId, $iIndex);
      $sChunkContent = $this->oLocalAccessor->file_get_contents($sChunkPath);
      if ($this->oLocalAccessor->file_put_contents($sFinalPath, $sChunkContent, FILE_APPEND) < 1) {
        return false;
      }

      // suppression du chunk pour libérer de l'espace
      if ($this->oLocalAccessor->unlink($sChunkPath) != true) {
        return false;
      }
    }

    return true;
  }

  /**
   * Undocumented function
   *
   * @param  string $sTransactionId
   * @return bool
   */
  public function clearTransaction(string $sTransactionId): bool {
    $sPath = $this->getTransactionPath($sTransactionId);

    if (!$this->oLocalAccessor->file_exists($sPath)){
      return true;
    }

    return $this->oLocalAccessor->clearDir($sPath) ?? false;
  }

}
