<?php

namespace App\Module\Vitis\WebService\Service;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;

/**
 * Service permettant de gérer les moteurs gtf et fme.
 *
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */
class JobsService {
  /**
   * @var string chemin vers le fichier qui contient les jobs
   */
  protected const JOBS_FILE_PATH = '@prop(engine_dir)/vitis/conf/jobs.json';

  /**
   * @var string api Path pour auto syncs les groupes
   */
  private const AUTO_SYNC_API_ROUTE = '/vitis/domains/autosyncgroups';

  /**
   * @var array structure du job par défaut pour l'auto sync
   */
  private const AUTO_SYNC_JOB = [
    'period' => '',
    'api_route' => self::AUTO_SYNC_API_ROUTE,
    'api_method' => 'POST',
    'api_params' => [],
  ];

  /**
   * @var Properties
   */
  protected $oPropertiesSrvc;

  /**
   * @var LoggerManager
   */
  protected $oLoggerManager;

  /**
   * @var LocalAccessor
   */
  protected $oLocalFileAccessor;

  /**
   * @param Properties $oPropertiesSrvc Service permettant d'accéder aux properties
   * @param LoggerManager $oLoggerManager Service permettant d'accéder aux loggers
   * @param LocalAccessor $oLocalFileAccessor Service pour accéder aux fichier sur le disque
   */
  public function __construct(Properties $oPropertiesSrvc, LoggerManager $oLoggerManager, LocalAccessor $oLocalFileAccessor) {
    $this->oPropertiesSrvc = $oPropertiesSrvc;
    $this->oLoggerManager = $oLoggerManager;
    $this->oLocalFileAccessor = $oLocalFileAccessor;
  }

  /**
   * programme synchronisation des groupes
   * @param string $sCronExpression expression cron à utiliser
   * @return bool true si l'ajout s'est bien passé false sinon
   */
  public function addAutosyncgroups(string $sCronExpression): bool {
    if (empty($sCronExpression)) {
      $this->oLoggerManager->get('API_ERROR')->error('Empty Cron expression to Sync groups');
      return false;
    }

    $aContent = $this->readJobsFile();

    if ($aContent == null) {
      return false;
    }

    if (!isset($aContent['jobs'])) {
      $aContent['jobs'] = [];
    }
    if (!is_array($aContent['jobs'])) {
      $aContent['jobs'] = [];
    }

    $aAutomationStruct = self::AUTO_SYNC_JOB;
    $aAutomationStruct['period'] = $sCronExpression;
    $bExists = false;
    foreach ($aContent['jobs'] as $iKey => $aJob) {
      // si l'appel à l'API existe met le à jour
      if (isset($aJob['api_route']) && $aJob['api_route'] === self::AUTO_SYNC_API_ROUTE) {
        $aContent['jobs'][$iKey] = $aAutomationStruct;
        $bExists = true;
      }
    }
    if (!$bExists) {
      // ajoute l'appel à l'API
      $aContent['jobs'][] = $aAutomationStruct;
    }

    return $this->writeJobsFile($aContent);
  }

  /**
   * déprogramme synchronisation des groupes
   * @return bool true la suppression est réussi
   */
  public function removeAutosyncgroups(): bool {
    $aContent = $this->readJobsFile();

    if ($aContent == null) {
      return false;
    }

    if (!isset($aContent['jobs'])) {
      return false;
    }
    if (!is_array($aContent['jobs'])) {
      return false;
    }
    $aJobs = [];
    foreach ($aContent['jobs'] as $aJob) {
      // si l'appel à l'API existe
      if (isset($aJob['api_route']) && $aJob['api_route'] === self::AUTO_SYNC_API_ROUTE) {
        // on ne le copie pas
        continue;
      }
      $aJobs[] = $aJob;
    }

    $aContent['jobs'] = $aJobs;

    return $this->writeJobsFile($aContent);
  }

  /**
   * Lit et décode le fichier jobs.json
   * @return array|null retourne un array contenant le jobs ou null en cas d'erreur
   */
  protected function readJobsFile(): ?array{
    $sJobsFilePath = $this->oPropertiesSrvc->replacePropertiesInString(self::JOBS_FILE_PATH);

    if (!$this->oLocalFileAccessor->file_exists($sJobsFilePath)) {
      $this->oLoggerManager->get('API_ERROR')->error('Unable to find file jobs.json : ' . $sJobsFilePath);
      return null;
    }

    return json_decode($this->oLocalFileAccessor->file_get_contents($sJobsFilePath), true);
  }

  /**
   * Encode et écrit le tableau dans jobs.json
   * @return bool
   */
  protected function writeJobsFile(array $aContent) : bool {
    $sJobsFilePath = $this->oPropertiesSrvc->replacePropertiesInString(self::JOBS_FILE_PATH);
    $sContent = json_encode($aContent, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES);

    if ($this->oLocalFileAccessor->file_put_contents($sJobsFilePath, $sContent) !== null) {
      return true;
    } else {
      $this->oLoggerManager->get('API_ERROR')->error('Error while updating jobs.json : ' . $sJobsFilePath);
      return false;
    }
  }
}
