<?php

namespace App\Module\Vitis\WebService\Service\Logger;

use App\Module\Vitis\WebService\Query\DateTimeTypeWithMicroTime;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\DateField;
use Doctrine\DBAL\Logging\SQLLogger;
use Monolog\Logger;

/**
 * A SQL logger that logs to the standard Monolog/logger.
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */
class CustomSqlLogger implements SQLLogger {
  /**
   * @var Properties
   */
  private $oPropertiesSrvc;
  /**
   * @var array
   */
  private $aMessages = [];
  /**
   * @var array
   */
  private $aTmpMessages = [];
  /**
   * @var string
   */
  private $sUsername = '';
  /**
   * @var Logger
   */
  private $oLogger;
  /**
   * @var bool
   */
  private $bTransaction = false;
  /**
   * @var bool
   */
  private $bDebug = false;

  /**
   * @param Logger $oLogger logger à utiliser pour les requêtes SQL
   * @param string $sUsername nom de l'utilisateur connecté
   * @param Properties $PropertiesSrvc Un objet pour accéder aux properties de l'application
   */
  public function __construct(Logger $oLogger, string $sUsername, Properties $PropertiesSrvc) {
    $this->oLogger = $oLogger;
    $this->sUsername = $sUsername;
    // Sauvegarde du service de properties
    $this->oPropertiesSrvc = $PropertiesSrvc;
    $this->bDebug = ('DEBUG' === $this->oPropertiesSrvc->get('log_level'));
  }

  /**
   * {@inheritdoc}
   */
  public function startQuery($sql, ?array $params = null, ?array $types = null): void {
    // Passage en mode transactionnel
    if ($sql == '"START TRANSACTION"' && !$this->bDebug) {
      $this->aTmpMessages = [];
      $this->bTransaction = true;
      return;
    }
    // validation de la transaction
    if ($sql == '"COMMIT"' && !$this->bDebug) {
      $this->aMessages = array_merge($this->aMessages, $this->aTmpMessages);
      $this->bTransaction = false;
      return;
    }

    $aParams = (isset($params)) ? $params : [];
    $sBindSql = $sql;

    // bind le sql si nécessaire pour le log
    foreach ($aParams as $i => $mParam) {
      $sParam = $mParam;

      $sType = (!empty($types) && isset($types[$i])) ? $types[$i] : gettype($mParam);

      switch ($sType) {
        case 'integer':
        case 'float':
          break;
        case 'string':
          if (!is_numeric($mParam) && empty($types)) {
            $sParam = "'" . $mParam . "'";
          }
          break;
        case 'array':
          $sParam = "'" . implode("','", $mParam) . "'";
          break;
        case 'datetime':
          $sParam = (isset($mParam)) ? $mParam->format(DateField::DEFAULT_FORMAT) : 'null';
          break;
        case 'udatetime':
          $sParam = (isset($mParam)) ? $mParam->format(DateTimeTypeWithMicroTime::POSTRES_DATE_FORMAT) : 'null';
          break;
        case 'json':
          $sParam = (isset($mParam)) ? json_encode($mParam) : 'null';
          break;
        default:
          error_log('type not known (' . $sType . ')');
          break;
      }

      $iPosition = strpos($sBindSql, "?");
      if ($iPosition > -1) {
        $sBindSql = substr_replace($sBindSql, $sParam, $iPosition, 1);
      } else {
        $sBindSql = str_replace(':' . $i, $sParam, $sBindSql);
      }
    }
    // définition du niveau de log
    $sLevel = 'info';
    if (strpos($sql, 'SELECT') === 0 || (in_array($sql, ['"START TRANSACTION"', '"COMMIT"']) && $this->bDebug)) {
      $sLevel = 'debug';
    }

    $aMessage = [
      'login' => $this->sUsername,
      'request' => $sBindSql,
      'parameters' => $params,
      'real_request' => $sql,
      'level' => $sLevel,
      'start' => microtime(true),
    ];

    // si une transaction est demmaré on stock de manière temporaire les requête en attendant le Commit
    if ($this->bTransaction && !$this->bDebug) {
      $this->aTmpMessages[] = $aMessage;
    } else {
      $this->aMessages[] = $aMessage;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function stopQuery(): void {
    // Ecriture de toutes les requêtes dans logs
    foreach ($this->aMessages as $aMessage) {
      $sMethod = $aMessage['level'];

      $sMessage = json_encode([
        'login' => $aMessage['login'],
        'request' => $aMessage['request'],
        'parameters' => $aMessage['parameters'],
        'real_request' => $aMessage['real_request'],
        'time_exec' => ((microtime(true) - $aMessage['start']) * 1000) . ' ms',
      ]);

      $this->oLogger->$sMethod($sMessage);
    }

    $this->aMessages = [];
  }
}
