<?php

namespace App\Module\Vitis\WebService\Service\Logger;

use App\Module\Vitis\WebService\Service\Configuration\Module;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Logger\CustomStreamHandler;
use Monolog\Formatter\LineFormatter;
use Monolog\Logger;
use Monolog\Processor\IntrospectionProcessor;
use Monolog\Processor\ProcessIdProcessor;
use Monolog\Processor\PsrLogMessageProcessor;
use Monolog\Processor\WebProcessor;

/**
 * Permet d'instancier les différents logger
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class LoggerManager {
  /**
   * Pattern pour générer les lignes de logs
   * extra :  Ensemble des données généré par les processors
   * context : Ensemble des variables lors de la détection d'une erreur
   * @var string
   */
  private const LINE_FORMAT = "[%datetime%] | %level_name% | %message% | %extra% | %context%\n";
  /**
   * Pattern pour générer la date
   * @var string
   */
  private const LINE_DATE_FORMAT = "Y-m-d H:i:sP";
  /**
   * @var Logger[]
   */
  private $aLogger = [];

  /**
   * @var Properties
   */
  private $oPropertiesSrvc;

  /**
   * @var array
   */
  private $aLogLevel = [
    'DEBUG' => Logger::DEBUG,
    'INFO' => Logger::INFO,
    'NOTICE' => Logger::NOTICE,
    'WARNING' => Logger::WARNING,
    'ERROR' => Logger::ERROR,
    'CRITICAL' => Logger::CRITICAL,
    'ALERT' => Logger::ALERT,
    'EMERGENCY' => Logger::EMERGENCY,
  ];
  /**
   * Pattern pour Configurer les cannaux
   * @var array
   */
  private $aConfigLogger = [];

  /**
   * @param Properties $PropertiesSrvc Un objet pour accéder aux properties de l'application
   * @param Module $aModuleConfigurator Service de configuration des modules seule la partie monolog_channels est utile
   * ex :             "Monolog_Channels" : {
   *                         "COMMON_DEBUG" : {
   *                                "if" : "@prop(debug) === true",
   *                                "path" : "@prop(vas_dir)/var/logs/@date/web/debug.log",
   *                                "level" : "INFO",
   *                                "processors" : [
   *                                    "Monolog\\Processor\\PsrLogMessageProcessor",
   *                                    "Monolog\\Processor\\IntrospectionProcessor",
   *                                    "Monolog\\Processor\\ProcessIdProcessor",
   *                                    "Monolog\\Processor\\WebProcessor"
   *                                ]
   *                          }
   *                  }
   */
  public function __construct(Properties $PropertiesSrvc, Module $aModuleConfigurator) {
    // Sauvegarde du service de properties
    $this->oPropertiesSrvc = $PropertiesSrvc;
    // récupération de la configuration des modules
    $aModuleConfig = $aModuleConfigurator->getConfiguration();
    // Enregistrement de la configuration de canaux de logs
    if (isset($aModuleConfig['Monolog_Channels'])) {
      $this->aConfigLogger = array_merge($this->aConfigLogger, $aModuleConfig['Monolog_Channels']);
    }

    // rédéfinition de la sortie de la fonction error_log
    $sDateReplacer = date($this->oPropertiesSrvc->get('log_period'));
    $sErrorLogDir = $this->oPropertiesSrvc->get('vas_dir') . '/var/log/' . $sDateReplacer . '/php';
    // création du dossier pour que que php puisse écrire dans le fichier
    if (!file_exists($sErrorLogDir)) {
      mkdir($sErrorLogDir, 0777, true);
    }
    $sErrorLogFile = $sErrorLogDir . '/error.log';
    ini_set('error_log', $sErrorLogFile);
  }

  /**
   * Permet de générer un canal pour produire des logs.
   * @param string $sName nom du canal
   * @param string $sPath Chemin vers le fichier de log du canal
   * @param string $sLevel Niveau de log minimum à envoyer dans le canal
   * @param callable[] $aProcessors Ensemble de classe permettant de formater les extras
   * @return Logger
   */
  private function createLogger(string $sName, string $sPath, string $sLevel = 'INFO', array $aProcessors = []): Logger {
    // create a log channel
    $this->aLogger[$sName] = new Logger($sName, [], $aProcessors);
    $oHandler = new CustomStreamHandler($this->oPropertiesSrvc, $sPath);
    $oFormatter = new LineFormatter(self::LINE_FORMAT, self::LINE_DATE_FORMAT, false, true);
    $oHandler->setFormatter($oFormatter);
    $this->aLogger[$sName]->pushHandler($oHandler);

    return $this->aLogger[$sName];
  }

  /**
   * Retourne un objet Monolog\Logger utilisant le canal demandé ou null si le canal n'existe pas
   * @param string $sName nom du canal
   * @return Logger
   */
  public function get(string $sName): Logger {
    // s'il existe on le retourne
    if (isset($this->aLogger[$sName])) {
      return $this->aLogger[$sName];
    } // s'il n'existe pas mais qu'il est commun on le créé
    else if (isset($this->aConfigLogger[$sName])) {
      // process du chemin
      $sDateReplacer = date($this->oPropertiesSrvc->get('log_period'));
      $sPath = $this->oPropertiesSrvc->replacePropertiesInString($this->aConfigLogger[$sName]['path']);
      $sPath = str_replace('@date', $sDateReplacer, $sPath);
      // Définition du niveau
      $sLevel = (isset($this->aConfigLogger[$sName]['level'])) ? $this->aConfigLogger[$sName]['level'] : 'INFO';
      // Initialisation des processors
      $aProcessors = [];
      if (isset($this->aConfigLogger[$sName]['processors'])) {
        foreach ($this->aConfigLogger[$sName]['processors'] as $sProcessor) {
          $aProcessors[] = new $sProcessor();
        }
      }
      // création du logger
      return $this->createLogger($sName, $sPath, $sLevel, $aProcessors);
    } 
    // si la clé n'est pas trouvé on retourne le logger par défaut
    return $this->get('API_ERROR');
  }

}

?>