<?php

namespace App\Module\Vitis\WebService\Service\Mail;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Mail\SMTPConnexion;

class Mail {
  /**
   * @var string[]
   */
  private const ALIAS_SEPARATORS = [';', "\r\n", "\r", "\n", ' '];
  /**
   * @var string
   */
  private const SEPARATOR = ',';
  /**
   * @var string
   */
  private $sSubject;
  /**
   * @var boolean
   */
  private $bUseHTML = true;
  /**
   * @var string
   */
  private $sBody;
  /**
   * @var array
   */
  private $aAttachments = [];
  /**
   * @var array
   */
  private $aTo = [];
  /**
   * @var array
   */
  private $aCc = [];
  /**
   * @var array
   */
  private $aCci = [];
  /**
   * @var array
   */
  private $aReply = [];
  /**
   * @var Properties
   */
  private $oProperties;
  /**
   * @var \Monolog\Logger
   */
  private $oErrorLogger;
  /**
   * @var \Monolog\Logger
   */
  private $oNotifyLogger;
  /**
   * @var LoggerManager
   */
  private $oLoggerManager;
  /**
   * @var \App\Module\Vitis\WebService\Service\FileAccessor\FileAccessorInterface
   */
  private $oFileAccessor;
  /**
   * @var \App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor
   */
  private $oLocalAccessor;
  /**
   * @var array
   */
  private $aSmtpConfiguration = [];

  /**
   * @param Properties $oProperties service d'accés aux properties
   * @param LoggerManager $oLogger Service d'accès aux logs
   * @param FileManager $oFileManager Service d'accès aux fichiers
   */
  public function __construct(Properties $oProperties, LoggerManager $oLogger, FileManager $oFileManager) {
    $this->oErrorLogger = $oLogger->get('API_ERROR');
    $this->oNotifyLogger = $oLogger->get('NOTIFY_INFO');
    $this->oLoggerManager = $oLogger;
    $this->oProperties = $oProperties;
    $this->oFileAccessor = $oFileManager->getFileAccessor();
    $this->oLocalAccessor = $oFileManager->getLocalFileAccessor();
  }

  /**
   *Charge un email depuis la structure
   *@param array $aMailStruct Structure of the email :
   *    aMailStruct.to           (string|array) => liste des destinataires,
   *    aMailStruct.cc           (string|array) => liste des copies,
   *    aMailStruct.cci          (string|array) => liste des copies cachées,
   *    aMailStruct.subject      (string)       => Sujet du mail
   *    aMailStruct.body         (string)       => Corps du mail
   *    aMailStruct.aAttachments (string|array) => Liste des chemins vers les PJ
   *    aMailStruct.bUseHTML     (boolean)      => Défini si le body est au format HTML
   *    aMailStruct.smtp_login   (string|null)  => Login de connexion au SMTP (optionnel)
   *    aMailStruct.smtp_password(string|null)  => password de connexion au SMTP (optionnel)
   * @return void
   */
  public function loadFromArray(array $aMailStruct): void {
    // set du sujet
    $this->setSubject($aMailStruct["subject"]);

    $this->bUseHTML = $aMailStruct["bUseHTML"] ?? true;

    // corps du mail
    $this->sBody = $aMailStruct["body"];

    $this->manageAttachment($aMailStruct["aAttachments"]);

    $this->aTo = (is_array($aMailStruct['to'])) ? $aMailStruct['to'] : $this->splitDestinationList($aMailStruct['to']);
    $this->aCc = (is_array($aMailStruct['cc'])) ? $aMailStruct['cc'] : $this->splitDestinationList($aMailStruct['cc']);
    $this->aCci = (is_array($aMailStruct['cci'])) ? $aMailStruct['cci'] : $this->splitDestinationList($aMailStruct['cci']);

    if (isset($aMailStruct['smtp_login'])) {
      $this->setSmtpLogin($aMailStruct['smtp_login']);
    }
    if (isset($aMailStruct['smtp_password'])) {
      $this->setSmtpPassword($aMailStruct['smtp_password']);

    }
  }
  /**
   * défini le sujet du mail et ajoute su tag en properties devant
   * @param string $sSubject sujet du mail
   * @return self
   */
  public function setSubject(string $sSubject): self {
    // concaténation du tag avec le sujet du mail
    $this->sSubject = $sSubject;
    return $this;
  }
  /**
   * Ajoute un destinataire au mail
   * @param string $sMail destinataire à ajouter
   * @return self
   */
  public function addTo(string $sMail): self {
    $this->aTo[] = $sMail;
    return $this;
  }
  /**
   * Ajoute un destinataire en cc au mail
   * @param string $sMail destinataire à ajouter
   * @return self
   */
  public function addCc(string $sMail): self {
    $this->aCc[] = $sMail;
    return $this;
  }
  /**
   * Ajoute un destinataire en cci au mail
   * @param string $sMail destinataire à ajouter
   * @return self
   */
  public function addCci(string $sMail): self {
    $this->aCci[] = $sMail;
    return $this;
  }
  /**
   * Ajoute un destinataire en replyTo au mail
   * @param string $sMail destinataire à ajouter
   * @return self
   */
  public function addReply(string $sMail): self {
    $this->aReply[] = $sMail;
    return $this;
  }
  /**
   * Ajoute une PJ
   * @param string $sPath chemin vers la pièce jointe
   * @param string $sFileName nom du ficheir à afficher dans le mail
   * @return self
   */
  public function addAttachment(string $sPath, string $sFileName = ''): self {
    $sFileName = (!empty($sFileName)) ? $sFileName : $this->oFileAccessor->getFileName($sPath);
    $this->aAttachments[] = ['path' => $sPath, 'name' => $sFileName];
    return $this;
  }
  /**
   * défini si le body est au format HTML
   * @param boolean $bIshtml défini si le corp du mail est en HTML
   * @return self
   */
  public function isHtml(bool $bIshtml): self {
    $this->bUseHTML = $bIshtml;
    return $this;
  }
  /**
   * défini le corp du mail
   * @param string $sBody Corp du mail
   * @return self
   */
  public function setBody(string $sBody): self {
    $this->sBody = $sBody;
    return $this;
  }
  /**
   * défini le Login de connexion au SMTP (surcharge la config en properties)
   * @param string $sLogin login SMTP à utiliser
   * @return self
   */
  public function setSmtpLogin(string $sLogin): self {
    $this->aSmtpConfiguration['smtp_login'] = $sLogin;
    return $this;
  }
  /**
   * défini le Password de connexion au SMTP (surcharge la config en properties)
   * @param string $sPassword Mot de passe à utiliser
   * @return self
   */
  public function setSmtpPassword(string $sPassword): self {
    $this->aSmtpConfiguration['smtp_password'] = $sPassword;
    return $this;
  }
  /**
   * défini la Configuration SMTP spécifique (surcharge la config) en properties
   * @param array $aSmtpConfiguration Paramètres de la connexion SMTP.
   * @return self
   */
  public function setSmtpConfiguration(array $aSmtpConfiguration): self {
    $this->aSmtpConfiguration = $aSmtpConfiguration;
    return $this;
  }
  /**
   * Divise une chaine contenant plusieurs mails séparé par , en un tableau de mail
   * @param string $sList Liste de mails à séparer
   * @return array Liste des mails
   */
  private static function splitDestinationList(?string $sList): array{
    if (empty($sList)) {
      return [];
    }

    $sUniformList = str_replace(self::ALIAS_SEPARATORS, self::SEPARATOR, $sList);
    return explode(self::SEPARATOR, $sUniformList);
  }
  /**
   * Analyse le paramètre attachments pour récupérer un tableau propre
   * @param string[]|string|null $aAttachments Liste des pièces jointes
   * @return array Liste des PJ propre
   */
  private function manageAttachment($aAttachments = null): array{
    if (empty($aAttachments)) {
      return [];
    }
    if (is_array($aAttachments)) {
      foreach ($aAttachments as $sAttachment) {
        $this->addAttachment($sAttachment);
      }
    } else {
      $this->addAttachment($aAttachments);
    }
    return $this->aAttachments;
  }

  /**
   * Envoi le mail configuré
   * @return bool true si le mail est partie false sinon
   */
  public function send(): bool {
    $sTmpDir = $this->oLocalAccessor->getTempDir();
    $bReturn = false;
    try {
      $oSMTPConnexion = new SMTPConnexion($this->oProperties, $this->oLoggerManager);
      // Configuration SMTP optionnelle.
      if (!empty($this->aSmtpConfiguration)) {
        $oSMTPConnexion->setConfiguration($this->aSmtpConfiguration);
      }

      $oMailer = $oSMTPConnexion->connect();
      if ($oMailer === null) {
        throw new \Exception('SMTP connection failed');
      }
      // ajout des destinataires
      foreach ($this->aTo as $sTo) {
        $oMailer->addAddress($sTo);
      }
      foreach ($this->aCc as $sCC) {
        $oMailer->addCC($sCC);
      }
      foreach ($this->aCci as $sCCi) {
        $oMailer->addBCC($sCCi);
      }
      foreach ($this->aReply as $sReply) {
        $oMailer->addReplyTo($sReply);
      }
      // set du sujet
      $oMailer->Subject = $this->oProperties->get('mail_tag_subject') . $this->sSubject;
      // set du body
      $oMailer->isHTML($this->bUseHTML);
      $oMailer->Body = ($this->bUseHTML) ? '<html><body>' . $this->sBody . '</body></html>' : $this->sBody;
      // gestion des PJ
      foreach ($this->aAttachments as $aAttachment) {
        $sAttachment = $aAttachment['path'];
        if ($this->oFileAccessor->file_exists($sAttachment)) {
          if ($this->oProperties->get('file_system') !== Properties::LOCAL_FS) {
            $sTmpFile = $sTmpDir . "/" . $this->oFileAccessor->getFileName($sAttachment);
            if (!$this->oLocalAccessor->is_dir($sTmpDir)) {
              $this->oLocalAccessor->mkdir($sTmpDir, 0777, true);
            }
            $this->oFileAccessor->copyDistantToLocal($sAttachment, $sTmpFile);
            $oMailer->addAttachment($sTmpFile, $aAttachment['name']);
          } else {
            $oMailer->addAttachment($sAttachment, $aAttachment['name']);
          }
          //$mail->addAttachment($path, $name, $encoding = 'base64', $type = 'application/octet-stream');
        }
      }
      if ($oMailer->send()) {
        $this->oNotifyLogger->info('Send mail to : ' . implode(', ', $this->aTo), $this->getMail());
        $bReturn = true;
      } else {
        throw new \Exception('SMTP connection succesfull but cannot send mail');
      }
    } catch (\Exception $e) {
      $this->oNotifyLogger->error('Cannot send mail to : ' . implode(',', $this->aTo) . ' ' . $e->getMessage(), $this->getMail());
    } finally {
      // nettoyage des PJ
      if ($this->oProperties->get('file_system') !== Properties::LOCAL_FS) {
        $this->oLocalAccessor->clearDir($sTmpDir);
      }
      return $bReturn;
    }
  }
  /**
   * retourne une structure standard pour le mail (logs)
   * @return array
   */
  private function getMail(): array{
    return [
      'to' => $this->aTo,
      'cc' => $this->aCc,
      'cci' => $this->aCci,
      'subject' => $this->sSubject,
      'attachments' => count($this->aAttachments),
    ];
  }
}
