<?php

namespace App\Module\Vitis\WebService\Service\PrintService;

use App\Module\Vitis\WebService\Service\VitisSystem;

class PrintTemplate extends AbstractPrinter {
  /**
   * Evalue un code HTML en tuilisant un scope json
   * @param string $sHtmlBody HTML à évaluer
   * @param string $sFormat A4, A3, A2, A1, A0
   * @param string $sOutputFormat pdf, image, vectorpdf, imagepdf
   * @param string $sOrientation portrait, landscape
   * @param array $aScope donnée utilisé dans le template HTML
   * @param array $aIncludes images à inclure dans l'HTML
   * @param string $sDestFilePath chemin du fichier de sortie
   * @return string|null HTML évalué ou null en cas d'erreur
   */
  public function print(string $sHtmlBody, string $sFormat, string $sOutputFormat, string $sOrientation, array $aScope, array $aIncludes, array $aMaps, string $sDestFilePath): ?string {
    try {
      // génération du tableau des arguments
      $aArguments = $this->prepareArguments($sHtmlBody, $sFormat, $sOutputFormat, $sOrientation, $aScope, $aIncludes, $aMaps);
      //exécution de la commande
      $sReturn = $this->callPrinter($aArguments);
      $this->oLocalFileAccessor->file_put_contents($this->sTempDir . '/print.log', $sReturn);
      $this->oDebugLogger->debug($sReturn);

      if ($this->oLocalFileAccessor->file_exists($this->sTempDir . '/rendered.pdf')) {
        $this->oLocalFileAccessor->copy($this->sTempDir . '/rendered.pdf', $sDestFilePath);
      } else {
        $this->oErrorLogger->error('Error while printing');
        $this->oErrorLogger->error($sReturn);
        return null;
      }

      if (!empty($this->sTempDir)) {
        if ($this->oLocalFileAccessor->file_exists($this->sTempDir)) {
          $this->oLocalFileAccessor->clearDir($this->sTempDir);
        }
      }

    } catch (\Exception $e) {
      $this->oErrorLogger->error('Impossible to print HTML : ' . $e->getMessage());
      return null;
    }
    return $sDestFilePath;
  }

  /**
   * Evalue un code HTML en tuilisant un scope json
   * @param string $sJsonBody Json à évaluer
   * @param array $aScope donnée utilisé dans le template Json
   * @param string $sDestFilePath chemin du fichier de sortie
   * @return string|null Json évalué ou null en cas d'erreur
   */
  public function printExcel(string $sJsonBody, array $aScope, string $sDestFilePath): ?string {
    try {
      // génération du tableau des arguments
      $aArguments = $this->prepareArgumentsExcel($sJsonBody, $aScope);
      //exécution de la commande
      if (VitisSystem::runOnWindows()) {
        $sCommand = self::NODE_SCRIPT_PATH['command_win'];
      } else {
        $sCommand = self::NODE_SCRIPT_PATH['command_lin'];
      }
      $sNodeScriptPath = $this->oProperties->replacePropertiesInString($sCommand);
      if ($sNodeScriptPath === null) {
        throw new \Exception('Cannot process path to node script');
      }
      $sReturn = $this->call($sNodeScriptPath, $aArguments);
      $this->oLocalFileAccessor->file_put_contents($this->sTempDir . '/print.log', $sReturn);
      $this->oDebugLogger->debug($sReturn);

      if ($this->oLocalFileAccessor->file_exists($this->sTempDir . '/rendered.xlsx')) {
        $this->oLocalFileAccessor->copy($this->sTempDir . '/rendered.xlsx', $sDestFilePath);
      } else {
        $this->oErrorLogger->error('Error while printing');
        $this->oErrorLogger->error($sReturn);
        return null;
      }
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Impossible to print XSLX : ' . $e->getMessage());
      return null;
    } finally {
      if (!empty($this->sTempDir)) {
        if ($this->oLocalFileAccessor->file_exists($this->sTempDir)) {
          $this->oLocalFileAccessor->clearDir($this->sTempDir);
        }
      }
    }
    return $sDestFilePath;
  }

  /**
   * Prépare les arguments pour la CLI
   * @param string $sHtmlBody HTML à évaluer
   * @param string $sFormat A4, A3, A2, A1, A0
   * @param string $sOutputFormat pdf, image, vectorpdf, imagepdf
   * @param string $sOrientation portrait, landscape
   * @param array $aScope donnée utilisé dans le template HTML
   * @param array $aIncludes images à inclure dans l'HTML
   * @throws \Exception si le dossier temporaire ne peut pas être créé
   * @return array tableau d'argument à utiliser pour appeler phantomjs
   */
  private function prepareArguments(string $sHtmlBody, string $sFormat, string $sOutputFormat, string $sOrientation, array $aScope, array $aIncludes, array $aMaps): array{
    $this->sTempDir = $this->oLocalFileAccessor->getTempDir();
    if ($this->oLocalFileAccessor->mkdir($this->sTempDir, 0777, true) !== true) {
      throw new \Exception('Cannot create Tempdir : ' . $this->sTempDir);
    }
    // génération d'un dossier temporaire
    $aArguments = ['printTemplate'];
    // chemin vers l'entrée HTML
    $sHtmlPath = $this->sTempDir . '/input.html';
    $this->oLocalFileAccessor->file_put_contents($sHtmlPath, $sHtmlBody);
    $aArguments[] = '--HTMLInputFileName="' . $sHtmlPath . '"';
    // chemin vers la sortie PDF
    $aArguments[] = '--TmpFolder="' . $this->sTempDir . '"';
    $aArguments[] = '--OutputFileName="' . $this->sTempDir . '/rendered.pdf' . '"';
    // chemin vers les includes
    $sIncludesPath = $this->sTempDir . '/includes.json';
    $this->oLocalFileAccessor->file_put_contents($sIncludesPath, json_encode($aIncludes));
    $aArguments[] = '--JsonIncludes="' . $sIncludesPath . '"';
    // chemin vers les maps
    $sMapsPath = $this->sTempDir . '/maps.json';
    $this->oLocalFileAccessor->file_put_contents($sMapsPath, json_encode($aMaps));
    $aArguments[] = '--JsonMaps="' . $sMapsPath . '"';
    // chemin vers le scope
    $sScopePath = $this->sTempDir . '/scope.json';
    $this->oLocalFileAccessor->file_put_contents($sScopePath, json_encode($aScope));
    $aArguments[] = '--JsonScope="' . $sScopePath . '"';
    $aArguments[] = '--Format="' . $sFormat . '"';
    $aArguments[] = '--OutputFormat="' . $sOutputFormat . '"';
    $aArguments[] = '--Orientation="' . $sOrientation . '"';

    return $aArguments;
  }

  /**
   * Prépare les arguments pour la CLI
   * @param string $sJsonBody JSON à évaluer
   * @param array $aScope donnée utilisé dans le template JSON
   * @throws \Exception si le dossier temporaire ne peut pas être créé
   * @return array tableau d'argument à utiliser pour appeler phantomjs
   */
  private function prepareArgumentsExcel(string $sJsonBody, array $aScope): array{
    $this->sTempDir = $this->oLocalFileAccessor->getTempDir();
    if ($this->oLocalFileAccessor->mkdir($this->sTempDir, 0777, true) !== true) {
      throw new \Exception('Cannot create Tempdir : ' . $this->sTempDir);
    }
    // génération d'un dossier temporaire
    $aArguments = ['printExcel'];
    // chemin vers l'entrée JSON
    $sHtmlPath = $this->sTempDir . '/input.json';
    $this->oLocalFileAccessor->file_put_contents($sHtmlPath, $sJsonBody);
    $aArguments[] = '--JsonTemplate="' . $sHtmlPath . '"';
    // chemin vers la sortie XLSX
    $aArguments[] = '--OutputFileName="' . $this->sTempDir . '/rendered.xlsx' . '"';
    // chemin vers le scope
    $sScopePath = $this->sTempDir . '/scope.json';
    $this->oLocalFileAccessor->file_put_contents($sScopePath, json_encode($aScope));
    $aArguments[] = '--JsonScope="' . $sScopePath . '"';
    return $aArguments;
  }

}
