<?php

namespace App\Module\Vitis\WebService\Service;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\FileAccessor\LfsService;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Ramsey\Uuid\Uuid;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Service encapsulant le service Request de symfony.
 * L'autowire/configure utilise ce service en singleton,
 * il est instacié à la réception de la requête et
 * c'est toujours le même qui est passé dans les paramètres.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */
class VitisRequest {
  /**
   * @var string
   */
  public const TOKEN_HEADER = 'Authorization';

  /**
   * @var Request
   */
  private $oRequest;
  /**
   * @var \Monolog\Logger
   */
  private $oLogger;
  /**
   * @var Properties
   */
  private $oProperties;
  /**
   * @var LocalAccessor
   */
  private $oLocalFileAccessor;
  /**
   * @var LfsService
   */
  private $oLfsService;
  /**
   * @var boolean
   */
  private $bAlreadyParsed = false;
  /**
   * @param RequestStack $oRequestStack Service symfony contenant la requête HTTP
   * @param LoggerManager $oLogger Service d'accès aux logs
   * @param Properties $properties service d'accés aux properties
   */
  public function __construct(RequestStack $oRequestStack, LoggerManager $oLogger, Properties $properties, FileManager $oFileManager) {
    $this->oRequest = $oRequestStack->getCurrentRequest();
    $this->oLogger = $oLogger->get('API_ERROR');
    $this->oProperties = $properties;
    $this->oLocalFileAccessor = $oFileManager->getLocalFileAccessor();
    $this->oLfsService = $oFileManager->getLfsService();
  }

  /**
   * Extrait les paramètres de la requête.
   * @param bool $bNoDataInBody Permet de bypass la récupération du body
   * @throws \InvalidArgumentException quand un des services utilisés est manquant
   * @return array Liste des paramètres de la requête
   */
  public function getData(bool $bNoDataInBody = false): array{
    if ($this->bAlreadyParsed) {
      return $this->oRequest->request->all();
    } else {
      return $this->extractData($bNoDataInBody);
    }
  }

  /**
   * Extrait les paramètres de la requête.
   * @param bool $bNoDataInBody Permet de bypass la récupération du body
   * @throws \InvalidArgumentException quand un des services utilisés est manquant
   * @return array Liste des paramètres de la requête
   */
  public function extractData(bool $bNoDataInBody = false): array{
    // Methode POST
    $aData = $this->oRequest->request->all(); // array assoc
    // Methode GET
    $aQueryData = $this->oRequest->query->all(); // array assoc

    $aData = (!empty($aData)) ? $aData : [];

    if (!empty($aQueryData)) {
      $aData = array_merge($aData, $aQueryData);
    }
    // si on a pas besoin de parser le corps de la requête on retourne les données
    // sinon on parse le corps de la requête avant de retourner les données
    if (!$bNoDataInBody) {
      $aData = $this->requestBodyExtractor($aData);

      // surcharge de la structure de fichier si elle n'est pas vide
      $this->multipleFileStructCorrecter();
      $this->manageTransactionalFile($aData);
      $this->addPictureSizeIntoFileStruct($aData);
    }
    // on remplace le contenu de la requête
    $this->oRequest->request->replace($aData);
    $this->bAlreadyParsed = true;
    return $aData;
  }

  /**
   * Défini si la requête a déjà été parsée.
   * @return boolean true si la requête a déjà été parsée false sinon
   */
  public function alreadyParsed(): bool {
    return $this->bAlreadyParsed;
  }

  /**
   * Extrait les paramètres du corps de la requête.
   * @param array $aData Données déjà extraite de la requête
   * @throws \InvalidArgumentException quand un des services utilisés est manquant
   * @return array Liste des paramètres de la requête modifiée
   */
  private function requestBodyExtractor(array $aData): array{
    // récupération du corps de la requête
    $sContentData = $this->oRequest->getContent(); // text brut
    // si le body est vide on ne lance pas le traitement
    if (empty($sContentData)) {
      return $aData;
    }

    // extraction du content type
    $sContentType = $this->oRequest->headers->get('content-type') ?? '';

    if (strpos($sContentType, ';') > -1) {
      $aContentType = explode(';', $sContentType);
      $sContentType = trim($aContentType[0]);
    }

    switch ($sContentType) {
      case 'application/json':
      case 'text/plain':
        $aPostData = json_decode($sContentData, true);
        if (is_null($aPostData)) {
          $this->oLogger->error('Error occured while Decoding JSON : ' . json_last_error_msg(), [$sContentData]);
        } else {
          $aData = array_merge($aData, $aPostData);
        }
        break;
      case 'application/x-www-form-urlencoded':
        // rien à faire data est déjà rempli avec une array key/value POST et PUT
        // $sContentData contient une chaine key1=value1&key2=value2
        break;
      case 'multipart/form-data':
        // rien à faire data est déjà rempli avec une array key/value POST
        // en PUT il faut traiter le corps de la reqête
        $aData = $this->parseMultipartBody($aData, $sContentData);
        break;
      default:
        $aData = [];
        $this->oLogger->error($sContentType . ' not accepted');
        break;
    }
    return $aData;
  }

  /**
   * Extrait la structure de $_FILES en multifile pour correspondre à nos besoins
   * @return void
   */
  private static function multipleFileStructCorrecter(): void {
    // pour chaque fichier dans le structure de fichier
    foreach ($_FILES as $key => $value) {
      if (!isset($value['name'])) {
        continue;
      }
      // Fichier multiple
      if (is_array($value['name'])) {
        $aFiles = [];
        for ($i = 0; $i < count($value['name']); $i++) {
          $aFile = [];
          foreach ($value as $key2 => $value2) {
            $aFile[$key2] = $value[$key2][$i];
          }
          array_push($aFiles, $aFile);
        }
        $_FILES[$key] = $aFiles;
      }
    }
  }

  /**
   * finalise la structure de fichier pour les images
   * @return void
   */
  private static function addPictureSizeIntoFileStruct(array $aValues): void {
    foreach ($_FILES as $sKeyField => $aFileValue) {
      // si le paramètre width ou height est défini
      if (isset($aValues[$sKeyField . '_width']) || isset($aValues[$sKeyField . '_height'])) {
        // Fichier multiple
        if (!array_key_exists('name', $aFileValue)) {
          for ($i = 0; $i < count($aFileValue); $i++) {
            if (isset($aValues[$sKeyField . '_width'])) {
              $_FILES[$sKeyField][$i]['width'] = $aValues[$sKeyField . '_width'];
            }
            if (isset($aValues[$sKeyField . '_height'])) {
              $_FILES[$sKeyField][$i]['height'] = $aValues[$sKeyField . '_height'];
            }
          }
        } else {
          if (isset($aValues[$sKeyField . '_width'])) {
            $_FILES[$sKeyField]['width'] = $aValues[$sKeyField . '_width'];
          }
          if (isset($aValues[$sKeyField . '_height'])) {
            $_FILES[$sKeyField]['height'] = $aValues[$sKeyField . '_height'];
          }
        }
      }
    }
  }

  /**
   * Extrait les paramètres du corps de la requête si elle est de type multipart/form-data.
   * @param array $aData Données déjà extraite de la requête
   * @param string $sContentData Corps de la requête
   * @throws \InvalidArgumentException quand un des services utilisés est manquant
   * @return array Liste des paramètres de la requête modifiée
   */
  private function parseMultipartBody(array $aData, string $sContentData): array{
    $sBoundary = null;
    $sContentType = $this->oRequest->headers->get('content-type') ?? '';

    if (strpos($sContentType, ';') > -1) {
      $aContentType = explode(';', $sContentType);
      $sContentType = trim($aContentType[0]);
      $sBoundary = '--' . trim(str_replace('boundary=', '', $aContentType[1]));
    }

    if (!empty($sContentData) && !empty($sBoundary)) {
      $aFormData = explode($sBoundary, $sContentData);
      //$aFormData = is_array($aFormData) ? $aFormData : [];
      foreach ($aFormData as $sPart) {
        if (!empty($sPart) && $sPart != "--\r\n") {
          // extraction du header et du corps
          $sPart = trim($sPart, "\r\n");
          //list($sRawHeaders, $sBody);
          $aParts = explode("\r\n\r\n", $sPart, 2);
          if (count($aParts) < 1) {
            continue;
          }
          // si le body n'est pas set sa valeur est null
          $sBody = isset($aParts[1]) ? $aParts[1] : null;
          // traitement des headers
          $sRawHeaders = explode("\r\n", $aParts[0]);
          $aHeaders = [];
          foreach ($sRawHeaders as $sHeader) {
            list($name, $value) = explode(':', $sHeader, 2);
            $aHeaders[strtolower($name)] = trim($value);
          }
          // extraction du nom du paramètre et traitement des fichiers en PUT
          if (isset($aHeaders['content-disposition'])) {
            preg_match('/^(.+); *name="([^"]+)"(; *filename="([^"]+)")?/', $aHeaders['content-disposition'], $matches);
            list(, $type, $name) = $matches;
            // nom de fichier en 4eme partie
            $sFilename = (isset($matches[4])) ? $matches[4] : null;
            if (!empty($sFilename)) {
              $sTempDir = $this->oLocalFileAccessor->getTempDir();
              $sTempFile = $sTempDir . '/' . Uuid::uuid1()->toString();
              if ($this->oLocalFileAccessor->mkdir($sTempDir, 0777, true) === true) {
                $iFileSize = $this->oLocalFileAccessor->file_put_contents($sTempFile, $sBody);
                // traitement des fichiers en PUT
                if (strpos($name, '[]') > -1) {
                  $name = str_replace('[]', '', $name);
                  if (!isset($_FILES[$name])) {
                    $_FILES[$name] = [];
                  }
                  $_FILES[$name][] = [
                    'name' => $sFilename,
                    'type' => $aHeaders['content-type'],
                    'tmp_name' => $sTempFile,
                    'error' => 0,
                    'size' => $iFileSize,
                  ];
                } else {
                  $_FILES[$name] = [
                    'name' => $sFilename,
                    'type' => $aHeaders['content-type'],
                    'tmp_name' => $sTempFile,
                    'error' => 0,
                    'size' => $iFileSize,
                  ];
                }
              }
            } else {
              if (strpos($name, '[]') > -1) {
                $name = str_replace('[]', '', $name);
                if (!isset($aData[$name])) {
                  $aData[$name] = [];
                }
                if (!empty(trim($sBody))) {
                  $aData[$name][] = $sBody;
                }
              } else {
                if (!empty(trim($sBody)) || $sBody === null || $sBody === '0') {
                  $aData[$name] = $sBody;
                }
              }
            }
          }
        }
      }
    }
    return $aData;
  }

  /**
   * Permet de parser la structure qui sotck les transactions et de les placer dans $_FILES
   *
   * @param  array $aData
   * @return void
   */
  private function manageTransactionalFile(array $aData): void {
    // si le dictionnaire de transaction n'existe pas on coupe
    if (!array_key_exists('vitis_transaction_files', $aData)) {
      return;
    }

    $aMapTransaction = $aData['vitis_transaction_files'];
    // gestion en cas de json
    if (is_string($aMapTransaction)) {
      $aMapTransaction = json_decode($aMapTransaction, true);
    }
    // on parcours le tableau de transaction
    foreach ($aMapTransaction as $sFieldName => $aTransactionId) {
      foreach ($aTransactionId as $sTransactionId) {
        $sPath = $this->oLfsService->getFilePath($sTransactionId);

        if (empty($sPath)) {
          continue;
        }

        $aTransactionFileStruct = [
          'name' => $this->oLfsService->getFileName($sTransactionId),
          'type' => $this->oLfsService->getFileMimeType($sTransactionId),
          'tmp_name' => $sPath,
          'error' => 0,
          'size' => $this->oLfsService->getFileSize($sTransactionId),
          'transaction' => $sTransactionId,
        ];

        // gestion en cas de fichier multiple
        if (count($aTransactionId) == 1) {
          $_FILES[$sFieldName] = $aTransactionFileStruct;
        } else {
          if (!isset($_FILES[$sFieldName])){
            $_FILES[$sFieldName] = [];
          }

          $_FILES[$sFieldName][] = $aTransactionFileStruct;
        }

      }
    }
  }

  /**
   * Retourne le token de la requête s'il est présent
   * @return string|null Le token s'il est là, null sinon
   */
  public function getToken(): ?string {
    if ($this->oRequest->headers->has(self::TOKEN_HEADER)) {
      return $this->oRequest->headers->get(self::TOKEN_HEADER);
    }
    return null;
  }
  /**
   * Retourne le nom de la route utilisée
   * @return string identifiant de la route dans route.json
   */
  public function getRouteName(): string {
    return $this->oRequest->attributes->get('_route');
  }
  /**
   * Retourne l'ip de provenance de la requête'
   * @return string|null l'ip du client
   */
  public function getClientIp(): ?string {
    return $this->oRequest->getClientIp();
  }
  /**
   * Retourne l'objet requête de symfony
   * @return Request serviec contenant la requête de symfony
   */
  public function getRequest(): Request {
    return $this->oRequest;
  }
  /**
   * Retourne Method
   * @return string Method
   */
  public function getMethod(): string {
    return $this->oRequest->getMethod();
  }
}
