/* global ol */

/**
 * Class Map
 * @param {object} opt_options
 */
PrintMap = function (opt_options) {

    var this_ = this;

    oProperties.print = oProperties.print ? oProperties.print : {};

    this.mapSize = [document.getElementById('map').offsetWidth, document.getElementById('map').offsetHeight];

    this.tileSize = (isDef(window.oProperties.print.tile_size) && window.oProperties.print.features_zoom >= 0) ? window.oProperties.print.tile_size : 2048;

    this.cacheSize = 0;

    this.featuresZoom = (parseFloat(opt_options.featuresZoom) >= 0) ? opt_options.featuresZoom : (isDef(window.oProperties.print.features_zoom) && window.oProperties.print.features_zoom >= 0) ? window.oProperties.print.features_zoom : 100;

    this.resolutionCoeff = (parseFloat(opt_options.resolutionCoeff) > 0) ? opt_options.resolutionCoeff : 1;

    this.defaultScale = 1000;

    this.apiUrl = opt_options.apiUrl;

    this.token = opt_options.token;

    this.mapDefinition = opt_options.mapJSON;

    this.mapJsonParser = new MapJSON({
        'properties': window.oProperties
    });

    this.view = this.mapJsonParser.getViewFromDef(this.mapDefinition, {
        'size': this.mapSize,
        'tileSize': [this.tileSize, this.tileSize]
    });

    this.features = isDef(opt_options.features) ? this.getFeaturesFromEWKT(opt_options.features) : null;

    this.measureFeatures = isDef(opt_options.measureFeatures) ? this.getFeaturesFromGeoJSON(opt_options.measureFeatures) : null;

    this.extent = this.getExtent(opt_options);

    this.tileGrid = this.getTileGridFromDefinition(this.tileSize);

    this.layers = this.mapJsonParser.getLayersFromDef(this.mapDefinition, {
        'size': this.mapSize,
        'tileSize': [this.tileSize, this.tileSize]
    });

    this.map = this.setMap(this.layers, this.view);

    this.featuresOverlay = this.setFeaturesOverlay(this.map);

    this.scaleTools = new Scale(this.map, this.resolutionCoeff);

    // Ajoute les features
    if (isDef(this.features)) {
        this.featuresOverlay.getSource().addFeatures(this.features);
    }

    // Ajoute les measureFeatures
    if (isDef(this.measureFeatures)) {

        // Ajout de la couche vecteur
        this.measureFeaturesOverlay = this.setMeasureFeaturesOverlay(this.map);

        // Ajout des features dans la couche
        this.addMeasureFeatures(this.measureFeatures, this.measureFeaturesOverlay);
    }

    // Centre la carte sur son étendue
    this.setMapExtent(this.extent);

    // Supprime tous les controls par défaut
    this.removeMapControls();

    // Taille des icones en fonction de la résolution demandée
    this.setMapSymbolsResolution(this.map, this.resolutionCoeff)

    // Si aucune étendue a été donnée et que seul un point a été donné, alors zoom sur this.defaultScale
    if (!isDef(opt_options.extent) && !isDef(this.features)) {
        if (this.features.length === 1 && this.features[0].getGeometry().getType() === 'Point') {
            this.scaleTools.setScale(this.defaultScale);
        }
    }

    // Renseigne l'échelle en cours
    tryCallPhantom({
        'cmd': 'setScale',
        'scale': this.scaleTools.getScale({
            'pretty': true
        })
    });
};

/**
 * Get the extent to use
 * @param {object} opt_options
 * @returns {array} extent
 */
PrintMap.prototype.getExtent = function (opt_options) {

    if (isDef(opt_options.extent)) {
        if (opt_options.extent.length === 4) {
            opt_options.extent[0] = Number(opt_options.extent[0]);
            opt_options.extent[1] = Number(opt_options.extent[1]);
            opt_options.extent[2] = Number(opt_options.extent[2]);
            opt_options.extent[3] = Number(opt_options.extent[3]);
            return opt_options.extent;
        }
    }

    if (isDef(opt_options.features)) {
        return this.getExtentFromFeatures(this.features);
    }

    // Si on arrive ici c'est qu'il n'y a aucune possibilité de trouver une étendue valable
    callError('No way to find extent: Neigther extent nor features given on a acceptable format');
};

/**
 * Get the extent of the features
 * @param {array<ol.Feature>} aFeatures
 * @returns {Array} Extent
 */
PrintMap.prototype.getExtentFromFeatures = function (aFeatures) {

    var totalExtent = [];

    for (var i = 0; i < aFeatures.length; i++) {
        var geomExtent = aFeatures[i].getGeometry().getExtent();

        if (i === 0) {
            totalExtent = geomExtent;
            continue;
        }

        if (geomExtent[0] < totalExtent[0])
            totalExtent[0] = geomExtent[0];
        if (geomExtent[1] < totalExtent[1])
            totalExtent[1] = geomExtent[1];
        if (geomExtent[2] > totalExtent[2])
            totalExtent[2] = geomExtent[2];
        if (geomExtent[3] > totalExtent[3])
            totalExtent[3] = geomExtent[3];

        delete geomExtent;
    }

    var xDist = totalExtent[2] - totalExtent[0];
    var yDist = totalExtent[3] - totalExtent[1];

    totalExtent[0] -= (xDist * this.featuresZoom / 100) / 2;
    totalExtent[1] -= (yDist * this.featuresZoom / 100) / 2;
    totalExtent[2] += (xDist * this.featuresZoom / 100) / 2;
    totalExtent[3] += (yDist * this.featuresZoom / 100) / 2;

    // Par mesure de sécurité
    if (totalExtent.length < 4) {
        totalExtent = [0, 10, 0, 10];
    }

    // En cas de simple point par exemple
    if (totalExtent[0] === totalExtent[2]) {
        totalExtent[0] = totalExtent[0] - totalExtent[0] * 0.0001;
        totalExtent[2] = totalExtent[2] + totalExtent[2] * 0.0001;
    }
    if (totalExtent[1] === totalExtent[3]) {
        totalExtent[1] = totalExtent[1] - totalExtent[1] * 0.0001;
        totalExtent[3] = totalExtent[3] + totalExtent[3] * 0.0001;
    }

    return totalExtent;
};

/**
 * Get a tilegrid grom a tile size
 * @param {number} tileSize
 * @returns {ol.tilegrid.TileGrid}
 */
PrintMap.prototype.getTileGridFromDefinition = function (tileSize) {

    var projExtent = this.view.getProjection().getExtent();
    var startResolution = ol.extent.getWidth(projExtent) / 256;
    var resolutions = new Array(22);

    for (var i = 0, ii = resolutions.length; i < ii; ++i) {
        resolutions[i] = startResolution / Math.pow(2, i);
    }

    var tileGrid = new ol.tilegrid.TileGrid({
        origin: ol.extent.getTopLeft(this.extent),
        extent: projExtent,
        resolutions: resolutions,
        tileSize: [tileSize, tileSize]
    });

    return tileGrid;
};

/**
 * Remove the map controls
 */
PrintMap.prototype.removeMapControls = function () {
    var aControls = this.map.getControls().getArray();

    for (var i = 0; i < aControls.length; i++) {
        this.map.removeControl(aControls[i]);
    }
};

/**
 * Fit the map on thie given extent
 * @param {type} extent
 */
PrintMap.prototype.setMapExtent = function (extent) {
    this.map.getView().fit(extent, {
        constrainResolution: false,
        nearest: true
    });
};

/**
 * Set the app map
 * @param {array<ol.layer.Base>} aLayers
 * @param {ol.View} olView
 * @returns {ol.Map}
 */
PrintMap.prototype.setMap = function (aLayers, olView) {

    var map = new ol.Map({
        layers: aLayers,
        target: 'map',
        view: olView
    });

    return map;
};

/**
 * Set the map features overlay
 * @param {ol.Map} olMap
 * @returns {ol.layer.Vector}
 */
PrintMap.prototype.setFeaturesOverlay = function (olMap) {

    var featuresOverlay = new ol.layer.Vector({
        map: olMap,
        source: new ol.source.Vector({}),
        style: new ol.style.Style({
            fill: new ol.style.Fill({
                color: 'rgba(255, 255, 255, 0.3)'
            }),
            stroke: new ol.style.Stroke({
                color: '#d9534f',
                width: 3
            }),
            image: new ol.style.Circle({
                radius: 7,
                fill: new ol.style.Fill({
                    color: '#d9534f'
                })
            })
        })
    });

    return featuresOverlay;
};

/**
 * Set the map measure features overlay
 * @param {ol.Map} olMap
 * @returns {ol.layer.Vector}
 */
PrintMap.prototype.setMeasureFeaturesOverlay = function (olMap) {

    var featuresOverlay = new ol.layer.Vector({
        map: olMap,
        source: new ol.source.Vector({}),
        style: new ol.style.Style({
            fill: new ol.style.Fill({
                color: 'rgba(255, 255, 255, 0.2)'
            }),
            stroke: new ol.style.Stroke({
                color: '#ffcc33',
                width: 2
            }),
            image: new ol.style.Circle({
                radius: 7,
                fill: new ol.style.Fill({
                    color: '#ffcc33'
                }),
                stroke: new ol.style.Stroke({
                    color: '#FFFFFF'
                })
            })
        })
    });

    return featuresOverlay;
};

/**
 * Get the olFeatures from an array of EWKT features
 * @param {array<string>} aEWKTFeatures Array of EWKT features
 * @returns {array<ol.Feature>}
 */
PrintMap.prototype.getFeaturesFromEWKT = function (aEWKTFeatures) {

    var aFeatures = [];
    var proj = this.view.getProjection().getCode();

    for (var i = 0; i < aEWKTFeatures.length; i++) {
        aFeatures.push(new ol.Feature({
            geometry: this.getGeomFromEWKT(aEWKTFeatures[i], proj)
        }));
    }

    return aFeatures;
};

/**
 * Return true if EWKTGeom is an EWKT geometry
 * @param {string} EWKTGeom
 * @returns {boolean}
 */
PrintMap.prototype.isEWKTGeom = function (EWKTGeom) {

    if (EWKTGeom.substr(0, 5).toUpperCase() !== 'SRID=')
        return false;

    var WKTFormat = new ol.format.WKT();
    var WKTGeom = EWKTGeom.substr(EWKTGeom.indexOf(';') + 1);

    try {
        WKTFormat.readGeometry(WKTGeom);
    } catch (e) {
        console.error('readGeometry failed: ', EWKTGeom);
        return false;
    }

    return true;
};

/**
 * Return an OpenLayers geometry from a EWKT geom
 * @param {string} EWKTGeom
 * @param {string} proj
 * @returns {ol.geom.Geometry}
 */
PrintMap.prototype.getGeomFromEWKT = function (EWKTGeom, proj) {

    if (typeof EWKTGeom !== 'string') {
        console.error('EWKTGeom is not a string: ', EWKTGeom);
        return null;
    }
    if (typeof proj !== 'string') {
        console.error('proj is not a string: ', proj);
        return null;
    }

    if (this.isEWKTGeom(EWKTGeom)) {

        var WKTFormat = new ol.format.WKT();
        var WKTGeom = EWKTGeom.substr(EWKTGeom.indexOf(';') + 1);
        var EWKTGeomProj = 'EPSG:' + EWKTGeom.slice(5, EWKTGeom.indexOf(';'));

        var oGeom = WKTFormat.readGeometry(WKTGeom, {
            dataProjection: EWKTGeomProj,
            featureProjection: proj
        });

        return oGeom;

    } else {
        console.error('Geom is not an EWKT Geometry: ', EWKTGeom);
        return null;
    }

};

/**
 * Get the olFeatures from an array of GeoJSON features
 * @param {string} sGeoJSONFeatures Array of GeoJSON features
 * @returns {array<ol.Feature>}
 */
PrintMap.prototype.getFeaturesFromGeoJSON = function (sGeoJSONFeatures) {

    if (!isDef(sGeoJSONFeatures)) {
        callLog('error: sGeoJSONFeatures is not defined');
        return null;
    }

    if (!isDef(JSON.parse(sGeoJSONFeatures))) {
        callLog('error: sGeoJSONFeatures is not valid JSON');
        return null;
    }

    var proj = this.view.getProjection().getCode();

    oGeoJSON = new ol.format.GeoJSON({
        defaultDataProjection: proj,
        featureProjection: proj
    });

    var aFeatures = []
    try {
        aFeatures = oGeoJSON.readFeatures(sGeoJSONFeatures);
    } catch (e) {
        callLog('error: ' + e);
    }

    return aFeatures;
};

/**
 * For WMS layers set MAP.RESOLUTION and MAP.DEFRESOLUTION to rise the symbols size
 *
 * @param  {object} olMap
 * @param  {integer} resolutionCoeff
 */
PrintMap.prototype.setMapSymbolsResolution = function (olMap, resolutionCoeff) {

    var newRes = resolutionCoeff * 72;
    var aLayers = olMap.getLayers().getArray();
    for (var i = 0; i < aLayers.length; i++) {
        if (aLayers[i].get('type') === 'imagewms' || aLayers[i].get('type') === 'tilewms') {
            aLayers[i].getSource().updateParams({
                'MAP.RESOLUTION': newRes,
                'MAP.DEFRESOLUTION': 72
            });
        }
    }
}

/**
 * addMeasureFeatures function - add the measure features
 *
 * @param  {array<ol.Feature>} aFeatures
 */
PrintMap.prototype.addMeasureFeatures = function (aFeatures, oVectorLayer) {

    callLog('addMeasureFeatures, ' + aFeatures);

    // Ajoute les features
    for (var i = 0; i < aFeatures.length; i++) {

        // Cercles
        if (aFeatures[i].get('Type') === 'Cercle') {
            if (isDef(aFeatures[i].get('geom_center')) &&
                isDef(aFeatures[i].get('geom_radius'))) {
                aFeatures[i].setGeometry(new ol.geom.Circle(
                    aFeatures[i].get('geom_center'),
                    aFeatures[i].get('geom_radius')));
            }
        }

        oVectorLayer.getSource().addFeature(aFeatures[i]);
    }

    // Ajoute les tooltip
    for (var i = 0; i < aFeatures.length; i++) {
        this.createMeasureTooltip(aFeatures[i]);
    }
}

/**
 * createMeasureTooltip function - Create a measure feature tooltip
 *
 * @param  {ol.Feature} olFeature
 */
PrintMap.prototype.createMeasureTooltip = function (olFeature) {

    oMeasureTooltipElement = document.createElement('div');
    oMeasureTooltipElement.className = 'measure-tooltip measure-tooltip-measure';
    oMeasureTooltip = new ol.Overlay({
        element: oMeasureTooltipElement,
        offset: [0, -15],
        positioning: 'bottom-center'
    });
    oMeasureTooltip.set('type', 'measure');

    if (olFeature !== undefined)
        oMeasureTooltip.set('feature', olFeature);

    this.map.addOverlay(oMeasureTooltip);

    oMeasureTooltipElement.className = 'measure-tooltip measure-tooltip-static';
    oMeasureTooltipElement.style.fontSize = 12 * this.resolutionCoeff + 'px';
    oMeasureTooltip.setOffset([0.8, -13]);

    this.updateTooltip(olFeature, oMeasureTooltip);
}

/**
 * getTooltipPosition function - Get the position to be use on a tooltip
 *
 * @param  {ol.Geom} geom
 * @return {array} position
 */
PrintMap.prototype.getTooltipPosition = function (geom) {

    if (geom.getType() === 'Polygon') {
        var position = geom.getInteriorPoint().getLastCoordinate();
    } else if (geom.getType() === 'LineString') {
        var position = geom.getLastCoordinate();
    } else if (geom.getType() === 'Circle') {
        var position = geom.getCenter();
    } else if (geom.getType() === 'Point') {
        var position = geom.getLastCoordinate();
    } else if (geom.getType() === 'GeometryCollection') {
        var position = null;
    } else {
        var position = geom.getLastCoordinate();
    }

    return position;
};

/**
 * updateTooltip function - Update the feature tooltip text and position
 *
 * @param  {ol.feature} olFeature
 * @param  {ol.Overlay} oOverlay
 */
PrintMap.prototype.updateTooltip = function (olFeature, oOverlay) {

    if (olFeature.get('Type') === 'Ligne') {
        oOverlay.getElement().innerHTML = olFeature.get('Longueur');
    } else if (olFeature.get('Type') === 'Polygone') {
        oOverlay.getElement().innerHTML = olFeature.get('Superficie');
    } else if (olFeature.get('Type') === 'Cercle') {
        oOverlay.getElement().innerHTML = "Périmètre: " + olFeature.get('Perimètre');
    } else if (olFeature.get('Type') === 'Point') {

        var proj = this.view.getProjection().getCode();
        var coords = ol.proj.transform(olFeature.get('Coordonnées'), proj, olFeature.get('Projection'));
        var htmlInfos = document.createElement('div');
        htmlInfos.innerHTML = olFeature.get('Projection');
        htmlInfos.innerHTML += '<br>';
        htmlInfos.innerHTML += 'X = ' + parseFloat(coords[0]).toPrecision(9);
        htmlInfos.innerHTML += '<br>';
        htmlInfos.innerHTML += 'Y = ' + parseFloat(coords[1]).toPrecision(9);
        oOverlay.getElement().appendChild(htmlInfos);
    }

    var position = this.getTooltipPosition(olFeature.getGeometry());
    if (isDef(position)) {
        oMeasureTooltip.setPosition(position);
    }
};